# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 

# JobQueue v0.1
#
# This initial revision simply keeps jobs in RAM, using the most trivial priority scheme
# possible -- sorted by sequential job ID. A more complex version might use the HD for storage
# of jobs, manage deletion of old jobs, keep a limit on the number of queued jobs and block
# when the limit is exceeded, and report various statistics.
 
import heap, os, tarfile, mutex, array
import zipfile, py_compile, random
from threading import Thread
from Queue import Queue
from os import path
from time import sleep
from CHOMP import BinaryOutputArchive, BinaryInputArchive
from LazyEvaluation import delay, force
from Serialize import *
 
MAX_CLONE	= 10
MAX_RUN 	= 100
INIT_WUS        = 200  

class JobQueueDriver(Thread):
	def __init__(self, job_queue, project_id, data_dir, message_dir, 
		     delay=5, jobs_per_wu=10, wu_per_batch=20, wu_max_length=100.0,
		     local_files=[]):

		self.dir_template       = path.join(data_dir, 'PROJ%d', 'RUN%d', 'CLONE%d')
		self.wu_template 	= path.join(self.dir_template, 'priority%d.swu')
		self.return_template	= path.join(self.dir_template, 'priority%d.shn')
                
                self.init_ready_msg     = path.join(message_dir, 'SHARPEN_init_ready')
                self.init_complete_msg  = path.join(message_dir, 'SHARPEN_init_complete')
		self.ready_msg 		= path.join(message_dir, 'SHARPEN_ready_to_receive')
		self.sent_msg 		= path.join(message_dir, 'SHARPEN_work_units_sent')
		self.update_file 	= path.join(message_dir, 'WU_update')
 
		self.delay = delay
		self.jobs_per_wu = jobs_per_wu
		self.wu_per_batch = wu_per_batch
		self.wu_max_length = wu_max_length
		self.queue = job_queue
		self.local_files = local_files
 
		self.project = project_id
		self.run_number = 0
		self.clone_number = 0
 
		self.dead = False
		self.waiting_wus = []
		
		Thread.__init__(self)
		self.setDaemon(True)

	def kill(self):
		self.dead = True

	def clear_server_data(self):
		for i in xrange(MAX_RUN):
			for j in xrange(MAX_CLONE):
				data_dir = self.dir_template % (self.project, i, j)
				for file in os.listdir(data_dir):
					os.unlink( path.join(data_dir, file) )
 
	def run(self):
                # for now just use 10 for priority
                priority = 10

		# pre-construct a tar archive for efficiency
		tar_pkg = tarfile.open( 'package.tar', 'w' )
		zip_scripts = zipfile.ZipFile( 'mods', 'w' )
		
		for file in self.local_files:
			base, ext = path.splitext(file)
			if ext == '.py':
				py_compile.compile(file, dfile=base)
				zip_scripts.write(base + '.pyc') 
			else:
				tar_pkg.add(file)

		zip_scripts.close()
		tar_pkg.add('mods')
		tar_pkg.close()

		# Now enter the main loop looking for updates
                print '-*-JobQueue-*- Entering Main loop'
                while not self.dead:
			## TODO: control flow of this loop is really confusing. Fix it.
			sleep(self.delay)

			# check for returned work units
			returned_wus = []
			for wu_desc in self.waiting_wus:
				wu_file = self.return_template % wu_desc
				if path.isfile(wu_file):
					returned_wus.append(wu_desc)
					print '-*-JobQueue-*- received WU ', wu_desc
					self.queue.wu_has_returned(wu_file, wu_desc)
			for rwu in returned_wus:
				self.waiting_wus.remove(rwu)

			# check for semaphore files
			if path.isfile(self.init_ready_msg):
				numUnits = INIT_WUS
				msg = 'init'
				print '-*-JobQueue-*- Starting initial WU population'
			elif path.isfile(self.ready_msg):
				numUnits = self.wu_per_batch
				msg = 'update'
			else:
				# server is not ready to receive WU updates yet
				continue
			
			# if we've gotten this far, write out new WUs
			uf = open(self.update_file, 'w')
 
			# write out a batch of work units
			for i in xrange(numUnits):
				wu_desc = (self.project, self.run_number, self.clone_number, priority)

				# keep filling the WU with jobs until it's full (will sleep until enough jobs are available)
				success = self.queue.create_WU( wu_desc,
							      	self.jobs_per_wu,
						      		self.wu_max_length,
						     		'wudata.serialized',
						      		self.delay )

				# if create_WU returns False, that means it hasn't written out any jobs.
				# that indicates that it's completely out of remote jobs, and the queue has been flushed.
				# (note: it may still have some local jobs to process!)

				if not success:
					if i == 0:
						# if we haven't even written out any WUs at all, just bail
						write_wu_sent = False
						break
					else:
						# otherwise, send out whatever we have so far
						print '-*-JobQueue-*- JobQueue is empty -- writing incomplete batch of jobs.'
						write_wu_sent = True
						break
				else:
					write_wu_sent = True

				self.waiting_wus.append( wu_desc )
				
				# assemble the serialized data and Python scripts into a full work unit
				os.system('cp package.tar %s' % self.wu_template % wu_desc)
                                wu = tarfile.open( self.wu_template % wu_desc, 'a' )
				
				wu.add('wudata.serialized')
				wu.close()
				os.remove('wudata.serialized')

				uf.write('(%d,%d,%d,%d)\n' % (self.project, self.run_number, priority, self.clone_number))
 
				self.clone_number += 1
				if self.clone_number == MAX_CLONE:
					self.run_number += 1
					self.clone_number = 0
 
					if self.run_number == MAX_RUN:
						print '-*-JobQueue-*- WARNING: looping around to run0/clone0'
						self.run_number = 0

				print 'Deposited a WU to disk (%d,%d,%d)' % (self.project, self.run_number, self.clone_number)
 
			uf.close()
 
			# write out the WU_sent message
			if write_wu_sent:
				if msg == 'init':
					print '-*-JobQueue-*- Finished initial WU population'
					os.remove(self.init_ready_msg)
				   	os.mknod(self.init_complete_msg)
				else:
					print '-*-JobQueue-*- Updated server with ', numUnits, ' WUs'
					os.remove(self.ready_msg)
				    	os.mknod(self.sent_msg)

class ResultNotAvailable(Exception):
	def __init__(self, job):
		self.job = job
 
	def __str__(self):
		return "Tried to access data from job %d, but that data hasn't arrived yet." % self.job.id

class DistributedJob:
	def __init__(self, ID, fn, args, seed=None):
		self.id = ID
		self.fn = fn
		self.args = args	
		self.seed = seed

		self.dependents = []
		self.result = '!!notavail!!'

		if self.seed is None:
			self.seed = random.randint(0, 2**31-1)

	def force_args(self):
		try:
			self.args = [force(arg) for arg in self.args]
		except ResultNotAvailable, e:
			e.job.dependents.append(self)
			raise
 
	def write_to_archive(self, archive):
		Serialize(archive, '%d %d %d %s %s' % ( self.id, len(self.args), self.seed, self.fn.__module__, self.fn.__name__ ) )
		for arg in self.args:
			Serialize(archive, arg)
 
	def get_result(self):
		if self.result == '!!notavail!!':
			raise ResultNotAvailable(self)
		return self.result
 
	def _wu_has_returned(self, archive):
		self.result = Deserialize(archive)

class LocalJob(DistributedJob):
	def __init__(self, ID, fn, args, seed=None):
		DistributedJob.__init__(self, ID, fn, args, seed)

	def execute(self):
		__import__(self.fn.__module__).random_seed = self.seed
		self.result = self.fn(*self.args)
	
class LocalJobThread(Thread):
	def __init__(self, jq):
		self.jq = jq
		self.q = Queue()

		Thread.__init__(self)
		self.setDaemon(True)
	
	def run(self):
		while True:
			job = self.q.get()
			job.execute()

			while not self.jq.mutex.testandset(): sleep(1)

			for dependent in job.dependents:
				self.jq.pending.push( (dependent.id, dependent) )
				if not isinstance(dependent, LocalJob):
					self.jq.n_dependent_remote_jobs -= 1

			self.jq.mutex.unlock()

class DistributedFunctor:
	def __init__(self, jq, fn, local):
		self.fn = fn
		self.jq = jq
		self.local = local
	def __call__(self, *args, **kw):
		return self.jq.distribute(self.fn, self.local, *args, **kw)
 
class JobQueue:
	def __init__(self):
		self.pending = heap.Heap()
		self.waiting = {}
		self.wu_job_map = {}
		self.mutex = mutex.mutex()
		self.free_id = 0
		self.flush_queue = False
		self.n_dependent_remote_jobs = 0

		self.lj_thread = LocalJobThread(self)
		self.lj_thread.start() 

	def flush(self, flush_val=True):
		self.flush_queue = flush_val

	def is_completely_empty(self):
		return (self.flush_queue and not self.pending and self.n_dependent_remote_jobs == 0)
 
	def distribute(self, fn, local, *args, **kw):
		while not self.mutex.testandset():
			sleep(1)
		
		if not local:
			job = DistributedJob(self.free_id, fn, args)
		else:
			job = LocalJob(self.free_id, fn, args)

		self.pending.push( (self.free_id, job) )
 
		self.free_id += 1
		self.mutex.unlock()
		return delay(job.get_result)

	def distribute_function(self, func, local=False):
		return DistributedFunctor(self, func, local)		
 
	def distribute_module(self, module, local=False):
		for key, value in module.__dict__.items():
			if callable(value):
				module.__dict__[key] = DistributedFunctor(self, value, local)
		return module
 
	def create_WU(self, wu_desc, n_jobs, max_work, filename, delay=5):
		# FIXME: Get rid of the ugly n_jobs hack in the next version of the core
		archive = CHOMP.BinaryStringOutputArchive()

                # Save the number of jobs in the archive
		# Can't actually do this now because the true value of n_jobs is not known in advance!
                # CHOMP.SaveToArchive(archive, '', repr(n_jobs))
		
		# Save the total amount of work to be done (in standardized seconds)
		CHOMP.SaveToArchive(archive, '', repr(max_work))
 
		bundled_jobs = 0
		job_list = []

		while bundled_jobs < n_jobs:
			if not self.pending:
				if self.flush_queue and self.n_dependent_remote_jobs == 0:
					# if we're completely out of jobs, send out the last work unit
					if bundled_jobs == 0:
						return False
					else:
						print '-*-JobQueue-*- Sent final WU -- JobQueue is now empty.'
						break
				else:
					# otherwise, we can wait for more jobs to come in
					sleep(delay)
					continue

			while not self.mutex.testandset():
				sleep(1)

			id, job = self.pending.popmin()
		
			try:
				job.force_args()
			
			except ResultNotAvailable:
				if not isinstance(job, LocalJob):
					self.n_dependent_remote_jobs += 1
			else:
				if isinstance(job, LocalJob):
					self.lj_thread.q.put(job)
				else: # a remote job
					job.write_to_archive(archive)
					self.waiting[id] = job
					bundled_jobs += 1
					job_list.append(id)
		
			self.mutex.unlock()
		
		self.wu_job_map[wu_desc] = job_list

		# splice in the number of jobs to the start of the archive... ugh
		hdr_len = len(CHOMP.BinaryStringOutputArchive().GetString())
		njobs_arc = CHOMP.BinaryStringOutputArchive()
		CHOMP.SaveToArchive(njobs_arc, '', repr(bundled_jobs))
		f = open(filename, 'w')
		f.write(njobs_arc.GetString())
		f.write(archive.GetString()[hdr_len:])

		return True
 
	def wu_has_returned(self, filename, wu_desc):
		wu_file = open(filename)
		job_list = self.wu_job_map[wu_desc]
		
		total_work = 0.0
		n_jobs = 0
	
		while True:
			try:
				ints = array.array('i')
				floats = array.array('f')
				ints.read(wu_file, 3)
				floats.read(wu_file, 1)

				data_length, hash, id = ints[0], ints[1], ints[2]
				work = floats[0]
				
				result_data = wu_file.read(data_length)
				archive = CHOMP.BinaryStringInputArchive(result_data)	

				# test for an error/traceback in the current job
				if work < 0:
					error_data = Deserialize(archive)
					print error_data
					continue

				job = self.waiting[id]
				del self.waiting[id]
				job._wu_has_returned(archive)

				while not self.mutex.testandset():
					sleep(1)
			 
				for dependent in job.dependents:
					self.pending.push( (dependent.id, dependent) )
					if not isinstance(dependent, LocalJob):
						self.n_dependent_remote_jobs -= 1

				self.mutex.unlock()
 
				job.dependents = []
				job_list.remove(id)

				total_work += work
				n_jobs += 1
				
			except EOFError:
				break
			except ValueError:
				break
			except RuntimeError:
				break

		print '-*-JobQueue-*-', n_jobs, 'jobs;', total_work, 'standardized seconds.'

		### TODO: put in some error logging if total_work turns out to be too small ###

		while not self.mutex.testandset():
			sleep(1)
		
		for id in job_list:
			#print 'job', id, 'not done; re-enqueuing'
			self.pending.push( (id, self.waiting[id]) )
			del self.waiting[id]

		self.mutex.unlock()
	
		del self.wu_job_map[wu_desc]



