# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 

import heap, os, random, time
import CHOMP, MySQLdb, logging

from os import path
from Queue import Queue
from threading import Thread, currentThread, Semaphore
from Serialize import *
from functools import update_wrapper

## Job States ##
MISSING = 0
PENDING = 1
WAITING = 2
AVAILABLE = 3
DISCARDED = 4
## End Job States ##

## nocache decorator ##
def nocache(f):
	f.nocache = True
	return update_wrapper(f,f)

## aggregate decorator ##
def aggregate_values(*args):
	def aux(f):
		f.aggregate_values = args
		return update_wrapper(f, f)
	return aux

#### Utility classes ####

class JobResultReference:
	def __init__(self, id):
		self.id = id

class DistributeFunctor:
	def __init__(self, jq, fn, local, project):
		self.fn = fn
		self.jq = jq
		self.local = local
		self.project = project

	def __call__(self, *args, **kw):
		if not kw.has_key('shn_jq_kw_project'):
			kw['shn_jq_kw_project'] = self.project

		return self.jq.run_job(self.fn, self.local, *args, **kw)

class DatabaseConnection:
	def __init__(self, hostname='localhost', dbname='jobqueue_runtime'):
		self.hostname = hostname
		self.dbname = dbname
		self.connections = {}
		self.queries = {}

	def create_connection(self):
		db = MySQLdb.connect(host=self.hostname, db=self.dbname)
		c = db.cursor()
		self.connections[ currentThread().getName() ] = c
		self.queries[ currentThread().getName() ] = []

	def execi(self, query, *args):
		# executes a query using a connection local to the current thread
		#self.c.execute('COMMIT;')
		self.c.execute(query, *args)
		return self.c.fetchall()

	def execd(self, query, *args, **kw):
		# instead of running immediately, this function caches a query until
		# db.commit() is called. This is useful because we want to send updates to
		# the database in short bursts to avoid transaction timout errors.

		try: many = kw['many']
		except: many = False

		self.q.append((query, args, many))

	def commit(self):
		self.c.execute('START TRANSACTION;')
		for query, args, many in self.q:
			if not many:
				self.c.execute(query, *args)
			else:
				self.c.executemany(query, *args)
		
		self.c.execute('COMMIT;')
		self.queries[ currentThread().getName() ] = []

	def get_connection(self):
		return self.connections[ currentThread().getName() ]

	def get_querylist(self):
		return self.queries[ currentThread().getName() ]

	c = property(get_connection)
	q = property(get_querylist)


#### Job classes ####

class JobQueueJob:
	def __init__(self, ID, session_ID, fn, args, kw_args, seed=None, project=None):
		self.id = ID
		self.session_id = session_ID
		self.project = project
		self.fn = fn
		self.args = list(args)
		self.kw_args = kw_args
		self.seed = seed

		try: 
			self.nocache = fn.nocache
		except: 
			self.nocache = False
	
		try:
			self.aggregate_args = []
			self.aggregate_kw_args = []
			for varname in fn.aggregate_values:
				varindex = fn.code.co_varnames.index(varname)
				if varindex < len(self.args):
					self.aggregate_args.append(varindex)
				else:
					self.aggregate_kw_args.append(varname)
		except:
			pass
	
		if self.seed is None:
			self.seed = random.randint(0, 2**31-1)

	def resolve_args(self, jq, immediate_id=None, immediate_result=None):
		unresolved = []
		for i in xrange(len(self.args)):
			arg = self.args[i]
			aggregate = i in self.aggregate_args
			resolved, result = self._resolve_arg(jq, arg, immediate_id, immediate_result, aggregate)
			
			if not resolved:
				unresolved.append(arg.id)
			else:
				self.args[i] = result
		
		for key, arg in self.kw_args.items():
			aggregate = key in self.aggregate_kw_args
			resolved, result = self._resolve_arg(jq, arg, immediate_id, immediate_result, aggregate)

			if not resolved:
				unresolved.append(arg.id)
			else:
				self.kw_args[key] = result

		return unresolved

	def _resolve_arg(self, jq, arg, immediate_id, immediate_result, aggregate=False):
		if isinstance(arg, JobResultReference):
			# if we've been passed an immediate result,
			# we can bypass the database for that job
			if arg.id == immediate_id:
				return (True, immediate_result)

			# check to see if there's an entry in the results
			# hash, which caches results in memory temporarily
			try:
				return (True, jq.results[arg.id])
			except: pass
			
			# if all else fails, we'll have to look in the DB	
			arg_status = jq.get_job_status(arg.id)
			assert(not arg_status in (MISSING, DISCARDED))
			
			if arg_status in (PENDING, WAITING):
				return (False, None)

			elif arg_status == AVAILABLE:
				return (True, jq.get_job_results(arg.id))
				## TODO: optimize this to avoid unnecessary parsing/re-serializing
				## when the result is just going to be passed into a remote job

		## drill into lists when aggregate is true
		elif aggregate and ( isinstance(arg, list) or isinstance(arg, tuple) ):
			full_result = []
			for sub_arg in arg:
				resolved, result = self._resolve_arg(jq, sub_arg, immediate_id, immediate_result, True)
				if not resolved:
					return (False, None)
				else:
					full_result.append(result)
			return (True, full_result)
		
		## drill into dicts when aggregate is true
		elif aggregate and isinstance(arg, dict):		
			full_result = {}
			for key, sub_arg in arg.items():
				resolved, result = self._resolve_arg(jq, sub_arg, immediate_id, immediate_result, True)
				if not resolved:
					return (False, None)
				else:
					full_result[key] = result
			return (True, full_result)

		## all other values are returned as-is
		else:
			return (True, arg)



class DistributedJob(JobQueueJob):
	def __init__(self, *args, **kwargs):
		JobQueueJob.__init__(self, *args, **kwargs)
		
		if self.fn:
			# we need to strip off the 'deployed' stuff from the module
			self.fn_mod = '.'.join( self.fn.__module__.split('.')[2:] )
			self.fn_name = self.fn.__name__

	def write_to_archive(self, archive):
		Serialize(archive, (self.id, self.session_id, self.seed))
		Serialize(archive, (self.fn_mod, self.fn_name))
		Serialize(archive, self.args)
		Serialize(archive, self.kw_args)
	
	@classmethod
	def load_from_archive(cls, archive):
		(id, sid, seed) = Deserialize(archive)
		(fnmod, fnname) = Deserialize(archive)
		args = Deserialize(archive)
		kw_args = Deserialize(archive)

		job = DistributedJob(id, sid, None, args, kw_args, seed=seed)
		job.fn_mod = fnmod
		job.fn_name = fnname

		return job
 
class LocalJob(JobQueueJob):
	def execute(self):
		__import__(self.fn.__module__).random_seed = self.seed
		return self.fn(*self.args, **self.kw_args)


#### JobQueue ####
### WARNING: Not safe to use across threads! ###

class JobQueue:
	def __init__(self, db, session_id=0, max_jobs_in_memory=-1, max_results_in_memory=-1):
		self.session_id = session_id
		self.db = db

		self.remote_queue = heap.Heap()
		self.local_queue = heap.Heap()
		self.dependent = {}
		self.results = {}

		self.queue_sem = Semaphore()
		self.free_id = 0
		
		self.flush_queue = False
		self.n_jobs_in_memory = 0
		self.max_jobs_in_memory = max_jobs_in_memory
		self.max_results_in_memory = max_results_in_memory
	
		currentThread().setName('JobQueue_Main_Thread')
		self.db.create_connection()
		self._init_dbase()
		
		logging.info('JobQueue initialized for session %d.' % self.session_id)

	### Synchronization Utilities ###

	def lock_queue(self):
		self.queue_sem.acquire()

	def unlock_queue(self):
		self.queue_sem.release()

	
	### Database Initialization ###
	
	def _init_dbase(self):
		self._create_jobs_table()	

	def _create_jobs_table(self):
		# first check to see if the table for this session is already present
		results = self.db.execi('SHOW TABLES;')
		for result in results:
			if result[0] == ( 'jobs_s%d' % self.session_id ):
				logging.info('JobQueue Resuming from database for session %d.' % self.session_id)
				return
	
		logging.info('JobQueue Starting new database for session %d.' % self.session_id)
	
		# if the table isn't present, we need to create it.
		self.db.execi("""CREATE TABLE jobs_s%d 
			( 	ID			BIGINT UNSIGNED,
				Status 			TINYINT UNSIGNED,
				Return_Count		TINYINT UNSIGNED,
				Returns_Required	TINYINT UNSIGNED,
				Dependents		LONGBLOB,
				Result			LONGBLOB,
				Label			VARCHAR(100),
				KEY		index_ID (ID)
			) ENGINE = INNODB;""" % self.session_id)

	### Database Management ###
	
	def reset_session(self):
		logging.warning('Clearing out database for JQ session %d. (User Request)' % self.session_id)
		self.db.execi('DROP TABLE jobs_s%d;' % self.session_id )
		self._init_dbase()

	def get_job_data(self, id, field):
		return self.db.execi('SELECT %s FROM jobs_s%d WHERE ID = %d;' % (field, self.session_id, id))

	def set_job_data(self, id, field, data):
		self.db.execd('UPDATE jobs_s%d SET %s = %%s WHERE ID = %d;' % 
				(self.session_id, field, id), data)

	def insert_job(self, id, name=' ', returns_required=1):
		self.db.execd('INSERT INTO jobs_s%d VALUES (%d, %d, 0, %d, %%s, %%s, %s);' % 
				(self.session_id, id, PENDING, returns_required,name), 
				(pickle.dumps(set()), ""))

	
	### Database Accessors ###

	def get_job_status(self, id):
		data = self.get_job_data(id, 'Status')
		if len(data) == 0:
			return MISSING
		else:
			return data[0][0]

	def set_job_status(self, id, status):
		self.set_job_data(id, 'Status', str(status))

	def get_job_results(self, id):
		str_result = self.get_job_data(id, 'Result')[0][0]
		arc = CHOMP.BinaryStringInputArchive(str_result)
		return Deserialize(arc)

	def set_job_results(self, id, result):
		arc = CHOMP.BinaryStringOutputArchive()
		Serialize(arc, result)

		data = self.get_job_data(id, 'Result')[0][0]
		if data != "" and data != arc.GetString():
			logging.critical('Result mismatch in job %d!' % id)
			## TODO: Do something more helpful for mismatches (e.g. put in 'Mismatches' table)

		self.set_job_data(id, 'Result', arc.GetString())

	def get_job_dependents(self, id):
		str_deps = self.get_job_data(id, 'Dependents')[0][0]
		return pickle.loads(str_deps)

	def set_job_dependents(self, id, deps):
		self.set_job_data(id, 'Dependents', pickle.dumps(deps))

	def add_job_dependent(self, id, dep):
		deps = self.get_job_dependents(id)
		deps.add(dep)
		self.set_job_dependents(id, deps)

	def get_job_return_count(self, id):
		return self.get_job_data(id, 'Return_Count')[0][0]

	def inc_job_return_count(self, id):
		count = self.get_job_return_count(id)
		self.set_job_data(id, 'Return_Count', count+1)

	def get_job_required_returns(self, id):
		return self.get_job_data(id, 'Returns_Required')[0][0]

	### Main interface ###

	def flush(self, flush_val=True):
		logging.info('Master script terminated -- flushing JobQueue.')
		self.flush_queue = flush_val

	def distribute_function(self, func, local=False, project=None):
		return DistributeFunctor(self, func, local, project)		
 
	def distribute_module(self, module, local=False, project=None):
		for key, value in module.__dict__.items():
			if callable(value) and not isinstance(value, DistributeFunctor):
				module.__dict__[key] = DistributeFunctor(self, value, local, project)
		return module

	def run_job(self, fn, local, *args, **kw):
		# grab some optional arguments
		try: 
			project = kw['shn_jq_kw_project']
			del kw['shn_jq_kw_project']
		except: 
			project = None
		
		id = self.free_id
		status = self.get_job_status(id)

		# add new job to the database
		if status == MISSING:
			self.insert_job(id, 1)	
			self.db.commit()

		if status in (MISSING, PENDING):
			# make sure not to load too many jobs into memory
			self.n_jobs_in_memory += 1
			if self.max_jobs_in_memory > 0 and self.n_jobs_in_memory > self.max_jobs_in_memory:
				try:
					while self.n_jobs_in_memory > self.max_jobs_in_memory:
						print 'WARNING: In-memory job limit exceeded. Blocking until more WUs have been sent out.'
						print 'WARNING: Signal a keyboard interrupt (Ctrl-C) to override and continue.' 			
						time.sleep(60)

				except KeyboardInterrupt:
					self.max_jobs_in_memory = -1

			# if we're ok with memory, create the new job
			if not local:
				job = DistributedJob(id, self.session_id, fn, args, kw, project=project)
				logging.debug('New remote job: %d (%s)' % (id, job.fn.__name__) )
				self.set_job_data(id, 'Label', str(job.fn.__name__)) # helpful for debugging 
			else:
				job = LocalJob(id, self.session_id, fn, args, kw, project=project)
				logging.debug('New local job: %d (%s)' % (id, job.fn.__name__) )
				self.set_job_data(id, 'Label', str(job.fn.__name__)) # helpful for debugging
	
			# we need to lock before the call to resolve_args to prevent
			# a race condition with the WU Receiver thread (see notify_job_dependents)	
			self.lock_queue()

			# figure out which jobs this job depends on
			required_jobs = job.resolve_args(self)

			if required_jobs != []:
				# if it depends on something, we don't put it in the queues
				# but in a special hash for later execution
				for req_job in required_jobs:
					self.add_job_dependent(req_job, id)

				self.dependent[id] = job
				self.db.commit()
				logging.debug('Postponing job %d: depends on jobs %s.' % (id, str(required_jobs)))
			else:
				# if the job can execute immediately, we put it in the appropriate queue
				job_priority = id
				if not local:
					self.remote_queue.push( (job_priority, job) )
				else:
					self.local_queue.push( (job_priority, job) )

			self.unlock_queue()

		elif status == AVAILABLE:
			logging.debug('Skipping cached job: %d' % id)
		elif status == DISCARDED:
			logging.debug('Skipping discarded job: %d' % id) 

		self.free_id += 1
		return JobResultReference(id)

	### JobQueueDriver Interface ###

	def pop_job(self, local):
		if local: 
			queue = self.local_queue
		else: 
			queue = self.remote_queue

		self.lock_queue()
		if not queue:
			self.unlock_queue()
			return None
			
		priority, job = queue.popmin()
		self.unlock_queue()
		
		if not job.nocache and self.get_job_status(job.id) == PENDING:
			self.set_job_status(job.id, WAITING)
		
		self.n_jobs_in_memory -= 1
		logging.debug('Processing job %d.' % job.id)

		return job
	
	def reinsert_job(self, job):
		self.lock_queue()
		self.n_jobs_in_memory += 1

		if isinstance(job, DistributedJob):
			self.remote_queue.push( (job.id, job) )
		elif isinstance(job, LocalJob):
			self.local_queue.push( (job.id, job) )

		self.unlock_queue()
		self.set_job_status(job.id, PENDING)

	def job_returned(self, id, result, nocache=False):
		self.add_result(id, result, nocache)
		
		if nocache:
			self.notify_job_dependents(id, result)
		else:
			# otherwise we'll need to deal with the database
			status = self.get_job_status(id)
			if status == MISSING:
				logging.error('Job %d returned while status was MISSING. Looks like db inconsistency.' % id)

			if status == AVAILABLE:
				self.set_job_results(id, result)
				logging.debug('Job %d verified.' % id)

			elif status in (WAITING, PENDING):
				self.set_job_results(id, result)
				return_count = self.get_job_return_count(id)
				self.inc_job_return_count(id)

				if return_count + 1 >= self.get_job_required_returns(id):
					self.set_job_status(id, AVAILABLE)
					self.db.commit()
					logging.debug('Job %d now AVAILABLE.' % id)

					self.notify_job_dependents(id, result)

	def notify_job_dependents(self, id, result):
		# we need to lock before the call to get_job_dependents to prevent
		# a race condition with the Main JobQueue thread (see run_job)
		self.lock_queue()
		job_dependents = self.get_job_dependents(id)
		self.unlock_queue()

		for dep_id in job_dependents:
			dep_status = self.get_job_status(dep_id)
			if dep_status == PENDING:
				while not self.dependent.has_key(dep_id):
					time.sleep(1)
				dep = self.dependent[dep_id]

				required_jobs = dep.resolve_args(self, immediate_id=id, immediate_result=result)
				if required_jobs == []:
					logging.debug('Job %d is now able to run.' % dep_id)

					del self.dependent[dep.id]
					self.n_jobs_in_memory -= 1
					self.reinsert_job(dep)

	def add_result(self, id, result, nocache):
		if self.max_results_in_memory >= 0 and len(self.results) > self.max_results_in_memory and not nocache:
			logging.warning('Maximum memory-cached results reached. Falling back to database.')
			return
		
		self.results[id] = result


