// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "amino_acid.h"

#include <sstream>

/** \file aa_properties.cpp
 * \see amino_acid.h
 */

///////////////////////////////////////////////////////////////////////////
//////                         DATA LOADERS                          //////
///////////////////////////////////////////////////////////////////////////

Vector3 ReadVector3(std::istringstream& istream)
{
	FLOAT x, y, z;
	istream >> x >> y >> z;
	return Vector3(x, y, z);
}

std::string ReadString(std::istringstream& istream) 
{
	int n;
	istream >> n;

	// discard everything up to and including the '!' separator
	while (istream.get() != '!');
	
	// now read in the actual string
	char* str = new char[n];
	for (int i = 0; i < n; ++i)
		str[i] = istream.get();

	// transfer the data over to a std::string	
	std::string sstr(str, n);
	delete[] str;
	
	return sstr;
}

#include "codegen/output/aa_props_data.hxx"

///////////////////////////////////////////////////////////////////////////
//////                           UTILITIES                           //////
///////////////////////////////////////////////////////////////////////////

template <typename Type>
void Init2D(std::vector< std::vector<Type> >& array, int n, int m)
{
	for (int i = 0; i < n; ++i)
		array.push_back( std::vector<Type>(m) );
}

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

AminoAcid::AminoAcid()
{
	// initialize the 2-D arrays
	Init2D(acceptorBase, 	MAX_ACCEPTORS, 		2);
	Init2D(chiAtoms, 	MAX_DIHEDRAL_ANGLES, 	N_CHI_ATOMS);
	Init2D(chiDependent,	MAX_RESIDUE_ATOMS,	MAX_DIHEDRAL_ANGLES);
	Init2D(templateAtom,	MAX_RESIDUE_ATOMS, 	N_TEMPLATE_ATOMS);
	Init2D(atomOrders,	N_ATOM_ORDERINGS, 	MAX_RESIDUE_ATOMS);
	Init2D(bondLengths, MAX_RESIDUE_ATOMS,	MAX_RESIDUE_ATOMS);
	
	// initialize topology graph
	topoGraph = BondGraph();
	tg_IndexMap = boost::get(boost::vertex_index, topoGraph);

	acceptorList.resize(MAX_ACCEPTORS);
	templateCoords.resize(MAX_RESIDUE_ATOMS);
	atomType.resize(MAX_RESIDUE_ATOMS);
	atomFlags.resize(MAX_RESIDUE_ATOMS);
	centroidAtoms.resize(MAX_CENTROID_ATOMS);
}

AminoAcidDirectory::Pointer AminoAcidDirectory::Create(AtomTypes::ConstPointer atypes)
{
	return AminoAcidDirectory::Pointer(new AminoAcidDirectory(atypes));
}

AminoAcidDirectory::AminoAcidDirectory(AtomTypes::ConstPointer atomTypes)
{
	// data_str is declared in aa_props_data.hxx
	std::istringstream istream(data_str);

	// load the data for each amino acid
	#include "codegen/output/aa_props_init.hxx"
	
	const int nativeOrder = CHOMP_ATOM_ORDER;

	for (int aai = 0; aai < N_AMINO_ACIDS; ++aai)
	{
		AminoAcid& aa = data_[aai];
		
		for (int orderingScheme = 0; orderingScheme < N_ATOM_ORDERINGS; ++orderingScheme)
			for (int atom = 0; atom < aa.nTotalAtoms; ++atom)
			{
				// initialize the indexOf maps for atom order
				aa.indexOf_[orderingScheme][ aa.atomOrders[orderingScheme][atom] ] = atom;
			}

		for (int atom = 0; atom < aa.nTotalAtoms; ++atom)
		{	
			std::string const& atomName = aa.atomOrders[nativeOrder][atom];

			// initialize atom type info
			aa.atomType[atom] = atomTypes->ToType(aa.name3Char + ":" + atomName, AtomType::RAW_TYPE);
				
			// initialize atom flags
			aa.atomFlags[atom] = GetAtomFlags(aa, atom, atomName, atomTypes);

			// add atoms as vertices of topology graph
			boost::add_vertex(aa.topoGraph);
		}

		// add edges between bonded atoms
		for (int atom = 0; atom < aa.nTotalAtoms; ++atom)
			for (int atom2 = atom+1; atom2 < aa.nTotalAtoms; atom2++)
				if (aa.bondLengths[atom][atom2] == 1)
					boost::add_edge(atom, atom2, aa.topoGraph);	
	}

}

///////////////////////////////////////////////////////////////////////////
//////                       UTILITY FUNCTIONS                       //////
///////////////////////////////////////////////////////////////////////////

int AminoAcidDirectory::GetAtomFlags(AminoAcid const& aa, int atom, 
				     std::string const& atomName, 
				     AtomTypes::ConstPointer atomTypes)
{
	static const std::string backboneAtoms[] = { "C", "O", "N", "H", "CA", "HA", "1HA", "2HA", "OXT" };
	int flags = 0;

	// flag the hydrogen bond acceptors
	for (int i = 0; i < aa.nHBAcceptors; ++i)
		if (atom == aa.acceptorList[i])
			flags |= ATM_FLAG_HB_ACCEPTOR;

	// set the side chain/backbone flag
	for (int a = 0; a < 9; ++a)
		if (atomName == backboneAtoms[a])
			flags |= ATM_FLAG_BB;

	// flag the hydrogens and polar hydrogens
	if (atomName[0] == 'H' || 
	    atomName.length() > 1 && atomName[1] == 'H')
	{
		flags |= ATM_FLAG_H;

		if (atomTypes->GetType(aa.atomType[atom], AtomType::ROSETTA_TYPE) == AtomType::Rosetta::POL_H ||
		    atomTypes->GetType(aa.atomType[atom], AtomType::ROSETTA_TYPE) == AtomType::Rosetta::BB_HN)
		{
			flags |= ATM_FLAG_POL_H;
		}
	}
			
	return flags;
}

///////////////////////////////////////////////////////////////////////////
//////                           ACCESSORS                           //////
///////////////////////////////////////////////////////////////////////////

AminoAcid const* AminoAcidDirectory::Get( AminoAcidIndex index ) const
{
	return &data_[index];
}

AminoAcid const* AminoAcidDirectory::Get(std::string const& aaName) const
{
	std::map<std::string, AminoAcid*>::const_iterator aai = byName_.find(aaName);

	if (aai == byName_.end())
		throw std::string("AminoAcidDirectory.Get: Amino acid " + aaName + "not found.");
	
	return aai->second;
}

///////////////////////////////////////////////////////////////////////////
//////                           AminoAcid                           //////
///////////////////////////////////////////////////////////////////////////

int AminoAcid::indexOf(int atomOrdering, std::string const& atom) const
{
	if (atomOrdering < 0 || atomOrdering >= N_ATOM_ORDERINGS)
		throw std::string("AminoAcid.indexOf: Atom ordering " + to_str(atomOrdering) + " not found.");

	std::map<std::string, int>::const_iterator i = indexOf_[atomOrdering].find(atom);

	if (i == indexOf_[atomOrdering].end())
		throw std::string("AminoAcid.indexOf: Atom " + atom + "not found.");
	
	return i->second;
}

std::vector<blitz::TinyVector<int,3> > AminoAcid::getBondAngleAtoms(void) const
{
	std::vector<blitz::TinyVector<int,3> > angle_triplets;
	int angle_atoms[3];
	
	//Iterate through the topology graph
	for (VertexRange i = boost::vertices(topoGraph); i.first != i.second; ++i.first)
	{
		angle_atoms[0] = boost::get(tg_IndexMap, *i.first);
		for (AdjacencyRange j = boost::adjacent_vertices(*i.first, topoGraph); j.first != j.second; ++j.first)
		{
			angle_atoms[1] = boost::get(tg_IndexMap, *j.first);
			for (AdjacencyRange k = boost::adjacent_vertices(*j.first, topoGraph); k.first != k.second; ++k.first)
			{
				angle_atoms[2] = boost::get(tg_IndexMap, *k.first);
				if (angle_atoms[0] != angle_atoms[2])
				{
					blitz::TinyVector<int,3> angle = blitz::TinyVector<int,3>(angle_atoms[0], angle_atoms[1], angle_atoms[2]);
					if (!tupleDone(angle, angle_triplets))
						angle_triplets.push_back(angle);
				}
			}
		}
	}	

	return angle_triplets;
}

std::vector<blitz::TinyVector<int,4> > AminoAcid::getTorsionAtoms(void) const
{
	std::vector<blitz::TinyVector<int,4> > torsion_quadruplets;
	int torsion_atoms[4];
	
	//Iterate through topology graph
	for (VertexRange i = boost::vertices(topoGraph); i.first != i.second; ++i.first)
	{
		torsion_atoms[0] = boost::get(tg_IndexMap, *i.first);
		for (AdjacencyRange j = boost::adjacent_vertices(*i.first, topoGraph); j.first != j.second; ++j.first)
		{
			torsion_atoms[1] = boost::get(tg_IndexMap, *j.first);
			for (AdjacencyRange k = boost::adjacent_vertices(*j.first, topoGraph); k.first != k.second; ++k.first)
			{
				torsion_atoms[2] = boost::get(tg_IndexMap, *k.first);
				if (torsion_atoms[0] != torsion_atoms[2])
				{
					for (AdjacencyRange l = boost::adjacent_vertices(*k.first, topoGraph); l.first != l.second; ++l.first)
					{
						torsion_atoms[3] = boost::get(tg_IndexMap, *l.first);
						if (torsion_atoms[1] != torsion_atoms[3])
						{
							blitz::TinyVector<int,4> torsion = blitz::TinyVector<int,4>(torsion_atoms[0], torsion_atoms[1], torsion_atoms[2], torsion_atoms[3]);
							if (!tupleDone(torsion, torsion_quadruplets))
								torsion_quadruplets.push_back(torsion);
						}
					}
				}
			}
		}
	}
	
	return torsion_quadruplets;
}


std::vector<blitz::TinyVector<int, 4> > AminoAcid::getImptorsAtoms(AtomTypes::ConstPointer atomTypes) const
{
	std::vector<blitz::TinyVector<int,4> > imptors_quadruplets;
	
	//Iterate through topology graph
	for (VertexRange i = boost::vertices(topoGraph); i.first != i.second; ++i.first)
	{
		int atom_index = boost::get(tg_IndexMap, *i.first);
		//improper torsion if central atom is SP2 hybridized (trigonal planar structure) or "RING" hybridized
		if (atomTypes->GetType(atomType[atom_index], AtomType::HYBRIDIZATION) == AtomType::Hybridization::SP2 ||
		    atomTypes->GetType(atomType[atom_index], AtomType::HYBRIDIZATION) == AtomType::Hybridization::RING)
		{
			//intra-residue improper torsion atoms do not include backbone nitrogen and carbon, since they fall under inter-residue improper torsion
//			if (atom_index != indexOf(CHOMP_ATOM_ORDER, "C") && atom_index != indexOf(CHOMP_ATOM_ORDER, "N"))
//			{
				int imptors_atoms[3], k = 0;
				for (AdjacencyRange j = boost::adjacent_vertices(*i.first, topoGraph); j.first != j.second; ++j.first)
				{
					imptors_atoms[k] = boost::get(tg_IndexMap, *j.first);
					k++;
				}
				if (k == 3)
				{
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[0], imptors_atoms[1], atom_index, imptors_atoms[2]));
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[0], imptors_atoms[2], atom_index, imptors_atoms[1]));
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[1], imptors_atoms[0], atom_index, imptors_atoms[2]));
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[1], imptors_atoms[2], atom_index, imptors_atoms[0]));
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[2], imptors_atoms[1], atom_index, imptors_atoms[0]));
					imptors_quadruplets.push_back(blitz::TinyVector<int,4> (imptors_atoms[2], imptors_atoms[0], atom_index, imptors_atoms[1]));
				}
//			}
		}
	}
	
	return imptors_quadruplets;
}

