// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Jason Paryani, Cheng Hong 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "general.h"
#include "system.h"
#include "ccd.h"

#include <blitz/tinyvec-et.h>
#include <cmath>

#define PI 3.14159265
///// FIXME: Move these constants somewhere else
///// Note: these constants match the parameters from Rosetta refold_ns.cc
FLOAT n_a = 1.458;
FLOAT a_c = 1.523;
FLOAT n_c = 1.329;
FLOAT c_o = 1.231;
FLOAT a_c_n = (PI * 116.2/180);
FLOAT c_n_a = PI * 121.7/180;
FLOAT n_c_o = (PI * 123.0/180); // annoying that the Rosetta angles add up to 360.9
FLOAT h_n_c = PI * 120.0/180;   // hydrogen geometry specified elsewhere? 
// precalculate for building virtual NCa 
FLOAT cos_of_pi_minus_a_c_n = cos(PI - a_c_n); 
FLOAT sin_of_pi_minus_a_c_n = sin(PI - a_c_n);
FLOAT cos_of_c_n_a_minus_a_c_n = cos(c_n_a - a_c_n); 
FLOAT sin_of_c_n_a_minus_a_c_n = sin(c_n_a - a_c_n); 
// precalculate for building virtual CO
FLOAT cos_of_pi_minus_h_n_c = cos(PI - h_n_c); 
FLOAT sin_of_pi_minus_h_n_c = sin(PI - h_n_c); 
FLOAT cos_of_n_c_o_minus_h_n_c = cos(n_c_o - h_n_c); 
FLOAT sin_of_n_c_o_minus_h_n_c = sin(n_c_o - h_n_c); 
//float bugcheck = 0;

// FIXME: Major annoyance is that C, O, N, CA do not always have the same indices for all residues, 
// Also, atomH will not work for Prolines 
int atomC;
int atomO;
int atomN;
int atomCA;
int atomH;

using std::vector;
using std::string;

typedef blitz::TinyVector<Vector3, 3> Matrix3;

FLOAT getDistance(Vector3 const& a, Vector3 const& b)
{
  FLOAT x = a[0] - b[0];
  FLOAT y = a[1] - b[1];
  FLOAT z = a[2] - b[2];
  return sqrt(x * x + y * y + z * z);
}

FLOAT getLength(Vector3 const& T)
{
  return sqrt(T[0] * T[0] + T[1] * T[1] + T[2] * T[2]);
}

Vector3 normalize(Vector3 const& T)
{
  Vector3 temp = Vector3(T);
  FLOAT ilen = 1.0 / getLength(temp);
  temp[0] *= ilen;
  temp[1] *= ilen;
  temp[2] *= ilen;

  return temp;
}

Vector3 dot(Matrix3 const& a, Vector3 const& b)
{
  Vector3 result;
  result[0] = a[0][0] * b[0] + a[1][0] * b[1] + a[2][0] * b[2];
  result[1] = a[0][1] * b[0] + a[1][1] * b[1] + a[2][1] * b[2];
  result[2] = a[0][2] * b[0] + a[1][2] * b[1] + a[2][2] * b[2];
  return result;
}

void updateatom(Residue& res, int atom, Matrix3 const& rotmatrix, Vector3 const& launchpad)
{
  Vector3 newcoord = dot(rotmatrix, res.GetAtomCoords()[atom] - launchpad) + launchpad;
  res.SetAtomCoords(atom, newcoord);
}

vector<Vector3> idealnextNCa(vector<Vector3> const& residue)
{
  //printf("Computing coordinates for N,Ca:\n");
  Vector3 ntermC = residue[atomC];
  Vector3 ntermO = residue[atomO];
  Vector3 ntermCA = residue[atomCA];
  //printf("Cx %.3f Ox %.3f CAx %.3f\n",ntermC[0],ntermO[0],ntermCA[0]);
  Vector3 xc_xa = ntermC - ntermCA;
  Vector3 xo_xc = ntermO - ntermC;
  Vector3 along_c_a = normalize(xc_xa);
  Vector3 orthogonal = normalize(blitz::cross(xc_xa, blitz::cross(xc_xa, xo_xc)));
  FLOAT part1 = n_c * cos_of_pi_minus_a_c_n;
  FLOAT part2 = n_c * sin_of_pi_minus_a_c_n;
  FLOAT part3 = n_a * cos_of_c_n_a_minus_a_c_n;
  FLOAT part4 = n_a * sin_of_c_n_a_minus_a_c_n;
  Vector3 newN = ntermC + part1 * along_c_a + part2 * orthogonal;
  Vector3 newCa = newN + part3 * along_c_a + part4 * orthogonal;
  vector<Vector3> temp;
  temp.push_back(newN);
  temp.push_back(newCa);
  return temp;
}


vector<Vector3> idealpreviousCO(vector<Vector3> const& residue)
{
  //printf("Computing coordinates for CO:\n");
  Vector3 ctermN = residue[atomN];
  Vector3 ctermH = residue[atomH];
  Vector3 ctermCA = residue[atomCA];
  Vector3 ctermC = residue[atomC];
  //printf("Nx %.3f Hx %.3f CAx %.3f Cx %.3f\n",ctermN[0],ctermH[0],ctermCA[0],ctermC[0]);
  Vector3 xn_xh = ctermN - ctermH;
  Vector3 xn_xa = ctermN - ctermCA;
  Vector3 along_n_h = normalize(xn_xh);
  Vector3 ortho1 = normalize(blitz::cross(xn_xh, xn_xa));
  Vector3 ortho2 = normalize(blitz::cross(xn_xh, ortho1));
  FLOAT part1 = n_c * cos_of_pi_minus_h_n_c;
  FLOAT part2 = n_c * sin_of_pi_minus_h_n_c;
  FLOAT part3 = c_o * cos_of_n_c_o_minus_h_n_c;
  FLOAT part4 = c_o * sin_of_n_c_o_minus_h_n_c;
  Vector3 newC = ctermN + part1 * along_n_h - part2 * ortho2;
  Vector3 newO = newC + part3 * along_n_h - part4 * ortho2;
  vector<Vector3> temp;
  temp.push_back(newC);
  temp.push_back(newO);
  return temp;
}


Vector3 getaxis(vector<Vector3> const& residue, string const& type)
{
  if(type == "phi")
  {
    Vector3 tmpCA = residue[atomCA];
    Vector3 tmpN = residue[atomN];
    Vector3 tmpvec = tmpCA - tmpN;
    return normalize(tmpvec);
  }
  else
  {
    Vector3 tmpCA = residue[atomCA];
    Vector3 tmpC = residue[atomC];
    Vector3 tmpvec = tmpC - tmpCA;
    return normalize(tmpvec);
  }
}

Vector3 getlaunchatom(vector<Vector3> const& residue, string const& type)
{
  if(type == "phi")
     return residue[atomCA];
  else
     return residue[atomC];
}

void rmatrixu(Vector3 const& axis, FLOAT angle, Matrix3* result)
{
 FLOAT cosa  = cos(angle);
 FLOAT sina  = sin(angle);
	
 (*result)[0][0] = cosa + (1 - cosa) * axis[0] * axis[0];
 (*result)[0][1] = (1 - cosa) * axis[1] *axis[0] + sina * axis[2];
 (*result)[0][2] = (1 - cosa) * axis[2] * axis[0] - sina * axis[1];
 (*result)[1][0] = (1 - cosa) * axis[1] * axis[0] - sina * axis[2];
 (*result)[1][1] = cosa + (1 - cosa) * axis[1] * axis[1];
 (*result)[1][2] = (1 - cosa) * axis[1] *axis[2] + sina * axis[0];
 (*result)[2][0] = (1 - cosa) * axis[2] * axis[0] + sina * axis[1];
 (*result)[2][1] = (1 - cosa) * axis[1] *axis[2] - sina * axis[0];
 (*result)[2][2] = cosa + (1 - cosa) * axis[2] * axis[2];
}


FLOAT SSD(Vector3 const& a1, Vector3 const& a2, Vector3 const& a3, Vector3 const& a4, 
          Vector3 const& b1, Vector3 const& b2, Vector3 const& b3, Vector3 const& b4)
{
  FLOAT tmpdist = 0;
  FLOAT tmpd = getDistance(a1, b1);
  tmpdist += tmpd * tmpd;
  tmpd = getDistance(a2, b2);
  tmpdist += tmpd * tmpd;
  tmpd = getDistance(a3, b3);
  tmpdist += tmpd * tmpd;
  tmpd = getDistance(a4, b4);
  tmpdist += tmpd * tmpd;
  return tmpdist;
}

FLOAT SSDverbose(Vector3 const& a1, Vector3 const& a2, Vector3 const& a3, 
          Vector3 const& b1, Vector3 const& b2, Vector3 const& b3)
{
  FLOAT tmpdist = 0;
  FLOAT tmpd = getDistance(a1, b1);
  printf("SSD: (%f,%f,%f) and (%f,%f,%f) = %f\n", a1[0],a1[1],a1[2],b1[0],b1[1],b1[2],tmpd); 
  tmpdist += tmpd * tmpd;
  tmpd = getDistance(a2, b2);
  printf("SSD: (%f,%f,%f) and (%f,%f,%f) = %f\n", a2[0],a2[1],a2[2],b2[0],b2[1],b2[2],tmpd); 
  tmpdist += tmpd * tmpd;
  tmpd = getDistance(a3, b3);
  printf("SSD: (%f,%f,%f) and (%f,%f,%f) = %f\n", a3[0],a3[1],a3[2],b3[0],b3[1],b3[2],tmpd); 
  tmpdist += tmpd * tmpd;
  return tmpdist;
}

// Not sure if struct is passed by value or reference
System::Pointer RebuildLoopsCCD(System::ConstPointer input, vector<int> const& chainbreaks, CCDParameters const& parameters)
{
	System::Pointer system = input->Clone();
	// Values that used be global, load values passed
	int NUMCCDROUNDS = parameters.NUMCCDROUNDS;
	int dofweightedflag = parameters.dofweightedflag;
	int fray = parameters.fray;
	int dt = parameters.dt;
	FLOAT convergedalpha = parameters.convergedalpha;	
	FLOAT SSDconverge = parameters.SSDconverge;
	FLOAT currentSSD = 0;

      	foreach (int nextbreak, chainbreaks)
	{
		vector<int> Ndofresi;
		vector<string> Ndoftype;
		vector<int> Cdofresi;
		vector<string> Cdoftype;
		vector<int> dofresi;
		vector<string> doftype;
		vector<FLOAT> dofweight;
		for(int dof = 0; dof < fray * 2; dof++)
		  {
		    Ndofresi.push_back(nextbreak - dof/2 - 1);
		    Cdofresi.push_back(nextbreak + int(dof/2));

		    ////// To reduce move size for "base" degrees of freedom far from the break 
		    ////// use weights to take smaller steps far from the break

		    dofweight.push_back(1 - FLOAT(dof)/int(fray*2));

		    //////   Because this variant of CCD alternates between updating DOF on either side
		    //////   of the chain break we need two stacks of DOF types 

		    if (dof%2==0)
		      {
			Ndoftype.push_back("psi");
			Cdoftype.push_back("phi");
		      }
		    else
		      {
			Ndoftype.push_back("phi");
			Cdoftype.push_back("psi");
		      }

		  }

		//////   Declare pointers to the important positions	
		Vector3 mobileC;
		Vector3 mobileO;
		Vector3 mobileN;
		Vector3 mobileCa;
		Vector3 anchorC;
		Vector3 anchorO;
		Vector3 anchorN;
		Vector3 anchorCa;
		vector<Vector3> virtualCO;
		vector<Vector3> virtualNCa;
		Residue& res1 = system->GetResidue(nextbreak - 1);
		Residue& res2 = system->GetResidue(nextbreak);
		vector<Vector3> const& atoms1 = res1.GetAtomCoords();
		vector<Vector3> const& atoms2 = res2.GetAtomCoords();


		//////   Each round of CCD consists of updates to either the N-terminal or C-terminal branch	

		for(int CCD = 0; CCD < NUMCCDROUNDS; CCD++)
		  {
		    //////  printf("Entered CCD loop, round %d.\n", CCD);



		    /////////////////////////////////////////////////////
		    //////  STAGE 1. Set mobile and anchor atoms  ///////
		    /////////////////////////////////////////////////////

		    //////  For even rounds, move the high branch (C-terminal)	
		    if(CCD % 2 == 0)
		      {

			////// LOW BRANCH ANCHOR: fixed target atoms (C,O,N,Ca) have lower residue numbers
			////// The C and O are real, but we must invent a position for the N and Ca

			atomC  =  res1.AtomIndex("C");
			atomO  =  res1.AtomIndex("O");
			atomCA =  res1.AtomIndex("CA");
			anchorC = atoms1[atomC];
			anchorO = atoms1[atomO];
			virtualNCa = idealnextNCa(atoms1);
			anchorN = virtualNCa[0];
			anchorCa = virtualNCa[1];

			////// HIGH BRANCH MOBILE: mobile atoms (C,O,N,Ca) have higher residue numbers 
			////// The C and O are virtual while the N and Ca are real
			atomN  = res2.AtomIndex("N");
			if (res2.GetAminoAcid().name3Char == "PRO") {
				//printf("Using Proline CD in place of missing H\n");
				atomH = res2.AtomIndex("CD");
			}
			else {
				atomH =  res2.AtomIndex("H");
			}
			atomCA =  res2.AtomIndex("CA");
			atomC  = res2.AtomIndex("C");
			virtualCO = idealpreviousCO(atoms2);
			mobileC = virtualCO[0];
			mobileO = virtualCO[1];
			mobileN = atoms2[atomN];
			mobileCa = atoms2[atomCA];

			//printf("Moving the high branch toward the low branch\n");
			////// For the rest of the branch update use Cdofresi stack of phis,psis
			dofresi = Cdofresi;
			doftype = Cdoftype;
		      }
		    //////  For odd rounds, move the low branch (N-terminal) 
		    else
		      {
			////// LOW BRANCH MOBILE: mobile atoms (C,O,N,Ca) have lower residue numbers 
			////// The C and O are real, but we must invent a position for the N and Ca
			atomC  =  res1.AtomIndex("C");
			atomO  =  res1.AtomIndex("O");
			atomCA =  res1.AtomIndex("CA");
			mobileC = atoms1[atomC];
			mobileO = atoms1[atomO];
			virtualNCa = idealnextNCa(atoms1);
			mobileN = virtualNCa[0];
			mobileCa = virtualNCa[1];

			////// HIGH BRANCH ANCHOR: fixed target atoms (C,O,N,Ca) have higher residue numbers 
			////// The C and O are virtual while the N and Ca are real
			atomN  = res2.AtomIndex("N");
			if (res2.GetAminoAcid().name3Char == "PRO") {
				//printf("Using Proline CD in place of missing H\n");
				atomH = res2.AtomIndex("CD");
			}
			else {
				atomH =  res2.AtomIndex("H");
			}
			atomCA =  res2.AtomIndex("CA");
			atomC  = res2.AtomIndex("C");
			virtualCO = idealpreviousCO(atoms2);
			anchorC = virtualCO[0];
			anchorO = virtualCO[1];
			anchorN = atoms2[atomN];
			anchorCa = atoms2[atomCA];

			//printf("Moving the low branch toward the high branch\n");
			////// For the rest of the branch update use Ndofresi stack of phis,psis
			dofresi = Ndofresi;
			doftype = Ndoftype;

		      }

		      if (0) {
                        printf("mobileC %f %f %f\n",mobileC[0],mobileC[1],mobileC[2]);
                        printf("mobileO %f %f %f\n",mobileO[0],mobileO[1],mobileO[2]);
                        printf("mobileN %f %f %f\n",mobileN[0],mobileN[1],mobileN[2]);
                        printf("mobileCa %f %f %f\n",mobileCa[0],mobileCa[1],mobileCa[2]);
                        printf("anchorC %f %f %f\n",anchorC[0],anchorC[1],anchorC[2]);
                        printf("anchorO %f %f %f\n",anchorO[0],anchorO[1],anchorO[2]);
                        printf("anchorN %f %f %f\n",anchorN[0],anchorN[1],anchorN[2]);
                        printf("anchorCa %f %f %f\n",anchorCa[0],anchorCa[1],anchorCa[2]);
		      }



		    ////////////////////////////////////////////////////////////////////////////////
		    //////  STAGE 2. Walk through degrees of freedom and update coordinates  ///////
		    ////////////////////////////////////////////////////////////////////////////////

		    for(int dof = 0; dof < int(fray * 2); dof++)
		      {
			//////   Current residue - note that this may go up or down w.r.t. the dof index
			//////   In other words, we walk through a list of degrees of freedom, but that
			//////   might be going in either direction when it comes to the residue number.
			//////   "type" determines if we are currently working on a phi or psi update

			Residue& resi = system->GetResidue(dofresi[dof]);
			string type = doftype[dof];
			//std::cout <<"Working on "<<resi.GetAminoAcid().name3Char<<resi.GetID()<<" "<<type<<std::endl;


		        ////////////////////////////////////////////////////////////////////
		        //////  STAGE 2A. Set the atom indices for this residue  ///////////
		        ////////////////////////////////////////////////////////////////////

			if (resi.GetAminoAcid().name3Char == "PRO") {
			    if (type == "phi") { 
			        //std::cout << "Skipping PRO phi for "<< resi.GetID() << std::endl;
			        continue;
			    }
                            //printf("Using Proline CD in place of missing H\n");
                            atomH = resi.AtomIndex("CD");
			}
			else {
                            atomH =  resi.AtomIndex("H");
			}
			atomC  =  resi.AtomIndex("C");
			atomO  =  resi.AtomIndex("O");
			atomCA =  resi.AtomIndex("CA");
			atomN  =  resi.AtomIndex("N");


		        /////////////////////////////////////////////////////////////////////////////
		        //////  STAGE 2B. Get the "weight" associated with this rotation  ///////////
		        /////////////////////////////////////////////////////////////////////////////

			//////   If dofweightedflag is set, use variable step weights for the different DOF
			//////   Far from the break, with a big lever arm, you might want to make smaller steps
			FLOAT dofW;
			if(dofweightedflag == 1)
			  dofW = dofweight[dof];
			else
			  dofW = 1;



		        ////////////////////////////////////////////////////////////////////////
		        //////  STAGE 2C. Calculate the center and axis of rotation  ///////////
		        ////////////////////////////////////////////////////////////////////////

			//////   For the current residue and torsion type lookup the axis of rotation 
			//////   If we are working on a phi, the axis is normalized (Ca - N)->, launch from Ca
			//////   If we are working on a psi, the axis is normalized (C - Ca)->, launch from C
			//////   The launch atom is the last atom that will not change position when changing the DOF
			//////   For numerical reasons we actually rebuild coordinates relative to a virtual location
			//////   0.5A backwards along the rotation axis 

			Vector3 axis = getaxis(resi.GetAtomCoords(), type);
			Vector3 launchatom = getlaunchatom(resi.GetAtomCoords(), type);
			Vector3 launchpad = launchatom - (axis / 2.0);

			//printf("axis %.2f %.2f %.2f\n",axis[0],axis[1],axis[2]);
			//printf("launchpad %.2f %.2f %.2f\n",launchpad[0],launchpad[1],launchpad[2]);



		        ////////////////////////////////////////////////////////////////////////////////
		        //////  STAGE 2D. Do the trigonometry to find the optimal rotation  ////////////
		        ////////////////////////////////////////////////////////////////////////////////

			//////   Vector nomenclature follows Canutescu and Dunbrack
			//////   F1, F2, F3, and F4 are vectors to the fixed target positions
			//////   M1, M2, M3, and M4 are vectors to the mobile positions
			//////   O1, O2, O3, and O4 are the projection of the M's onto the rotation axis
			//////   R1, R2, R3, and R4 are the component of each M perpendicular to the rotation axis
			//////   S1, S2, S3, and S4 are orthogonal to the axis and each R
			//////   f1, f2, f3, and f4 are the component of each F perpendicular to the rotation axis
			

			//////   C positions
			float includeC = 1;
			Vector3 M1 = mobileC - launchpad;
			Vector3 O1 = blitz::dot(M1, axis) * axis;
			Vector3 R1 = M1 - O1;
			Vector3 F1 = anchorC - launchpad;
			Vector3 f1 = F1 - O1;
			Vector3 normR1 = R1; //just to initialize
			Vector3 normS1 = R1; //just to initialize
			FLOAT lR1 = getLength(R1);
			if(lR1 < 0.0001) { 
			  includeC = 0;
                          //printf("mobile C is on the rotation axis\n");
			  //continue;  ////  If no deviation, move on. Prevents normalization nan
			}
			else {
			  normR1 = normalize(R1);
			  Vector3 S1 = blitz::cross(normR1, axis);
			  if(getLength(S1) < 0.0001) { 
			    includeC = 0;
                            printf("S1 %f\n",getLength(S1));
			    //continue;  ////  If no deviation, move on. Prevents normalization nan
			  }
			  else {
			    normS1 = normalize(S1);
			  }
			}


			//////   O positions
			float includeO = 1;
			Vector3 M2 = mobileO - launchpad;
			Vector3 O2 = blitz::dot(M2, axis) * axis;
			Vector3 R2 = M2 - O2;
			Vector3 F2 = anchorO - launchpad;
			Vector3 f2 = F2 - O2;
			Vector3 normR2 = R2; //just to initialize
			Vector3 normS2 = R2; //just to initialize
			FLOAT lR2 = getLength(R2);
			if(lR2 < 0.0001) { 
			  includeO = 0;
                          //printf("mobile O is on the rotation axis\n");
			  //continue;  ////  If no deviation, move on. Prevents normalization nan
			}
			else {
			  normR2 = normalize(R2);
			  Vector3 S2 = blitz::cross(normR2, axis);
			  if(getLength(S2) < 0.0001) { 
			    includeO = 0;
                            printf("S2 %f\n",getLength(S2));
			    //continue;  ////  If no deviation, move on. Prevents normalization nan
			  }
			  else {
			    normS2 = normalize(S2);
			  }
			}


			//////   N positions
			float includeN = 1;
			Vector3 M3 = mobileN - launchpad;
			Vector3 O3 = blitz::dot(M3, axis) * axis;
			Vector3 R3 = M3 - O3;
			Vector3 F3 = anchorN - launchpad;
			Vector3 f3 = F3 - O3;
			Vector3 normR3 = R3; //just to initialize
			Vector3 normS3 = R3; //just to initialize
			FLOAT lR3 = getLength(R3);
			if(lR3 < 0.0001) { 
			  includeN = 0;
                          //printf("mobile N is on the rotation axis\n");
			  //continue;  ////  If no deviation, move on. Prevents normalization nan
			}
			else {
			  normR3 = normalize(R3);
			  Vector3 S3 = blitz::cross(normR3, axis);
			  if(getLength(S3) < 0.0001) { 
			    includeN = 0;
                            printf("S3 %f\n",getLength(S3));
			    //continue;  ////  If no deviation, move on. Prevents normalization nan
			  }
			  else {
			    normS3 = normalize(S3);
			  }
			}



			//////   Ca positions
			int includeCA = 1;
			Vector3 M4 = mobileCa - launchpad;
			Vector3 O4 = blitz::dot(M4, axis) * axis;
			Vector3 R4 = M4 - O4;
			Vector3 F4 = anchorCa - launchpad;
			Vector3 f4 = F4 - O4;
			Vector3 normR4 = R4; //just to initialize
			Vector3 normS4 = R4; //just to initialize
			FLOAT lR4 = getLength(R4);
			if(lR4 < 0.0001) { 
			  includeCA = 0;
                          //printf("mobile Ca is on the rotation axis\n");
			  //continue;  ////  If no deviation, move on. Prevents normalization nan
			}
			else {
			  normR4 = normalize(R4);
			  Vector3 S4 = blitz::cross(normR4, axis);
			  if(getLength(S4) < 0.0001) { 
			    includeCA = 0;
                            printf("S4 %f\n",getLength(S4));
			    //continue;  ////  If no deviation, move on. Prevents normalization nan
			  }
			  else {
			    normS4 = normalize(S4);
			  }
			}
		


			// If at least one mobile/target pair is not on the rotation axis we can proceed
			if (includeC + includeO + includeN + includeCA < 1) {
			  printf("No productive rotations. Skipping");
			  continue;
			} 

			//////   The rotation that would best overlap the anchor and mobile C,O,N,Ca groups
 			//////   can be computed using a big fraction
			FLOAT numer = 0;
			FLOAT denom = 0;
			if (includeC) {
			  numer = blitz::dot(f1, normS1)*lR1;
			  denom = blitz::dot(f1, normR1)*lR1;
			}
			if (includeO) {
			  numer += blitz::dot(f2, normS2)*lR2;
			  denom += blitz::dot(f2, normR2)*lR2;
			}
			if (includeN) {
			  numer += blitz::dot(f3, normS3)*lR3;
			  denom += blitz::dot(f3, normR3)*lR3;
			}
			if (includeCA) {
			  numer += blitz::dot(f4, normS4)*lR4;
			  denom += blitz::dot(f4, normR4)*lR4;
			}

			if (0) {
                        printf("1 %f 2 %f 3 %f 4 %f\n",lR1,lR2,lR3,lR4);
                        printf("f1 %f %f %f\n",f1[0],f1[1],f1[2]);
                        printf("f2 %f %f %f\n",f2[0],f2[1],f2[2]);
                        printf("f3 %f %f %f\n",f3[0],f3[1],f3[2]);
                        printf("f4 %f %f %f\n",f4[0],f4[1],f4[2]);
                        printf("normS1 %f %f %f\n",normS1[0],normS1[1],normS1[2]);
                        printf("normS2 %f %f %f\n",normS2[0],normS2[1],normS2[2]);
                        printf("normS3 %f %f %f\n",normS3[0],normS3[1],normS3[2]);
                        printf("normS4 %f %f %f\n",normS4[0],normS4[1],normS4[2]);
                        printf("normR1 %f %f %f\n",normR1[0],normR1[1],normR1[2]);
                        printf("normR2 %f %f %f\n",normR2[0],normR2[1],normR2[2]);
                        printf("normR3 %f %f %f\n",normR3[0],normR3[1],normR3[2]);
                        printf("normR4 %f %f %f\n",normR4[0],normR4[1],normR4[2]);
			}

			//////   Well, we want the arctangent of that fraction. . .	
			//////   alpha is that optimal torsion
			FLOAT alpha = 0;
			if(CCD % 2 == 0 && dof == 0)
			  {
			   alpha = atan2(numer, denom)*180/PI;
			  }
			else
			  alpha = -atan2(numer, denom)*180/PI;
			// Commenting out the check below because it wouldn't compile on the mac
			//if (isnan(alpha)) {
                        //    printf("CCD failure: NaN. Returning input\n");
                        //    return input->Clone();
			//}



		        ////////////////////////////////////////////////////////////////////////////////
		        //////  STAGE 2E. Given the optimal rotation, determine the move  //////////////
		        ////////////////////////////////////////////////////////////////////////////////

			//////   If the DOF is already there, move on
			if(fabs(alpha) < convergedalpha) {
			  //printf("NO ROTATION for phi %d #############\n",resi.GetID());
			  continue;
			}

			//////   Otherwise, step towards alpha
			FLOAT twist;
			if(alpha > dt * dofW)
			  twist = dt * dofW;
			else if(alpha < -dt * dofW)
			  twist = -dt*dofW;
			else
			  twist = alpha;

			//////   Summarize the change
			//if (type == "phi")
			//	printf("ROTATING by %.2f (out of %.2f) for phi %d #############\n",twist,alpha,resi.GetID());
			//if (type == "psi")
			//	printf("ROTATING by %.2f (out of %.2f) for psi %d #############\n",twist,alpha,resi.GetID());



		        ////////////////////////////////////////////////////////////////////////////////
		        //////  STAGE 2F. Given the transformation, update "downstream" chain  /////////
		        ////////////////////////////////////////////////////////////////////////////////

			//////   Convert the quaternion (axis and rotation) to a rotation matrix
			Matrix3 rotmatrix;
			rmatrixu(axis, PI*twist/180, &rotmatrix);

			if (CCD % 2 == 1)
			{
				//printf("\tUpdate low branch\n");
				if (type == "phi") {
					for (int j = 0; j < (int)resi.GetAtomCoords().size(); ++j) {
						// For phi, don't update amide N-H-Ca 
						if ((j == atomN)||(j == atomCA)||(j == atomH)) { continue; } 
						updateatom(resi, j, rotmatrix, launchpad);
					}
				}
				else {
					// For psi, only update O
					updateatom(resi, atomO, rotmatrix, launchpad);
				}
				for (int i = resi.GetID() + 1; i < nextbreak; ++i) {
					for (int j = 0; j < (int)system->GetResidue(i).GetAtomCoords().size(); ++j) {
						updateatom(system->GetResidue(i), j, rotmatrix, launchpad);
					}
				}
			}
			else
			{
				//printf("\tUpdate high branch\n");
				if (type == "psi") {
					for (int j = 0; j < (int)resi.GetAtomCoords().size(); ++j) {
						// For psi, don't update carbonyl C-O 
						if ((j == atomC)||(j == atomO)) { continue; } 
						updateatom(resi, j, rotmatrix, launchpad);
					}
				}
				else {
					// For phi, only update H for phi 
					updateatom(resi, atomH, rotmatrix, launchpad);
				}
				for (int i = nextbreak; i < resi.GetID(); ++i) {
					for (int j = 0; j < (int)system->GetResidue(i).GetAtomCoords().size(); ++j) {
						updateatom(system->GetResidue(i), j, rotmatrix, launchpad);
					}
				}
			}



		        ////////////////////////////////////////////////////////
		        //////  STAGE 2G. Update the virtual positions  ////////
		        ////////////////////////////////////////////////////////

			if(CCD % 2 == 1)
			  { 
			    // Low branch was mobile, refresh the virtual N,Ca
                            atomC  =  res1.AtomIndex("C");
                            atomO  =  res1.AtomIndex("O");
                            atomCA =  res1.AtomIndex("CA");
			    virtualNCa = idealnextNCa(atoms1);
			    mobileN = virtualNCa[0];
			    mobileCa = virtualNCa[1];
			  }
			else
			  { 
			    // High branch was mobile, refresh the virtual C,O
                            if (res2.GetAminoAcid().name3Char == "PRO") {
                                //printf("Using Proline CD in place of missing H\n");
                                atomH = res2.AtomIndex("CD");
                            }
                            else {
                                atomH =  res2.AtomIndex("H");
                            }
                            atomN  = res2.AtomIndex("N");
                            atomCA =  res2.AtomIndex("CA");
                            atomC  = res2.AtomIndex("C");
			    virtualCO = idealpreviousCO(atoms2);
			    mobileC = virtualCO[0];
			    mobileO = virtualCO[1];
			  }



		        /////////////////////////////////////////////////
		        //////  STAGE 2H. Check for convergence  ////////
		        /////////////////////////////////////////////////

			currentSSD = SSD(mobileC, mobileO, mobileN, mobileCa, anchorC, anchorO, anchorN, anchorCa);
			//printf("currentSSD: %f, SSDconverge: %f\n", currentSSD, SSDconverge);			
			if(currentSSD < SSDconverge) {
			    //printf("CCD converged: SSD = %f\n",currentSSD);
			    goto NextChainBreak;
			}


		      }
		  }
NextChainBreak:
		continue; // we break out to here when CCD converges, and skip to the next chainbreak
	}

	return system;
}



