#! /usr/bin/python
# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 


"""
This module creates boilerplate code for amino acid data. To create a new amino acid,
you can add it to the AminoAcidList.py file and put an entry in aa_props/. Note that
if this new amino acid is a simple variant of an amino acid that already exists, it
is best to do this in some kind of automated fashion to avoid duplicating data.

Also note that if you add a new amino acid, this will implicitly add a new set of raw
atom types to the atom type system. You'll need to associate each raw type with a 
specific type in each of the existing atom type systems; currently, all atom type systems
are derived automatically from the Rosetta system, so you only need to add mappings to 
rosetta.py. Regardless, you will also need to run ./generate_atom_types_code.py

One thing to note about this system is that it generates a big string of data 
(data_str) and bakes it right into the source. The reason for this is that originally
we generated code to load the data directly, as in 

data_[ALA].something[0] = 1;
data_[ALA].something[1] = 10;
... etc ...

but we found that this method doesn't scale well -- GCC chokes on large datasets
presented this way and uses many gigabytes of RAM trying to compile this kind of source.
The "right thing" to do to solve this would be to load the data at runtime from a data
file instead, but that would be an interface-breaking change. For now, it's OK just
to load the data in a kind of roundabout way from the source file.
"""

import pickle
from string import Template
from aa_props.AminoAcid import AminoAcidInfo
from aa_props.Format import CreateFormatter
from AminoAcidList import aa_list

def generate_enum(outfile):
	outfile.write('#define N_AMINO_ACIDS ' + str(len(aa_list)) + '\n')
	outfile.write('#define N_ATOM_ORDERINGS ' + str(len(AminoAcidInfo.atomOrderEnums)) + '\n\n')

	outfile.write('enum AminoAcidIndex { ' + ', '.join(aa_list) + ' };\n')
	outfile.write('enum AtomOrderClass { ' + ', '.join(AminoAcidInfo.atomOrderEnums.values()) + ' };\n')
	
def generate_init_fun(outfile):
	for aa in aa_list:
		outfile.write('Load%s(data_[%s], istream);\n' % (aa, aa))
	
	outfile.write('\n')

	for aa in aa_list:
		outfile.write('byName_["' + aa + '"] = &data_[' + aa + '];\n')
		
def generate_aa_fn(outfile, aa, data, format):
	reprs = []
	
	outfile.write('void Load%s(AminoAcid& data, std::istringstream& istream)\n' % aa)
	outfile.write('{\n')

	# write out the amino acid name data
	outfile.write('\tdata.aminoAcid = ' + aa + ';\n')
	outfile.write('\tdata.name3Char = "' + aa + '";\n')

	# change the atom order key names to their enum equivalents
	atomOrders = {}
	for orderType in data['atomOrders'].keys():
		orderEnumType = AminoAcidInfo.atomOrderEnums[orderType]
		orderIndex = AminoAcidInfo.atomOrderEnums.values().index(orderEnumType)

		atomOrders[orderIndex] = data['atomOrders'][orderType]

	# now we don't need the original data
	del data['atomOrders']

	# write out the atom ordering data
	repr, reader = format(aa, 'data.atomOrders', atomOrders, False)
	reprs.append(repr)
	outfile.write(reader)
	
	# write out all the normal data
	for item, values in data.items():
		repr, reader = format(aa, 'data.' + item, values)
		reprs.append(repr)
		outfile.write(reader)

	outfile.write('\n}\n\n')

	return reprs

def generate_all_aa_fns(outfile, aa_data, formatters):
	reprs = []
	for aa in aa_list:
		data = aa_data[aa]
		reprs += generate_aa_fn(outfile, aa, data, formatters)
	
	outfile.write('char* data_str =\n')
	for repr in reprs:
		outfile.write(repr)
	outfile.write(';\n')
	
if __name__ == '__main__':
	
	aa_data = {}
	for aa in aa_list:
		m = __import__('aa_props.' + aa, {}, {}, [0])
		aa_data[aa] = m.AAData().data

	order_data = {}
	for aa in aa_list: 
		order_data[aa] = aa_data[aa]['atomOrders']['trie']

	format = CreateFormatter(order_data)

	generate_enum(     open('output/aa_props_enum.hxx',  'w') )
	generate_init_fun( open('output/aa_props_init.hxx',  'w') )

	generate_all_aa_fns( open('output/aa_props_data.hxx', 'w'), aa_data, format )


