#! /usr/bin/python
# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 


"""
This module generates the multiple atom type systems used in CHOMP.

The basic idea here is that different categorizations of atoms are useful
in different circumstances. For instance, Rosetta uses a 26-type system
(also implemented in CHOMP), which does not include a specific type for
beta carbons. Rosetta simultaneously uses a different ad-hoc type system
to categorize atoms when deciding whether or not to count pairwise interactions
between neighboring residues (here implemented as the CountPair type system).

Using this module, the creation and maintenance of multiple conflicting
atom categorizations is made simple and clean. We start with a fundamental
'raw' type system, which at this point has a different type for every atom
in every amino acid. We then have a many-to-one mapping that translates
this raw type into every other atom type we may use. This module takes
care of generating this mapping as well as support code needed for it to work.

To add a new type system, simply add a new Python script to the atom_types
directory, following the examples in rosetta.py and countpair.py. Then,
run this script and it will automatically detect your addition and include
it in the generated C++ code.
"""

import os
import pickle
import sys

from AminoAcidList import aa_list

def DeletePyc():
	"""
	The .pyc files Python makes are annoying, so we delete them.
	"""
	for item in os.listdir('atom_types'):
		if item[0] != '.':
			base, ext = tuple(item.split('.'))
			if ext == 'pyc':
				os.unlink( os.path.join('atom_types', item) )

def MakeTypes(AAlist):
	"""
	Walk through each module in the atom_types package and add
	its type system to the type dictionary.
	"""
	main_lst = []	# to contain the list of raw atom types

	# generate the raw atom type list
	for AA in AAlist:
		m = __import__('aa_props.' + AA, {}, {}, [0])
		atoms = m.AAData().data['atomOrders']['trie']

		for atom in atoms:
			main_lst.append('%s:%s' % (AA, atom))

 	types = []
	DeletePyc()

	# walk through each module, adding its type to the types dictionary
	for item in os.listdir('atom_types'):
		if item != '__init__.py' and item[0] != '.':
			f = item.strip('.py')
			m = __import__('atom_types.' + f, {}, {}, [0])

			types.append(m.AtomType(main_lst))
	DeletePyc()

	return main_lst, types

def GenerateStringConversions(name, lst, outfile):
	"""
	Generate the enumeration -> string structure
	included in the output for convenience (debugging, etc).
	"""
	outfile.write('\tstd::string toStringData%s[] = {\n' % name)
	for type in lst:
		outfile.write('\t\t"%s",\n' % type)
	
	outfile.write('\t};\n\n')
	outfile.write('''
		toString_[AtomType::%s] = std::vector<std::string>(
				toStringData%s, toStringData%s + %d);
	\n''' % (name, name, name, len(lst)))

def GenerateTypeTable(types, data, main_lst, outfile):
	"""
	Generate the actual mapping between raw types and types within
	specific type systems.
	"""
	
	# make the full type table
	outfile.write('\tint conversionData[N_RAW_ATOM_TYPES][N_ATOM_TYPE_SYSTEMS] = {\n')
	for i in xrange(len(main_lst)):
		raw_type = main_lst[i]
		outfile.write('\t\t{ ')
		for type in types:
			outfile.write('%d, ' % data[type][raw_type])
		outfile.write('},\n')
	outfile.write('\t};\n\n')

def GenerateHeader(types, main_lst, headerfile):
	"""
	Create the header file to be included everywhere the atom type systems
	are referenced. This includes declarations for the various relevant 
	enumerations of atom types and type systems.
	"""

	headerfile.write('#define N_RAW_ATOM_TYPES %d\n' % len(main_lst))
	headerfile.write('#define N_ATOM_TYPE_SYSTEMS %d\n\n' % len(types))

	headerfile.write('namespace AtomType {\n\n')

	# make the enum of type systems
	headerfile.write('enum TypeSystem {\n')
	for type in types:
		headerfile.write('\t' + type.enumName + ',\n')
	headerfile.write('};\n')

	# write the type system symbol enums
	for type in types:
		headerfile.write('struct ' + type.name + ' {\n\n')

		#make the type symbol enumeration
		headerfile.write('enum {\n')
		for atomtype in type.typeList:
			headerfile.write('\t%s,\n' % type.enumFn(atomtype))
		headerfile.write('};\n};\n\n')
	
	headerfile.write('\n\n}\n')

if __name__ == '__main__':
	# open the output files
	outfile = open('output/atom_types_impl.hxx','w')
	headerfile = open('output/atom_types.hxx', 'w')

	# create the type info
	main_lst,types = MakeTypes(aa_list) 

	# create the header (non-impl) file
	GenerateHeader(types, main_lst, headerfile)

	# now for each type, generate the output for that type
	data = {}
	for type in types:
		GenerateStringConversions(type.enumName, type.typeList, outfile)
		
		# create the data used in GenerateTypeTable
		data[type] = {}
		for key, value in type.types.items():
			for raw_type in value:
				data[type][raw_type] = type.typeList.index(key)

	GenerateTypeTable(types, data, main_lst, outfile)


