#! /usr/bin/python
# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 


""" 
This module generates several .hxx files related to the DataDirectory. You can run
it from the command line with no arguments.

data_tables.hxx is included in the data_table.h header file and contains declarations
for all the data tables in the cgs input file.

data_table_dispatch.hxx is included in the corresponding source file, and simply
dispatches the loading of the tables on a given input string.

data_tables_dispatch_binary_input/output.hxx are similar to the above, but contain
code for saving/loading each table to and from a binary image on the disk.
"""

import sys, os
import DataTableList

def WriteDataTableCode(data):
	"""
	Takes in a list of array declarations and produces the contents of
	the data_tables and data_table_dispatch files. It's probably easiest 
	to just look at those files to see what this does.
	"""

	# output files:
	tableHeaderFile   = 'output/data_tables.hxx'
	tableDispatchFile = 'output/data_table_dispatch.hxx'
	tableBinaryFileI  = 'output/data_table_dispatch_binary_input.hxx' 
	tableBinaryFileO  = 'output/data_table_dispatch_binary_output.hxx' 
	
	items = []
	for decl in data:
		# split the declarations up into their parts
		tokens = decl.split('>')
		signature = tokens[0].split('<')[1]	
		data_type, n_dims = signature.split(',')
		name = tokens[-1].strip();

		items.append( (data_type, n_dims, name) )

	WriteTableDeclarations(items, open(tableHeaderFile, 'w'))
	WriteTableDispatch(items, open(tableDispatchFile, 'w'))
	WriteTableDispatchBinaryInput(items, open(tableBinaryFileI, 'w'))
	WriteTableDispatchBinaryOutput(items, open(tableBinaryFileO, 'w'))

def WriteTableDeclarations(table, fout):
	"""
	Outputs the declaration of a struct containing the data
	tables in the form:
	blitz::Array<-,-> -;
	"""

	# write the accessors
	for entry in table:
		fout.write('\tblitz::Array<%s,%s> const& Get%s() const { return %s_; }\n' % (entry[0], entry[1], entry[2], entry[2]))
	for entry in table:
		fout.write('\tvoid Set%s(blitz::Array<%s,%s> const& value) { %s_ = value; }\n' % (entry[2], entry[0], entry[1], entry[2])) 
	fout.write('private:\n')

	# write the declarations
	for entry in table:
		fout.write('\tblitz::Array<%s,%s> %s_;\n' % entry)

	fout.close()

def WriteTableDispatch(table, fout):
	"""
	Outputs a series of if-statements to compare an input string
	to the name of each table in the input.
	"""
	
	# the first entry doesn't get an 'else' before the 'if'
	firstEntry = True

	for entry in table:
		if not firstEntry: fout.write('else ')
		firstEntry = False

		fout.write('if (tableName == "%s")\n' % entry[2])
		fout.write('{\n\t')
		fout.write('file >> %s_;' % entry[2])
		fout.write('\n}\n')

	fout.write('else\n\t throw std::string("Data table %s not recognized");\n\n' % entry[2])

	fout.close()

def WriteTableDispatchBinaryInput(table, fout):
	"""
	Outputs some code to load a binary image of a table from the disk.
	"""
	
	# the first entry doesn't get an 'else' before the 'if'
	firstEntry = True

	for entry in table:
		if not firstEntry: fout.write('else ')
		firstEntry = False

		fout.write('if (tableName == "%s")\n' % entry[2])
		fout.write('{\n\t')
		fout.write('TinyVector<int, %s> shape;\n\t' % entry[1])
		fout.write('for (int i = 0; i < %s; ++i)\n\t' % entry[1])
		fout.write('\tshape(i) = dims[i];\n\n\t')
		fout.write('%s_.reference(Array<%s, %s>( (%s*)table_buffer, shape, deleteDataWhenDone ));\n' %	\
				(entry[2], entry[0], entry[1], entry[0]) )
		fout.write('\n}\n')

	fout.write('else\n\t throw std::string("Data table " + tableName + " not recognized");\n\n')

	fout.close()

def WriteTableDispatchBinaryOutput(table, fout):
	"""
	Outputs some code to save a binary image of a table to the disk.
	"""
	for entry in table:
		fout.write('CompileSingle(file, "%s", %s_);\n' % (entry[2], entry[2]) )
	fout.close()

if __name__ == '__main__':
	WriteDataTableCode(DataTableList.data)


