// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "data_table.h"
#include "general.h"

#include <iostream>
#include <fstream>
#include <ctype.h>

/** \file data_table.cpp
 * \see data_table.h
 */

using namespace blitz;

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

DataDirectory::Pointer DataDirectory::Create(std::string const& directoryFile)
{
	return DataDirectory::Pointer( new DataDirectory(directoryFile) );
}

DataDirectory::DataDirectory(std::string const& directoryFile)
{
	// check if the directory file is a binary image
	std::ifstream file;
	file.open(directoryFile.c_str(), std::ios::binary);

	char buffer[256];
	file.read(buffer, 6);
	buffer[6] = '\0';

	if (!strcmp(buffer, "DDIMAG"))
	{
		compiled_ = true;

		while (!file.eof())
		{
			int name_length, data_length, n_dims;
			int dims[10];
			
			// read in the header
			file.read( (char*)&name_length, sizeof(int) );
			file.read( (char*)&data_length, sizeof(int) );
			file.read( (char*)&n_dims, sizeof(int) );
		
			for (int i = 0; i < n_dims; ++i)
				file.read( (char*)&(dims[i]), sizeof(int) );

			// read in the table name
			if (name_length >= 255)
				throw std::string("DataDirectory image file is corrupted.\n");

			file.read( buffer, name_length );
			buffer[name_length] = '\0';	
			std::string tableName = buffer;

			// read in all the data
			char* table_buffer = new char[data_length];
			file.read( table_buffer, data_length );
			
			// templated code
			#include "codegen/output/data_table_dispatch_binary_input.hxx"
		}				
	}
	else
	{
		compiled_ = false;
		file.close();

		SetDirectoryFile(directoryFile);
		ReloadAllDataTables();
	}
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC INTERFACE                       ///////
///////////////////////////////////////////////////////////////////////////

void DataDirectory::SetDirectoryFile(std::string const& fileName)
{
	if (compiled_)
		throw std::string("DataDirectory.SetDirectoryFile: This function is not available in compiled mode.");

	std::ifstream file;
	file.open(fileName.c_str());

	// FIXME: not portable
	std::string path = fileName.substr( 0, fileName.find_last_of("/") );
	if (path == fileName)
		path = ".";

	if (!file.is_open())
		throw std::string("DataDirectory.SetDirectoryFile: Failed to open data directory file " + fileName);

	while (!file.eof())
	{
		// strip off any leading whitespece
		while (isspace(file.peek())) file.get();
		
		// if there is a '#' character, ignore the rest of the line
		// cause it's a comment
		if (file.peek() == '#')
			file.ignore(1024, '\n');

		// otherwise, the format should be <table name> <file name>
		std::string tableName, tableFile;
		file >> tableName >> tableFile;

		// FIXME: not portable
		// store the entry into a tableName -> fileName map
		if (tableName != "" && tableFile != "")
			dataDirectory_[tableName] = path + "/" + tableFile;
	}

	file.close();
}

void DataDirectory::ReloadDataTable(std::string const& tableName,
	               	            std::string const& fileName)
{
	if (compiled_)
		throw std::string("DataDirectory.ReloadDataTable: This function is not available in compiled mode.");

	std::ifstream file;

	// if the fileName isn't given, it should be already present in the data directory.
	// if it isn't, there's going to be an error.
	std::string const& realFileName = 
		(fileName != "") ? fileName : dataDirectory_[tableName];

	// FIXME: check for empty fileName (bad table name)
	
	file.open(realFileName.c_str());

	if (!file.is_open())
		throw "DataDirectory.ReloadDataTable: Failed to open data table file " + realFileName + ".\n"
		      + "Check your data directory file for errors.";
	
	// the data table files contain comment headers, so discard everything until we find
	// the end of the comments
	bool comments = true;
	while (comments)
	{
		// strip off any leading whitespece
		while (isspace(file.peek())) file.get();
		
		// if there is a '-' character, this may be the end of the comments
		if (file.peek() == '-')
		{
			std::string line;
			std::getline(file, line);

			// take out any whitespace and dashes from the line
			std::remove(line.begin(), line.end(), ' ');
			std::remove(line.begin(), line.end(), '\t');
			std::remove(line.begin(), line.end(), '-');

			if (line.substr(0, 14) == "DATASTARTSHERE")
				comments = false;
		}
		else file.get();
	}

	// templated code
	#include "codegen/output/data_table_dispatch.hxx"

	file.close();
}

void DataDirectory::ReloadAllDataTables()
{
	if (compiled_)
		throw std::string("DataDirectory.ReloadAllDataTables: This function is not available in compiled mode.");

	// for each entry in the data directory
	std::map<std::string, std::string>::const_iterator i;
	for (i = dataDirectory_.begin(); i != dataDirectory_.end(); ++i)
		ReloadDataTable(i->first, i->second);
}

template<typename ArrayType>
void CompileSingle(std::ofstream& file, std::string name, ArrayType const& table)
{
	int name_length = (int)name.size();
	int data_length = (int)(table.size() * sizeof(typename ArrayType::T_numtype));

	int n_dims = table.dimensions();
	int dims[10];
	for (int i = 0; i < n_dims; ++i)
		dims[i] = table.extent(i);

	file.write( (char*)&name_length, sizeof(int) );
	file.write( (char*)&data_length, sizeof(int) );
	file.write( (char*)&n_dims, sizeof(int) );

	for (int i = 0; i < n_dims; ++i)
		file.write( (char*)&(dims[i]), sizeof(int) );

	file.write( (char*)name.data(), name_length );
	file.write( (char*)table.data(), data_length );
}

void DataDirectory::Compile(std::string const& outputFileName) const
{
	std::ofstream file(outputFileName.c_str(), std::ios::binary);
	char hdr[] = "DDIMAG";
	file.write(hdr, 6);

	// templated code
	#include "codegen/output/data_table_dispatch_binary_output.hxx"
}


