// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "dunbrack.h"
#include "interpolate.h"
#include "amino_acid.h"

#include <fstream>
#include <list>

using namespace CHOMPMath;

///////////////////////////////////////////////////////////////////////////
//////                     AUXILIARY STRUCTURES                     ///////
///////////////////////////////////////////////////////////////////////////

struct PercentData
{
	PercentData(int _nRot, FLOAT _percent): 
		 nRot(_nRot), percent(_percent) {}

	int nRot;
	FLOAT percent;
};

bool PercentDataCmp(PercentData const& pd1, PercentData const& pd2)
{
	return pd1.percent > pd2.percent;
}

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

DunbrackLibrary::ConstPointer DunbrackLibrary::Create(AminoAcidDirectory::ConstPointer aadir, std::string const& fileName)
{
	return DunbrackLibrary::ConstPointer( new DunbrackLibrary(aadir, fileName) );
}

DunbrackLibrary::DunbrackLibrary(AminoAcidDirectory::ConstPointer aadir, std::string const& fileName)
				: aadir_(aadir)
{
	// check if the input file is a binary image
	std::ifstream is;
	is.open(fileName.c_str(), std::ios::binary);

	if (!is.is_open())
		throw "DunbrackLibrary: failed to open " + fileName;

	char buffer[256];
	is.read(buffer, 8);
	buffer[8] = '\0';

	if (!strcmp(buffer, "DLIBIMAG"))
	{
		LoadFromBinary(is);
		return;
	}

	// if not, we need to parse a plaintext input
	is.seekg(0, std::ios::beg);
	LoadAuxiliaryFiles(fileName);

	const int nAminoAcids = 20;
	const int chiZeroEntries = 2;

	int nEntries, nBins, maxRotamers;
	is >> nEntries >> nBins >> maxRotamers;

	if (is.fail())
		throw "DunbrackLibrary: In first line of Dunbrack library at " + fileName + ":\n"
		      "Expected '<# of entries> <# of angle bins> <max. # of rotamers>\n";

	// resize the arrays:
	packedIndex_.resize(nAminoAcids, nBins, nBins, maxRotamers);
	percentOrderIndex_.resize(nAminoAcids, nBins, nBins, maxRotamers);
	chiData_.resize(nEntries + chiZeroEntries);
	sdData_.resize(nEntries + chiZeroEntries);
	probData_.resize(nEntries + chiZeroEntries);

	int packedI = 0;
	char line[110];
	char aa_name[4];
	AminoAcidIndex aa_index;
	int phi, psi, last_phi = -1000, last_psi = -1000;
	int phiBin = 0, psiBin = 0;
	int nRot = 0;
	std::list<PercentData> pdlist;

	is.getline(line, 110);

	for (; packedI < nEntries; ++packedI)
	{
		is.getline(line, 110);
		if (is.fail()) break;

		std::sscanf(line, "%4s", aa_name);
		aa_index = aadir_->Get(aa_name)->aminoAcid;

		std::sscanf(line + 5,  "%4d", &phi);
		std::sscanf(line + 10, "%4d", &psi);
	
		if (phi > 175 || psi > 175)
			continue;

		std::sscanf(line + 33, "%8f", &probData_(packedI)   );

		std::sscanf(line + 43, "%6f", &chiData_(packedI)(0) );
		std::sscanf(line + 51, "%6f", &chiData_(packedI)(1) );
		std::sscanf(line + 59, "%6f", &chiData_(packedI)(2) );
		std::sscanf(line + 67, "%6f", &chiData_(packedI)(3) );

		std::sscanf(line + 77, "%6f", &sdData_(packedI)(0) );
		std::sscanf(line + 85, "%6f", &sdData_(packedI)(1) );
		std::sscanf(line + 93, "%6f", &sdData_(packedI)(2) );
		std::sscanf(line +101, "%6f", &sdData_(packedI)(3) );

		if (last_phi == -1000)
		{
			last_phi = phi;
			last_psi = psi;
		}

		if (is.eof()) break;

		if (last_phi != phi || last_psi != psi)
		{
			last_phi = phi;
			last_psi = psi;

			pdlist.sort(PercentDataCmp);
			
			std::list<PercentData>::iterator p_i = pdlist.begin();
			int i = 0;
			for (; i < nRot; ++i, ++p_i)
				percentOrderIndex_((int)aa_index, phiBin, psiBin, i) = p_i->nRot;

			pdlist.clear();
			nRot = 0;
		}
		
		phiBin = (phi + 180)/10;
		psiBin = (psi + 180)/10;

		pdlist.push_back(PercentData(nRot, probData_(packedI)));		
		packedIndex_((int)aa_index, phiBin, psiBin, nRot++) = packedI;
	}

	if (packedI != nEntries)
		throw "DunbrackLibrary: Error reading Dunbrack library at entry # " + to_str(packedI);

	// fill in GLY and ALA since they don't appear in the Dunbrack file
	blitz::Range angleRange(0, nBins - 1);
	blitz::Range entryRange(packedI, packedI + 1);

	packedIndex_(ALA, angleRange, angleRange, 0) = packedI;
	packedIndex_(GLY, angleRange, angleRange, 0) = packedI + 1;

	chiData_(entryRange) = 0.0;
	sdData_(entryRange)  = 0.0;
	probData_(entryRange) = 1.0;
}

void DunbrackLibrary::LoadAuxiliaryFiles(std::string const& fileName)
{
	std::string revlook = fileName + ".revlookup";

	std::ifstream is;
	is.open(revlook.c_str());

	if (!is.is_open())
		throw "DunbrackLibrary: Failed to open Dunbrack reverse lookup table at " + revlook;

	is >> revLookup_;
	is.close();

	std::string maxrot = fileName + ".maxrot";
	is.open(maxrot.c_str());

	if (!is.is_open())
		throw "DunbrackLibrary: Failed to open Dunbrack rotamer count table at " + maxrot;

	is >> maxRotType_;
	is.close();
}

template<typename T_numtype, int N_rank>
void LoadArray(std::ifstream& file, blitz::Array<T_numtype, N_rank>& array)
{
	int dims[N_rank];
	for (int i = 0; i < N_rank; ++i)
		file.read( (char*)&(dims[i]), sizeof(int) );

	blitz::TinyVector<int, N_rank> shape;
	for (int i = 0; i < N_rank; ++i)
		shape(i) = dims[i];

	int data_length;
	file.read( (char*)&data_length, sizeof(int) );
	
	char* table_buffer = new char[data_length];
	file.read( table_buffer, data_length );

	array.reference(blitz::Array<T_numtype, N_rank>( 
			(T_numtype*)table_buffer, shape, blitz::duplicateData ));

	delete[] table_buffer;
}

template<typename T_numtype, int N_rank>
void SaveArray(std::ofstream& file, typename blitz::Array<T_numtype, N_rank> const& array)
{
	for (int i = 0; i < N_rank; ++i)
		file.write( (char*)&(array.shape()(i)), sizeof(int) );

	int data_length = (int)(array.size() * sizeof(T_numtype));
	file.write( (char*)&data_length, sizeof(int) );
	file.write( (char*)array.data(), data_length );
}

void DunbrackLibrary::LoadFromBinary(std::ifstream& file)
{
	LoadArray<Vector4, 1>(file, chiData_);
	LoadArray<Vector4, 1>(file, sdData_);
	LoadArray<FLOAT, 1>(file, probData_);

	LoadArray<int, 4>(file, packedIndex_);
	LoadArray<int, 4>(file, percentOrderIndex_);

	LoadArray<int, 5>(file, revLookup_);
	LoadArray<int, 2>(file, maxRotType_);
}

void DunbrackLibrary::SaveToBinary(std::ofstream& file) const
{
	SaveArray<Vector4, 1>(file, chiData_);
	SaveArray<Vector4, 1>(file, sdData_);
	SaveArray<FLOAT, 1>(file, probData_);

	SaveArray<int, 4>(file, packedIndex_);
	SaveArray<int, 4>(file, percentOrderIndex_);

	SaveArray<int, 5>(file, revLookup_);
	SaveArray<int, 2>(file, maxRotType_);
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC INTERFACE                       ///////
///////////////////////////////////////////////////////////////////////////

int DunbrackLibrary::GetNRotByPercent(int aminoAcid, int iPhi, int iPsi, int percentOrder) const
{
	return percentOrderIndex_(aminoAcid, iPhi, iPsi, percentOrder);
}

int DunbrackLibrary::GetNRotByDunbrackID(int aminoAcid, blitz::TinyVector<int, 4> const& dunIndex) const
{
	int nRot = 0;
	int multiplier = 1;

	for (int i = 3; i >= 0; --i)
		if ( dunIndex(i) != 0 )
		{
			nRot += ( dunIndex(i) - 1 ) * multiplier;		
			multiplier *= maxRotType_(aminoAcid, i);
		}
	return nRot;
}

DunbrackLibrary::IntAngleBins DunbrackLibrary::GetAngleBins(FLOAT phi, FLOAT psi, bool start, bool end) const
{
	DunbrackLibrary::IntAngleBins i;
	int extra;

	i.phi = int(phi + 185.0);
	extra = -i.phi / 360;
	if (i.phi < 0) ++extra;
	i.phi = (i.phi+extra*360)/10 % 36;

	i.psi = int(psi + 185.0);
	extra = -i.psi / 360;
	if (i.psi < 0) ++extra;
	i.psi = (i.psi+extra*360)/10 % 36;

	if (start) i.phi = 9;
	if (end)   i.psi = 31;
	
	return i;
}
	
DunbrackLibrary::AngleBinData DunbrackLibrary::GetInterpData(FLOAT phi, FLOAT psi, bool start, bool end) const
{
	// Implementation note from Rosetta:
	//  Get bins appropriate for bilinear interpolation. To work with the
	//   dunbrack maps, we add 180 to reflect that bin 1 is -180 degrees,
	//   and we add 5 (binrange/2.) to reflect that dunbrack's
	//   bins are labeled at the *center* of the binrange rather
	//   than the min value.

	DunbrackLibrary::AngleBinData result;

	const int bins = 36;

	GetAngleInterpBins(phi+185.0, bins, &result.phi1, &result.phi2, &result.dphi);
	GetAngleInterpBins(psi+185.0, bins, &result.psi1, &result.psi2, &result.dpsi);
	
	//  adjust edges
	if ( start ) 
	{
		result.phi1 = 9;
		result.phi2 = 9; 
		result.dphi = 0.0;
	}
	else if ( end )
	{
		result.psi1 = 31;
		result.psi2 = 31;
		result.dpsi = 0.0;
	}
	
	return result;
}

Vector4 DunbrackLibrary::GetInterpChi(int aminoAcid, DunbrackLibrary::AngleBinData const& interpData, int nRot) const
{
	int nChi = aadir_->Get((AminoAcidIndex)aminoAcid)->nChi;

	int i11 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi1, nRot);
	int i12 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi2, nRot);
	int i21 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi1, nRot);
	int i22 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi2, nRot);
	
	FLOAT dummy;
	Vector4 result = 0.0;

	for (int i = 0; i < nChi; ++i)
	{
		InterpBilinearAngles(chiData_(i11)(i), chiData_(i21)(i),
				     chiData_(i12)(i), chiData_(i22)(i),
				     interpData.dphi, interpData.dpsi,
				     36, &(result(i)), &dummy, &dummy);
	}

	return result;
}


Vector4 DunbrackLibrary::GetInterpSD(int aminoAcid, DunbrackLibrary::AngleBinData const& interpData, int nRot) const
{
	int nChi = aadir_->Get((AminoAcidIndex)aminoAcid)->nChi;

	int i11 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi1, nRot);
	int i12 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi2, nRot);
	int i21 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi1, nRot);
	int i22 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi2, nRot);
	
	FLOAT dummy;
	Vector4 result = 0.0;

	for (int i = 0; i < nChi; ++i)
	{
		InterpBilinear(sdData_(i11)(i), sdData_(i21)(i),
			       sdData_(i12)(i), sdData_(i22)(i),
			       interpData.dphi, interpData.dpsi,
			       36, &(result(i)), &dummy, &dummy);
	}

	return result;
}

FLOAT DunbrackLibrary::GetInterpPercent(int aminoAcid, DunbrackLibrary::AngleBinData const& interpData, int nRot) const
{
	int i11 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi1, nRot);
	int i12 = packedIndex_(aminoAcid, interpData.phi1, interpData.psi2, nRot);
	int i21 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi1, nRot);
	int i22 = packedIndex_(aminoAcid, interpData.phi2, interpData.psi2, nRot);
	
	FLOAT val, dummy;
	
	InterpBilinear(probData_(i11), probData_(i21),
		       probData_(i12), probData_(i22),
		       interpData.dphi, interpData.dpsi,
		       36, &val, &dummy, &dummy);
	
	return val;
}

int DunbrackLibrary::GetBaseRotamer(int aminoAcid, Vector4 const& chi) const
{
	blitz::TinyVector<int, 4> dunRotID = 0;
	const int nChi = aadir_->Get((AminoAcidIndex)aminoAcid)->nChi;
	
	for (int chi_i = 0; chi_i < nChi; ++chi_i)
	{
		int variant = 0;
	
		// special case handling for certain amino acids...
		if (aminoAcid == ASN && chi_i == 1)
		{
			if       ( dunRotID(0) == 1 ) variant = 0;
			else if  ( dunRotID(0) == 2 ) variant = 1;
			else	 		      variant = 2;
		}
		else if (aminoAcid == GLN && chi_i == 2)
		{
			if 	 ( dunRotID(1) == 2 ) variant = 0;
			else			      variant = 1;
		}

		for (int div = 0; div < revLookup_.extent(3); ++div)
		{
			int bound = revLookup_(aminoAcid, chi_i, variant, div, 0);
			int rot_n = revLookup_(aminoAcid, chi_i, variant, div, 1);

			FLOAT f_bound = static_cast<FLOAT>(bound);
			if ( chi(chi_i) <= f_bound )
			{
				dunRotID(chi_i) = rot_n;
				break;
			}
			else if ( bound == 180 )
				break;
		}
	}

	return GetNRotByDunbrackID(aminoAcid, dunRotID);
}

void DunbrackLibrary::Compile(std::string const& outputFileName) const
{
	std::ofstream file(outputFileName.c_str(), std::ios::binary);
	char hdr[] = "DLIBIMAG";
	file.write(hdr, 8);
	
	SaveToBinary(file);
}


