// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "energy_graph.h"

#include <iostream>
#include <fstream>

// NOTE: As of 1/13/07, MSVC 9.0 and boost's SVN trunk fail to compile
// the foreach macro when it's combined with the boost::vertices or
// boost::out_edges functions. I've reverted those instances of foreach
// to normal for loops for now.

///////////////////////////////////////////////////////////////////////////
//////                        LOCAL TYPEDEFS                        ///////
///////////////////////////////////////////////////////////////////////////

typedef std::pair<EnergyGraph::out_edge_iterator, 
		  EnergyGraph::out_edge_iterator> EdgeIterPair;
typedef std::pair<EnergyGraph::vertex_iterator, 
		  EnergyGraph::vertex_iterator> VertIterPair;

//////////////////////////////////////////////////////////////////////////
/////                       SIMPLE ENERGY GRAPH                      /////
//////////////////////////////////////////////////////////////////////////

void EnergyGraph::AddVertex(int ID, FLOAT energy)
{
	boost::add_vertex(EnergyGraphNode(ID, energy), *this);
}

void EnergyGraph::AddEdge(int ID1, int ID2, FLOAT energy)
{
	EnergyGraph::vertex_iterator v1 = FindVertexIterator(ID1),
				     v2 = FindVertexIterator(ID2);

	boost::add_edge(*v1, *v2, energy, *this);
}

void EnergyGraph::SetVertexEnergy(int ID, FLOAT newEnergy)
{
	(*this)[*FindVertexIterator(ID)].energy = newEnergy;
}

void EnergyGraph::SetEdgeEnergy(int ID1, int ID2, FLOAT newEnergy)
{
	(*this)[FindEdgeDescriptor(ID1, ID2)] = newEnergy;
}

void EnergyGraph::CapGraphAt(FLOAT cap)
{
        VertIterPair verts = boost::vertices(*this);
        EnergyGraph::vertex_iterator v;

        //Cap vertex energies
        for (v = verts.first; v != verts.second; ++v)
        {
                if ( (*this)[*v].energy > cap)
                        (*this)[*v].energy = cap;
                else if ( (*this)[*v].energy < -cap)
                        (*this)[*v].energy = -cap;
        }

        //Cap edge energies
        for (v = verts.first; v != verts.second; ++v)
        {
                EdgeIterPair outEdges = boost::out_edges(*v, *this);
                for (EnergyGraph::out_edge_iterator e = outEdges.first; e != outEdges.second; ++e)
                {
                        if ( (*this)[*e] > cap)
                                (*this)[*e] = cap;
                        else if ( (*this)[*e] < -cap)
                                (*this)[*e] = -cap;
                }
        }
}

FLOAT EnergyGraph::ComputeEnergy(std::vector<EnergyGraph::vertex_iterator> const& stateVector, 
								 bool noActivate)
{
	FLOAT result = 0.0; //bgEnergy;

	// set active flags if necessary
	if (!noActivate)
	{
		VertIterPair verts = boost::vertices(*this);
		EnergyGraph::vertex_iterator v;
		
		for(v = verts.first; v != verts.second; ++v)
			(*this)[*v].active = false;

		foreach(EnergyGraph::vertex_iterator i, stateVector)
			(*this)[*i].active = true;
	}

	foreach(EnergyGraph::vertex_iterator i, stateVector)
	{
		EnergyGraph::vertex_descriptor u = *i;
		result += (*this)[u].energy;

		// and add all its relevant edges
		EdgeIterPair outEdges = boost::out_edges(u, *this);
		for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
		{
			EnergyGraph::edge_descriptor e = *ei;

			EnergyGraph::vertex_descriptor v = boost::target(e, *this);
			if ((*this)[v].active)
				result += (*this)[e] * 0.5;
		}
	}

	return result;
}

FLOAT EnergyGraph::ComputeEnergy(std::vector<int> const& stateVector, bool noActivate)
{
	std::vector<EnergyGraph::vertex_iterator> iters;
	FindVertexIterators(stateVector, iters);
	return ComputeEnergy(iters, noActivate);
}

FLOAT EnergyGraph::DiffEnergy(EnergyGraph::vertex_iterator IDIn, 
			      EnergyGraph::vertex_iterator IDOut,
			      RuntimeStats::Pointer rstats) const
{
	unsigned long diffops = 0;	// for runtime stats

	FLOAT result = (*this)[*IDIn].energy - (*this)[*IDOut].energy;

	// add in the edges for the incoming vertex
	EdgeIterPair outEdges = boost::out_edges(*IDIn, *this);
	int count = 0;
	for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		EnergyGraph::vertex_descriptor u = boost::target(e, *this);
		if ( (*this)[u].active )
			result += (*this)[e];

		++diffops;
	}

	// subtract out the edges for the outgoing vertex
	outEdges = boost::out_edges(*IDOut, *this);
	for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		EnergyGraph::vertex_descriptor u = boost::target(e, *this);
		if ( (*this)[u].active )
			result -= (*this)[e];

		++diffops;
	}

	if (rstats.get() != NULL)
		rstats->Increment(RSTATS_SIMANNEAL_DIFFENERGY, diffops);

	return result;
}

void EnergyGraph::DeleteNodes(IDSet const& nodeIDs)
{	
	std::pair<int, int> id_set;
	foreach (id_set, nodeIDs)
	{
		for (int i = 0; i < id_set.second; ++i)
		{
			const int node = i + id_set.first;

			EnergyGraph::vertex_iterator v = FindVertexIterator(node);
			boost::clear_vertex(*v, *this);
			boost::remove_vertex(*v, *this);
		}
	}
}

EnergyGraph::vertex_iterator EnergyGraph::FindVertexIterator(int ID) const
{
	VertIterPair verts = boost::vertices(*this);
	EnergyGraph::vertex_iterator v;

	for(v = verts.first; v != verts.second; ++v)
		if ( (*this)[*v].elementID == ID )
			return v;
	
	throw std::string("FindVertex: Vertex ID not found in graph.");
}

void EnergyGraph::FindVertexIterators(std::vector<int> const& IDs, 
		                      std::vector<EnergyGraph::vertex_iterator>& iteratorsOut) const
{
	iteratorsOut.resize(IDs.size());
	std::vector< bool > found(IDs.size());
	for (int i = 0; i < (int)IDs.size(); ++i)
		found[i] = false;

	std::pair<EnergyGraph::vertex_iterator, 
		  EnergyGraph::vertex_iterator> vi = boost::vertices(*this);

	bool all_done = false;
	for (; vi.first != vi.second && !all_done; ++vi.first)
	{
		for (int i = 0; i < (int)IDs.size(); ++i)
		{
			if ( (*this)[*vi.first].elementID == IDs[i] ) 
			{
				found[i] = true;
				iteratorsOut[i] = vi.first;
			}
		}

		all_done = true;
		for (int i = 0; i < (int)IDs.size(); ++i)
			if (!found[i])
				all_done = false;
	}

	if (!all_done)
		throw std::string("FindVertex: Vertex ID not found in graph.");
}

EnergyGraph::edge_descriptor EnergyGraph::FindEdgeDescriptor( int vID1, int vID2 ) const
{
	EnergyGraph::vertex_iterator v1 = FindVertexIterator(vID1),
				     v2 = FindVertexIterator(vID2);

	EdgeIterPair outEdges = boost::out_edges(*v1, *this);
	for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		EnergyGraph::vertex_descriptor u = boost::target(e, *this);
		if ( u == *v2 )
			return e;
	}

	throw std::string("FindEdge: Edge not found.");
}

std::vector<int> EnergyGraph::GetVertexIDs() const
{
	VertIterPair verts = boost::vertices(*this);
	EnergyGraph::vertex_iterator v;
	std::vector<int> IDs;

	for (v = verts.first; v != verts.second; ++v)
		IDs.push_back((*this)[*v].elementID);

	return IDs;
}

std::vector<int> EnergyGraph::GetActiveVertexIDs() const
{
	VertIterPair verts = boost::vertices(*this);
	EnergyGraph::vertex_iterator v;
	std::vector<int> IDs;

	for (v = verts.first; v != verts.second; ++v) {
		if ((*this)[*v].active)	
			IDs.push_back((*this)[*v].elementID);
	}

	return IDs;
}

std::vector<int> EnergyGraph::GetEdgeDestinationIDs( int ID, FLOAT cutoff=0 ) const 
{
	std::vector<int> result;

	EnergyGraph::vertex_iterator v = FindVertexIterator(ID);
	EdgeIterPair outEdges = boost::out_edges(*v, *this);
	for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		EnergyGraph::vertex_descriptor u = boost::target(e, *this);
		// Given a cutoff, return destinations for edges where abs magnitude exceeds the cutoff
		if ((cutoff==0) or ( std::abs((*this)[e]) > cutoff )) {
			result.push_back( (*this)[u].elementID );
		}
	}

	return result;
}

FLOAT EnergyGraph::GetVertexEnergy( int ID ) const
{
	return (*this)[*FindVertexIterator(ID)].energy;
}

FLOAT EnergyGraph::GetEdgeEnergy( int ID1, int ID2 ) const
{
	return (*this)[FindEdgeDescriptor(ID1, ID2)];
}

std::vector< std::pair< int, FLOAT > > 
	EnergyGraph::GetVertexEdges( int ID ) const
{
	std::vector< std::pair< int, FLOAT > > result;

	EnergyGraph::vertex_iterator v = FindVertexIterator(ID);
	EdgeIterPair outEdges = boost::out_edges(*v, *this);
	for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		EnergyGraph::vertex_descriptor u = boost::target(e, *this);
		result.push_back( std::make_pair( (*this)[u].elementID, (*this)[e] ));
	}

	return result;
}

//////////////////////////////////////////////////////////////////////////
/////                        ENERGY HYPERGRAPH                       /////
//////////////////////////////////////////////////////////////////////////

void EnergyHyperGraph::AddVertex(int ID, FLOAT energy)
{
	EnergyGraph::AddVertex(ID, energy);
}

void EnergyHyperGraph::AddEdge(std::vector<int> const& IDs, FLOAT energy)
{
	EnergyGraph::vertex_descriptor u = 
		boost::add_vertex( EnergyGraphNode(-1, energy), *this );

	std::vector<EnergyGraph::vertex_iterator> iters;
	FindVertexIterators(IDs, iters);

	foreach (EnergyGraph::vertex_iterator v, iters)
		boost::add_edge(u, *v, 0.0, *this);
}

void EnergyHyperGraph::SetVertexEnergy(int ID, FLOAT newEnergy)
{
	EnergyGraph::SetVertexEnergy(ID, newEnergy);
}

void EnergyHyperGraph::SetEdgeEnergy(std::vector<int> const& IDs, FLOAT newEnergy)
{
	(*this)[FindEdgeDescriptor(IDs)].energy = newEnergy;
}

FLOAT EnergyHyperGraph::ComputeEnergy(std::vector<EnergyGraph::vertex_iterator> const& stateVector,
									  bool noActivate)
{
	FLOAT result = 0.0; //bgEnergy;

	// set active flags if necessary
	if (!noActivate)
	{
		VertIterPair verts = boost::vertices(*this);
		EnergyGraph::vertex_iterator v;
			
		for(v = verts.first; v != verts.second; ++v)
			(*this)[*v].active = false;

		foreach(EnergyGraph::vertex_iterator i, stateVector)
			(*this)[*i].active = true;
	}

	foreach(EnergyGraph::vertex_iterator i, stateVector)
	{
		EnergyGraph::vertex_descriptor u = *i;
		result += (*this)[u].energy;

		// and add all its relevant edges
		EdgeIterPair outEdges = boost::out_edges(u, *this);
		for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
		{
			EnergyGraph::edge_descriptor e = *ei;
			
			// this vertex represents a hyperedge
			EnergyGraph::vertex_descriptor v = boost::target(e, *this);
			assert( (*this)[v].elementID == -1 );
			
			// now check to see if all the nodes connected by the hyperedge are in the current state
			bool all_nodes_used = true;
			FLOAT edge_degree = 0.0;
			
			EdgeIterPair outEdges2 = boost::out_edges(v, *this);
			for (EnergyGraph::out_edge_iterator ei2 = outEdges2.first; ei2 != outEdges2.second; ++ei2)
			{
				EnergyGraph::edge_descriptor e2 = *ei2;
				
				// this vertex represents a data node
				EnergyGraph::vertex_descriptor v2 = boost::target(e2, *this);
				if ( !(*this)[v2].active )
				{
					all_nodes_used = false;
					break;
				}

				edge_degree += 1.0;
			}

			// if all nodes are present, add the energy
			if (all_nodes_used)
				result += (*this)[v].energy / edge_degree;
		}
	}

	return result;
}

FLOAT EnergyHyperGraph::ComputeEnergy(std::vector<int> const& stateVector, bool noActivate)
{
	std::vector<EnergyGraph::vertex_iterator> iters;
	FindVertexIterators(stateVector, iters);
	return EnergyHyperGraph::ComputeEnergy(iters, noActivate);
}

FLOAT EnergyHyperGraph::DiffEnergy(EnergyGraph::vertex_iterator IDIn, 
				   EnergyGraph::vertex_iterator IDOut,
				   RuntimeStats::Pointer rstats) const
{
	unsigned long diffops = 0;		// for runtime stats

	FLOAT result = (*this)[*IDIn].energy - (*this)[*IDOut].energy;

	// add in the edges for the incoming vertex
	EdgeIterPair outEdges = boost::out_edges(*IDIn, *this);
	for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;

		// this vertex represents a hyperedge
		EnergyGraph::vertex_descriptor v = boost::target(e, *this);
		assert( (*this)[v].elementID == -1 );

		// check to see if all the *other* nodes connected by the hyperedge are in the current state
		bool all_nodes_used = true;
		EdgeIterPair outEdges2 = boost::out_edges(v, *this);
		for (EnergyGraph::out_edge_iterator ei2 = outEdges2.first; ei2 != outEdges2.second; ++ei2)
		{
			++diffops;
			EnergyGraph::edge_descriptor e2 = *ei2;
		
			// this vertex represents a data node
			EnergyGraph::vertex_descriptor v2 = boost::target(e2, *this);
			if ( !(*this)[v2].active && v2 != *IDIn )
			{
				all_nodes_used = false;
				break;
			}
		}

		// if all nodes are present, add the energy
		if (all_nodes_used)
			result += (*this)[v].energy;
	}

	// subtract out the edges for the outgoing vertex
	outEdges = boost::out_edges(*IDOut, *this);
	for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::edge_descriptor e = *ei;
		
		// this vertex represents a hyperedge
		EnergyGraph::vertex_descriptor v = boost::target(e, *this);
		assert( (*this)[v].elementID == -1 );

		// check to see if all the nodes connected by the hyperedge are in the current state
		bool all_nodes_used = true;
		EdgeIterPair outEdges2 = boost::out_edges(v, *this);
		for (EnergyGraph::out_edge_iterator ei2 = outEdges2.first; ei2 != outEdges2.second; ++ei2)
		{
			++diffops;
			EnergyGraph::edge_descriptor e2 = *ei2;
		
			// this vertex represents a data node
			EnergyGraph::vertex_descriptor v2 = boost::target(e2, *this);
			if ( !(*this)[v2].active )
			{
				all_nodes_used = false;
				break;
			}
		}

		// if all nodes are present, subtract the energy
		if (all_nodes_used)
			result -= (*this)[v].energy;
	}

	if (rstats.get() != NULL)
		rstats->Increment(RSTATS_SIMANNEAL_DIFFENERGY_HG, diffops);
	
	return result;
}

void EnergyHyperGraph::DeleteNodes(IDSet const& nodeIDs)
{	
	std::pair<int, int> id_set;
	foreach (id_set, nodeIDs)
	{
		for (int i = 0; i < id_set.second; ++i)
		{
			const int node = i + id_set.first;

			EnergyGraph::vertex_iterator v = FindVertexIterator(node);
			
			EdgeIterPair outEdges = boost::out_edges(*v, *this);
			for(EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
			{
				EnergyGraph::edge_descriptor e = *ei;
				
				EnergyGraph::vertex_descriptor v2 = boost::target(e, *this);
				assert( (*this)[v2].elementID == -1 );
			
				boost::clear_vertex(v2, *this);
				boost::remove_vertex(v2, *this);
			}

			boost::clear_vertex(*v, *this);
			boost::remove_vertex(*v, *this);
		}
	}
}

FLOAT EnergyHyperGraph::GetVertexEnergy( int ID ) const
{
	return EnergyGraph::GetVertexEnergy(ID);
}

std::vector<int> EnergyHyperGraph::GetVertexIDs() const
{
        VertIterPair verts = boost::vertices(*this);
        EnergyGraph::vertex_iterator v;
        std::vector<int> IDs;

        for (v = verts.first; v != verts.second; ++v)
                IDs.push_back((*this)[*v].elementID);

        return IDs;
}


FLOAT EnergyHyperGraph::GetEdgeEnergy( std::vector<int> const& IDs ) const
{
	return (*this)[FindEdgeDescriptor(IDs)].energy;	
}

std::vector< std::pair< std::vector<int>, FLOAT > > 
	EnergyHyperGraph::GetVertexEdges( int ID ) const
{
	std::vector< std::pair< std::vector<int>, FLOAT > > result;

	EnergyGraph::vertex_iterator v = FindVertexIterator(ID);
	EdgeIterPair outEdges = boost::out_edges(*v, *this);
	for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
	{
		EnergyGraph::vertex_descriptor u = boost::target(*ei, *this);
		std::vector<int> IDlist;

		EdgeIterPair outEdges2 = boost::out_edges(u, *this);
		for (EnergyGraph::out_edge_iterator ei2 = outEdges2.first; ei2 != outEdges2.second; ++ei2)
		{
			EnergyGraph::vertex_descriptor w = boost::target(*ei2, *this);
			if ( (*this)[w].elementID != ID )
				IDlist.push_back( (*this)[w].elementID );
		}

		result.push_back( std::make_pair( IDlist, (*this)[u].energy ) );
	}

	return result;
}

EnergyGraph::vertex_descriptor 
	EnergyHyperGraph::FindEdgeDescriptor( std::vector<int> const& IDs ) const
{
	VertIterPair verts = boost::vertices(*this);
	EnergyGraph::vertex_iterator v;

	for(v = verts.first; v != verts.second; ++v)
		if ( (*this)[*v].elementID == -1 )
		{
			std::vector< bool > found(IDs.size());
			for (int i = 0; i < (int)IDs.size(); ++i)
				found[i] = false;

			EdgeIterPair outEdges = boost::out_edges(*v, *this);
			for (EnergyGraph::out_edge_iterator ei = outEdges.first; ei != outEdges.second; ++ei)
			{
				EnergyGraph::edge_descriptor e = *ei;
				EnergyGraph::vertex_descriptor u = boost::target(e, *this);
			
				bool is_in_ids = false;
				for (int i = 0; i < (int)IDs.size(); ++i)
				{
					if ( (*this)[u].elementID == IDs[i] ) 
					{
						found[i] = true;
						is_in_ids = true;
						break;
					}
				}
				
				if (!is_in_ids)
				{
					found[0] = false;
					break;
				}
			}

			bool all_found = true;
			for (int i = 0; i < (int)IDs.size(); ++i)
				if (!found[i])
					all_found = false;

			if (all_found)
				return *v;
		}

	throw std::string("FindEdge: Edge not found.");
}
	

