// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "neighbor_map.h"

#include <fstream>
#include <cassert>

#include <blitz/tinyvec-et.h>

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

NeighborMap::Pointer NeighborMap::Create(std::vector<System::ConstPointer> const& systems, 
					 blitz::Array<FLOAT, 2> const& cutoffs)
{
	return NeighborMap::Pointer(new NeighborMap(systems, cutoffs));
}

NeighborMap::Pointer NeighborMap::Create(System::ConstPointer system, blitz::Array<FLOAT, 2> const& cutoffs)
{
	std::vector<System::ConstPointer> temp;
	temp.push_back(system);
	
	return NeighborMap::Pointer(new NeighborMap(temp, cutoffs));
}

NeighborMap::NeighborMap (std::vector<System::ConstPointer> const& systems, 
			  blitz::Array<FLOAT, 2> const& cutoffs)
	: data_(systems, false), nNeighbors_(systems, 1)
{
	std::vector<System::ConstPointer>::const_iterator sys1, sys2;

	const FLOAT dist2_cutoff = 100.0;

	for (sys1 = systems.begin(); sys1 != systems.end(); ++sys1)
	{
		blitz::Array<int, 1>& nNeighbors1 = nNeighbors_(*sys1);

		for (sys2 = sys1; sys2 != systems.end(); ++sys2)
		{
			blitz::Array<int, 1>& nNeighbors2 = nNeighbors_(*sys2);
			blitz::Array<bool, 2>& data = data_(*sys1, *sys2);

			foreach (Residue const& res1, **sys1)
				foreach (Residue const& res2, **sys2)
				{
					const int storageIndex1 = (*sys1)->GetStorageIndex(res1);
					const int storageIndex2 = (*sys2)->GetStorageIndex(res2);

					if (sys1 == sys2 && res1.GetID() == res2.GetID())
					{
						data( storageIndex1, storageIndex2 ) = false;
						continue;
					}
				
					const FLOAT pc1 = cutoffs((int)res1.GetAminoAcid().aminoAcid, 
								  (int)res2.GetAminoAcid().aminoAcid ) + 1.0;

					Vector3 const& v1 = res1.GetAtomCoords()[res1.GetAminoAcid().neighborAtom];
					Vector3 const& v2 = res2.GetAtomCoords()[res2.GetAminoAcid().neighborAtom];

					const Vector3 v12 = v1 - v2;

					const FLOAT dist2 = blitz::dot(v12, v12);

					data( storageIndex1, storageIndex2 ) = (dist2 <= pc1 * pc1);
					
					if (dist2 < dist2_cutoff)
					{
						++nNeighbors1( storageIndex1 );
						if (sys1 != sys2)
							++nNeighbors2( storageIndex2 );
					}
				}
		}
	}
}

///////////////////////////////////////////////////////////////////////////
//////                           ACCESSORS                           //////
///////////////////////////////////////////////////////////////////////////

bool NeighborMap::AreNeighbors (Residue const& res1, Residue const& res2) const
{
	try
	{
		System::ConstPointer s1 = res1.GetSystem(), s2 = res2.GetSystem();
		return data_(s1, s2, s1->GetStorageIndex(res1), s2->GetStorageIndex(res2));
	}
	catch (std::string error)
	{
		throw std::string("AreNeighbors: Input residues are not in this NeighborMap.");
	}
}

int NeighborMap::NumNeighbors (Residue const& res) const
{
	try
	{
		System::ConstPointer s = res.GetSystem();
		return nNeighbors_(s, s->GetStorageIndex(res));
	}
	catch (std::string error)
	{
		throw std::string("NumNeighbors: Input residue is not in this NeighborMap.");
	}
}

///////////////////////////////////////////////////////////////////////////
//////                              I/O                              //////
///////////////////////////////////////////////////////////////////////////

std::ostream& operator<< (std::ostream& os, NeighborMap::ConstPointer map)
{
	return os << "N_NEIGHBORS:\n" << map->nNeighbors_ << "NEIGHBOR_MAP:\n" << map->data_ << std::endl;
}

std::istream& operator>> (std::istream& is, NeighborMap::Pointer& map)
{
	std::string dummy;
	is >> dummy;
	is >> map->nNeighbors_;
	is >> dummy;
	is >> map->data_;
	return is;
}

std::ostream& operator<< (std::ostream& os, NeighborMap::Pointer map)
{
	return os << NeighborMap::ConstPointer(map);
}


