// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Chris Kennelly 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "opt_base_simanneal.h"
#include <ctime>

typedef boost::variate_generator<boost::mt19937&, boost::uniform_real<> > RealGenerator;
typedef boost::variate_generator<boost::mt19937&, boost::uniform_int<> > IntGenerator;

///////////////////////////////////////////////////////////////////////////
//////                      PARAMETER ACCESSORS                      //////
///////////////////////////////////////////////////////////////////////////

std::vector<FLOAT> OptimizationBaseSimulatedAnnealing::GetSchedule() const
{
	return schedule_;
}

void OptimizationBaseSimulatedAnnealing::SetSchedule(std::vector<FLOAT> const& schedule)
{
	schedule_ = schedule;

	lowTemp_  = *std::min(schedule.begin(), schedule.end());
	highTemp_ = *std::max(schedule.begin(), schedule.end());
}

FLOAT OptimizationBaseSimulatedAnnealing::GetJumpCutoff() const
{
	return jumpCutoff_;
}

void OptimizationBaseSimulatedAnnealing::SetJumpCutoff(FLOAT cutoff)
{
	// FIXME: unreasonable cutoffs can cause infinite loops
	jumpCutoff_ = cutoff;
}

FLOAT OptimizationBaseSimulatedAnnealing::GetIterationMultiplier() const
{
	return iterMultiplier_;
}

void OptimizationBaseSimulatedAnnealing::SetIterationMultiplier(FLOAT multiplier)
{
	// FIXME: add validation here
	iterMultiplier_ = multiplier;
}

///////////////////////////////////////////////////////////////////////////
//////                        IMPLEMENTATION                        ///////
///////////////////////////////////////////////////////////////////////////

OptimizationBaseSimulatedAnnealing::OptimizationBaseSimulatedAnnealing()
	: OptimizationBase()
{
	lowTemp_	= 0.3;
	highTemp_	= 100.0;
	jumpCutoff_	= -1.0;

	// initialize with a default schedule
	for (int i = 0; i < 20; ++i)
		schedule_.push_back( (highTemp_ - lowTemp_) * std::exp(-float(i)) + lowTemp_ );

	iterMultiplier_ = 5.0;
}

FLOAT OptimizationBaseSimulatedAnnealing::GetTemperature(bool quench)
{
	if (quench) 
	{
		scheduleIndex_++;
		return lowTemp_;
	}

	if (scheduleIndex_ >= 4)
	{
		int epoch = (int)energies_.size();

		FLOAT avgEnergy = (energies_[epoch - 4] + energies_[epoch - 3] + energies_[epoch - 2]) / 3.0;
		if (energies_[epoch - 1] - avgEnergy > jumpCutoff_)
		{
			scheduleIndex_ = 1;
			return schedule_.front();
		}
	}

	return schedule_[scheduleIndex_++];
}

std::pair<int, int> OptimizationBaseSimulatedAnnealing::PickMove(bool quench)
{
	if (quench)
	{
		std::pair<int, int> IDRange = searchSpace_->DimensionIDRange( freeDimList_[quenchDim_] );

		if (quenchID_ == IDRange.second - IDRange.first)
		{
			if (++quenchDim_ == (int)freeDimList_.size())
				quenchDim_ = 0;
			quenchID_ = 0;
 		}

		return std::make_pair( freeDimList_[quenchDim_], quenchID_++);
	}
	else
	{
		IntGenerator dimGen(mtrand_, boost::uniform_int<>(0, (unsigned int)freeDimList_.size() - 1));
		int randomDim = dimGen();
		
		std::pair<int, int> IDRange = searchSpace_->DimensionIDRange( freeDimList_[randomDim] );

		IntGenerator idGen(mtrand_, boost::uniform_int<>(0, IDRange.second - IDRange.first));
		return std::make_pair( freeDimList_[randomDim], idGen() );
	}
}

bool OptimizationBaseSimulatedAnnealing::KeepChange(FLOAT prevEnergy, FLOAT diff, FLOAT temperature, bool quench)
{
	FLOAT lnprob = 0.0;
	FLOAT beta = 1.0/temperature;

	if (diff < 0.0)
		return true;

	if (!quench)
	{
		lnprob = beta * diff;
			
		if (lnprob < 10.0)
		{
			FLOAT prob = std::exp(-lnprob);

			RealGenerator pgen(mtrand_, boost::uniform_real<>(0, 1));
			if (pgen() < prob)
				return true;
		}
	}

	return false;
}

