// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "opt_faster.h"
#include <ctime>
#include <stdio.h>
#include <algorithm>

///////////////////////////////////////////////////////////////////////////
//////                      PARAMETER ACCESSORS                      //////
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//////                        IMPLEMENTATION                        ///////
///////////////////////////////////////////////////////////////////////////

FasterPacker::FasterPacker()
	: OptimizationBase()
{
	Randomize_	= true;
	SPRrounds_	= 2;
	SimultaneousFixes_ = 1;
	EdgeCutoff_ = 1;
}

std::pair< std::vector<int>, FLOAT >
	FasterPacker::Pack(
                    DiscreteSearchSpace::ConstPointer searchSpace,
                    EnergyGraph* graph, 
		    std::vector<int> const* initialState,
                    EnergyCombList::Pointer energyList, int listSize)
{
        // Check on the energy list - should be in reverse order
        if ((energyList.get() != NULL) && (!(*energyList).isDescOrder()))
            (*energyList).reverse();

	// Optimization starts from an optional initialState
	searchSpace_ = searchSpace;
	InitState(*graph, initialState? *initialState : std::vector<int>());

	// Energies is a list of all the best energies encountered
	//energies_.clear();  
	//energies_.push_back(graph->ComputeEnergy(stateVector_));

        // Keep the best solution seen so far
	//FLOAT bestE = energies_.back();
	FLOAT bestE = graph->ComputeEnergy(stateVector_);
        
        // Vector for storing temporary combinatorial states
        std::vector<int> tempVec;

        // If asked to store a list of solutions (EnergyCombList), add the first solution to the list
        if (energyList.get() != NULL) {
            // Populate a vector with the element ID numbers for each energy graph node
            foreach(EnergyGraph::vertex_iterator i, stateVector_)
                tempVec.push_back((*graph)[*i].elementID);

            // Add the vector to the energy combination list
            //(*energyList).push(EnergyComb(tempVec, energies_.back()));
            (*energyList).push(EnergyComb(tempVec, bestE));
        }

	///////////////////////////////////////////////////////////////////////////////////////////////////
	//////////   Lookups for converting rotamers and  residues : first_rot, last_rot, rot2res   ///////
	///////////////////////////////////////////////////////////////////////////////////////////////////

	// Convenient to have a vector of int nodes rather than vertex iterators
	std::vector<int> Nodes = graph->GetVertexIDs();
	int finalrot = Nodes.back();

	// Optionally, shuffle the order
	if (Randomize_) { std::random_shuffle(Nodes.begin(),Nodes.end()); }

	// Store the first rotamer for each residue	
	std::vector<int> first_rot;
	foreach(EnergyGraph::vertex_iterator rot, stateIDStart_) { 
		first_rot.push_back((*graph)[*rot].elementID);
	}

	// Store the last rotamer for each residue	
	std::vector<int> last_rot;
	for(unsigned int z=1; z<first_rot.size(); z++) {
		last_rot.push_back(first_rot[z] - 1);
	} 
	last_rot.push_back(finalrot);

	// Verbose check of correct rotamer ranges
	//for(unsigned int resid=0;resid<first_rot.size();resid++) {
	//	std::cout << "Residue index " << resid << ": Searching rotamers " << first_rot[resid] <<"-"<< last_rot[resid] << "\n";
	//}


	// Construct an array for easy lookup of a residue index given the rotamer
	std::vector<int> rot2res;
	for(unsigned int i=0;i <= Nodes.size() + first_rot[0];i++) {  // tricky line
		int resindex = -1;
		// For each rotamer, run through each residue's starting rotamer to find the correct one 
		foreach(unsigned int startrot, first_rot) {
			if (startrot > i) { break; }
			resindex += 1;		
		} 
		//std::cout << i << " " << resindex << "\n";
		rot2res.push_back(resindex);
	}
	//foreach(int n, rot2res) {
	//	std::cout << n << " ";
	//} 
	//std::cout << "\n";


	///////////////////////////////////////////////////
	//////////   Setup variables for SPR   ////////////
	///////////////////////////////////////////////////

        // Populate a vector with the element ID numbers for each energy graph node
	std::vector<int> combo;
        foreach(EnergyGraph::vertex_iterator i, stateVector_)
            combo.push_back((*graph)[*i].elementID);
	FLOAT currentE = graph->ComputeEnergy(combo);  // Batch moves must improve upon currentE
	//std::cout << "Original E " << currentE << "\n";
	std::vector<int> bestcombo = combo; // Any successful batch moves get stored in bestcombo


	 
	for (int SPRround=0;SPRround < SPRrounds_;SPRround++) {
		//std::cout << "SPR round "<<SPRround<<" =========================================================\n";

		/////////////////////////////////////////////////////////
		//  Perturb the system by fixing each rotamer in turn  //
		/////////////////////////////////////////////////////////

		foreach(int n, Nodes) {
		//for(int n=50;n<=55;n++) {

			std::vector<int> perturb = combo;

			// Construct vertex iterators for the before and after perturbation
			EnergyGraph::vertex_iterator unperturbedVertex = stateVector_[rot2res[n]];
			EnergyGraph::vertex_iterator perturbedVertex   = stateIDStart_[rot2res[n]];
			while ((*graph)[*perturbedVertex].elementID < n) { ++perturbedVertex; } 
			if (unperturbedVertex==perturbedVertex) { continue; }

                        (*graph)[*perturbedVertex].active = true;
                        (*graph)[*unperturbedVertex].active = false;
			//std::cout << "\nPerturb res " << rot2res[n] << " from rot " << (*graph)[*unperturbedVertex].elementID
			//std::cout << " to " << (*graph)[*perturbedVertex].elementID << "\n";

			perturb[rot2res[n]] = n;  // Make the rotamer perturbation
			std::vector<int> batchmove = perturb; // To aggregate best rotamers into a batch move 


			/////////////////////////////////////////////
			//  Identify interacting partner residues  //
			/////////////////////////////////////////////

			// Find unique interacting residues: TODO: add cutoff parameter to ignore small interaction partners 
                        std::vector<int> partner_rots = graph->GetEdgeDestinationIDs(n,EdgeCutoff_);
                        std::vector<int> partners;
			foreach(int v, partner_rots) {
				partners.push_back(rot2res[v]);
			} 
			std::vector<int>::iterator partner_it;
			partner_it = std::unique(partners.begin(),partners.end());
			partners.resize( partner_it - partners.begin() );
			//foreach(int v, partners) { std::cout << v << " "; }
			//std::cout << "\n";

			// Optionally, shuffle the order of partners 
			if (Randomize_) { std::random_shuffle(partners.begin(),partners.end()); }
			

			//////////////////////////////////////////////////////////////
			//  Find best rotamer for any interacting partner residues  //
			//////////////////////////////////////////////////////////////

			foreach(int resn, partners) {
				//std::cout << "Residue " << resn << ": rotamer options " << first_rot[resn] <<"-"<< last_rot[resn] << "\n";
				EnergyGraph::vertex_iterator oldVertex = stateVector_[resn];
                		EnergyGraph::vertex_iterator newVertex = stateIDStart_[resn];
                		EnergyGraph::vertex_iterator bestVertex = stateVector_[resn];
				FLOAT bestrotdiff = 0;

				// Walk through the rotamer alternatives
				for(int rotID=first_rot[resn];rotID <= last_rot[resn];rotID++) { 
                			if (newVertex == oldVertex) {
						++newVertex;
                        			continue;
					}
                			FLOAT diff = graph->DiffEnergy(newVertex, oldVertex);
					//std::cout << "Trial switch from " << (*graph)[*oldVertex].elementID << " to ";
					//std::cout << (*graph)[*newVertex].elementID << " dE = " << diff << "\n";
					if (diff < bestrotdiff) {
						bestVertex = newVertex;	
						bestrotdiff = diff;
						batchmove[resn] = rotID;
					}

					++newVertex;
				}
				//if (bestrotdiff!=0) {
				//    std::cout << "Best rotamer relaxation (res "<<resn<<"): " << bestrotdiff << " for ";
				//    std::cout << (*graph)[*oldVertex].elementID << " => " << (*graph)[*bestVertex].elementID << "\n";
				//}
			}


			//////////////////////////////////////////////////////////////////
			//  Determine whether to accept or reject the batch relaxation  //
			//////////////////////////////////////////////////////////////////

			FLOAT batchE = graph->ComputeEnergy(batchmove);
			if (batchE < currentE) {
				//std::cout << "Accepting batch move " << currentE << " => " << batchE << "\n";
				currentE = batchE;
				bestcombo = batchmove;
				bestE = batchE;

				// For each change in batchmove, update the graph and the stateVector
				for(uint i=0;i<batchmove.size();i++) {
					if (batchmove[i]!=combo[i]) {
						EnergyGraph::vertex_iterator unrelaxedVertex = stateVector_[i];
						EnergyGraph::vertex_iterator relaxedVertex   = stateIDStart_[i];
						while ((*graph)[*relaxedVertex].elementID < batchmove[i]) { ++relaxedVertex; } 
                        			(*graph)[*relaxedVertex].active = true;
                        			(*graph)[*unrelaxedVertex].active = false;
						//std::cout << "Switching res " << i << " from rot " << (*graph)[*unrelaxedVertex].elementID << " to " << (*graph)[*relaxedVertex].elementID << "\n";
						stateVector_[i] = relaxedVertex;
					}
				} 


				// If maintaining a list of solutions, add to this list
        			if (energyList.get() != NULL) {
        			    // Populate a vector with the element ID numbers for each energy graph node
                               	    tempVec.clear();
        			    foreach(EnergyGraph::vertex_iterator i, stateVector_)
        			        tempVec.push_back((*graph)[*i].elementID);
	
        			    // Add the vector to the energy combination list
        			    //(*energyList).push(EnergyComb(tempVec, energies_.back()));
        			    (*energyList).push(EnergyComb(tempVec, bestE));

	        	            // Trim the list if we're over the limit
                	       	    if ((*energyList).size() > listSize) {
        			        (*energyList).pop();
				    }
        			}

				combo = batchmove;
			}
			else {  // Rejecting the batch move, discard the graph perturbation
				//std::cout << "Rejecting batch move " << currentE << " => " << batchE << "\n";
                        	(*graph)[*perturbedVertex].active = false;
                        	(*graph)[*unperturbedVertex].active = true;
			}
		} 

	}				

        // If applicable, reverse the list
        if (energyList.get() != NULL) {
            (*energyList).reverse();
	}

	return std::make_pair(bestcombo, bestE);
}

void FasterPacker::SetEdgeCutoff(FLOAT cutoff) {
	EdgeCutoff_ = cutoff;
}

FLOAT FasterPacker::GetEdgeCutoff() {
	return EdgeCutoff_;
}

void FasterPacker::SetRandomize(bool Randomize) {
	Randomize_ = Randomize;
}

bool FasterPacker::GetRandomize() {
	return Randomize_;
}

void FasterPacker::SetSPRrounds(int roundCount) {
	// IMPLEMENT some form of verifications
	SPRrounds_ = roundCount;
}

int FasterPacker::GetSPRrounds() {
	return SPRrounds_;
}

void FasterPacker::SetSimultaneousFixes(int fixCount) {
	// IMPLEMENT some form of verifications
	SimultaneousFixes_ = fixCount;
}

int FasterPacker::GetSimultaneousFixes() {
	return SimultaneousFixes_;
}

