// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Chris Kennelly 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "opt_montecarlo_fbf.h"
#include <ctime>

typedef boost::variate_generator<boost::mt19937&, boost::uniform_real<> > RealGenerator;
typedef boost::variate_generator<boost::mt19937&, boost::uniform_int<> > IntGenerator;

///////////////////////////////////////////////////////////////////////////
//////                      PARAMETER ACCESSORS                      //////
///////////////////////////////////////////////////////////////////////////

int MonteCarloFBFPacker::GetMaxFrames()
{
	return maxFrames_;
}

void MonteCarloFBFPacker::SetMaxFrames(int maxFrames)
{
	maxFrames_	= maxFrames;
}


FLOAT MonteCarloFBFPacker::GetIterationMultiplier() const
{
	return iterMultiplier_;
}

void MonteCarloFBFPacker::SetIterationMultiplier(FLOAT multiplier)
{
	// FIXME: add validation here
	iterMultiplier_ = multiplier;
}

FLOAT MonteCarloFBFPacker::GetTemperature()
{
	return temperature_;
}

void MonteCarloFBFPacker::SetTemperature(FLOAT temperature)
{
	temperature_ = temperature;
}

///////////////////////////////////////////////////////////////////////////
//////                        IMPLEMENTATION                        ///////
///////////////////////////////////////////////////////////////////////////

MonteCarloFBFPacker::MonteCarloFBFPacker()
	: temperature_(1), mtrand_(42u)
{
	iterMultiplier_ = 5.0;
	SeedWithTime();
}

void MonteCarloFBFPacker::InitState(EnergyGraph& graph, std::vector<int> const& initialState)
{
	stateVector_.clear();
	stateIDStart_.clear();
	freeDimList_.clear();
	
	stateVector_.resize(searchSpace_->NumDimensions());
	stateIDStart_.resize(searchSpace_->NumDimensions());

	std::vector<int>::const_iterator initialID = initialState.begin();

	for (int d = 0; d < searchSpace_->NumDimensions(); ++d)
	{
		std::pair<EnergyGraph::vertex_iterator, 
			  EnergyGraph::vertex_iterator> vi = boost::vertices(graph);
	
		std::pair<int, int> IDRange = searchSpace_->DimensionIDRange(d);

		for (; vi.first != vi.second; ++vi.first)
		{
			if ( graph[*vi.first].elementID == IDRange.first ) 
			{
				stateIDStart_[d] = vi.first;
				break;
			}
		}

		int newIDOffset;
		if (initialID != initialState.end())
		{
			if (*initialID < IDRange.first ||
			    *initialID > IDRange.second)
			    throw std::string("Invalid initial element ID.");

			newIDOffset = *initialID - IDRange.first;
			++initialID;
		}
		else
		{
			IntGenerator igen(mtrand_, boost::uniform_int<>(0, IDRange.second - IDRange.first));
			newIDOffset = igen();
		}

		stateVector_[d] = stateIDStart_[d];
		for (int i = 0; i < newIDOffset; ++i)
			++stateVector_[d];

		if (IDRange.second > IDRange.first)
			freeDimList_.push_back(d);
	}

	if (freeDimList_.empty())
		throw std::string("MonteCarloFBFPacker error: No free dimensions.");
}

std::pair<int, int> MonteCarloFBFPacker::PickMove()
{
	IntGenerator dimGen(mtrand_, boost::uniform_int<>(0, (unsigned int)freeDimList_.size() - 1));
	int randomDim = dimGen();
		
	std::pair<int, int> IDRange = searchSpace_->DimensionIDRange( freeDimList_[randomDim] );

	IntGenerator idGen(mtrand_, boost::uniform_int<>(0, IDRange.second - IDRange.first));
	return std::make_pair( freeDimList_[randomDim], idGen() );
}


bool MonteCarloFBFPacker::KeepChange(FLOAT prevEnergy, FLOAT diff, FLOAT temperature)
{

	if (diff < 0.0)
		return true;

	if (temperature < 0.001)
		return false;

	FLOAT lnprob = 0.0;
	FLOAT beta = 1.0/temperature;
	lnprob = beta * diff;
	
	//if ( prevEnergy > 1.0 )
	//	lnprob /= prevEnergy;
	
	if (lnprob < 10.0)
	{
		FLOAT prob = std::exp(-lnprob);

		RealGenerator pgen(mtrand_, boost::uniform_real<>(0, 1));
		if (pgen() < prob)
			return true;
	}

	return false;
}


std::pair< std::vector<int>, FLOAT >
	MonteCarloFBFPacker::Pack(
                    DiscreteSearchSpace::ConstPointer searchSpace,
                    EnergyGraph* graph, 
		    std::vector<int> const* initialState,
                    EnergyCombList::Pointer energyList, int listSize)
{
        // Check on the energy list - should be in reverse order
        if ((energyList.get() != NULL) && (!(*energyList).isDescOrder()))
            (*energyList).reverse();

	searchSpace_ = searchSpace;
	energies_.clear();

	const int N_ITERATIONS	= (int)(searchSpace_->NumElements() * iterMultiplier_);
	
	InitState(*graph, initialState? *initialState : std::vector<int>());
	energies_.push_back(graph->ComputeEnergy(stateVector_));

        // Keep the best solution seen so far
	std::vector<EnergyGraph::vertex_iterator> bestState = stateVector_;
	FLOAT bestEnergy = energies_.back();
        
        // vector for storing temporary states
        std::vector<int> tempVec;

        // Add the first solution to the list if necceary
        if (energyList.get() != NULL) {
            // populate the vector
            foreach(EnergyGraph::vertex_iterator i, stateVector_)
                tempVec.push_back((*graph)[*i].elementID);

            // add the solution to the list
            (*energyList).push(EnergyComb(tempVec, energies_.back()));
        }

	// record runtime stats, if necessary
	if (stats_.get() != NULL) {
		stats_->Increment(RSTATS_MONTECARLO_ITER, N_ITERATIONS);
	}

	FLOAT temperature = GetTemperature();
	int frame = 1;	

	for (int i = 0; i < N_ITERATIONS; ++i)
	{
		std::pair<int, int> move = PickMove();
			
		EnergyGraph::vertex_iterator newVertex = stateIDStart_[move.first];
		EnergyGraph::vertex_iterator oldVertex = stateVector_[move.first];
		for (int j = 0; j < move.second; ++j)
			++newVertex;
		
		if (newVertex == oldVertex)
			continue;

		FLOAT diff = graph->DiffEnergy(newVertex, oldVertex);
		if (KeepChange(energies_.back(), diff, temperature) || frame == maxFrames_)
		{
			energies_.back() += diff;
		
			(*graph)[*newVertex].active = true;
			(*graph)[*oldVertex].active = false;

			stateVector_[move.first] = newVertex;	
				
                        // Add to the list if necessary
                        if (energyList.get() != NULL) {
                        	// remove old state
                                tempVec.clear();

                                // format the vector
                                foreach(EnergyGraph::vertex_iterator i, stateVector_)
                                    tempVec.push_back((*graph)[*i].elementID);

                                // push the solution
                                (*energyList).push(EnergyComb(tempVec, energies_.back()));

                                // pop if we're over the limit
                                if ((*energyList).size() > listSize)
                                    (*energyList).pop();
               		}

			// keep the new best energy
                        if (energies_.back() < bestEnergy)
			{
				bestEnergy = energies_.back();
				bestState = stateVector_;
			}

			// grab current solution as "the best" if leaving...
			if (frame == maxFrames_) {
				bestEnergy = energies_.back();
				bestState = stateVector_;

				break;
			}
		}

		frame++;

		// grabbing current best
		if (frame == maxFrames_) {		
			break;
		}

		frame++;
	}

	energies_.push_back( energies_.back() );
       
	// update best energy
	bestEnergy	= graph->ComputeEnergy(bestState);

        // format the best solution for output
	std::vector<int> result;
	foreach(EnergyGraph::vertex_iterator i, bestState)
		result.push_back((*graph)[*i].elementID);

        // reverse the list if necessary
        if (energyList.get() != NULL)
            (*energyList).reverse();

	return std::make_pair(result, bestEnergy);
}

void MonteCarloFBFPacker::Seed( unsigned int seed )
{
	mtrand_.seed( (boost::mt19937::result_type)seed );
}

void MonteCarloFBFPacker::SeedWithTime()
{
	mtrand_.seed( (unsigned int)time(0) );
}


void MonteCarloFBFPacker::RecordRuntimeStats(RuntimeStats::Pointer stats) {
	stats_	= stats;
}

