// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Chris Kennelly 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "opt_montecarlo_mg.h"
#include <ctime>

///////////////////////////////////////////////////////////////////////////
//////                      PARAMETER ACCESSORS                      //////
///////////////////////////////////////////////////////////////////////////

void MGMonteCarloPacker::AddGraph(EnergyGraph* graph) {
	graphs_.push_back(graph);
}

void MGMonteCarloPacker::ClearGraphs() {
	graphs_.clear();
}

///////////////////////////////////////////////////////////////////////////
//////                        IMPLEMENTATION                        ///////
///////////////////////////////////////////////////////////////////////////

MGMonteCarloPacker::MGMonteCarloPacker()
	: OptimizationBaseMonteCarlo()
{
	graphs_ = std::vector<EnergyGraph*>();
}

std::pair< std::vector<int>, FLOAT >
	MGMonteCarloPacker::Pack(
                    DiscreteSearchSpace::ConstPointer searchSpace,
                    /* EnergyGraph* graph, */
		    std::vector<int> const* initialState,
                    EnergyCombList::Pointer energyList, int listSize)
{
	// IMPLEMENT: Perform a sanity check on the list of graphs
	int numberOfGraphs = graphs_.size();

	// Grab the 1st graph for use in situations unrelated to the energy
	EnergyGraph* graph = graphs_ [0];

        // Check on the energy list - should be in reverse order
        if ((energyList.get() != NULL) && (!(*energyList).isDescOrder()))
            (*energyList).reverse();

	searchSpace_ = searchSpace;
	energies_.clear();

	const int N_ITERATIONS	= (int)(searchSpace_->NumElements() * iterMultiplier_);
	
	InitState(*graph, initialState? *initialState : std::vector<int>());

	FLOAT highestEnergy;
	int highestEnergyGraph;

	for (int currentGraph = 0; currentGraph < numberOfGraphs; currentGraph++) 
	{
		EnergyGraph* workingGraph = graphs_ [currentGraph];
		FLOAT workingEnergy = workingGraph->ComputeEnergy(stateVector_);

		if (currentGraph == 0 || highestEnergy < workingEnergy) {
			highestEnergy = workingEnergy;
			highestEnergyGraph = currentGraph;
		}
	}

	energies_.push_back(highestEnergy);

        // Keep the best solution seen so far
	std::vector<EnergyGraph::vertex_iterator> bestState = stateVector_;
	FLOAT bestEnergy = highestEnergy;
	int bestEnergyGraph = highestEnergyGraph;
        
        // vector for storing temporary states
        std::vector<int> tempVec;

        // Add the first solution to the list if necceary
        if (energyList.get() != NULL) {
            // populate the vector
            foreach(EnergyGraph::vertex_iterator i, stateVector_)
                tempVec.push_back((*graph)[*i].elementID);

            // add the solution to the list
            (*energyList).push(EnergyComb(tempVec, energies_.back()));
        }

	// record runtime stats, if necessary
	if (stats_.get() != NULL) {
		stats_->Increment(RSTATS_MONTECARLO_ITER, N_ITERATIONS);
	}

	FLOAT temperature = GetTemperature();
	std::vector<EnergyGraph::vertex_iterator> trialVector = std::vector<EnergyGraph::vertex_iterator>();
	
	for (int i = 0; i < N_ITERATIONS; ++i)
	{
		std::pair<int, int> move = PickMove();
			
		EnergyGraph::vertex_iterator newVertex = stateIDStart_[move.first];
		EnergyGraph::vertex_iterator oldVertex = stateVector_[move.first];
		for (int j = 0; j < move.second; ++j)
			++newVertex;
		
		if (newVertex == oldVertex)
			continue;

		trialVector.clear();
		for (int currentResidue = 0; currentResidue < searchSpace_->NumDimensions(); currentResidue++) {
			if (currentResidue == move.first) {
				trialVector.push_back(newVertex);
			} else {
				trialVector.push_back(stateVector_ [currentResidue]);
			}
		}

		FLOAT trialEnergy = 0;
		int trialEnergyGraph = 0;

		for (int currentGraph = 0; currentGraph < numberOfGraphs; currentGraph++) {
			EnergyGraph* workingGraph = graphs_ [currentGraph];
			FLOAT workingEnergy = workingGraph->ComputeEnergy(trialVector);

			if (currentGraph == 0 || trialEnergy < workingEnergy) {
				trialEnergy = workingEnergy;
				trialEnergyGraph = currentGraph;
			}
		}

		FLOAT diff = (trialEnergy - energies_.back());
		if (KeepChange(energies_.back(), diff, temperature))
		{
			energies_.back() += diff;
		
			(*graph)[*newVertex].active = true;
			(*graph)[*oldVertex].active = false;

			stateVector_[move.first] = newVertex;	
				
                        // Add to the list if necessary
                        if (energyList.get() != NULL) {
                        	// remove old state
                                tempVec.clear();

                                // format the vector
                                foreach(EnergyGraph::vertex_iterator i, stateVector_)
                                    tempVec.push_back((*graph)[*i].elementID);

                                // push the solution
                                (*energyList).push(EnergyComb(tempVec, energies_.back()));

                                // pop if we're over the limit
                                if ((*energyList).size() > listSize)
                                    (*energyList).pop();
               		}

			// keep the new best energy
                        if (energies_.back() < bestEnergy)
			{
				bestEnergy = trialEnergy;
				bestState = stateVector_;
				bestEnergyGraph = trialEnergyGraph;
			}
		}
	}

	energies_.push_back( energies_.back() );
       
        // format the best solution for output
	std::vector<int> result;
	foreach(EnergyGraph::vertex_iterator i, bestState)
		result.push_back((*graph)[*i].elementID);

        // reverse the list if necessary
        if (energyList.get() != NULL)
            (*energyList).reverse();

	return std::make_pair(result, bestEnergy);
}

