// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "opt_simanneal.h"
#include <ctime>

///////////////////////////////////////////////////////////////////////////
//////                      PARAMETER ACCESSORS                      //////
///////////////////////////////////////////////////////////////////////////

int SimulatedAnnealingPacker::GetMaxFrames()
{
	return maxFrames_;
}

void SimulatedAnnealingPacker::SetMaxFrames(int maxFrames)
{
	maxFrames_ = maxFrames;
}

FLOAT SimulatedAnnealingPacker::GetLastTemperature()
{
	return lastTemperature_;
}

///////////////////////////////////////////////////////////////////////////
//////                        IMPLEMENTATION                        ///////
///////////////////////////////////////////////////////////////////////////

SimulatedAnnealingPacker::SimulatedAnnealingPacker()
	: OptimizationBaseSimulatedAnnealing()
{

}

std::pair< std::vector<int>, FLOAT >
	SimulatedAnnealingPacker::Pack(
                    DiscreteSearchSpace::ConstPointer searchSpace,
                    EnergyGraph* graph, 
		    std::vector<int> const* initialState,
                    EnergyCombList::Pointer energyList, int listSize)
{
        // Check on the energy list - should be in reverse order
        if ((energyList.get() != NULL) && (!(*energyList).isDescOrder()))
            (*energyList).reverse();

	searchSpace_ = searchSpace;
	energies_.clear();
	scheduleIndex_ = 0;

	quenchDim_ = 0;
	quenchID_ = 0;

	const int N_ITERATIONS	= (int)(searchSpace_->NumElements() * iterMultiplier_);
	
	InitState(*graph, initialState? *initialState : std::vector<int>());
	energies_.push_back(graph->ComputeEnergy(stateVector_));

        // Keep the best solution seen so far
	std::vector<EnergyGraph::vertex_iterator> bestState = stateVector_;
	FLOAT bestEnergy = energies_.back();
        
        // vector for storing temporary states
        std::vector<int> tempVec;

        // Add the first solution to the list if necceary
        if (energyList.get() != NULL) {
            // populate the vector
            foreach(EnergyGraph::vertex_iterator i, stateVector_)
                tempVec.push_back((*graph)[*i].elementID);

            // add the solution to the list
            (*energyList).push(EnergyComb(tempVec, energies_.back()));
        }

	int frame = 0;

	for (int epoch = 0; epoch < (int)schedule_.size(); ++epoch)
	{
		// record runtime stats, if necessary
		if (stats_.get() != NULL) {
			if (energyList.get() != NULL && listSize > 0) {
				stats_->Increment(RSTATS_SIMANNEAL_LISTOP,
						  (unsigned long long)
						  (log((double)listSize) * N_ITERATIONS));
			}
			stats_->Increment(RSTATS_SIMANNEAL_ITER, N_ITERATIONS);
		}

		bool quench = (epoch == (int)schedule_.size() - 1);
		FLOAT temperature = GetTemperature(quench);
	
		for (int i = 0; i < N_ITERATIONS; ++i)
		{
			std::pair<int, int> move = PickMove(quench);
			
			EnergyGraph::vertex_iterator newVertex = stateIDStart_[move.first];
			EnergyGraph::vertex_iterator oldVertex = stateVector_[move.first];
			for (int j = 0; j < move.second; ++j)
				++newVertex;
		
			if (newVertex == oldVertex)
				continue;

			frame++;

			FLOAT diff = graph->DiffEnergy(newVertex, oldVertex, stats_);
			if (KeepChange(energies_.back(), diff, temperature, quench) || frame == maxFrames_)
			{
				energies_.back() += diff;
		
				(*graph)[*newVertex].active = true;
				(*graph)[*oldVertex].active = false;

				stateVector_[move.first] = newVertex;	
				
                                // Add to the list if necessary
                                if (energyList.get() != NULL) {
                                    // remove old state
                                    tempVec.clear();

                                    // format the vector
                                    foreach(EnergyGraph::vertex_iterator i, stateVector_)
                                        tempVec.push_back((*graph)[*i].elementID);

                                    // push the solution
                                    (*energyList).push(EnergyComb(tempVec, energies_.back()));

                                    // pop if we're over the limit
                                    if ((*energyList).size() > listSize)
                                        (*energyList).pop();
                                }

				// keep the new best energy
                                if (energies_.back() < bestEnergy)
				{
					bestEnergy = energies_.back();
					bestState = stateVector_;
				}

				if (frame == maxFrames_) {
					bestEnergy = graph->ComputeEnergy(stateVector_);
					bestState = stateVector_;

					break;
				}
			}

			frame++;

			if (frame == maxFrames_) {
				break;
			}			
		}

		energies_.push_back( energies_.back() );
		if (frame == maxFrames_) {
			lastTemperature_	= temperature;
			break;
		}
	}

	// update the best energy
	bestEnergy = graph->ComputeEnergy(bestState);
       
        // format the best solution for output
	std::vector<int> result;
	foreach(EnergyGraph::vertex_iterator i, bestState)
		result.push_back((*graph)[*i].elementID);

        // reverse the list if necessary
        if (energyList.get() != NULL)
            (*energyList).reverse();

	return std::make_pair(result, bestEnergy);
}

