// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "pdb.h"

#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <boost/algorithm/string/trim.hpp>

namespace PDB
{

///////////////////////////////////////////////////////////////////////////
//////                       PARSER FUNCTIONS                       ///////
///////////////////////////////////////////////////////////////////////////

template <typename T>
T Extract(std::string str, T defaultVal = T())
{
	try
	{
		boost::trim(str);
		return boost::lexical_cast<T>(str);
	}
	catch ( boost::bad_lexical_cast )
	{
		// TODO: maybe if we get a good tiered logging facility, it
		// would be a good idea to put a warning here that a field
		// was missing
		
		return defaultVal;
	}
}

bool DummyStringRecordHandler(std::string const& dummy) { return true; }
bool DummyAtomRecordHandler(AtomRecord const& dummy) { return true; }

Reader::Reader() : repeatLastRecord_(false)
{
	ConnectRecordSignal( boost::function< bool (std::string const&)>(DummyStringRecordHandler) );
	ConnectRecordSignal( boost::function< bool (AtomRecord  const&)>(DummyAtomRecordHandler) );
}

bool Reader::ParseAtomRecord(std::istream& is)
{
	if (repeatLastRecord_)
		return AtomRecordSignal(atomRecord_);

	std::string text;
	std::getline(is, text);

	try
	{
		atomRecord_.serial	= Extract<uint>	 	( text.substr( 6,  5) );
		atomRecord_.name	= Extract<std::string> 	( text.substr(12,  4) );
		atomRecord_.altLoc	= Extract<std::string> 	( text.substr(16,  1) );
		atomRecord_.resName	= Extract<std::string> 	( text.substr(17,  3) );
		atomRecord_.chainID	= Extract<std::string>	( text.substr(21,  1) );
		atomRecord_.resSeq	= Extract<uint>		( text.substr(22,  4) );
		atomRecord_.iCode	= Extract<std::string>	( text.substr(26,  1) );
		atomRecord_.position(0)	= Extract<FLOAT>	( text.substr(30,  8) );
		atomRecord_.position(1)	= Extract<FLOAT>	( text.substr(38,  8) );
		atomRecord_.position(2)	= Extract<FLOAT>	( text.substr(46,  8) );
		atomRecord_.occupancy	= Extract<FLOAT>	( text.substr(54,  6) );
		atomRecord_.tempFactor	= Extract<FLOAT>	( text.substr(60,  6) );
		atomRecord_.element	= Extract<std::string>	( text.substr(76,  2) );
		atomRecord_.charge	= Extract<std::string>	( text.substr(78,  2) );
	}
	catch (std::out_of_range)
	{
		// ignore too-short lines for now
	}

	return AtomRecordSignal(atomRecord_);	
}

std::istream& operator>> (std::istream& is, Reader& reader)
{
	if (!reader.repeatLastRecord_)
		reader.recordName_ = reader.NextRecordType(is);

	if (reader.recordName_ == "ATOM")
		reader.repeatLastRecord_ = !reader.ParseAtomRecord(is);
	else
	{
		// we don't have a parser for any other record, so send all others
		// intact to the Default signal
		if (reader.recordName_ != "EOF")
		{
			if (!reader.repeatLastRecord_)
				std::getline(is, reader.stringRecord_);
			
			reader.repeatLastRecord_ = !reader.DefaultRecordSignal(reader.stringRecord_);
		}
	}

	return is;
}

std::string Reader::NextRecordType (std::istream& is) const
{
	std::string result = "";

	// skip over any whitespace lines
	while (isspace(is.peek())) is.get();
	
	std::streampos originalPos = is.tellg();
	
	if (is.eof())
		return "EOF";

	char buffer[7];
	is.read(buffer, 6);
	is.clear();	          // clear out potential fail and eof flags
	
	buffer[is.gcount()] = '\0';
	is.seekg(originalPos);
	
	result = buffer;
	boost::trim(result);
	
	return result;
}

///////////////////////////////////////////////////////////////////////////
//////                       WRITER FUNCTIONS                       ///////
///////////////////////////////////////////////////////////////////////////

std::ostream& operator<< (std::ostream& os, Writer const& writer)
{
	foreach ( std::string const& line, writer.text_ )
		os << line << std::endl;

	return os;
}

void Writer::WriteRecord (std::string const& r)
{
	text_.push_back(r);
}

void Writer::WriteRecord (AtomRecord const& r)
{
	uint serial = autoSerial_? atomRecordIndex_ : r.serial;
	++atomRecordIndex_;

	// TODO: This validation code is going to end up being duplicated
	// for every WriteRecord function, so it may be worthwhile to eventually
	// factor it out into some sort of automated record-checking system

	const uint  BOUND_5DIGITS 	= 100000;
	const uint  BOUND_4DIGITS 	= 10000;
	
	const FLOAT BOUND_4DIGITS_F 	= 10000.0;
	const FLOAT BOUND_4DIGITS_NEG_F = -1000.0;
	
	const FLOAT BOUND_3DIGITS_F 	= 1000.0;
	const FLOAT BOUND_3DIGITS_NEG_F = -100.0;
	
	boost::format serialErr("serial %d too large (must be 5 digits or less)");
	boost::format resSeqErr("residue seqpos %d too large (must be 4 digits or less)");
	boost::format posErr   ("position(%d) (%8.3f) too large (must be 4 digits or less before the decimal point)");
	boost::format occErr   ("occupancy (%8.3f) too large (must be 3 digits or less before the decimal point)");
	boost::format tempErr  ("tempFactor (%8.3f) too large (must be 3 digits or less before the decimal point)");

	// check for invalid decimal arguments
	if (serial >= BOUND_5DIGITS)
		throw boost::str( serialErr % serial );
	if (r.resSeq >= BOUND_4DIGITS)
		throw boost::str( resSeqErr % r.resSeq );
	
	// check for invalid float arguments
	for (int i = 0; i < 3; ++i)
		if (r.position(i) >= BOUND_4DIGITS_F || r.position(i) <= BOUND_4DIGITS_NEG_F)
			throw boost::str( posErr % (i + 1) % r.position(i) );

	if (r.tempFactor >= BOUND_3DIGITS_F || r.tempFactor <= BOUND_3DIGITS_NEG_F)
		throw boost::str( tempErr % r.tempFactor);
	if (r.occupancy >=  BOUND_3DIGITS_F || r.occupancy <=  BOUND_3DIGITS_NEG_F)
		throw boost::str( occErr % r.occupancy);
	
	text_.push_back( 
	   boost::str(
	   boost::format("ATOM  %5d %-4.4s%1.1s%3.3s %1.1s%4d%1.1s   %8.3f%8.3f%8.3f%6.2f%6.2f          %2.2s%2.2s") 
	   
	   % serial % r.name % r.altLoc % r.resName % r.chainID % r.resSeq % r.iCode
	   
	   % r.position(0) % r.position(1) % r.position(2)
	   
	   % r.occupancy % r.tempFactor % r.element % r.charge ) );
}

}


