// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "residue.h"
#include "vectors.h"
#include "system.h"
#include "atom_type.h"

#include <blitz/tinyvec-et.h>

#include <boost/lexical_cast.hpp>
#include <boost/bind.hpp>
#include <boost/signals.hpp>
#include <boost/format.hpp>

#include <string>
#include <cassert>
#include <cctype>

using namespace CHOMPMath;

///////////////////////////////////////////////////////////////////////////
//////                       PDBResidueFactory                       //////
///////////////////////////////////////////////////////////////////////////

PDBResidueFactory::PDBResidueFactory( 
	System::ConstPointer system, 
	std::istream& is, 
	PDB::Reader& reader,
	Residue const* prevRes)
	    : system_(system), is_(is), reader_(reader), prevRes_(prevRes)
{
	// Setup the reader to call HandleAtomRecord when it finds an atom record
	boost::function<bool (PDB::AtomRecord const&)> handler = 
		boost::bind(&PDBResidueFactory::HandleAtomRecord, this, _1);

	connection_ = reader.ConnectRecordSignal(handler);

	state_ = STATE_STARTING;
	aa_ = 0;
	atomIndex_ = 0;
	succeeded_ = true;
}

PDBResidueFactory::~PDBResidueFactory()
{
	connection_.disconnect();
}

Residue PDBResidueFactory::ConstructResidue()
{
	while (state_ != STATE_DONE && !is_.eof())
		is_ >> reader_;

	if (state_ != STATE_DONE)
		FinishResidue();

	return residue_;
}

bool PDBResidueFactory::HandleAtomRecord( PDB::AtomRecord const& record )
{
	atomSerial_ = record.serial;
	int atom;

	switch (state_)
	{
		// If the residue name is not recognized, go into a skipping state until the residue name changes

		case STATE_SKIP:
			if (record.resName == skipResidue_)
				break;
			
			// otherwise fall through
			state_ = STATE_STARTING;

		case STATE_STARTING:
			
			try
			{
				aa_ = system_->GetAminoAcidData()->Get(record.resName);
			}
			catch (std::string) // residue not found!
			{
				std::cout << "WARNING: Skipping unknown residue " << record.resName << ".\n";
				skipResidue_ = record.resName;
				
				state_ = STATE_SKIP;
				break;
			}

			residue_.system_	= system_;
			residue_.id_ 		= record.resSeq;
			residue_.aminoAcid_	= aa_;

			residue_.atoms_.resize(aa_->nTotalAtoms);
			residue_.hbOrient_.resize(aa_->nTotalAtoms);

			atomInitialized_.resize(aa_->nTotalAtoms);
			for (int i = 0; i < aa_->nTotalAtoms; ++i)
				atomInitialized_[i] = false;
	
			state_ = STATE_READING_ATOMS;

			// fall through to next case
		case STATE_READING_ATOMS:
		
			// check if the residue is over in the PDB
			if (record.resName != aa_->name3Char ||
			    record.resSeq  != (uint)residue_.id_)
			{
				FinishResidue();
				return false;	// signal that we want this record processed again later
			}

			// TODO: Protonated AA variants should be handled explicitly
			// as separate amino acid classes. In the meantime, we have a hack 
			// to be able to parse protonated variants but pretend they're normal.
			// Currently, all this function does is return true if it recognizes hydrogen names
			// corresponding to protonated forms of ASP, GLU, or HIS 
			if (ProtonatedAAHack(aa_->aminoAcid, record.name))
				break;

			// Find where to put the next atom by looking up the index for that atom name.
			// CHOMP_ATOM_ORDER is the trie order determined by the aa_props/XXX.py files
			try
			{
				atom = aa_->indexOf(CHOMP_ATOM_ORDER, record.name);
			}
			catch (std::string e)
			{
				std::cerr << ErrorHeader() << e << std::endl;
				std::cerr << "...skipping residue and continuing...\n";
				
				succeeded_ = false;
				return true;
			}

			residue_.atoms_[atom] = record.position;

			atomInitialized_[atom] = true;

			// TODO: Handle atom alternate locations properly... for now, 
			// alternate atoms are simply skipped over

			break;

		case STATE_DONE:
			throw ErrorHeader() + "something strange happened.";
	}

	return true;
}

void PDBResidueFactory::FinishResidue()
{
	bool rebuildEntireSidechain = false;

	int remaining = std::count(atomInitialized_.begin(), atomInitialized_.end(), false);
		
	// find the position of the backbone hydrogen, if any	
	int bHpos = -1;
	try
	{
		bHpos = residue_.AtomIndex("H");
	}
	catch (...)
	{
		// this amino acid has no backbone H
	}
	
	// If there are uninitialized atoms; try to regenerate them from a template
	// FIXME: potential bug -- should probably check that all template atoms are 
	// initialized! though this is currently a nonissue since only hydrogens
	// can be regenerated this way and they can't be template atoms
	if (remaining != 0)
		for (int atom = 0; atom < aa_->nTotalAtoms; ++atom)
			if (!atomInitialized_[atom])
			{
				bool success = false;
								
				// backbone H atoms need a bit of special handling.
				success = (atom == bHpos) ?
					residue_.PlaceHN(prevRes_) :
					residue_.PlaceAtomFromTemplate(atom);
				
				if (success)
				{
					atomInitialized_[atom] = true;
					--remaining;
				}

				// If any non-templatable side-chain atoms are missing, we can
				// rebuild the whole side-chain from the template, discarding
				// any native chi angles.
				else if ( !(residue_.GetAminoAcid().atomFlags[atom] & ATM_FLAG_BB) )
				{
					rebuildEntireSidechain = true;
					--remaining;
				}
			}
	
	if (remaining == 0)
		state_ = STATE_DONE;
	else
	{
		// There are uninitialized atoms (backbone atoms were missing) 
		std::string error = ErrorHeader() + 
			"ERROR: Unexpected end of residue. Missing the following atoms:\n";

		for (int atom = 0; atom < aa_->nTotalAtoms; ++atom)
			if (!atomInitialized_[atom])
				error += aa_->atomOrders[CHOMP_ATOM_ORDER][atom] + "\n";

		std::cerr << error << "... skipping residue and continuing ...\n";

		state_ = STATE_DONE;
		succeeded_ = false;
	}

	if (rebuildEntireSidechain)
	{
		residue_.UseTemplateSideChain();
		std::cerr << ErrorHeader() << "WARNING: Missing heavy atoms. Rebuilding entire sidechain from template.\n";
	}
}

bool PDBResidueFactory::ProtonatedAAHack(AminoAcidIndex aa, std::string const& atom)
{
	// Return true if the passed atom belongs to a protonated variant

	if (aa == ASP) return atom == "HD1"  || atom == "HD2" ||
			      atom == "1HD1" || atom == "1HD2";

	if (aa == GLU) return atom == "HE1"  || atom == "HE2" || 
			      atom == "1HE1" || atom == "1HE2";

	if (aa == HIS) return atom == "HD1" || atom == "1HD1";

	return false;
}

std::string PDBResidueFactory::ErrorHeader() 
{
	return boost::str( boost::format(
		"While parsing %3s at seqpos %d, atom %d: " ) 
		% aa_->name3Char % residue_.id_ % atomSerial_ );
}

bool PDBResidueFactory::Succeeded()
{
	return succeeded_;
}

///////////////////////////////////////////////////////////////////////////
//////                            Residue                            //////
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

Residue::Residue()
	: id_(-1), aminoAcid_(0), 
	  phiPsiDirty_(true), 	chiDirty_(true),
	  centroidDirty_(true), hbOrientDirty_(true)
{
}

///////////////////////////////////////////////////////////////////////////
//////                       CHECKED ACCESSORS                       //////
///////////////////////////////////////////////////////////////////////////

std::vector<Vector3> const& Residue::GetOrientationVectors()
{
	ComputeOrientationVectors();
	return hbOrient_;
}

std::vector<Vector3> const& Residue::GetOrientationVectors() const
{
	if (hbOrientDirty_)
		throw std::string( "GetOrientationVectors(): Atom coords have been changed on a const Residue." );
	return hbOrient_;
}

Vector3 const& Residue::GetCentroid()
{
	ComputeCentroid();
	return centroid_;
}

Vector3 const& Residue::GetCentroid() const
{
	if (centroidDirty_)
		throw std::string( "GetCentroid(): Atom coords have been changed on a const Residue." );
	return centroid_;
}

Vector2 Residue::GetPhiPsi()
{
	ComputePhiPsi();
	return Vector2(phi_, psi_);
}

Vector2 Residue::GetPhiPsi() const
{
	if (phiPsiDirty_)
		throw std::string( "GetPhiPsi(): Atom coords have been changed on a const Residue." );
	return Vector2(phi_, psi_);
}

void Residue::UpdatePhiPsi()
{
	phiPsiDirty_ = true;
	ComputePhiPsi();
}

Vector4 const& Residue::GetChi()
{
	ComputeChi();
	return chi_;
}

Vector4 const& Residue::GetChi() const
{
	if (chiDirty_)
		throw std::string( "GetChi(): Atom coords have been changed on a const Residue." );
	return chi_;
}

int Residue::AtomIndex(std::string const& atomName) const
{
	return aminoAcid_->indexOf(CHOMP_ATOM_ORDER, atomName);
}

FLOAT Residue::GetOmega() const
{
	return ComputeOmega();
}

FLOAT Residue::GetBondAngle(blitz::TinyVector<int, 3> angleAtoms) const
{
	return ComputeBondAngle(angleAtoms);
}

FLOAT Residue::GetTorsionAngle(blitz::TinyVector<int, 4> torsionAtoms) const
{
	return ComputeTorsionAngle(torsionAtoms);
}

FLOAT Residue::GetImptorsAngle(blitz::TinyVector<int, 4> imptorsAtoms) const
{
	return ComputeImptorsAngle(imptorsAtoms);
}

///////////////////////////////////////////////////////////////////////////
//////                      PRIVATE MEMBER FNS                      ///////
///////////////////////////////////////////////////////////////////////////

void Residue::ComputePhiPsi()
{
	// Compute the backbone dihedral angles Phi and Psi as applicable

	if (!phiPsiDirty_) return;
	
	const FLOAT NULL_PHI = -90.0;
	const FLOAT NULL_PSI = 130.0;
	const FLOAT chainBreakDist2 = 9.0;

	Vector3 myN  = atoms_[AtomIndex("N")];
	Vector3 myC  = atoms_[AtomIndex("C")];
	Vector3 myCA = atoms_[AtomIndex("CA")];

	System::ConstPointer system = system_.lock();

	// Phi is undefined for the first residue or if there is a chain break. Otherwise,
	// calculate the dihedral using the carbon position from the previous residue.

	if (system->IsFirstInChain(*this))
		phi_ = NULL_PHI;	
	else
	{
		Residue const& prevRes = system->PrevResidue(*this);
		Vector3 prevC = prevRes.atoms_[ prevRes.AtomIndex("C") ];

		Vector3 bondC_N = myN - prevC;
		FLOAT   bondLen2 = blitz::dot(bondC_N, bondC_N);

		if (bondLen2 > chainBreakDist2)
			phi_ = NULL_PHI;
		else
			phi_ = CalculateDihedralAngle( prevC, myN, myCA, myC );
	}

	// Psi is undefined for the last residue or if there is a chain break. Otherwise,
	// calculate the dihedral using the nitrogen position from the next residue.

	if (system_.lock()->IsLastInChain(*this))
		psi_ = NULL_PSI;
	else
	{
		Residue const& nextRes = system->NextResidue(*this);
		Vector3 nextN = nextRes.atoms_[ nextRes.AtomIndex("N") ];
	
		Vector3 bondC_N = nextN - myC;
		FLOAT   bondLen2 = blitz::dot(bondC_N, bondC_N);

		if (bondLen2 > chainBreakDist2)
			psi_ = NULL_PSI;
		else
			psi_ = CalculateDihedralAngle( myN, myCA, myC, nextN );
	}

	phiPsiDirty_ = false;
}

void Residue::ComputeChi()
{
	// Compute the sidechain dihedral angles Chi1, Chi2, Chi3, and Chi4 as applicable

	if (!chiDirty_) return;

	for (int chi_i = 0; chi_i < 4; ++chi_i)
	{
		if (chi_i >= aminoAcid_->nChi)
		{
			chi_(chi_i) = FLOAT(0.0);
			continue;
		}

		int c0 = aminoAcid_->chiAtoms[chi_i][0];
		int c1 = aminoAcid_->chiAtoms[chi_i][1];
		int c2 = aminoAcid_->chiAtoms[chi_i][2];
		int c3 = aminoAcid_->chiAtoms[chi_i][3];

		chi_(chi_i) = CalculateDihedralAngle(atoms_[c0], atoms_[c1], 
			                     atoms_[c2], atoms_[c3]);
	}

	chiDirty_ = false;
}

void Residue::ComputeOrientationVectors()
{
	// The vectors from a h-bond donor to the hydrogen, or from the acceptor to its base(s)
	// are needed to compute hydrogen bond energy

	if (!hbOrientDirty_) return;

	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
	{
		if (aminoAcid_->atomFlags[i] & ATM_FLAG_POL_H)
		{
			// If atom i happens to be a polar hydrogen, backtrack to find the last non-hydrogen.
			// The ordering of atoms should guarantee that the last non-hydrogen atom is the
			// heavy atom bonded to this hydrogen

			int baseIndex = i;
			while (aminoAcid_->atomFlags[baseIndex] & ATM_FLAG_H)
				--baseIndex;

			Vector3 vec = atoms_[baseIndex] - atoms_[i];
			FLOAT len = std::sqrt(blitz::dot(vec, vec));

			hbOrient_[i] = vec / len;
		}
		else if (aminoAcid_->atomFlags[i] & ATM_FLAG_HB_ACCEPTOR)
		{
			Vector3 v2;

			// If atom i happens to be a h-bond acceptor we must find the base atom(s)
			// Walk through the list of acceptors, until acceptor a refers to atom i
			// Now the integer a may also be used as an index to extract the acceptor base
			// If the second number stored for that acceptorBase is -1, then the base is
			// simply the atom indicated by the first number.
			// Otherwise, the second number is a positive integer referring to a second
			// reference atom bonded to the acceptor.
			// The base atoms for HIS 'ND1' are 'CG' and 'CE1'

			for (int a = 0; a < aminoAcid_->nHBAcceptors; ++a)
				if (aminoAcid_->acceptorList[a] == i)
				{
					if (aminoAcid_->acceptorBase[a][1] == -1)
						v2 = atoms_[aminoAcid_->acceptorBase[a][0]];
					else
						v2 = 0.5 * (atoms_[aminoAcid_->acceptorBase[a][0]] + 
					    		    atoms_[aminoAcid_->acceptorBase[a][1]]);
				}

			Vector3 vec = atoms_[i] - v2;
			FLOAT len = std::sqrt(blitz::dot(vec, vec));

			hbOrient_[i] = vec / len;
		}
	}

	hbOrientDirty_ = false;
}

void Residue::ComputeCentroid()
{
	if (!centroidDirty_) return;

	centroid_ = 0.0;
	for (int i = 0; i < aminoAcid_->nCentroidAtoms; ++i)
		centroid_ += atoms_[ aminoAcid_->centroidAtoms[i] ];

	centroid_ *= 1.0 / static_cast<FLOAT>(aminoAcid_->nCentroidAtoms);

	centroidDirty_ = false;
}

FLOAT Residue::ComputeOmega() const
{
	const FLOAT null_Omega = 180.0;
	const FLOAT chainBreakDist2 = 9.0;

	Vector3 myN  = atoms_[AtomIndex("N")];
	Vector3 myC  = atoms_[AtomIndex("C")];
	Vector3 myCA = atoms_[AtomIndex("CA")];
	
	System::ConstPointer system = system_.lock();
	
	if (system->IsLastInChain(*this))
		return null_Omega;
	else
	{
		Residue const& nextRes = system->NextResidue(*this);
		Vector3 nextN = nextRes.atoms_[ nextRes.AtomIndex("N") ];
		Vector3 nextCA = nextRes.atoms_[ nextRes.AtomIndex("CA") ];
	
		Vector3 bondC_N = myC - nextN;
		FLOAT   bondLen2 = blitz::dot(bondC_N, bondC_N);

		if (bondLen2 > chainBreakDist2)
			return null_Omega;
		else
			return CalculateDihedralAngle( myCA, myC, nextN, nextCA );
	}
}

FLOAT Residue::ComputeBondAngle(blitz::TinyVector<int, 3> angleAtoms) const
{
	Vector3 const& a1 = atoms_[angleAtoms[0]],
				a2 = atoms_[angleAtoms[1]],
				a3 = atoms_[angleAtoms[2]];
	
	return CalculateVectorAngle(a1 - a2, a3 - a2);
}

FLOAT Residue::ComputeTorsionAngle(blitz::TinyVector<int, 4> torsionAtoms) const
{
	int a1 = torsionAtoms[0], a2 = torsionAtoms[1], a3 = torsionAtoms[2], a4 = torsionAtoms[3];
		
	return CalculateDihedralAngle( atoms_[a1], atoms_[a2], atoms_[a3], atoms_[a4] );
}

FLOAT Residue::ComputeImptorsAngle(blitz::TinyVector<int, 4> imptorsAtoms) const
{
	int a1 = imptorsAtoms[0], a2 = imptorsAtoms[1], a3 = imptorsAtoms[2], a4 = imptorsAtoms[3];
	
	return CalculateOutOfPlaneAngle( atoms_[a1], atoms_[a2], atoms_[a3], atoms_[a4] );
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC MEMBER FNS                       //////
///////////////////////////////////////////////////////////////////////////

void Residue::UpdateCoords()
{
	ComputePhiPsi();
	ComputeChi();
	ComputeCentroid();
	ComputeOrientationVectors();
}

bool Residue::operator== (Residue const& rhs) const
{
	System::ConstPointer system1 = system_.lock();
	System::ConstPointer system2 = rhs.system_.lock();

	return (system1.get() == system2.get() &&
	        id_ == rhs.id_);
}

void Residue::SetAtomCoords(int atomIndex, Vector3 const& coords)
{
	atoms_[atomIndex] = coords;

	if (aminoAcid_->atomFlags[atomIndex] & ATM_FLAG_BB)
		phiPsiDirty_ = true;

	chiDirty_ 	= true;
	centroidDirty_ 	= true;
	hbOrientDirty_ 	= true;
}

void Residue::AlignBackbone(
	Vector3 const* src, Vector3* dest,
	AminoAcid const* srcaa, AminoAcid const* destaa)
{
	int dpos_N  = destaa->indexOf(CHOMP_ATOM_ORDER, "N");
	int dpos_C  = destaa->indexOf(CHOMP_ATOM_ORDER, "C");
	int dpos_CA = destaa->indexOf(CHOMP_ATOM_ORDER, "CA");

	int spos_N  = srcaa->indexOf(CHOMP_ATOM_ORDER, "N");
	int spos_C  = srcaa->indexOf(CHOMP_ATOM_ORDER, "C");
	int spos_CA = srcaa->indexOf(CHOMP_ATOM_ORDER, "CA");

	Matrix3 rotMatrix;
	Vector3 transVector;
	
	AlignVectors(dest[dpos_CA], dest[dpos_N], src[spos_CA], src[spos_N],
		      &rotMatrix, &transVector);
	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
	{
		dest[i] = VectorMatrixMultiply(dest[i], rotMatrix);
		dest[i] += transVector;
	}

	AlignVectorToPlane(dest[dpos_N], dest[dpos_CA], dest[dpos_C], src[spos_C], 
		       &rotMatrix, &transVector);
	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
		if (i != dpos_N && i != dpos_CA)
		{
			dest[i] = VectorMatrixMultiply(dest[i], rotMatrix);
			dest[i] += transVector;
		}
}

void Residue::AlignBackboneToResidue(Residue const& target)
{
	AlignBackbone(&(target.atoms_[0]), &atoms_[0], 
			target.aminoAcid_, aminoAcid_);

	phiPsiDirty_ = true;
	centroidDirty_ 	= true;
	hbOrientDirty_ 	= true;
}

void Residue::UseTemplateSideChain()
{
	Vector3 tempCoords[MAX_RESIDUE_ATOMS];
	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
		tempCoords[i] = aminoAcid_->templateCoords[i];

	AlignBackbone(&atoms_[0], tempCoords, aminoAcid_, aminoAcid_);
	
	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
		if (!(aminoAcid_->atomFlags[i] & ATM_FLAG_BB))
			atoms_[i] = tempCoords[i];

	chiDirty_ 	= true;
	centroidDirty_ 	= true;
	hbOrientDirty_ 	= true;
}

void Residue::Mutate(AminoAcid const* newAminoAcid, Residue const* prevRes)
{
	// prepare a copy of our old atom coords
	std::vector<Vector3> oldAtoms = atoms_;

	// save our old identity and switch to the new one
	AminoAcid const* oldAminoAcid = aminoAcid_;
	aminoAcid_ = newAminoAcid;

	// allocate storage for a new number of atoms
	atoms_.resize(newAminoAcid->nTotalAtoms);
	hbOrient_.resize(newAminoAcid->nTotalAtoms);

	// we'll be using the backbone of the old residue verbatim,
	// but we may need to shuffle the ordering of the atoms around
	// if the backbone atoms are in different orders in the two
	// residues.
	
	// first, we prepare an aligned template of the backbone for use
	// when the new amino acid has a backbone atom the old one doesn't
	std::vector<Vector3> templateAtoms = newAminoAcid->templateCoords;
	AlignBackbone(&atoms_[0], &templateAtoms[0], oldAminoAcid, newAminoAcid);

	// now we fill in our backbone coords
	std::string bb[] = { "C", "O", "N", "CA", "H", "HA", "1HA", "2HA", "OXT" };
	foreach (std::string const& bbAtom, bb)
	{
		int newIndex = -1, oldIndex = -1;

		try
		{
			newIndex = newAminoAcid->indexOf(CHOMP_ATOM_ORDER, bbAtom);
			oldIndex = oldAminoAcid->indexOf(CHOMP_ATOM_ORDER, bbAtom);
		}
		catch (...) { }

		if (newIndex != -1) 
		{
			if (oldIndex != -1)
			{
				// if we're lucky, we can copy over the atom directly
				atoms_[newIndex] = oldAtoms[oldIndex];
			}
			else 
			{
				// if we're unlucky, we have to load it from the template	
				// or use PlaceHN for H
				
				if (bbAtom == "H")
					PlaceHN(prevRes);
				else if (bbAtom == "OXT")
					PlaceOXT();
				else
					atoms_[newIndex] = templateAtoms[newIndex];
			}
		}				
	}

	// now set the mutated residue to use a generic, templated side chain
	UseTemplateSideChain();

	// finally, regenerate all secondary data
	UpdateCoords();
}

FLOAT Residue::RMSD(Residue const& rhs) const
{
	if (rhs.aminoAcid_->aminoAcid != this->aminoAcid_->aminoAcid)
		throw std::string("You can only take the RMSD of identical residues..!");

	Residue rhsClone = rhs.Clone(GetSystem()); 
	rhsClone.AlignBackboneToResidue(*this);

	FLOAT sum = 0.0;
	for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i)
	{
		Vector3 delta = atoms_[i] - rhsClone.atoms_[i];
		sum += blitz::dot(delta, delta);
	}

	return std::sqrt(sum) / (FLOAT)aminoAcid_->nTotalAtoms;
}

FLOAT Residue::StandardizeChi(FLOAT chi, int chiIndex) const
{
	AminoAcidIndex aa = aminoAcid_->aminoAcid;

	if ( chiIndex == 1 ) 
	{  
		if ( aa == PHE || aa == TYR ) 
			return WrapValue(chi - 60.0, 180.0) + 60.0;
		else if ( aa == ASP )
			return WrapValue(chi, 180.0);
	}
	else if ( chiIndex == 2 )
	{
		if ( aa == GLU )
			return WrapValue(chi, 180.0);
	}

	return WrapValue(chi, 360.0);
}

void Residue::StandardizeChis()
{
	Vector4 initialChi = GetChi();
	
	for (int chi_i = 0; chi_i < 4; ++chi_i)
	{
		if (chi_i >= aminoAcid_->nChi)
		{
			chi_(chi_i) = FLOAT(0.0);
			continue;
		}

		FLOAT newChi = StandardizeChi(chi_(chi_i), chi_i);
		chi_(chi_i) = newChi;
	}	
}

void Residue::SetChi(Vector4 const& chi)
{
	Vector4 initialChi = GetChi();
	
	for (int chi_i = 0; chi_i < 4; ++chi_i)
	{
		if (chi_i >= aminoAcid_->nChi)
		{
			chi_(chi_i) = FLOAT(0.0);
			continue;
		}

		FLOAT newChi = StandardizeChi(chi(chi_i), chi_i);

		int c1 = aminoAcid_->chiAtoms[chi_i][1];
		int c2 = aminoAcid_->chiAtoms[chi_i][2];

		FLOAT rotation = (newChi - initialChi(chi_i)) * 3.14159 / 180.0;

		Matrix3 rotMat;
		Vector3 transVec;

		RotateAboutVector(atoms_[c1], atoms_[c2], rotation, &rotMat, &transVec);
		
		for (int i = 0; i < aminoAcid_->nTotalAtoms; ++i) 
			if (aminoAcid_->chiDependent[i][chi_i])	
			{
				atoms_[i] = VectorMatrixMultiply(atoms_[i], rotMat);
				atoms_[i] += transVec;
			}
	
		chi_(chi_i) = newChi;
	}	

	// proline has its cyclical weirdness, so you can't place its hydrogens correctly
	// just by performing dihedral rotations (I think??). 
	// just manually regenerate them from the template data
	
	if (aminoAcid_->aminoAcid == PRO)
	{
		PlaceAtomFromTemplate(AtomIndex("1HD"));
		PlaceAtomFromTemplate(AtomIndex("2HD"));
		PlaceAtomFromTemplate(AtomIndex("1HG"));
		PlaceAtomFromTemplate(AtomIndex("2HG"));
	}

	chiDirty_ = false;
	centroidDirty_ = true;
	hbOrientDirty_ = true;
}

Residue Residue::Clone(System::ConstPointer newOwner) const
{
	Residue result = *this;	
	result.system_ = newOwner;
	return result;
}

Residue Residue::CloneWithChi(Vector4 const& chi, System::ConstPointer newOwner) const
{
	Residue result = *this;

	result.UseTemplateSideChain();
	result.SetChi(chi);
	result.UpdateCoords();

	result.system_ = newOwner; 
	return result;
}

bool Residue::PlaceAtomFromTemplate(int atomIndex)
{
	int tatom[4];
	Vector3 tmplt[4];
	
	if (aminoAcid_->templateAtom[atomIndex][0] == -1)
		return false;

	for (int i = 0; i < 3; ++i)
		tatom[i] = aminoAcid_->templateAtom[atomIndex][i];
	tatom[3] = atomIndex;

	for (int i = 0; i < 4; ++i)
		tmplt[i] = aminoAcid_->templateCoords[tatom[i]];

	Matrix3 rotMatrix;
	Vector3 transVector;
	
	AlignVectors(tmplt[0], tmplt[1], atoms_[tatom[0]], atoms_[tatom[1]],
		      &rotMatrix, &transVector);
	for (int i = 0; i < 4; ++i)
	{
		tmplt[i] = VectorMatrixMultiply(tmplt[i], rotMatrix);
		tmplt[i] += transVector;
	}

	AlignVectorToPlane(tmplt[0], tmplt[1], tmplt[2], atoms_[tatom[2]], 
		       &rotMatrix, &transVector);
	for (int i = 2; i < 4; ++i)
	{
		tmplt[i] = VectorMatrixMultiply(tmplt[i], rotMatrix);
		tmplt[i] += transVector;
	}

	atoms_[atomIndex] = tmplt[3];

	if (aminoAcid_->atomFlags[atomIndex] & ATM_FLAG_BB)
		phiPsiDirty_ = true;

	chiDirty_ 	= true;
	centroidDirty_ 	= true;
	hbOrientDirty_ 	= true;

	return true;
}

bool Residue::PlaceOXT()
{
    Vector3 CA = atoms_[ AtomIndex("CA") ];
    Vector3 C = atoms_[ AtomIndex("C") ];
    Vector3 O = atoms_[ AtomIndex("O") ];
    Vector3 CfromCA = C - CA;
    Vector3 OfromC = O - C;
    CfromCA /= std::sqrt( blitz::dot(CfromCA, CfromCA) );
    OfromC /= std::sqrt( blitz::dot(OfromC, OfromC) );
    Vector3 ortho1 = blitz::cross(CfromCA, OfromC);
    Vector3 ortho2 = blitz::cross(CfromCA, ortho1);
    ortho2 /= std::sqrt( blitz::dot(ortho2, ortho2) );

    // Bond length of 1.25, 126 degree angle between Oxygens
    atoms_[ AtomIndex("OXT") ]  = C  +  0.567 * CfromCA  +  1.114 * ortho2;
    return true;
}

bool Residue::PlaceHN(Residue const* prevRes)
{
	Vector3 N  = atoms_[ AtomIndex("N") ];
	Vector3 CA = atoms_[ AtomIndex("CA") ];
	Vector3 C;

	bool chainBreak = false;
	
	if (prevRes)
	{
		C  = prevRes->atoms_[ prevRes->AtomIndex("C") ];
		
		const FLOAT chainBreakDist2 = 9.0;
		if (blitz::dot(N - C, N - C) > chainBreakDist2)
			chainBreak = true;
	}
	else 
		chainBreak = true;

	if (chainBreak)
	{
		//atoms_[ AtomIndex("H") ] = atoms_[ AtomIndex("N") ] + Vector3(0.0,0.0,0.001);
		// In the absence of a preceding residue and defined phi value we can
		// place the amide hydrogen assuming a phi value of -75 (a median value)
		C = atoms_[ AtomIndex("C") ];
	
		Vector3 NfromCA = N - CA;
		Vector3 CAfromC = CA - C;
		NfromCA /= std::sqrt( blitz::dot(NfromCA, NfromCA) );
		CAfromC /= std::sqrt( blitz::dot(CAfromC, CAfromC) );
		Vector3 ortho1 = blitz::cross(CAfromC, NfromCA);
		Vector3 ortho2 = blitz::cross(NfromCA, ortho1);
		
		// Bond length of 1.01, 120 degree angle, -75 degree virtual phi
		atoms_[ AtomIndex("H") ]  = N  +  0.505 * NfromCA  +  0.847 * ortho1  +  0.220 * ortho2;
	}
	else
	{
		const FLOAT bondLength = 1.01;

		Vector3 CAtoN = CA - N;
		Vector3 CtoN  = C  - N;
		
		CAtoN /= std::sqrt( blitz::dot(CAtoN, CAtoN) );
		CtoN  /= std::sqrt( blitz::dot(CtoN,  CtoN)  );

		Vector3 NtoH = -(CAtoN + CtoN);

		NtoH /= std::sqrt( blitz::dot(NtoH, NtoH) );
		NtoH *= bondLength;

		atoms_[ AtomIndex("H") ] = N + NtoH;
	}

	return true;
}

void Residue::WriteToPDB(PDB::Writer& writer) const
{
	for (int atom = 0; atom < aminoAcid_->nTotalAtoms; ++atom)
	{
		std::string const& atomName = aminoAcid_->atomOrders[PDB_ATOM_ORDER][atom];
		int trieOrderAtom = aminoAcid_->indexOf(CHOMP_ATOM_ORDER, atomName);

		PDB::AtomRecord ar;

		ar.name	      = isdigit(atomName[0]) ? atomName : " " + atomName;
		ar.resName    = aminoAcid_->name3Char;
		ar.resSeq     = id_;
		ar.position   = atoms_[trieOrderAtom];
		ar.occupancy  = 1.0;
		ar.tempFactor = 0.0;

		try
		{	
			writer.WriteRecord(ar);
		}
		catch (std::string error)
		{
			throw boost::str( boost::format( 
				"While writing PDB ATOM record %d %s:%s: ")
				% ar.resSeq % ar.resName % ar.name ) + error;
		}
	}
}



