// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "rotamer_generator.h"

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

RotamerGenerator::RotamerGenerator(NeighborMap::ConstPointer nmap, DunbrackLibrary::ConstPointer dunLib)
				: nmap_(nmap), dunLib_(dunLib)
{
	SetupDefaultExtraChiData();
	
	// Rosetta default parameters
	params_.minRotamers 		= 2;
	
	params_.maxPercentBuried  	= 0.98;
	params_.maxPercentSurface 	= 0.95;
	
	params_.buriedCutoff 		= 18;
	params_.baseBuriedCutoff 	= 19;
	params_.maxRotamersBuried 	= 30;
	params_.maxRotamersSurface 	= 24;

	params_.bumpEnergyCutoff 	= 0.0;
	params_.extraChiSDCutoff	= 0.0;

	extraChiFlags_.resize(N_AMINO_ACIDS, N_BURIAL_TYPES, MAX_DIHEDRAL_ANGLES);
	extraChiFlags_ = 0;

	for (int aa = 0; aa < N_AMINO_ACIDS; ++aa)
		extraChiFlags_(aa, 1, 0) = EXTRA_CHI_1xSD;

	extraChiFlags_(int(PHE), 1, 1) = EXTRA_CHI_1xSD;
	extraChiFlags_(int(TRP), 1, 1) = EXTRA_CHI_1xSD;
	extraChiFlags_(int(TYR), 1, 1) = EXTRA_CHI_1xSD;
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC INTERFACE                       ///////
///////////////////////////////////////////////////////////////////////////

boost::shared_ptr<RotamerSet> RotamerGenerator::GenerateRotamers(Residue const& res, RotamerSet const* oldRotamers) const
{
	// no Dunbrack rotamers for ALA and GLY, so just create a single-rotamer rotamer set for them
	if (res.GetAminoAcid().aminoAcid == ALA || res.GetAminoAcid().aminoAcid == GLY)
		return boost::shared_ptr<RotamerSet>( new RotamerSet(res) );
	
	// otherwise, generate rotamers from the Dunbrack library
	int iPhi, iPsi;

	FLOAT phi = res.GetPhiPsi()(0);
	FLOAT psi = res.GetPhiPsi()(1);

	int firstRes = res.GetSystem()->IsFirstInChain(res);
	int lastRes  = res.GetSystem()->IsLastInChain(res);

	DunbrackLibrary::AngleBinData phipsi  = dunLib_->GetInterpData(phi, psi, firstRes, lastRes);
	DunbrackLibrary::IntAngleBins iphipsi = dunLib_->GetAngleBins(phi, psi, firstRes, lastRes);
	
	iPhi = iphipsi.phi;
	iPsi = iphipsi.psi;

	int maxRotamers =  (nmap_->NumNeighbors(res) >= params_.baseBuriedCutoff) ? 
				params_.maxRotamersBuried : params_.maxRotamersSurface;

	FLOAT maxPercent = (nmap_->NumNeighbors(res) >= params_.baseBuriedCutoff) ? 
				params_.maxPercentBuried : params_.maxPercentSurface;

	FLOAT	percentTotal = 0.0;
	int 	nBaseRotamers = 0;

	std::list<Residue> rotList;
	std::list<Residue>::iterator rotStart;

	FLOAT bestBumpEnergy = 100000.0;

	while ( (percentTotal < maxPercent || nBaseRotamers < params_.minRotamers) && 
		nBaseRotamers < maxRotamers)
	{
		int nRot = dunLib_->GetNRotByPercent(res.GetAminoAcid().aminoAcid, iPhi, iPsi, nBaseRotamers++);

		Vector4 chiValues = dunLib_->GetInterpChi(res.GetAminoAcid().aminoAcid, phipsi, nRot);
		Vector4 sdValues = dunLib_->GetInterpSD(res.GetAminoAcid().aminoAcid, phipsi, nRot);
		FLOAT   percent = dunLib_->GetInterpPercent(res.GetAminoAcid().aminoAcid, phipsi, nRot);

		rotList.push_back(res.CloneWithChi(chiValues, res.GetSystem() ));
		rotStart = rotList.end();
		--rotStart;
	
		GetExtraChi(res, chiValues, sdValues, &rotList);

		std::list<Residue>::iterator chi_i;
		for (chi_i = rotStart; chi_i != rotList.end();)
		{
			bool deleteThisChi = false;
		
			// test for low probability
			// TODO: probability cutoff not implemented

			// test for bumping
			// TODO: bump energy not implemented
			FLOAT bumpEnergy = 0.0;

			// make sure to keep at least the one best rotamer..!
			if (bumpEnergy > bestBumpEnergy && bumpEnergy > params_.bumpEnergyCutoff)
				deleteThisChi = true;
			else if (bumpEnergy < bestBumpEnergy)
				bestBumpEnergy = bumpEnergy;

			// place the centroid ??

			if (deleteThisChi)
				chi_i = rotList.erase(chi_i);
			else ++chi_i;
		}

		percentTotal += percent;
	}

	if (oldRotamers)
	{
		for (int i = 0; i < oldRotamers->NumRotamers(); ++i)
			rotList.push_back(res.CloneWithChi(oldRotamers->GetChi(i), res.GetSystem() ));
	}

	return boost::shared_ptr<RotamerSet>( new RotamerSet(res, rotList) );
}

void RotamerGenerator::GetExtraChi(Residue const& res, Vector4 const& baseChi, Vector4 const& baseSD, std::list<Residue>* rotList) const
{
	const int nchi = res.GetAminoAcid().nChi;
	const int burial = (nmap_->NumNeighbors(res) >= params_.buriedCutoff)? 1 : 0;

	std::list<FLOAT> totalChi[MAX_DIHEDRAL_ANGLES];
	
	for (int i = 0; i < MAX_DIHEDRAL_ANGLES; ++i)
		totalChi[i].push_back(baseChi(i));

	for (int i = 0; i < nchi; ++i)
	{
		if (baseSD(i) < params_.extraChiSDCutoff) continue;

		int flag = 1;
		for (int flag_i = 0; flag_i < N_EXTRA_CHI_TYPES; ++flag_i, flag *= 2)
		{
			if (extraChiFlags_((int)res.GetAminoAcid().aminoAcid, burial, i) & flag)
			{
				for (std::list<FLOAT>::const_iterator echi_i = extraChiData_[flag_i].begin();
				     echi_i != extraChiData_[flag_i].end(); ++echi_i)
				{
					totalChi[i].push_back(baseChi(i) + baseSD(i) * (*echi_i));
				}
			}
		}
	}

	int hChi = 0;
	std::list<FLOAT> const* extraH;

	if (res.GetAminoAcid().aminoAcid == SER || res.GetAminoAcid().aminoAcid == THR)
	{
		hChi = 1;
		extraH = !( extraChiFlags_((int)res.GetAminoAcid().aminoAcid, burial, hChi) & EXTRA_CHI_H )
				? &hydrogenChi1_ : &extraHydrogenChi1_;
	}
	else if (res.GetAminoAcid().aminoAcid == TYR)
	{
		hChi = 2;
		extraH = !( extraChiFlags_((int)res.GetAminoAcid().aminoAcid, burial, hChi) & EXTRA_CHI_H )
				? &hydrogenChi2_ : &extraHydrogenChi2_;
	}

	if (hChi)
		for (std::list<FLOAT>::const_iterator echi_i = extraH->begin();
		     echi_i != extraH->end(); ++echi_i)
		{
			totalChi[hChi].push_back(*echi_i);
		}

	std::list<FLOAT>::iterator chi0, chi1, chi2, chi3;
	for (chi0 = totalChi[0].begin(); chi0 != totalChi[0].end(); ++chi0)
		for (chi1 = totalChi[1].begin(); chi1 != totalChi[1].end(); ++chi1)
			for (chi2 = totalChi[2].begin(); chi2 != totalChi[2].end(); ++chi2)
				for (chi3 = totalChi[3].begin(); chi3 != totalChi[3].end(); ++chi3)
				{
					if (chi0 == totalChi[0].begin() && chi1 == totalChi[1].begin() &&
					    chi2 == totalChi[2].begin() && chi3 == totalChi[3].begin())
					    continue;
				
					rotList->push_back( res.CloneWithChi( Vector4(*chi0, *chi1, *chi2, *chi3) ,
							    res.GetSystem() ));
				}
}

void RotamerGenerator::SetupDefaultExtraChiData()
{
	const int 	EC_1xSD		= 0,
			EC_2xSD		= 1,
			EC_ARO		= 2,
			EC_HALF_ARO	= 3,
			EC_CYS		= 4;
	
	extraChiData_[EC_1xSD].push_back( +1.0 );
	extraChiData_[EC_1xSD].push_back( -1.0 );

	extraChiData_[EC_2xSD].push_back( +2.0 );
	extraChiData_[EC_2xSD].push_back( -2.0 );

	extraChiData_[EC_ARO].push_back( +0.66 );
	extraChiData_[EC_ARO].push_back( +0.33 );
	extraChiData_[EC_ARO].push_back( -0.33 );
	extraChiData_[EC_ARO].push_back( -0.66 );

	extraChiData_[EC_HALF_ARO].push_back( +0.5 );
	extraChiData_[EC_HALF_ARO].push_back( -0.5 );

	extraChiData_[EC_CYS].push_back( +0.25 );
	extraChiData_[EC_CYS].push_back( +0.5  );
	extraChiData_[EC_CYS].push_back( +0.75 );
	extraChiData_[EC_CYS].push_back( +1.25 );
        extraChiData_[EC_CYS].push_back( +1.5  );
	extraChiData_[EC_CYS].push_back( -0.25 );
	extraChiData_[EC_CYS].push_back( -0.5  );
	extraChiData_[EC_CYS].push_back( -0.75 );
	extraChiData_[EC_CYS].push_back( -1.25 );
	extraChiData_[EC_CYS].push_back( -1.5  );

	hydrogenChi1_.push_back(  60.0  );
	hydrogenChi1_.push_back( -60.0  );
	hydrogenChi1_.push_back(  180.0 );
		
	extraHydrogenChi1_.push_back(  60.0  );
	extraHydrogenChi1_.push_back(  40.0  );
	extraHydrogenChi1_.push_back(  80.0  );
	extraHydrogenChi1_.push_back(  80.0  );
	extraHydrogenChi1_.push_back(  160.0 );
	extraHydrogenChi1_.push_back(  200.0 );
	extraHydrogenChi1_.push_back( -60.0  );
	extraHydrogenChi1_.push_back( -40.0  );
	extraHydrogenChi1_.push_back( -80.0  );
	
	hydrogenChi2_.push_back( 180.0 );

	extraHydrogenChi2_.push_back(  20.0  );
	extraHydrogenChi2_.push_back(  20.0  );
	extraHydrogenChi2_.push_back(  180.0  );
	extraHydrogenChi2_.push_back(  200.0  );
	extraHydrogenChi2_.push_back(  160.0  );
}


