// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "general.h"
#include "rotamer_trie.h"

#include <blitz/tinyvec-et.h>

#include <cassert>
#include <string>
#include <map>

/** \file rotamer_trie.cpp
 *  \see rotamer_trie.h
 */

///////////////////////////////////////////////////////////////////////////
//////                        MISC. UTILITIES                        //////
///////////////////////////////////////////////////////////////////////////

const FLOAT SAME_CUTOFF = 0.0001;

bool ChiAngleCmp(Residue const* res1, Residue const* res2)
{
	Vector4 chi1 = res1->GetChi();
	Vector4 chi2 = res2->GetChi();

	for (int i = 0; i < 4; ++i)
		if ( std::abs(chi1(i) - chi2(i)) < SAME_CUTOFF )
			continue;
		else if (chi1(i) > chi2(i))
			return true;
		else 
			return false;

	return false;
}

inline bool ChiDependent(AminoAcid const& aa, Residue const& res, int chiIndex, int atom)
{
	// due to its weird connectivity, PRO needs special handling here
	if (chiIndex == 0 && aa.aminoAcid == PRO)
	{
		if (atom == res.AtomIndex("CD")  ||
	    	    atom == res.AtomIndex("1HD") ||
	   	    atom == res.AtomIndex("2HD"))
			return true;
	}

	// the following three amino acids have the interesting property that (according to
	// Rosetta, at least), the 'base' for their acceptor atoms can have a *deeper* branch
	// depth than the acceptor atoms themselves. That means those acceptors must have 
	// their branch depth artificially increased, or else their hb_orient vectors
	// will come out wrong in the trie.
	// and yes, this bug was a bitch to track down.
	if (aa.aminoAcid == SER || aa.aminoAcid == THR || aa.aminoAcid == TYR)
	{
		// find the acceptor base for this acceptor
		for (int a = 0; a < aa.nHBAcceptors; ++a)
			if (aa.acceptorList[a] == atom)
			{
				int abase = aa.acceptorBase[a][0];
				if ( aa.chiDependent[abase][chiIndex] )
					return true;
			}
	}

	// for all other cases, the normal chi-dependence data will do the trick
	return aa.chiDependent[atom][chiIndex];
}

///////////////////////////////////////////////////////////////////////////
//////                          RotamerSet                          ///////
///////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

RotamerSet::RotamerSet(Residue const& baseResidue, std::list<Residue> const& rotList)
	: baseResidue_(baseResidue)
{
	std::list<Residue const*> residues;
	foreach (Residue const& res, rotList)
		residues.push_back(&res);
	
	residues.sort(ChiAngleCmp);
	nRotamers_ = (int)residues.size();

	std::list<Residue const*>::iterator res_i;
	Residue const* lastRes = 0;
	int nNode = 0;

	nodes_.resize(nRotamers_ * MAX_RESIDUE_ATOMS);

	std::list<Residue const*> duplicateList;

	for (res_i = residues.begin(); res_i != residues.end(); ++res_i)
	{
		Residue const& res = **res_i;
		
		int firstNewChi = -1;
		if (lastRes)
		{
			int i;
			for (i = 0; i < res.GetAminoAcid().nChi; ++i)
			{
				if ( std::abs( lastRes->GetChi()[i] - res.GetChi()[i] ) > SAME_CUTOFF )
					break;
			}
			
			// discard duplicated chi angles
			if ( i == res.GetAminoAcid().nChi )
			{
				duplicateList.push_back(*res_i);
				continue;
			}

			firstNewChi = i;
		}
		lastRes = &res;

		BYTE hv_depth = 0;
		for (int i = 0; i < res.GetAminoAcid().nTotalAtoms; ++i)
		{
			if (firstNewChi == -1 || ChiDependent(res.GetAminoAcid(), res, firstNewChi, i))
			{
				nodes_(nNode).pos = res.GetAtomCoords()[i];
				nodes_(nNode).depth = i;	
				nodes_(nNode).hv_depth = hv_depth;
				nodes_(nNode).hb_orient = res.GetOrientationVectors()[i];

				nodes_(nNode).branch_depth = 0;
				for (int j = 0; j < 4; ++j)
					if (ChiDependent(res.GetAminoAcid(), res, j, i))
						++nodes_(nNode).branch_depth;
				++nNode;
			}

			// if this isn't a hydrogen and this isn't the last atom, increment the heavy
			// atom depth for the next atom
			if (i + 1 < res.GetAminoAcid().nTotalAtoms && !(res.GetAminoAcid().atomFlags[i+1] & ATM_FLAG_H))
				++hv_depth;
		}
	}

	nodes_.resizeAndPreserve(nNode);

	foreach (Residue const* dupRes, duplicateList)
		residues.remove(dupRes);

	chi_.resize( (int)residues.size());
	centroid_.resize( (int)residues.size());
	nRotamers_ = (int)residues.size();
	int i = 0;
	
	for (res_i = residues.begin(); res_i != residues.end(); ++res_i, ++i)
	{
		chi_(i) = (*res_i)->GetChi();
		centroid_(i) = (*res_i)->GetCentroid();
	}
}

RotamerSet::RotamerSet(Residue const& res)
	: baseResidue_(res)
{
	nodes_.resize(res.GetAminoAcid().nTotalAtoms);
	nRotamers_ = 1;

	BYTE hv_depth = 0;
	for (int i = 0; i < res.GetAminoAcid().nTotalAtoms; ++i)
	{
		nodes_(i).pos = res.GetAtomCoords()[i];
		nodes_(i).depth = i;	
		nodes_(i).hv_depth = hv_depth;
		nodes_(i).hb_orient = res.GetOrientationVectors()[i];

		nodes_(i).branch_depth = 0;
		for (int j = 0; j < 4; ++j)
			if (ChiDependent(res.GetAminoAcid(), res, j, i))
				++nodes_(i).branch_depth;

		// if this isn't a hydrogen and this isn't the last atom, increment the heavy
		// atom depth for the next atom
		if (i + 1 < res.GetAminoAcid().nTotalAtoms && !(res.GetAminoAcid().atomFlags[i+1] & ATM_FLAG_H))
			++hv_depth;
	}

	chi_.resize(1);
	centroid_.resize(1);
	
	chi_(0)      = res.GetChi();
	centroid_(0) = res.GetCentroid();
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC MEMBER FNS                       //////
///////////////////////////////////////////////////////////////////////////

void RotamerSet::PrintRotamers() const
{
	for (int i = 0; i < chi_.extent(0); ++i)
	{
		std::cout << "ROTAMER: " << baseResidue_.GetID() + 1 << " " << i + 1 << " ";
		for (int j = 0; j < 4; ++j)
			std::cout << chi_(i)(j) << " ";
		std::cout << std::endl;

		//std::cout << "CENTROID: " << baseResidue_.GetID() + 1 << " " << i + 1 << " " <<
		//	centroid_(i)(0) << " " << centroid_(i)(1) << " " << centroid_(i)(2) << std::endl;
	}
}

Residue RotamerSet::GetRotamer(int index) const
{
	if (index < 0 || index >= nRotamers_)
		throw "RotamerSet.GetRotamer: index " + to_str(index) + " out of range.";
	
	Residue res = baseResidue_;

	int nRot = 0;
	for (int i = 0; i < nodes_.extent(0); ++i)
	{
		RotamerNode const& node = nodes_(i);
		res.SetAtomCoords(node.depth, node.pos);

		if (nodes_(i).depth == res.GetAminoAcid().nTotalAtoms - 1)
		{
			if (nRot++ == index)
				break;
		}
	}

	return res;
}


