// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "rotamers.h"
#include "vectors.h"

#include <iostream>
bool printRotamersFull = false;

///////////////////////////////////////////////////////////////////////////
//////                         CONSTRUCTORS                         ///////
///////////////////////////////////////////////////////////////////////////

RotamerLibrary::Pointer RotamerLibrary::Create(System::ConstPointer system)
{
	std::vector<System::ConstPointer> systems;
	systems.push_back(system);

	return RotamerLibrary::Pointer( new RotamerLibrary(systems) );
}

RotamerLibrary::Pointer RotamerLibrary::Create(std::vector<System::ConstPointer> const& systems) 
{
	return RotamerLibrary::Pointer( new RotamerLibrary(systems) );
}

RotamerLibrary::RotamerLibrary(std::vector<System::ConstPointer> const& systems)
{
	nTotalRotamers_ = 0;
	nRotamerSets_ 	= 0;
	rotamerIndex_ 	= 0;
	systems_	= systems;

	for (int i = 0; i < (int)systems.size(); ++i)
	{
		RotamerSetData newRotamerData;
		newRotamerData.system 	= systems[i];

		foreach (Residue const& res, *systems[i])
		{
			newRotamerData.rotamers.push_back( boost::shared_ptr<RotamerSet>( new RotamerSet(res) ) );
			newRotamerData.baseIndices.push_back(rotamerIndex_++);

			nTotalRotamers_++;
		}
		rotamerData_[systems[i].get()] = newRotamerData;
	}

	GenerateRSetStorageIndices();
}

///////////////////////////////////////////////////////////////////////////
//////                       PUBLIC MEMBER FNS                       //////
///////////////////////////////////////////////////////////////////////////

RotamerLibrary::Pointer RotamerLibrary::Clone() const
{
	RotamerLibrary::Pointer clone = RotamerLibrary::Create(systems_);

	clone->rotamerData_ = rotamerData_;
	clone->nRotamerSets_ = nRotamerSets_;
	clone->nTotalRotamers_ = nTotalRotamers_;
	clone->rotamerIndex_ = rotamerIndex_;

	return clone;
}

///////////////////////////////////////////////////////////////////////////
//////                   ADDING/DELETING ROTAMERS                   ///////
///////////////////////////////////////////////////////////////////////////

void RotamerLibrary::GenerateRotamersWithoutInvalidation(Residue const& res, RotamerGenerator const& generator, bool discardOld)
{
	RotamerSetData& rset = GetRotamerSetData(res);
	int resIndex = rset.system->GetStorageIndex(res);
	RotamerSet const* oldRotamers = discardOld? 0 : rset.rotamers[resIndex].get();
	
	// add new rotamers
	rset.baseIndices[resIndex] = rotamerIndex_;
	rset.rotamers[resIndex] = generator.GenerateRotamers(res, oldRotamers);

	nTotalRotamers_ += rset.rotamers[resIndex]->NumRotamers();
	rotamerIndex_  += rset.rotamers[resIndex]->NumRotamers();
}

RotamerIDSet RotamerLibrary::GenerateRotamers(Residue const& res, RotamerGenerator const& generator, bool discardOld)
{
	RotamerIDSet invalidRotamers = InvalidateRotamers(res);
	GenerateRotamersWithoutInvalidation(res, generator, discardOld);
	
	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::GenerateRotamers(System::ConstPointer system, RotamerGenerator const& generator, bool discardOld)
{
	RotamerIDSet invalidRotamers = InvalidateRotamers(system);

	foreach (Residue const& res, *system)
		GenerateRotamersWithoutInvalidation(res, generator, discardOld);

	return invalidRotamers;
}

void RotamerLibrary::GenerateRotamers(RotamerGenerator const& generator, bool discardOld)
{
	foreach (System::ConstPointer sys, systems_)
		foreach (Residue const& res, *sys)
			GenerateRotamers(res, generator, discardOld);
}

RotamerIDSet RotamerLibrary::ReplaceSystem(System::ConstPointer oldSystem, System::ConstPointer newSystem)
{
	RotamerIDSet invalidRotamers = InvalidateRotamers(oldSystem);

	RotamerSetData rset = rotamerData_.find(oldSystem.get())->second;
	
	rset.rotamers.clear();
	rset.baseIndices.clear();

	foreach (Residue const& res, *newSystem)
	{
		rset.rotamers.push_back( boost::shared_ptr<RotamerSet>( new RotamerSet(res) ) );
		rset.baseIndices.push_back(rotamerIndex_++);

		nTotalRotamers_++;
	}

	rset.system = newSystem;
	rotamerData_.erase(oldSystem.get());
	rotamerData_[newSystem.get()] = rset;

	systems_.erase( remove(systems_.begin(), systems_.end(), oldSystem), systems_.end());
	systems_.push_back(newSystem);

	GenerateRSetStorageIndices();

	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::AddRotamers(Residue const& res, std::vector<Vector4> const& chis)
{
	RotamerIDSet invalidRotamers = InvalidateRotamers(res);

	std::list<Residue> rotList;

	std::pair<int, Vector4> oldChi;
	foreach (oldChi, GetChiMap(res))
		rotList.push_back(res.CloneWithChi(oldChi.second, res.GetSystem()));
	
	foreach (Vector4 const& chi, chis)
		rotList.push_back(res.CloneWithChi(chi, res.GetSystem()));

	RotamerSetData& rset = GetRotamerSetData(res);
	int resIndex = rset.system->GetStorageIndex(res);

	rset.baseIndices[resIndex] = rotamerIndex_;
	rset.rotamers[resIndex] = boost::shared_ptr<RotamerSet>( new RotamerSet(res, rotList) );

	nTotalRotamers_ += rset.rotamers[resIndex]->NumRotamers();
	rotamerIndex_  += rset.rotamers[resIndex]->NumRotamers();

	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::AddRotamers(RotamerLibrary::ConstPointer rlib, std::vector<int> const& ids)
{
	std::map< RotamerSet const*, std::vector<Vector4> > sortedChi;

	foreach (int id, ids)
	{
		int index = 0;
		RotamerSetData const& rsetd = rlib->FindRotamer(id, &index);

		RotamerSet const* rset = rsetd.rotamers[index].get();
		int baseIndex = rsetd.baseIndices[index]; 
		
		sortedChi[rset].push_back( rset->GetChi(id - baseIndex) );
	}

	RotamerIDSet invalidRotamers;

	std::map< RotamerSet const*, std::vector<Vector4> >::iterator i;
	for (i = sortedChi.begin(); i != sortedChi.end(); ++i)
	{
		Residue const& res = i->first->GetBaseResidue();

		if (std::find(systems_.begin(), systems_.end(), res.GetSystem()) == systems_.end())
			throw std::string("RotamerLibrary.AddRotamers: While importing rotamers from another RotamerLibrary, tried "
					  "to import a rotamer for a System not present in this RotamerLibrary.");

		RotamerIDSet resInvalidRotamers = AddRotamers(res, i->second);

		std::pair<int, int> entry;
		foreach (entry, resInvalidRotamers)
			invalidRotamers.push_back(entry);
	}

	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::DeleteRotamers(Residue const& res, std::vector<int> const& ids)
{
	RotamerIDSet invalidRotamers = InvalidateRotamers(res);

	RotamerSetData& rset = GetRotamerSetData(res);
	int resIndex = rset.system->GetStorageIndex(res);

	std::list<Residue> rotList;
	
	for (int i = 0; i < rset.rotamers[resIndex]->NumRotamers(); ++i)
	{
		int rotIndex = rset.baseIndices[resIndex] + i;

		if (std::find(ids.begin(), ids.end(), rotIndex) == ids.end())
			rotList.push_back(res.CloneWithChi( rset.rotamers[resIndex]->GetChi(i),
							    res.GetSystem() ));
	}

	rset.baseIndices[resIndex] = rotamerIndex_;
	rset.rotamers[resIndex] = boost::shared_ptr<RotamerSet>( new RotamerSet(res, rotList) );

	nTotalRotamers_ += rset.rotamers[resIndex]->NumRotamers();
	rotamerIndex_  += rset.rotamers[resIndex]->NumRotamers();

	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::DeleteRotamers(std::vector<int> const& ids)
{
	std::map< RotamerSet const*, std::vector<int> > sortedIDs;

	foreach (int id, ids)
	{
		int index = 0;
		RotamerSetData const& rsetd = FindRotamer(id, &index);

		sortedIDs[&*rsetd.rotamers[index]].push_back(id);
	}

	RotamerIDSet invalidRotamers;

	std::map< RotamerSet const*, std::vector<int> >::iterator i;
	for (i = sortedIDs.begin(); i != sortedIDs.end(); ++i)
	{
		Residue const& res = i->first->GetBaseResidue();

		RotamerIDSet resInvalidRotamers = DeleteRotamers(res, i->second);

		std::pair<int, int> entry;
		foreach (entry, resInvalidRotamers)
			invalidRotamers.push_back(entry);
	}

	return invalidRotamers;
}

///////////////////////////////////////////////////////////////////////////
//////                        OTHER ACCESSORS                        //////
///////////////////////////////////////////////////////////////////////////

Residue RotamerLibrary::GetRotamerResidue(int rotamerID) const
{
	int index = 0;
	RotamerSetData const& rset = FindRotamer(rotamerID, &index);

	return rset.rotamers[index]->GetRotamer( rotamerID - rset.baseIndices[index] );
}

std::pair<int, int> RotamerLibrary::GetRotamerIDRange(Residue const& res) const
{
	RotamerSetData const& rset = GetRotamerSetData(res);
	int resIndex = rset.system->GetStorageIndex(res);

	if (rset.rotamers[resIndex].get() == 0)
		throw std::string("GetRotamerIDRange: The rotamer set from this residue hasn't been generated yet.");

	const int baseIndex = rset.baseIndices[resIndex];
	return std::make_pair(baseIndex, baseIndex + rset.rotamers[resIndex]->NumRotamers() - 1);
}

int RotamerLibrary::GetRotamerID(RotamerSet const& rset, int rotIndex) const
{
	Residue const& res = rset.GetBaseResidue();
	RotamerSetData const& rdata = GetRotamerSetData(res);
	
	int resIndex = rdata.system->GetStorageIndex(res);
	return rdata.baseIndices[resIndex] + rotIndex;
}

int RotamerLibrary::GetStorageIndex(RotamerSet const& rset) const
{
	Residue const& res = rset.GetBaseResidue();
	return GetRotamerSetData(res).baseStorageIndex + res.GetSystem()->GetStorageIndex(res);
}

std::map<int, Vector4> RotamerLibrary::GetChiMap(Residue const& res) const
{
	RotamerSetData const& rdata = GetRotamerSetData(res);
	int resIndex = rdata.system->GetStorageIndex(res);

	std::map<int, Vector4> result;

	for (int i = 0; i < rdata.rotamers[resIndex]->NumRotamers(); ++i)
	{
		result[ rdata.baseIndices[resIndex] + i ] = rdata.rotamers[resIndex]->GetChi(i);
	}

	return result;
}

std::vector<Vector4> RotamerLibrary::GetRotamerChis(Residue const& res) const
{
	RotamerSetData const& rdata = GetRotamerSetData(res);
	int resIndex = rdata.system->GetStorageIndex(res);

	std::vector<Vector4> result;

	for (int i = 0; i < rdata.rotamers[resIndex]->NumRotamers(); ++i)
		result.push_back( rdata.rotamers[resIndex]->GetChi(i) );

	return result;
}

Vector4 RotamerLibrary::GetClosestChi(Residue const& res) const
{
	Vector4 currMinChi;
	FLOAT minDist = 1e10;

	Residue chiRes = res.Clone(res.GetSystem());

	std::pair<int, Vector4> libChi;
	foreach (libChi, GetChiMap(res))
	{
		chiRes.SetChi(libChi.second);
		chiRes.UpdateCoords();
		FLOAT dist = res.RMSD(chiRes);
		
		if (dist < minDist)
		{
			minDist = dist;
			currMinChi = libChi.second;
		}
	}

	return currMinChi;
}

System::Pointer RotamerLibrary::CreateSystem(System::ConstPointer baseSystem,
					     std::vector<int> const& rotamerIDs) const
{
	System::Pointer result = System::Create(baseSystem->GetAtomTypes(),
						baseSystem->GetAminoAcidData());

	RotamerDataMap::const_iterator rset_i = rotamerData_.find(baseSystem.get());
	if (rset_i == rotamerData_.end())
		throw std::string("CreateSystem: Base System not found.");
	
	if ((int)rotamerIDs.size() != nRotamerSets_)
		throw "CreateSystem: Rotamer ID array length (" + to_str(rotamerIDs.size()) 
		    + ") doesn't match NumRotamerSets (" + to_str(nRotamerSets_) + ").";

	RotamerSetData const& rset = rset_i->second;

	foreach (Residue const& res, *baseSystem)
	{
		int resIndex = baseSystem->GetStorageIndex(res);
		int rotIndex = rotamerIDs[resIndex + rset.baseStorageIndex] - rset.baseIndices[resIndex];

		Residue const& clone = rset.rotamers[resIndex]->GetRotamer(rotIndex).Clone(result);
		result->Append(clone);
	}

	return result;
}

std::vector<int> RotamerLibrary::TranslateRotamerIDs(RotamerLibrary::ConstPointer sourceLib, 
						     std::vector<int> const& rotamerIDs) const
{
	RotamerDataMap::const_iterator sourceRData, destRData;

	int IDindex = 0;
	std::vector<int> result;

	for(sourceRData = sourceLib->rotamerData_.begin(), destRData = rotamerData_.begin();
	    sourceRData != sourceLib->rotamerData_.end() && destRData != rotamerData_.end(); 
	    ++sourceRData, ++destRData)
	{
		RotamerSetData const& source = sourceRData->second;
		RotamerSetData const& dest = destRData->second;

		if (source.system != dest.system)
			throw std::string("TranslateRotamerIDs: Source and destination RotamerLibraries must have matching Systems.");
		
		for (int i = 0; i < (int)source.baseIndices.size(); ++i, ++IDindex)
		{
			if (rotamerIDs[IDindex] <   source.baseIndices[i] ||
			    rotamerIDs[IDindex] >=  source.baseIndices[i] + source.rotamers[i]->NumRotamers())
				throw std::string("TranslateRotamerIDs: Source IDs do not match source RotamerLibrary.");
				
			Vector4 const& chi = source.rotamers[i]->GetChi( rotamerIDs[IDindex] - source.baseIndices[i] );

			bool rotamer_found = false;
			for (int j = 0; j < dest.rotamers[i]->NumRotamers(); ++j)
			{
				Vector4 const& chi2 = dest.rotamers[i]->GetChi( j );
				int c = 0;
				for (; c < 4; ++c)
					if ( std::abs(chi(c) - chi2(c)) > 0.0001 )
						break;

				if (c == 4)
				{
					rotamer_found = true;
					result.push_back( dest.baseIndices[i] + j );
					break;
				}
			}

			if (!rotamer_found)
			{
				std::cout<<rotamerIDs[IDindex]<<"\n";
				throw std::string("TranslateRotamerIDs: Rotamer not found in dest RotamerLibrary.");
			}
		}
	}

	assert(result.size() == rotamerIDs.size());
	return result;
}

///////////////////////////////////////////////////////////////////////////
//////                           ITERATORS                           //////
///////////////////////////////////////////////////////////////////////////

std::pair<RotamerLibraryIterator, RotamerLibraryIterator> 
RotamerLibrary::SingleResidues() const
{
	// precondition: rotamerData_ should never be empty!

	RotamerDataMap::const_iterator begin = rotamerData_.begin();
	RotamerDataMap::const_iterator end1  = rotamerData_.end();
	--end1;

	return std::make_pair( RotamerLibraryIterator(begin, begin->second.rotamers.begin(), rotamerData_.end() ),
			       RotamerLibraryIterator(rotamerData_.end(), end1->second.rotamers.end(), rotamerData_.end()) );
}

std::pair<RotamerLibraryPairIterator, RotamerLibraryPairIterator>  
RotamerLibrary::PairResidues() const
{
	// precondition: rotamerData_ should never be empty!
	std::pair<RotamerLibraryIterator, RotamerLibraryIterator> singles = SingleResidues();

	RotamerLibraryIterator startPlus1 = singles.first;
	++startPlus1;

	return std::make_pair( RotamerLibraryPairIterator(singles.first, startPlus1, singles.second),
			       RotamerLibraryPairIterator(singles.second, singles.second, singles.second) );
}

//////////////////////////////////////////////////////////////////////////
/////                 DISCRETE SEARCH SPACE INTERFACE                /////
//////////////////////////////////////////////////////////////////////////

int RotamerLibrary::NumDimensions() const
{
	return NumRotamerSets();
}

int RotamerLibrary::NumElements() const
{
	return NumRotamers();
}

std::pair<int, int> RotamerLibrary::DimensionIDRange(int dimension) const
{
	if (dimension < 0)
		throw std::string("DimensionIDRange: Invalid dimension index.");

	foreach (System::ConstPointer sys, systems_)
	{
		if ((int)sys->size() <= dimension)
			dimension -= (int)sys->size();
		else
		{
			Residue const& res = sys->GetResidueByStorageIndex(dimension);
			return GetRotamerIDRange(res);
		}
	}

	throw std::string("DimensionIDRange: Invalid dimension index.");
}

///////////////////////////////////////////////////////////////////////////
//////                      PRIVATE MEMBER FNS                      ///////
///////////////////////////////////////////////////////////////////////////

RotamerIDSet RotamerLibrary::InvalidateRotamers(Residue const& res)
{
	RotamerIDSet invalidRotamers;
	
	RotamerSetData& rset = GetRotamerSetData(res);
	int resIndex = rset.system->GetStorageIndex(res);

	// remove old rotamers
	nTotalRotamers_ -= rset.rotamers[resIndex]->NumRotamers();

	invalidRotamers.push_back( std::make_pair(
					rset.baseIndices[resIndex],
					rset.rotamers[resIndex]->NumRotamers() ) );

	return invalidRotamers;
}

RotamerIDSet RotamerLibrary::InvalidateRotamers(System::ConstPointer oldSystem)
{
	RotamerIDSet invalidRotamers;

	RotamerDataMap::iterator rset_i = rotamerData_.find(oldSystem.get());
	if (rset_i == rotamerData_.end())
		throw std::string("InvalidateRotamers: Old system not found.");
	
	RotamerSetData& rset = rset_i->second;
	
	for (int i = 0; i < (int)rset.rotamers.size(); ++i)
		if (rset.rotamers[i].get() != 0)
		{
			nTotalRotamers_ -= rset.rotamers[i]->NumRotamers();

			invalidRotamers.push_back( std::make_pair(
							rset.baseIndices[i],
							rset.rotamers[i]->NumRotamers() ) );
		}

	return invalidRotamers;
}

RotamerSetData& RotamerLibrary::GetRotamerSetData(Residue const& res)
{
	RotamerDataMap::iterator rset_i = rotamerData_.find(res.GetSystem().get());
	if (rset_i == rotamerData_.end())
		throw std::string("GetRotamerSetData: System not found.");
	
	return rset_i->second;
}

RotamerSetData const& RotamerLibrary::GetRotamerSetData(Residue const& res) const
{
	RotamerDataMap::const_iterator rset_i = rotamerData_.find(res.GetSystem().get());
	if (rset_i == rotamerData_.end())
		throw std::string("GetRotamerSetData: System not found.");
	
	return rset_i->second;
}

RotamerSetData const& RotamerLibrary::FindRotamer(int rotamerID, int* indexOut) const
{
	for(RotamerDataMap::const_iterator rdata_i = rotamerData_.begin();
	    rdata_i != rotamerData_.end(); ++rdata_i)
	{
		RotamerSetData const& rset = rdata_i->second;

		for (int i = 0; i < (int)rset.baseIndices.size(); ++i)
		{
			if (rotamerID >= rset.baseIndices[i] &&
			    rotamerID <  rset.baseIndices[i] + rset.rotamers[i]->NumRotamers())
			{
				*indexOut = i;
				return rset;
			}
		}
	}

	throw "RotamerLibrary.GetFind: Rotamer index " + to_str(rotamerID) + " not found.";
}

void RotamerLibrary::GenerateRSetStorageIndices()
{
	nRotamerSets_ = 0;
	foreach (System::ConstPointer sys, systems_)
	{
		rotamerData_[sys.get()].baseStorageIndex = nRotamerSets_;
		nRotamerSets_ += (int)rotamerData_[sys.get()].rotamers.size();
	}
}


