// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "system.h"
#include "pdb.h"

#include <fstream>

#include <boost/weak_ptr.hpp>

////////////////////////////////////////////////////////////////////////////
////				CONSTRUCTORS 				////
////////////////////////////////////////////////////////////////////////////

System::Pointer System::Create(AtomTypes::ConstPointer atypes, AminoAcidDirectory::ConstPointer aadir)
{
	return System::Pointer(new System(atypes, aadir));
}

/*System::Pointer System::Join(std::vector<System::Pointer> const& systems)
{
	return Join( (std::vector<System::ConstPointer> const&)systems );
}*/

System::System(AtomTypes::ConstPointer atypes, AminoAcidDirectory::ConstPointer aadir)
		: firstIndex_(0), atypes_(atypes), aadir_(aadir)
{

}

// TODO: in the docs, warn that this doesn't check for chainbreaks!
// let's call the results of joining two separate chains together "undefined".
// IsLastInChain and IsFirstInChain certainly won't work, at least.
System::Pointer System::Join(std::vector<System::ConstPointer> const& systems)
{
	if (systems.empty())
		throw std::string("System.Join: Cannot join an empty set of systems.");

	System::Pointer result = Create(systems[0]->atypes_, systems[0]->aadir_);

	int nextID = 0;
	foreach (System::ConstPointer sys, systems)
		if (sys->size() != 0)
		{
			nextID = sys->residues_[0].GetID();
			break;
		}
	
	foreach (System::ConstPointer sys, systems)
	{
		if (sys->size() == 0) continue;

		if (sys->atypes_ != result->atypes_ || sys->aadir_ != result->aadir_)
			throw std::string("System.Join: System AtomTypes or AminoAcidDirectory mismatch.");

		foreach (Residue const& res, *sys)
		{
			result->residues_.push_back(res.Clone(result));
			result->residues_.back().id_ = nextID++;
		}
	}

	if (result->size() != 0)
		result->firstIndex_ = result->residues_[0].GetID();

	result->prevSys_ = systems.front()->prevSys_;
	result->nextSys_ = systems.back()->nextSys_;

	return result;
}

////////////////////////////////////////////////////////////////////////////
////				PUBLIC MEMBER FNS			////
////////////////////////////////////////////////////////////////////////////

void System::Append(Residue const& residue)
{
	if (residues_.empty())
		firstIndex_ = residue.GetID();
	else if (residue.GetID() != residues_.back().GetID() + 1)
	{
		throw std::string("Append: Non-consecutive residue index ") + to_str(residue.GetID());
	}

	residues_.push_back(residue);
	residues_.back().UpdateCoords();

	if (residues_.size() > 1)
	{
		Residue& prevResidue = *(residues_.end() - 2);
		prevResidue.UpdatePhiPsi();
	}
}

void System::Clear()
{
	firstIndex_ = 0;
	residues_.clear();
}
	
Residue& System::GetResidue(int index)
{	
	return GetResidueByStorageIndex(index - firstIndex_);
}

Residue const& System::GetResidue(int index) const
{
	return GetResidueByStorageIndex(index - firstIndex_);
}

Residue& System::GetResidueByStorageIndex(int storageIndex)
{	
	if (storageIndex < 0 || storageIndex >= (int)residues_.size())
		throw std::string("GetResidue: Index out of range.");

	return residues_[storageIndex];
}

Residue const& System::GetResidueByStorageIndex(int storageIndex) const
{
	if (storageIndex < 0 || storageIndex >= (int)residues_.size())
		throw std::string("GetResidue: Index out of range.");

	return residues_[storageIndex]; 
}

int System::GetStorageIndex(Residue const& res) const
{
	int storageIndex = res.GetID() - firstIndex_;
	if (res.GetSystem().get() != this || storageIndex < 0 || storageIndex >= (int)residues_.size())
		throw std::string("GetStorageIndex: Residue not found.");

	return storageIndex;
}

AminoAcidDirectory::ConstPointer System::GetAminoAcidData() const
{
	return aadir_;
}

AtomTypes::ConstPointer System::GetAtomTypes() const
{
	return atypes_;
}

bool System::IsFirstInChain(Residue const& res) const
{
	return GetStorageIndex(res) == 0 && (prevSys_.get() == 0 || prevSys_->size() == 0);
}

bool System::IsLastInChain(Residue const& res) const
{
	return GetStorageIndex(res) == (int)residues_.size() - 1 && (nextSys_.get() == 0 || nextSys_->size() == 0);
}

Residue& System::NextResidue(Residue const& res) 
{
	int storageIndex = GetStorageIndex(res);

	if (storageIndex == (int)residues_.size() - 1)
	{
		if (nextSys_.get() == 0 || nextSys_->size() == 0)
			throw "NextResidue: Residue " + to_str(res.GetID()) + " is the end of its chain.";
		return nextSys_->residues_[0];
	}
	
	return residues_[storageIndex+1];
}

Residue& System::PrevResidue(Residue const& res)
{
	int storageIndex = GetStorageIndex(res);

	if (storageIndex == 0)
	{
		if (prevSys_.get() == 0 || prevSys_->size() == 0)
			throw "PrevResidue: Residue " + to_str(res.GetID()) + " is the start of its chain.";
		return prevSys_->residues_[prevSys_->size() - 1];
	}
	
	return residues_[storageIndex-1];
}

// FIXME: code duplication

Residue const& System::NextResidue(Residue const& res) const 
{
	int storageIndex = GetStorageIndex(res);

	if (storageIndex == (int)residues_.size() - 1)
	{
		if (nextSys_.get() == 0 || nextSys_->size() == 0)
			throw "NextResidue: Residue " + to_str(res.GetID()) + " is the end of its chain.";
		return nextSys_->residues_[0];
	}
	
	return residues_[storageIndex+1];
}

Residue const& System::PrevResidue(Residue const& res) const
{
	int storageIndex = GetStorageIndex(res);

	if (storageIndex == 0)
	{
		if (prevSys_.get() == 0 || prevSys_->size() == 0)
			throw "PrevResidue: Residue " + to_str(res.GetID()) + " is the start of its chain.";
		return prevSys_->residues_[prevSys_->size() - 1];
	}
	
	return residues_[storageIndex-1];
}

System::Pointer System::Clone() const
{
	System::Pointer result = Create(atypes_, aadir_);
	*result = *this;

	for (ResidueList::iterator i = result->residues_.begin(); 
	     i != result->residues_.end(); ++i)
		*i = i->Clone(result);

	return result;
}

std::vector<System::Pointer> System::Split(std::vector<int> const& splitPoints) const
{
	std::vector<System::Pointer> result;
	std::vector<int> storageIndices;

	for (int i = 0; i < (int)splitPoints.size(); ++i)
	{
		result.push_back( Create(atypes_, aadir_) );
		storageIndices.push_back( GetStorageIndex( GetResidue(splitPoints[i]) ) );
	}
	result.push_back( Create(atypes_, aadir_) );

	std::sort(storageIndices.begin(), storageIndices.end());

	for (int i = 0; i < (int)storageIndices.size() + 1; ++i)
	{
		const int siStart = ( i == 0 ) ? 0 : storageIndices[i - 1];
		const int siEnd   = ( i == (int)storageIndices.size() ) ? (int)residues_.size() : storageIndices[i];

		for (int si = siStart; si < siEnd; ++si)
			result[i]->residues_.push_back(residues_[si].Clone(result[i]));

		if (siStart != siEnd)
			result[i]->firstIndex_ = result[i]->residues_[0].GetID();
		
		if (i != 0)
		{
			result[i]->prevSys_ = result[i-1];
			result[i-1]->nextSys_ = result[i];
		}
	}

	result.front()->prevSys_ = this->prevSys_;
	result.back()->nextSys_ = this->nextSys_;

	// we might end up with an empty system at the end of the list, so in that case
	// hook up the system *before* it to nextSys, or functions like IsLastInChain
	// may become confused.
	if (result.back()->size() == 0)
		result[ result.size() - 2 ]->nextSys_ = this->nextSys_;
		
	return result;
}

void System::MutateResidue( int id, AminoAcid const* newAminoAcid )
{
	int storageIndex = id - firstIndex_;
	
	if (storageIndex < 0 || storageIndex >= (int)residues_.size())
		throw std::string("MutateResidue: Invalid residue ID");

	Residue const& res = residues_[storageIndex];

	Residue const* prevRes = 0;
	if (storageIndex > 0)
		prevRes = &residues_[storageIndex - 1];

	// make a copy of the old residue with a new AA identity
	Residue mutatedResidue = res;
	mutatedResidue.Mutate(newAminoAcid, prevRes);
	
	// and copy the new residue into the old residue's location
	residues_[storageIndex] = mutatedResidue;
}

///////////////////////////////////////////////////////////////////////////
//////                              I/O                              //////
///////////////////////////////////////////////////////////////////////////

bool HandleDefaultRecord(std::string const& record)
{
	//std::cerr << "WARNING: SKIPPING UNKNOWN PDB RECORD '" << record.substr(0, 6) << "'\n";
	
	return true;
}		

//Parse the pdb file
std::istream& operator>> (std::istream& is, System::Pointer& sys)
{
	PDB::Reader reader;

	boost::function<bool (std::string const&)> handler(HandleDefaultRecord);
	boost::signals::connection c = reader.ConnectRecordSignal(handler);

	sys->residues_.clear();
	
	try
	{
		sys->firstIndex_ = 0;
		while (!is.eof())
		{
			// for now, ignore all non-ATOM records
			std::string nextRecord = reader.NextRecordType(is);
			while (nextRecord != "ATOM" && nextRecord != "EOF")
			{
				is >> reader;
				nextRecord = reader.NextRecordType(is);
			}

			// when we encounter an ATOM record, start
			// parsing into a Residue
			if (!is.eof())
			{
				Residue const* prevRes = 0;
				if (!sys->residues_.empty())
					prevRes = &sys->residues_.back();	
				
				PDBResidueFactory factory(sys, is, reader, prevRes);

				Residue res = factory.ConstructResidue();
				if (factory.Succeeded())
					sys->residues_.push_back(res);
			}

			if (!sys->residues_.empty())
				sys->firstIndex_ = sys->residues_[0].GetID();
		}

		if (!sys->residues_.empty())
		{
			// check that all the residues are in order
			for (int i = 1; i < (int)sys->residues_.size(); ++i)
			{
				int currIndex = sys->residues_[i-1].GetID() + 1;
				
				if (sys->residues_[i].GetID() != currIndex)
				{
					std::cerr << "WARNING: Non-consecutive residues found at seqpos " <<
						sys->residues_[i].GetID() << ". All subsequent residues will be re-numbered.\n";
					for (int j = i; j < (int)sys->residues_.size(); ++j)
						sys->residues_[j].SetID(currIndex++);

					break;
				}
			}

			foreach (Residue& res, sys->residues_)
			{
				res.UpdateCoords();
				res.StandardizeChis();
			}
		}
	}
	catch (std::string error)
	{
		std::cerr << "ERROR LOADING SYSTEM FROM PDB:\n" + error << std::endl;
		sys->Clear();
	}

	c.disconnect();

	return is;
}

std::ostream& operator<< (std::ostream& os, System::ConstPointer sys)
{
	try
	{
		PDB::Writer writer;
		foreach (Residue const& res, *sys)
			res.WriteToPDB(writer);

		os << writer;
	}
	catch (std::string error)
	{
		std::cerr << "ERROR WRITING SYSTEM TO PDB:\n" + error << std::endl;
	}

	return os;
}

std::ostream& operator<< (std::ostream& os, System::Pointer sys)
{
	return os << System::ConstPointer(sys);
}


