#include <iostream>
#include <fstream>

#include "some_chomp_headers.h"

///////////////////////////////////////////////////////////////////////////
//////                            GLOBALS                            //////
///////////////////////////////////////////////////////////////////////////

AtomTypes::ConstPointer 		atomTypes;
AminoAcidDirectory::ConstPointer 	aminoAcidDirectory;
DunbrackLibrary::ConstPointer 		dunbrackLibrary;
DataDirectory::ConstPointer 		dataDirectory;

///////////////////////////////////////////////////////////////////////////
//////                       HELPER FUNCTIONS                       ///////
///////////////////////////////////////////////////////////////////////////

void InitGlobals ()
{	
	atomTypes 		= AtomTypes::Create();
	aminoAcidDirectory 	= AminoAcidDirectory::Create(atomTypes);

	// If you are just going to run the executable from the dist directory these are OK
	dunbrackLibrary 	= DunbrackLibrary::Create(aminoAcidDirectory,"data/tables/dunbrack.bbdep"); 
	dataDirectory		= DataDirectory::Create("data/tables/data.dir");

	// Otherwise, to run the executable from any directory define full pathways here:
        //EnergyFunction::SetPluginPath("/home/csnow/code/sharpen_test/dist");
	//dunbrackLibrary 	= DunbrackLibrary::Create(aminoAcidDirectory, 
	//			"/home/csnow/code/sharpen_test/dist/data/tables/dunbrack.bbdep");
	//dataDirectory		= DataDirectory::Create("/home/csnow/code/sharpen_test/dist/data/tables/data.dir");
}

///////////////////////////////////////////////////////////////////////////
//////                          MAIN SCRIPT                          //////
///////////////////////////////////////////////////////////////////////////

int main (int argc, char** argv)
{
	if (argc < 3)
	{
		std::cout << "Usage: ./default_repack <input PDB> <output PDB>\n";
		return 1;
	}

	try
	{
		std::cout << "loading data...\n";
		InitGlobals();
		
		std::cout << "loading energy function...\n";
		EnergyFunction energyFunction("libCHOMP_ef_rosetta.so", atomTypes, dataDirectory, dunbrackLibrary);

		std::cout << "loading PDB...\n";
		System::Pointer sys = System::Create(atomTypes, aminoAcidDirectory);
		
		std::ifstream input(argv[1]);
		input >> sys;
	
		std::cout << "generating rotamers...\n";
		NeighborMap::ConstPointer nmap = NeighborMap::Create(sys, dataDirectory->GetPairCutoffTable());
		RotamerLibrary::Pointer rotLib = RotamerLibrary::Create(sys);
		
		RotamerGenerator rotGen(nmap, dunbrackLibrary);
		rotLib->GenerateRotamers(rotGen);
		
		RotamerIterationMap::ConstPointer iterMap = RotamerIterationMap::Create(rotLib);

		std::cout << "filling energy graph...\n";
		EnergyGraph energyGraph;
		energyFunction.FillEnergyGraph(rotLib, nmap, iterMap, &energyGraph);
		
		std::cout << "packing...\n";
		SimulatedAnnealingPacker sa;
		std::pair< std::vector<int>, float > result = sa.Pack(rotLib, &energyGraph);
		std::cout << "E = " << result.second << "\n";
		
		std::cout << "writing output...\n";
		System::ConstPointer sys2 = rotLib->CreateSystem(sys, result.first);

		std::ofstream output(argv[2]);
		output << sys2;
	}
	catch (std::string const& error)
	{	
		std::cout << "ERROR: " << error << std::endl;
		return 1;
	}
		
	return 0;
}

