// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include <string>
#include <utility>
#include <iostream>
#include <blitz/tinyvec-et.h>

#include "applicators.h"
#include "energy_terms.h"
#include "atom_pairwise.h"
#include "neighbor_map.h"
#include "data_table.h"

//////////////////////////////////////////////////////////////////////////
/////                       LOCAL DATA BUFFERS                       /////
//////////////////////////////////////////////////////////////////////////

blitz::Array<FLOAT, 2> 	AREnergies;  ///< Atom-Rotamer energies -- a temporary stack of energy values between a given atom and a set of rotamers
blitz::Array<FLOAT, 1> 	AEnergies;  ///< Atom-Atom energies -- a temporary stack of energy values to accumulate an entry in AREnergies

blitz::Array<bool, 2>  	SkipH;	    ///< An array that keeps track of whether we're skipping the hydrogens bonded to a given heavy atom

blitz::Array<bool, 2> 	HBChk;

//////////////////////////////////////////////////////////////////////////
/////                         LOCAL VARIABLES                        /////
//////////////////////////////////////////////////////////////////////////

// (these are global to make it easier to use functions in atom_pairwise_util.h without 
// passing in a ton of parameters, cluttering up the user-defined energy function code)

RotamerNode const* 	_atom1;
RotamerNode const* 	_atom2;
Residue     const* 	_res1;
Residue     const* 	_res2;
int			_res1_si;	// storage index for res1
int			_res2_si;	// storage index for res2
uint 			_flags1;
uint 			_flags2;


RotamerNode const* 	_hydrogen;
RotamerNode const* 	_acceptor;
Residue     const* 	_hres;
Residue     const* 	_ares;
int			_hres_si;	// storage index for hres
int			_ares_si;	// storage index for ares
uint 			_hflags;
uint 			_aflags;

int 			hv_index2;	// counts heavy atoms in rotamer set 2
int 			hdepth;		// keeps track of which heavy atom we're on in a rotamer

//////////////////////////////////////////////////////////////////////////
/////                          EXTERNAL DATA                         /////
//////////////////////////////////////////////////////////////////////////

extern unsigned int			rotamerCap_;
extern unsigned int			residueCap_;

extern DataDirectory::ConstPointer 	directory_;
extern AtomTypes::ConstPointer 		atypes_;
extern NeighborMap::ConstPointer	nmap_;

//////////////////////////////////////////////////////////////////////////
/////                            INCLUDES                            /////
//////////////////////////////////////////////////////////////////////////

#include "atom_pairwise_util.h"
#include "atom_pairwise_ef.h"

#include "hbond_util.h"
#include "hbond_ef.h"

//////////////////////////////////////////////////////////////////////////
/////                        PUBLIC INTERFACE                        /////
//////////////////////////////////////////////////////////////////////////

void Atom_Pairwise_Initialize()
{
	// set up the temporary data storage arrays
	AREnergies.resize(MAX_DIHEDRAL_ANGLES + 1,  rotamerCap_);
	AEnergies.resize (MAX_DIHEDRAL_ANGLES + 1);

	SkipH.resize(MAX_RESIDUE_HEAVY_ATOMS, MAX_RESIDUE_HEAVY_ATOMS * rotamerCap_);

	HBChk.resize(residueCap_, 2);
	HBChk = false;
}

bool Atom_Pairwise_TrieTrie(RotamerLibrary::ConstPointer rotlib,
			    RotamerSet const & trie1, 
		    	    RotamerSet const & trie2,
			    blitz::Array<FLOAT, 2>& output,
			    bool bbOnly)
{
	_res1 = &trie1.GetBaseResidue();
	_res2 = &trie2.GetBaseResidue();

	_res1_si = rotlib->GetStorageIndex(trie1);
	_res2_si = rotlib->GetStorageIndex(trie2);

	// bail out if these residues aren't neighbors
	if (!nmap_->AreNeighbors(*_res1, *_res2))
		return false;

	// pre-initialize some useful ranges (the data arrays are way oversized, so
	// it's useful to move around only the relevant subsets of them)
	//blitz::Range allRotamers1(0, trie1.NumRotamers() - 1);
	//blitz::Range allRotamers2(0, trie2.NumRotamers() - 1);

	//unsigned int altRotamerSize1 = trie1.NumRotamers();
	unsigned int altRotamerSize2 = trie2.NumRotamers();

	unsigned int rotamerStride = AREnergies.stride(blitz::firstDim);

	// we haven't decided to skip any hydrogens for these residues yet, so
	// clear out the array as it's currently full of old data
	//SkipH(blitz::Range(0, _res1->nHeavyAtoms - 1), 
	//      blitz::Range(0, _res2->nHeavyAtoms * trie2.NumRotamers() - 1)) = false;

	int _depth1 = 0;
	int _bdepth1 = 0;

	_atom1 = &trie1.GetTrieNodes()(0);
	const int nNodes1 = trie1.GetTrieNodes().extent(0);
	const int nNodes2 = trie2.GetTrieNodes().extent(0);

	const int maxDepth1 = _res1->GetAminoAcid().nTotalAtoms - 1;
	//const int maxDepth2 = _res2->GetAminoAcid().nTotalAtoms - 1;

	int nRot1 = 0;

	// for each atom in all the rotamers in residue 1
	for (int index1 = 0; index1 < nNodes1; ++index1, ++_atom1)
	{
		int _old_depth1 = _depth1;
		int _old_bdepth1 = _bdepth1;

		_depth1  = _atom1->depth;		// see rotamer_trie.h
		_bdepth1 = _atom1->branch_depth;
		//hdepth = _atom1->hv_depth;
		_flags1 = _res1->GetAminoAcid().atomFlags[_depth1];

		FLOAT* const ARDest = &AREnergies(_bdepth1, 0);
		
		if (_depth1 == 0) // if we're starting on a new residue, clear out the AREnergies array
			//AREnergies(_bdepth1, allRotamers2) = FLOAT(0.0);
			memset( ARDest, 0, altRotamerSize2 * sizeof(FLOAT) );

		else if (_bdepth1 > _old_bdepth1 || _depth1 <= _old_depth1)
		{
			// if we've just switched to a different rotamer, or we've moved to
			// a branching point in the trie structure, make a copy of the data
			// since the old data will be useful again later
		
			memcpy( ARDest, ARDest - rotamerStride, altRotamerSize2 * sizeof(FLOAT) );

			//	AREnergies(_bdepth1,   allRotamers2) = 
			//	AREnergies(_bdepth1-1, allRotamers2);
		}

		int nRot2 = 0;
	
		// if we're only scoring interactions between the sidechain atoms of the rotamers,
		// skip backbone atoms here
		if (bbOnly && !(_flags1 & ATM_FLAG_BB))
			continue;

		int _depth2 = 0;
		int _bdepth2 = 0;

		_atom2 = &trie2.GetTrieNodes()(0);
		const int maxDepth2 = _res2->GetAminoAcid().nTotalAtoms - 1;

		//hv_index2 = -1;
		// for each atom in all the rotamers in residue 2
		for (int index2 = 0; index2 < nNodes2; ++index2, ++_atom2)
		{
			int _old_depth2 = _depth2;
			int _old_bdepth2 = _bdepth2;

			_depth2  = _atom2->depth;
			_bdepth2 = _atom2->branch_depth;
			_flags2 = _res2->GetAminoAcid().atomFlags[_depth2];

			FLOAT* dataDest = &AEnergies(_bdepth2);

			if (_depth2 == 0) // if we're starting on a new residue, clear out the array
				*dataDest = FLOAT(0.0);
			else if (_bdepth2 > _old_bdepth2 || _depth2 <= _old_depth2)
			{
				// if we've just switched to a different rotamer, or we've moved
				// to a branching point in the trie structure, make a copy of the data
				// since the old data will be useful again later
				*dataDest = *(dataDest - 1);
			}
			
			// if we're only scoring interactions between the sidechain atoms of the rotamers,
			// skip backbone atoms here
			// FIXME: how does this interact with counters like hv_index2?
			if (bbOnly && !(_flags2 & ATM_FLAG_BB))
				continue;

			// keep track of which # heavy atom we're on for this residue (2)
			//if (!(_flags2 & ATM_FLAG_H))
			//	++hv_index2;

			// REAL ENERGY FUNCTION STARTS HERE
					
			FLOAT result = 0.0;
		
			Atom_PairwiseTerm<CombinedEnergyTerms>(&result, 0);
			if (FillHBondVars())
				HBondTerm<CombinedEnergyTerms>(&result, 0);

			*dataDest += result;
			// REAL ENERGY FUNCTION ENDS HERE

			// if we've finished scoring an atom-rotamer interaction, copy the data
			if (_depth2 == maxDepth2)
			{
				//AREnergies(_bdepth1, nRot2++) += *dataDest;
				ARDest[nRot2++] += *dataDest;
			}
		} // end for atom in all rotamers of residue 2

		// if we've finished scoring a rotamer-rotamer interaction, copy the data
		if (_depth1 == maxDepth1)
			//RREnergies(nRot1++, allRotamers2) = AREnergies(_bdepth1, allRotamers2);
			memcpy( &output(nRot1++, 0), ARDest, altRotamerSize2 * sizeof(FLOAT) );
	
	} // end for atom in all rotamers of residue 1

	return true;
}

FLOAT Atom_Pairwise_ResidueResidue(System::ConstPointer system,
				   Residue const& residue1, 
			       	   Residue const& residue2,
				   bool backboneOnly,
				   blitz::Array<FLOAT, 1>& energiesOut)
{
	FLOAT  total = 0.0;
	FLOAT* output = &(energiesOut(0));

	_res1 = &residue1;
	_res2 = &residue2;

	_res1_si = system->GetStorageIndex(residue1);
	_res2_si = system->GetStorageIndex(residue2);
	
	RotamerNode atom1, atom2;
	_atom1 = &atom1;
	_atom2 = &atom2;

	// bail out if these residues aren't neighbors
	if (!nmap_->AreNeighbors(*_res1, *_res2))
		return 0.0;

	// we haven't decided to skip any hydrogens for these residues yet, so
	// clear out the array as it's currently full of old data
	//SkipH(blitz::Range(0, _res1->nHeavyAtoms - 1), 
	//      blitz::Range(0, _res2->nHeavyAtoms - 1)) = false;

	//hdepth = -1;
	// for each atom in all the rotamers in residue 1
	for (int index1 = 0; index1 < _res1->GetAminoAcid().nTotalAtoms; ++index1)
	{
		atom1.pos = _res1->GetAtomCoords()[index1];
		atom1.hb_orient = _res1->GetOrientationVectors()[index1];
		atom1.depth = index1;
		_flags1 = _res1->GetAminoAcid().atomFlags[index1];
	
		// if we're only scoring backbone-backbone interactions, skip non-backbone atoms
		if (backboneOnly && !(_flags1 & ATM_FLAG_BB))
			continue;

		// keep track of which # heavy atom we're on for this residue (1)
		//if (!(_flags1 & ATM_FLAG_H))
		//	++hdepth;

		//hv_index2 = -1;
		// for each atom in residue 2
		for (int index2 = 0; index2 < _res2->GetAminoAcid().nTotalAtoms; ++index2)
		{
			atom2.pos = _res2->GetAtomCoords()[index2];
			atom2.hb_orient = _res2->GetOrientationVectors()[index2];
			atom2.depth = index2;
			_flags2 = _res2->GetAminoAcid().atomFlags[index2];
				
			// if we're only scoring backbone-backbone interactions, skip non-backbone atoms
			if (backboneOnly && !(_flags2 & ATM_FLAG_BB))
				continue;

			// keep track of which # heavy atom we're on for this residue (2)
			//if (!(_flags2 & ATM_FLAG_H))
			//	++hv_index2;

			// REAL ENERGY FUNCTION STARTS HERE
			
			Atom_PairwiseTerm<SeparateEnergyTerms>(&total, output);
			if (FillHBondVars())
				HBondTerm<SeparateEnergyTerms>(&total, output);

			// REAL ENERGY FUNCTION ENDS HERE
		
		} // end for atom in residue 2

	} // end for atom in all rotamers of residue 1

	return total;
}

bool Atom_Pairwise_IntraRes_Trie(RotamerLibrary::ConstPointer rotlib,
				 RotamerSet const & trie, 
				 blitz::Array<FLOAT, 1>& output)
{
	_res1 = &trie.GetBaseResidue();
	_res2 = &trie.GetBaseResidue();

	_res1_si = rotlib->GetStorageIndex(trie);	
	_res2_si = rotlib->GetStorageIndex(trie);	

	RotamerNode const* rotamer[MAX_RESIDUE_ATOMS];

	const int nNodes = trie.GetTrieNodes().extent(0);
	const int maxDepth = _res1->GetAminoAcid().nTotalAtoms - 1;

	RotamerNode const* _atom = &trie.GetTrieNodes()(0);
	int nRot = 0;

	for (int index = 0; index < nNodes; ++index, ++_atom)
	{
		rotamer[_atom->depth] = _atom;

		FLOAT sum = 0.0;
		if (_atom->depth == maxDepth)
		{
			for (int i = 0; i <= maxDepth; ++i)
			{
				_atom1 = rotamer[i];
				_flags1 = _res1->GetAminoAcid().atomFlags[_atom1->depth];

				for (int j = i + 1; j <= maxDepth; ++j)
				{
					_atom2 = rotamer[j];
					_flags2 = _res1->GetAminoAcid().atomFlags[_atom2->depth];

					Atom_Pairwise_IntraRes<CombinedEnergyTerms>(&sum, 0);
				}
			}

			output(nRot++) = sum;
		}
	}

	return true;
}

FLOAT Atom_Pairwise_IntraRes_Residue(System::ConstPointer system,
				     Residue const& res,
				     blitz::Array<FLOAT, 1>& energiesOut)
{
	_res1 = &res;
	_res2 = &res;

	_res1_si = system->GetStorageIndex(res);
	_res2_si = system->GetStorageIndex(res);
	
	RotamerNode atom1, atom2;
	_atom1 = &atom1;
	_atom2 = &atom2;

	FLOAT sum = 0.0;
	FLOAT* output = &(energiesOut(0));

	for (int i = 0; i < res.GetAminoAcid().nTotalAtoms; ++i)
	{
		atom1.pos = _res1->GetAtomCoords()[i];
		atom1.hb_orient = _res1->GetOrientationVectors()[i];
		atom1.depth = i;
		_flags1 = _res1->GetAminoAcid().atomFlags[i];

		for (int j = i + 1; j < res.GetAminoAcid().nTotalAtoms; ++j)
		{
			atom2.pos = _res2->GetAtomCoords()[j];
			atom2.hb_orient = _res2->GetOrientationVectors()[j];
			atom2.depth = j;
			_flags2 = _res2->GetAminoAcid().atomFlags[j];

			Atom_Pairwise_IntraRes<SeparateEnergyTerms>(&sum, output);
		}
	}

	return sum;
}


