// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 


#include "atom_pairwise_helpers.h"
#include "atom_pairwise_weight.h"

//////////////////////////////////////////////////////////////////////////
/////                           PARAMETERS                           /////
//////////////////////////////////////////////////////////////////////////

extern FLOAT parm_HCutoff;
extern FLOAT parm_ACutoff;

extern FLOAT parm_LJAttrWeight;
extern FLOAT parm_LJRepWeight;
extern FLOAT parm_SolvWeight;

extern FLOAT parm_LJRepWeight_IntraRes;

extern unsigned long long atomPairCount;
extern unsigned long long atomPairPassedCutoffCount;

//////////////////////////////////////////////////////////////////////////
/////                           INTERFACE                            /////
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
/////                          INTER-RESIDUE                         /////
//////////////////////////////////////////////////////////////////////////

template < void AddTerm(FLOAT, FLOAT*, FLOAT*, PairwiseEnergyTerm) >
inline void Atom_PairwiseTerm(FLOAT* total, FLOAT* output)
{
	// record a count of atom pairs for statistical purposes
	atomPairCount++;

	// isH == one of the atoms is a hydrogen
	bool isH = (Atom1_IsHydrogen() || Atom2_IsHydrogen());

	// if two heavy atoms are farther apart than a given threshold, you can tell CHOMP to
	// skip the evaluation of the pairwise terms for the bonded hydrogen atoms. 
	//if (isH && SkipThisHydrogen())
	//	return;

	FLOAT dist2 = Atoms_DistSquared();

	// see above
	//if (!isH && dist2 > parm_HCutoff)
	//	SkipBondedHydrogens();

	FLOAT atrE = 0.0, repE = 0.0, solvE = 0.0;

	if (dist2 < parm_ACutoff)
	{
		// record a count of cutoff-passing atom pairs for statistical purposes
		atomPairPassedCutoffCount++;

		// Rosetta reduces the weight of some backbone atom pairs. see atom_pairwise_weight.h.
		FLOAT weight = Rosetta_PairWeight();
		if (weight != 0.0)
		{
			int type1 = Atom1_GetType(AtomType::ROSETTA_TYPE),
	     		    type2 = Atom2_GetType(AtomType::ROSETTA_TYPE);
		
			int bin1;
			FLOAT frac;
			Atom_Dist2Interp(dist2, bin1, frac);

			if (isH) // hydrogens only get a VDW repulsive term
			{
				repE = weight * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetLJRepulsiveTable()); 
				repE *= parm_LJRepWeight;
			}
			else
			{	// heavy atoms get the full complement
				atrE = weight * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetLJAttractiveTable());
				repE = weight * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetLJRepulsiveTable());

				solvE = weight * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetSolvationTable1()) +
				        weight * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetSolvationTable2());
			
				atrE *= parm_LJAttrWeight;
				repE *= parm_LJRepWeight;
				solvE *= parm_SolvWeight;
			}
		}
	}

	AddTerm(atrE,  total, output, PTERM_AP_ATR);
	AddTerm(repE,  total, output, PTERM_AP_REP);
	AddTerm(solvE, total, output, PTERM_AP_SOLV);
}

//////////////////////////////////////////////////////////////////////////
/////                          INTRA-RESIDUE                         /////
//////////////////////////////////////////////////////////////////////////

template < void AddTerm(FLOAT, FLOAT*, FLOAT*, OneBodyEnergyTerm) >
inline void Atom_Pairwise_IntraRes(FLOAT* total, FLOAT* output)
{
	FLOAT ir_cp = directory_->GetIntraRes_CountPair()((int)_res1->GetAminoAcid().aminoAcid, 
				                         (int)_atom1->depth, (int)_atom2->depth);

	FLOAT repE = 0.0;
	if (ir_cp != 0.0)
	{
		FLOAT dist2 = Atoms_DistSquared();
	
		if (dist2 < parm_ACutoff)
		{
			int type1 = Atom1_GetType(AtomType::ROSETTA_TYPE),
	     		    type2 = Atom2_GetType(AtomType::ROSETTA_TYPE);
		
			int bin1;
			FLOAT frac;
			Atom_Dist2Interp(dist2, bin1, frac);

			repE = ir_cp * Atom_DistType2_Term(bin1, frac, type1, type2, directory_->GetLJRepulsiveTable()); 
		}		
	}

	AddTerm(repE * parm_LJRepWeight_IntraRes, total, output, STERM_AP_REP);
}


