// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROSETTA_COUNT_PAIR
#define IVL_ROSETTA_COUNT_PAIR

#include "../atom_pairwise.h"

/** \brief Emulates Rosetta's count_pair function.
 *
 * This function reduces the pairwise atom energies between certain backbone atoms
 * in neighboring residues. This is apparently necessary because these interaction
 * energies are already included in some per-residue energy terms. So, for a given
 * pair of atoms, this function returns a weight (None, Half, or Full) which should
 * then be used to scale the pairwise energy value for that pair.
 * If the result is None, you can just skip that pair altogether to save some
 * processing time.
 */

// TODO: This can easily be refactored into a simple table lookup, which should
// provide a pretty decent speed increase

inline FLOAT Rosetta_PairWeight()
{
	const uint ri1 = _res1->GetID(), ri2 = _res2->GetID();
	const uint type1 = _res1->GetAminoAcid().atomType[_atom1->depth], 
		   type2 = _res2->GetAminoAcid().atomType[_atom2->depth];

	const FLOAT wFull = FLOAT(1.0);
	const FLOAT wHalf = FLOAT(0.2);
	const FLOAT wNone = FLOAT(0.0);

	// don't count atoms in the same residue
	if (ri1 == ri2)
		return wNone;

	const uint rtype1 = atypes_->GetType(type1, AtomType::ROSETTA_TYPE),
		   rtype2 = atypes_->GetType(type2, AtomType::ROSETTA_TYPE);

	// Note that both halves of this if statement are pretty much the same.
	// This would be more elegant if I just used a swap(), but this is how
	// Rosetta does it (presumably for performance).
	// TODO: profile this later to see if it's really as dumb as it looks
	if (ri1 < ri2)
	{
		// this function doesn't deal with residues more than 2 apart
		if (ri2 - ri1 > 2)
			return wFull;

		if (ri2 - ri1 == 2)
		{
			// for 2-apart residues, the C-N interactions are down-weighted
			
			if (rtype1 == AtomType::Rosetta::BB_CO && 
			   (rtype2 == AtomType::Rosetta::BB_N || 
			    rtype2 == AtomType::Rosetta::PRO_N))
				return wHalf;
			else
				return wFull;
		}

		// for neighboring residues, the weights are looked up in a table
		return directory_->GetCountPairTable()(atypes_->GetType(type1, AtomType::COUNT_PAIR_TYPE),
		                                       atypes_->GetType(type2, AtomType::COUNT_PAIR_TYPE));
	}
	else	// same as the other half, but with ri1 and ri2 swapped
	{
		if (ri1 - ri2 > 2)
			return wFull;

		if (ri1 - ri2 == 2)
		{
			if (rtype2 == AtomType::Rosetta::BB_CO && 
			   (rtype1 == AtomType::Rosetta::BB_N || 
			    rtype1 == AtomType::Rosetta::PRO_N))
				return wHalf;
			else
				return wFull;
		}

		return directory_->GetCountPairTable()(atypes_->GetType(type2, AtomType::COUNT_PAIR_TYPE),
		                                       atypes_->GetType(type1, AtomType::COUNT_PAIR_TYPE));
	}
}

#endif


