// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

//////////////////////////////////////////////////////////////////////////
/////                        HBOND PARAMETERS                        /////
//////////////////////////////////////////////////////////////////////////

struct HBondComponents
{
	HBondComponents() {}

	HBondComponents(FLOAT _r, FLOAT _xD, FLOAT _xH):
		r(_r), xD(_xD), xH(_xH) 
	{
	}

	FLOAT r, xD, xH;
};

inline HBondComponents GetHBondParameters()
{
	Vector3 AH = _hydrogen->pos - _acceptor->pos;
	FLOAT r2 = blitz::dot(AH, AH);
	
	FLOAT r  = std::sqrt(r2);
	AH *= 1.0 / r;
					
	Vector3 HD  = _hydrogen->hb_orient;
	Vector3	A1A = _acceptor->hb_orient;

	FLOAT xD = blitz::dot(AH, HD);
	FLOAT xH = blitz::dot(A1A, AH);
	
	return HBondComponents(r, xD, xH);
}

//////////////////////////////////////////////////////////////////////////
/////                           HBOND TYPES                          /////
//////////////////////////////////////////////////////////////////////////

enum HBondType 
{
	HBOND_SC_SC,		// sidechain-sidechain	
	HBOND_SC_BB,		// sidechain-backbone
	HBOND_BB_BB_SR,		// backbone-backbone (<= srCutoff residues apart)
	HBOND_BB_BB_LR		// backbone-backbone (>  srCutoff residues apart)
};

enum HBondPolynomial
{
	HBPOLY_BB_HELIX,	// alpha-helix bond
	HBPOLY_BB_SHEET,	// beta-sheet bond
	HBPOLY_BB_SHORT,	// short-range bond (loop?)
	HBPOLY_SC_SP2_SR,	// SP2 hybridization, short-range
	HBPOLY_SC_SP2_LR,	// SP2 hybridization, long-range
	HBPOLY_SC_SP3_SR,	// SP3 hybridization, short-range
	HBPOLY_SC_SP3_LR,	// SP3 hybridization, long-range
	HBPOLY_RING_N_SR,	// ring-nitrogen hybridization, short-range
	HBPOLY_RING_N_LR	// ring-nitrogen hybridiztaion, long-range
};

//////////////////////////////////////////////////////////////////////////
/////                        HELPER FUNCTIONS                        /////
//////////////////////////////////////////////////////////////////////////

inline int GetHBondType(int resDist)
{
	if (!( Acceptor_IsBackbone() || Hydrogen_IsBackbone() ))
		return HBOND_SC_SC;
	
	else if (!( Acceptor_IsBackbone() && Hydrogen_IsBackbone() ))
		return HBOND_SC_BB;

	else if (resDist <= parm_HBond_srCutoff)
		return HBOND_BB_BB_SR;

	else    return HBOND_BB_BB_LR;
}

inline int GetHBondPolynomial(int hbType, int resDist, FLOAT r)
{
	scInterpolate = false;
	int type;

	if (hbType == HBOND_SC_SC || hbType == HBOND_SC_BB)
	{
		if (Acceptor_GetType(AtomType::HYBRIDIZATION) == AtomType::Hybridization::SP3)
			type = HBPOLY_SC_SP3_SR;
		else if (Acceptor_GetType(AtomType::HYBRIDIZATION) == AtomType::Hybridization::RING)
			type = HBPOLY_RING_N_SR;
		else
			type = HBPOLY_SC_SP2_SR;
		
		if ( r > parm_HBond_scInterpMax )
			type += 1;
		else if (r > parm_HBond_scInterpMin )
			scInterpolate = true;
	}
	else
	{
		if      (resDist == 4) type = HBPOLY_BB_HELIX;
		else if (resDist > 4)  type = HBPOLY_BB_SHEET;
		else		       type = HBPOLY_BB_SHORT;
	}

	return type;
}

inline void InterpolateSC(int polyType, HBondComponents params, HBondComponents* energies)
{
	FLOAT dfac = 1.0 / ( parm_HBond_scInterpMax - parm_HBond_scInterpMin );
	FLOAT frac = ( parm_HBond_scInterpMax - params.r ) * dfac;
	FLOAT rem = 1.0 - frac;

	int longRangeType = polyType + 1;

	FLOAT e_xh2 = Atom_FloatInt_Term(params.xH, longRangeType, directory_->GetHB_XHTable(),
		    			 1000.0/(parm_HBond_xHMax - parm_HBond_xHMin), parm_HBond_xHMin);
	FLOAT e_xd2 = Atom_FloatInt_Term(params.xD, longRangeType, directory_->GetHB_XDTable(),
	    				 1000.0/(parm_HBond_xDMax - parm_HBond_xDMin), parm_HBond_xDMin);
					
	energies->xD = frac * energies->xD + rem * e_xd2;
	energies->xH = frac * energies->xH + rem * e_xh2;
}

inline void FadeOut(FLOAT value, FLOAT max, FLOAT edge, FLOAT* out1, FLOAT *out2)
{
	FLOAT dfac = 1.0 / ( max - edge );
	FLOAT frac = ( max - value ) * dfac;

	*out1 *= frac;
	*out2 *= frac;
}

// TODO: figure this out and pull the numbers out into parameters
inline FLOAT assign_single_weight(int nb)
{
	if ( nb < 7 ) {
		return 0.2;
	} else if ( nb > 24 ) {
		return  1.0;
	} else {
		return (nb-2.75)/21.25;
	}
}

inline FLOAT get_hb_env_weight_lin(int nb1, int nb2)
{
	float weight1 = assign_single_weight(nb1);
	float weight2 = assign_single_weight(nb2);
	return (weight1+weight2)/2.0;
}



