// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ATOM_PAIRWISE_UTIL
#define IVL_ATOM_PAIRWISE_UTIL

#include "atom_pairwise.h"
#include "interpolate.h"

/** \file atom_pairwise_util.h
 *
 * \brief This file contains small utility functions to be used in the user-defined energy function code.
 * 
 * The functions here hide away some of the complexity involved in dealing with the data structures involved
 * in representing the rotamers efficiently. They are all inline, so they don't impose any overhead.
 */

/// \brief Returns the distance squared (in angstroms squared) between the two atoms currently being evaluated.
inline FLOAT Atoms_DistSquared()
{
	const Position temp = _atom1->pos - _atom2->pos;
	return blitz::dot(temp, temp);
}

/// \brief Returns the type of the atom currently being evaluated within the given type system.
///
/// \param typeSystem One of the enumerated type systems (see atom_types, etc.)
inline int Atom1_GetType(AtomType::TypeSystem typeSystem)
{
	return atypes_->GetType(_res1->GetAminoAcid().atomType[_atom1->depth], typeSystem); 
}

/// \brief Returns the type of the atom currently being evaluated within the given type system.
///
/// \param typeSystem One of the enumerated type systems (see atom_types, etc.)
inline int Atom2_GetType(AtomType::TypeSystem typeSystem)
{
	return atypes_->GetType(_res2->GetAminoAcid().atomType[_atom2->depth], typeSystem); 
}

/// \brief Returns true iff the atom currently being evaluated is a hydrogen.
inline bool Atom1_IsHydrogen()
{
	return _flags1 & ATM_FLAG_H;
}

/// \brief Returns true if the atom currently being evaluated is a hydrogen.
inline bool Atom2_IsHydrogen()
{
	return _flags2 & ATM_FLAG_H;
}

/// \brief Informs the iteration code that the subsequent hydrogens should not be evaluated.
/// NOTE: You should only call this if both atom1 and atom2 are heavy atoms!
inline void SkipBondedHydrogens()
{
	SkipH(hdepth, hv_index2) = true;
}

/// \brief Returns true if the current hydrogen should be skipped.
/// NOTE: You should call this only if either atom1 or atom2 is a hydrogen!
inline bool SkipThisHydrogen()
{
	return SkipH(hdepth, hv_index2);
}

/// \brief Returns the number of bonds separating two atoms
inline int GetNumBonds()
{
        int res1_ID = _res1->GetID();
        int res2_ID = _res2->GetID();

        //Intra-residue bond length lookup
        if (res1_ID == res2_ID)
                return _res1->GetAminoAcid().bondLengths[_atom1->depth][_atom2->depth];
        //Inter-residue bond length lookup
        //NOTE: assumes res1_ID < res2_ID
        else
        {
                int BondsToC = _res1->GetAminoAcid().bondLengths[_atom1->depth][_res1->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "C")],
                        BondsToN = _res2->GetAminoAcid().bondLengths[_atom2->depth][_res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "N")];

                return BondsToC + BondsToN + 3 * abs(res1_ID - res2_ID) - 2;
        }
}

#endif

