// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "atom_pairwise.h"
#include "rotamer_pairwise.h"
#include "rotamer_onebody.h"

#include "amino_acid.h"
#include "rotamers.h"
#include "energy_graph.h"
#include "system.h"
#include "atom_type.h"
#include "rotamer_iteration_map.h"
#include "data_table.h"
#include "energy_terms.h"
#include "parameters.h"
#include "runtime_stats.h"

#include <iostream>

/** \file energy_function/main.cpp
 *
 * \brief This is where the main scoring function is implemented.
 *
 * This file contains the main program flow logic for the energy function scoring
 * mechanism. It's very simple, so it can easily be edited to create a different
 * energy function.
 */

//////////////////////////////////////////////////////////////////////////
/////                           LOCAL DATA                           /////
//////////////////////////////////////////////////////////////////////////

unsigned int 			rotamerCap_;
unsigned int			residueCap_;

DataDirectory::ConstPointer 	directory_;
DunbrackLibrary::ConstPointer 	dunLib_;
AtomTypes::ConstPointer 	atypes_;
NeighborMap::ConstPointer	nmap_;
RuntimeStats::Pointer		rstats_;

blitz::Array<FLOAT, 2> REnergies2D_FA;   
blitz::Array<FLOAT, 2> REnergies2D_Rot;  

blitz::Array<FLOAT, 1> REnergies1D_FA;  
blitz::Array<FLOAT, 1> REnergies1D_Rot; 

unsigned long long atomPairCount;
unsigned long long atomPairPassedCutoffCount;
unsigned long long atomPairHBCount;

unsigned long long TorsionCount;
unsigned long long ImptorsCount;
unsigned long long bondstretchCount;
unsigned long long anglebendCount;
unsigned long long vdwCount;
unsigned long long coulCount;

const FLOAT epsilon = 0.0001;

///////////////////////////////////////////////////////////////////////////
//////                   PRIVATE HELPER FUNCTIONS                   ///////
///////////////////////////////////////////////////////////////////////////

unsigned long AllocateMemory(int rotamerCap, int residueCap)
{
	unsigned long mem1Dkb = rotamerCap * 2 * sizeof(FLOAT) / 1024;
	unsigned long mem2Dkb = mem1Dkb * rotamerCap;

	REnergies2D_FA.resize(rotamerCap, rotamerCap);
	REnergies2D_Rot.resize(rotamerCap, rotamerCap);

	REnergies1D_FA.resize(rotamerCap);
	REnergies1D_Rot.resize(rotamerCap);

	rotamerCap_ = rotamerCap;
	residueCap_ = residueCap;

	return mem1Dkb + mem2Dkb;
}

void AdjustMemoryCaps(System::ConstPointer sys)
{
	if (sys->size() > residueCap_)
		residueCap_ = (unsigned int)(sys->size() * 1.2);
}

void AdjustMemoryCaps(RotamerLibrary::ConstPointer rotLib)
{
	if ((unsigned int)rotLib->NumRotamerSets() > residueCap_)
		residueCap_ = (unsigned int)(rotLib->NumRotamerSets() * 1.2);

	int maxRots = 0;
	foreach (RotamerSet const& rset, rotLib->SingleResidues())
	{
		if (rset.NumRotamers() > maxRots)
			maxRots = rset.NumRotamers();
	}

	if ((unsigned int)maxRots > rotamerCap_)
	{
		unsigned long memKb = AllocateMemory( (unsigned int)( maxRots * 1.2 ), residueCap_ );

		std::cerr << "Rotamer library contains too many rotamers per residue.\n";
		std::cerr << "Re-initializing energy function with larger scratch memory space...\n";
		std::cerr << "   now using approx. " << memKb/1024 << "MB RAM.\n";
	}
}

void ResetOpCounts()
{
	atomPairCount = 0;
	atomPairHBCount = 0;
	atomPairPassedCutoffCount = 0;

	TorsionCount = 0;
	ImptorsCount = 0;
	bondstretchCount = 0;
	anglebendCount = 0;
	vdwCount = 0;
	coulCount = 0;
}

//////////////////////////////////////////////////////////////////////////
/////                        PUBLIC INTERFACE                        /////
//////////////////////////////////////////////////////////////////////////

extern "C" DllExport void Initialize(DataDirectory::ConstPointer directory,
				     DunbrackLibrary::ConstPointer dunLib,
				     AtomTypes::ConstPointer atypes,
				     RuntimeStats::Pointer rstats
					= RuntimeStats::Pointer())
{
	unsigned long memKb = AllocateMemory(MAX_RESIDUE_ROTAMERS, MAX_RESIDUES);
	//std::cerr << "Initializing energy function scratch memory... using approx. " 
	//	  << memKb/1024 << "MB RAM.\n";

	directory_ = directory;
	dunLib_ = dunLib;
	atypes_ = atypes;
	rstats_ = rstats;
}

extern "C" DllExport void ComputePairwiseEnergies(System::ConstPointer residues,
			     			  NeighborMap::ConstPointer nmap,
			     			  blitz::Array<FLOAT, 3>& output)
{
	nmap_ = nmap;

	blitz::Array<FLOAT, 1> temp;
	temp.resize(N_PAIRWISE_ENERGY_TERMS);

	output.resize(residues->size(), residues->size(), N_PAIRWISE_ENERGY_TERMS);
	output = 0.0;

	System::const_iterator res_i, res_i2;
	AdjustMemoryCaps(residues);
	Atom_Pairwise_Initialize();

//	ResetOpCounts();
	
	// compute backbone energies once to fill in the HBond tables (this is fast)
/*	for (res_i = residues->begin(); res_i != residues->end(); ++res_i)
	{
		for (res_i2 = res_i, ++res_i2; 
		     res_i2 != residues->end(); ++res_i2)
		{
			Atom_Pairwise_ResidueResidue(*res_i, *res_i2, true, temp);
		}
	}
*/
	for (res_i = residues->begin(); res_i != residues->end(); ++res_i)
		for (res_i2 = res_i, ++res_i2; res_i2 != residues->end(); ++res_i2)
		{
			temp = 0.0;
			Atom_Pairwise_ResidueResidue(*res_i, *res_i2, false, temp);
		 	Rotamer_Pairwise_ResidueResidue(*res_i, *res_i2, temp);

			output(	res_i->GetSystem()->GetStorageIndex(*res_i), 
				res_i2->GetSystem()->GetStorageIndex(*res_i2),
				blitz::Range::all()) = temp;
		}

	if (rstats_.get() != NULL)
	{
		rstats_->Increment(RSTATS_ATOM_PAIRWISE, atomPairCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_PASSED_CUTOFF, atomPairPassedCutoffCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_HB, atomPairHBCount);
	}
}

extern "C" DllExport void ComputeOneBodyEnergies(System::ConstPointer residues,
  	       					 NeighborMap::ConstPointer nmap,
						 blitz::Array<FLOAT, 2>& output)
{
	nmap_ = nmap;

	blitz::Array<FLOAT, 1> temp;
	temp.resize(N_ONEBODY_ENERGY_TERMS);
	
	output.resize(residues->size(), N_ONEBODY_ENERGY_TERMS);
	output = 0.0;

	AdjustMemoryCaps(residues);
	Atom_Pairwise_Initialize();

	System::const_iterator res_i;
	for (res_i = residues->begin(); res_i != residues->end(); ++res_i)
	{
		temp = 0.0;
		Rotamer_OneBody_Residue(*res_i, temp);
		Atom_Pairwise_IntraRes_Residue(*res_i, temp);

		output(res_i->GetSystem()->GetStorageIndex(*res_i), blitz::Range::all()) = temp;
	}

	if (rstats_.get() != NULL)
	{
		rstats_->Increment(RSTATS_ATOM_PAIRWISE, atomPairCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_PASSED_CUTOFF, atomPairPassedCutoffCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_HB, atomPairHBCount);
	}
}

extern "C" DllExport FLOAT ComputeScore(System::ConstPointer residues, 
					NeighborMap::ConstPointer nmap)
{
	ResetOpCounts();
	std::cout<<"\nNumber of residues: "<<residues->size()<<"\n";

	System::const_iterator res_i;
	int totalAtoms = 0;
	for (res_i = residues->begin(); res_i != residues->end(); ++res_i)
		totalAtoms += res_i->GetAminoAcid().nTotalAtoms;
	std::cout<<"Total number of atoms: "<<totalAtoms<<"\n\n";
	
	nmap_ = nmap;

	blitz::Array<FLOAT, 1> pairwise, onebody;
	pairwise.resize(N_PAIRWISE_ENERGY_TERMS);
	onebody.resize(N_ONEBODY_ENERGY_TERMS);
	pairwise = 0.0;
	onebody = 0.0;

	blitz::Array<FLOAT, 2> res_onebody;
	blitz::Array<FLOAT, 3> res_pairwise;
	
	ComputeOneBodyEnergies(residues, nmap, res_onebody);
	ComputePairwiseEnergies(residues, nmap, res_pairwise);

        std::cout<<"TorsionCount: "<<TorsionCount<<"\n";
        std::cout<<"ImptorsCount: "<<ImptorsCount<<"\n";
        std::cout<<"bondstretchCount: "<<bondstretchCount<<"\n";
        std::cout<<"anglebendCount: "<<anglebendCount<<"\n";
        std::cout<<"vdwCount: "<<vdwCount<<"\n";
        std::cout<<"coulCount: "<<coulCount<<"\n\n";
	
	for (int i = 0; i < (int)residues->size(); ++i)
		for (int j = i+1; j < (int)residues->size(); ++j)
			pairwise += res_pairwise(i,j, blitz::Range::all());
	
	for (int i = 0; i < (int)residues->size(); ++i)
		onebody += res_onebody(i, blitz::Range::all());
	
	FLOAT total = 0.0;
	for (int i = 0; i < N_PAIRWISE_ENERGY_TERMS; ++i)
		total += pairwise(i);
	
	for (int i = 0; i < N_ONEBODY_ENERGY_TERMS; ++i)
		total += onebody(i);
/*	print all one-body energies
	for (int i = 0; i < (int)residues->size(); ++i)
	{
		FLOAT blah = 0.0;
		for (int j = 0; j < res_onebody.extent(1); ++j)
			blah = blah + res_onebody(i,j);

		std::cout<<i<<"\t"<<blah<<"\n";
	}
	print all pairwise energies
	for (int i = 0; i < (int)residues->size(); ++i)
	{
		for (int j = i; j < (int)residues->size(); ++j)
		{
			FLOAT blah = 0.0;
			for (int k = 0; k < res_pairwise.extent(2); ++k)
				blah = blah + res_pairwise(i,j,k);
			std::cout<<i<<"\t"<<j<<"\t"<<blah<<"\n";
		}
	}
*/

	std::string e_terms[] = {"Coulombic", "Lennard Jones", "Bond Stretch", "Angle Bend", "Torsion", "Improper Torsion"};
	for (int i = 0; i < N_PAIRWISE_ENERGY_TERMS; ++i)
		std::cout<<e_terms[i]<<": "<<pairwise(i) + onebody(i)<<"\n";
	std::cout<<"\n";	

	return total;
}

extern "C" DllExport void MakeEnergyGraph(RotamerLibrary::ConstPointer rotamers,
					  NeighborMap::ConstPointer nmap,
					  RotamerIterationMap::ConstPointer rmap,
					  EnergyGraph* graphOut)
{
	FLOAT weight;
	nmap_ = nmap;
	
	System::const_iterator res_i, res_i2;
	std::pair<RotamerSet const*, RotamerSet const*> pair;

	AdjustMemoryCaps(rotamers);
	Atom_Pairwise_Initialize();

	ResetOpCounts();
	// compute backbone energies once to fill in the HBond tables (this is fast)
/*	foreach (pair, rotamers->PairResidues())
		if (rmap->CountPair(pair.first->GetBaseResidue(), pair.second->GetBaseResidue()) > epsilon)
			Atom_Pairwise_TrieTrie(*pair.first, *pair.second, REnergies2D_FA, true);*/

	// create all the vertices (rotamers) in the energy-graph, and assign them
	// single-body energies when needed. 

	foreach (RotamerSet const& single, rotamers->SingleResidues())
	{
		// if the single-body energy is not to be included in the graph, we still need the 
		// corresponding nodes but we can just set the energy to zero.
		weight = rmap->CountSingle(single.GetBaseResidue());
		if (weight < epsilon)
		{
			// make sure the nodes aren't already present in the graph, if we're going to 
			// add fresh ones regardless of the RotamerIterationMap

			for ( int i = 0; i < single.NumRotamers(); ++i )
			{
				int rotamer_id = rotamers->GetRotamerID(single, i);
				bool node_exists = false;
				foreach (EnergyGraph::vertex_descriptor v, boost::vertices(*graphOut))
				{
					if ( (*graphOut)[v].elementID == rotamer_id )
					{
						node_exists = true;
						break;
					}
				}

				if (!node_exists)
					boost::add_vertex(EnergyGraphNode(rotamer_id, 0.0), *graphOut);
			}

			continue;
		}
		
		// calculate all the data for one rotamer, store into matrices
		bool atomwise = Atom_Pairwise_IntraRes_Trie(single, REnergies1D_FA);
		bool resiwise = Rotamer_OneBody_Trie(single, REnergies1D_Rot);
		
		// convert the data to graph vertices
		for ( int i = 0; i < single.NumRotamers(); ++i )
		{
			FLOAT sum = 0.0;
			if (atomwise) sum += REnergies1D_FA(i);
			if (resiwise) sum += REnergies1D_Rot(i);
			boost::add_vertex(EnergyGraphNode(rotamers->GetRotamerID(single, i), sum*weight), 
					  *graphOut);
		}
	}	

	// now create all the edges (pairwise energies)
	foreach (pair, rotamers->PairResidues())
	{
		weight = rmap->CountPair(pair.first->GetBaseResidue(), pair.second->GetBaseResidue());
		if (weight < epsilon)
			continue;
	
		// calculate all the data for one rotamer pair, store in matrices
		bool atomwise = Atom_Pairwise_TrieTrie(*pair.first, *pair.second, REnergies2D_FA, false);
		bool resiwise = Rotamer_Pairwise_TrieTrie(*pair.first, *pair.second, REnergies2D_Rot);

		// convert the matrices to edges
		if (atomwise || resiwise)
		{
			// WARNING: (REALLY IMPORTANT)
			// For efficiency, this part of the code assumes that the graph stores
			// its nodes in the order they were placed with add_vertex, AND that rotamer
			// IDs are sequential for a residue!
			
			// yeah, this code is pretty confusing and completely breaks a bunch of
			// abstractions, but this part is so time-critical that it's pretty much
			// necessary to do so.

			const long firstID1 = rotamers->GetRotamerID(*pair.first, 0);
			const long firstID2 = rotamers->GetRotamerID(*pair.second, 0);
			const long lastID1 = firstID1 + pair.first->NumRotamers() - 1;
			const long lastID2 = firstID2 + pair.second->NumRotamers() - 1;

			bool entered_range1 = false;
			foreach(EnergyGraph::vertex_descriptor u, boost::vertices(*graphOut))
			{
				const long u_id = (*graphOut)[u].elementID;
				if (u_id < firstID1 || u_id > lastID1)
				{
					if (entered_range1)
						break;	// we have now exited the valid ID range, so we're done
					else
						continue; // we haven't yet entered the valid ID range
				}
				
				// U is now in the range of rotamers corresponding to residue 1
				entered_range1 = true;
			
				FLOAT* const RRData  = &REnergies2D_FA( (int)(u_id - firstID1), 0);
				FLOAT* const RRData2 = &REnergies2D_Rot( (int)(u_id - firstID1), 0);

				bool entered_range2 = false;
				foreach(EnergyGraph::vertex_descriptor v, boost::vertices(*graphOut))
				{
					const long v_id = (*graphOut)[v].elementID;
					if (v_id < firstID2 || v_id > lastID2)
					{
						if (entered_range2)
							break;
						else
							continue;
					}

					// V is now in the range of rotamers corresponding to residue 2
					entered_range2 = true;

					FLOAT sum = 0.0;
					if (atomwise) sum += RRData [v_id - firstID2];
					if (resiwise) sum += RRData2[v_id - firstID2];

					if (graphOut->IncludeEnergy(sum))
						boost::add_edge(u, v, sum*weight, *graphOut);
				}
			}
		}
	}

	graphOut->CapGraphAt(10000.0);

	if (rstats_.get() != NULL)
	{
		rstats_->Increment(RSTATS_ATOM_PAIRWISE, atomPairCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_PASSED_CUTOFF, atomPairPassedCutoffCount);
		rstats_->Increment(RSTATS_ATOM_PAIRWISE_HB, atomPairHBCount);
	}
}


