// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "atom_pairwise_helpers.h"
#include "atom_type_tables.h"

//////////////////////////////////////////////////////////////////////////
/////                           PARAMETERS                           /////
//////////////////////////////////////////////////////////////////////////

extern FLOAT parm_vdw14Weight;
extern FLOAT parm_coul_Cutoff;
extern FLOAT parm_vdw_Cutoff;

extern unsigned long long atomPairCount;
extern unsigned long long atomPairPassedCutoffCount;

extern unsigned long long coulCount;
extern unsigned long long vdwCount;
extern unsigned long long bondstretchCount;

//////////////////////////////////////////////////////////////////////////
/////                           INTERFACE                            /////
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
/////                          INTER-RESIDUE                         /////
//////////////////////////////////////////////////////////////////////////

template < void AddTerm(FLOAT, FLOAT*, FLOAT*, PairwiseEnergyTerm) >
inline void Atom_PairwiseTerm(FLOAT* total, FLOAT* output)
{
	// record a count of atom pairs for statistical purposes
	atomPairCount++;

	int atomType1 = Atom1_GetType(AtomType::OPLSaa_TYPE),
		atomType2 = Atom2_GetType(AtomType::OPLSaa_TYPE),
		interactionType1 = Atom1_GetType(AtomType::OPLSaa_INTERACTION_TYPE),
		interactionType2 = Atom2_GetType(AtomType::OPLSaa_INTERACTION_TYPE);
	
	FLOAT dist2 = Atoms_DistSquared();
	
	FLOAT coulE = 0.0, vdwE = 0.0, bstretchE = 0.0;
	
	//number of bond lengths separating the two atoms
	int numBondLengths = GetNumBonds();
	
	//nonbonded interaction weight:
	FLOAT nonBondWeight;
	
	if (numBondLengths < 3)			//Closer than 1,4 interactions not considered nonbonding
		nonBondWeight = 0.0;
	else if (numBondLengths == 3)	//weighted 1,4 interactions
		nonBondWeight = parm_vdw14Weight;
	else										//full weight for farther than 1,4 interactions
		nonBondWeight = 1.0;

	//Nonbonded Interactions


	if (nonBondWeight != 0.0)
	{
	        int bin;
        	FLOAT frac;
	        Atom_Dist2Interp(dist2, bin, frac, NB_DIST_BINS_PER_A2);
		// record a count of cutoff-passing atom pairs for statistical purposes
		atomPairPassedCutoffCount++;
			
		const float charge_ptr = directory_->GetOPLSaa_chargeTable()(atomType1, atomType2);
		coulE = nonBondWeight * charge_ptr / std::sqrt(dist2);
		if (coulE != 0)
			coulCount++;
	
		if (dist2 < parm_vdw_Cutoff)
		{	
			vdwE = nonBondWeight * Atom_DistType2_Term(bin, frac, interactionType1, interactionType2, directory_->GetOPLSaa_vdwTable());

			if (vdwE != 0)
				vdwCount++;
		}
	}
	
	//Bond Stretching
	if (numBondLengths == 1)
	{
		int bin;
		FLOAT frac;
		Atom_Dist2Interp(dist2, bin, frac, BS_DIST_BINS_PER_A2);
		//Obtain data table index for the pair of atoms
		int index = getIndex(bondstretch_tuples, blitz::TinyVector<int, 2>(interactionType1, interactionType2), bondstretch_size, true);
		//Do bond stretching energy lookup
		if (index != -1)
		{
			bondstretchCount++;
			bstretchE = Atom_DistType1_Term(bin, frac, index, directory_->GetOPLSaa_BondStretchTable());
		}
		//throw error if atom type tuple is not in the data table
		else
			throw std::string("CHOMP error: atom interaction types not in Bond Stretch data table\n");
	}
		
	AddTerm(coulE,  total, output, PTERM_COUL);
	AddTerm(vdwE,  total, output, PTERM_LJ);
	AddTerm(bstretchE, total, output, PTERM_BS);
}
	
//////////////////////////////////////////////////////////////////////////
/////                          INTRA-RESIDUE                         /////
//////////////////////////////////////////////////////////////////////////


//NOTE: Much of this code is copy and paste from Atom_PairwiseTerm(), because the calculations needed
//	for atom pairwise terms are much the same for inter-residue and intra-residue interactions.  The only
//	difference is the energy term that is added (one-body vs. pairwise).  I may combine the two functions,
//	or at least rewrite them so as not to have duplicated code.
template < void AddTerm(FLOAT, FLOAT*, FLOAT*, OneBodyEnergyTerm) >
inline void Atom_Pairwise_IntraRes(FLOAT* total, FLOAT* output)
{
	// record a count of atom pairs for statistical purposes
	atomPairCount++;

	int atomType1 = Atom1_GetType(AtomType::OPLSaa_TYPE),
		atomType2 = Atom2_GetType(AtomType::OPLSaa_TYPE),
		interactionType1 = Atom1_GetType(AtomType::OPLSaa_INTERACTION_TYPE),
		interactionType2 = Atom2_GetType(AtomType::OPLSaa_INTERACTION_TYPE);
	
	FLOAT dist2 = Atoms_DistSquared();
	
	FLOAT coulE = 0.0, vdwE = 0.0, bstretchE = 0.0;
	
	//number of bond lengths separating the two atoms
	int numBondLengths = GetNumBonds();
	
	//nonbonded interaction weight:
	FLOAT nonBondWeight;
	
	if (numBondLengths < 3)			//Closer than 1,4 interactions not considered nonbonding
		nonBondWeight = 0.0;
	else if (numBondLengths == 3)	//0.5 for 1,4 interactions
		nonBondWeight = parm_vdw14Weight;
	else										//full weight for farther than 1,4 interactions
		nonBondWeight = 1.0;
	
	//Nonbonded Interactions
	if (nonBondWeight != 0.0)
	{
	        int bin;
	        FLOAT frac;
	        Atom_Dist2Interp(dist2, bin, frac, NB_DIST_BINS_PER_A2);

		// record a count of cutoff-passing atom pairs for statistical purposes
		atomPairPassedCutoffCount++;
		
		if (dist2 < parm_coul_Cutoff)
		{
			const float charge_ptr = directory_->GetOPLSaa_chargeTable()(atomType1, atomType2);
			coulE = nonBondWeight * charge_ptr / std::sqrt(dist2);
			if (coulE != 0)
				coulCount++;
		}

		if (dist2 < parm_vdw_Cutoff)
		{
			vdwE = nonBondWeight * Atom_DistType2_Term(bin, frac, interactionType1, interactionType2, directory_->GetOPLSaa_vdwTable());
			if (vdwE != 0)
				vdwCount++;
		}
	}
	
	//Bond Stretching
	if (numBondLengths == 1)
	{
		int bin;
		FLOAT frac;
		Atom_Dist2Interp(dist2, bin, frac, BS_DIST_BINS_PER_A2);

		//Obtain data table index for the pair of atoms
		int index = getIndex(bondstretch_tuples, blitz::TinyVector<int, 2>(interactionType1, interactionType2), bondstretch_size, true);
		//Do bond stretching energy lookup
		if (index != -1)
		{
			bondstretchCount++;
			bstretchE = Atom_DistType1_Term(bin, frac, index, directory_->GetOPLSaa_BondStretchTable());
		}
		//throw error if atom type tuple is not in the data table
		else
			throw std::string("CHOMP error: atom interaction types not in bondstretch data table");
	}
		
	AddTerm(coulE,  total, output, OBTERM_COUL);
	AddTerm(vdwE,  total, output, OBTERM_LJ);
	AddTerm(bstretchE, total, output, OBTERM_BS);
}

