// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef ATOM_PAIRWISE_HELPERS_H
#define ATOM_PAIRWISE_HELPERS_H

const FLOAT NB_DIST_BINS_PER_A2 = 20.0;  ///< The number of bins in the data tables per angstrom squared of distance between atoms for nonbonded terms
const FLOAT BS_DIST_BINS_PER_A2  = 100.0; ///< The number of bins in the data tables per angstrom squared of distance between atoms for bondstretch terms

/** \brief Looks up an interpolated value from a 2D data table indexed by (index, Distance^2)
 *
 * \param bin1	The lower bin where the value lies
 * \param frac	The difference between the actual value and the bin value (fraction of a bin)
 * \param index	The row index of the table to look up
 * \param table The 2D data table to look up the value in
 *
 * \return FLOAT An interpolated value from the table
*/
inline FLOAT Atom_DistType1_Term(int bin1, FLOAT frac, int index, blitz::Array<FLOAT, 2> const& table)
{
	const FLOAT* bin1_ptr = &table(index, bin1);

	if (bin1 >= table.extent(1) - 1)
		return *bin1_ptr;

        const FLOAT* bin2_ptr = &table(index, bin1 + 1);

        return *bin1_ptr + frac * (*bin2_ptr - *bin1_ptr);
}

/** \brief Looks up an interpolated value from a 3D data table indexed by [Atom Type, Atom Type, Distance^2].
 *
 * \param bin1 The lower bin where the value lies
 * \param frac The difference between the actual value and the bin value (fraction of a bin)
 * \param type1 The type of the first atom (in whatever type system table is using)
 * \param type2 The type of the second atom (in whatever type system table is using)
 * \param table The 3D data table to look up the value in
 *
 * \return FLOAT An interpolated value from table
 */
inline FLOAT Atom_DistType2_Term(int bin1, FLOAT frac, int type1, int type2, blitz::Array<FLOAT, 3> const& table)
{
	const FLOAT* bin1_ptr = &table(type1, type2, bin1);
	
	if (bin1 >= table.extent(2) - 1)
		return *bin1_ptr;
	
	const FLOAT* bin2_ptr = &table(type1, type2, bin1 + 1);
	
	return *bin1_ptr + frac * (*bin2_ptr - *bin1_ptr);
}

/** \brief Discretizes the given distance^2 value into the proper bin
 *
 *  \param dist2 The distance squared between two atoms
 *  \param bin1 The variable to save the bin number
 *  \param frac The difference between the actual distance and the discretized distance (fraction of a bin)
 *  \param dist_bins_per_a2 The appropriate number of bins per angstrom squared of distance
 */
inline void Atom_Dist2Interp(FLOAT dist2, int& bin1, FLOAT& frac, const FLOAT dist_bins_per_a2)
{
	const FLOAT f_bin1 = (dist2 - 0.0) * dist_bins_per_a2;
	bin1    = static_cast<int>(f_bin1);
	frac   = f_bin1 - static_cast<float>(bin1);
}

#endif

