// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "general.h"
#include "atom_type_tables.h"

#define ANGLE_BINS_PER_REV 1440

extern Residue const* _res;
extern Residue const* _res1;
extern Residue const* _res2;

/** \brief Creates the inter-residue combined bond graph of the current two neighboring residues
 *
 * \param combined_graph	A pointer to the bond graph to create.
 * \param combined_IndexMap	A pointer to the index map that accompanies the bond graph
 */
inline void Create_InterRes_BondGraph(BondGraph* combined_graph, IndexMap* combined_IndexMap)
{
	//combined topology graph of the two residues, starts as just te first graph
	*combined_graph = BondGraph(_res1->GetAminoAcid().topoGraph);
	*combined_IndexMap = boost::get(boost::vertex_index, *combined_graph);
	
	int res1_nTotalAtoms = _res1->GetAminoAcid().nTotalAtoms,
		res2_nTotalAtoms = _res2->GetAminoAcid().nTotalAtoms;
	
	//add vertices of second graph to the combined graph
	for (int i = 0; i < res2_nTotalAtoms; i++)
		boost::add_vertex(*combined_graph);
	//add edges of second graph to the combined graph
	for (int i = res1_nTotalAtoms; i < res1_nTotalAtoms + res2_nTotalAtoms; i++)
		for (int j = i+1; j < res1_nTotalAtoms + res2_nTotalAtoms; j++)
			if (_res2->GetAminoAcid().bondLengths[i - res1_nTotalAtoms][j - res1_nTotalAtoms] == 1)
				boost::add_edge(i, j, *combined_graph);
	//add inter-residue C-N bond edge to the combined graph
	boost::add_edge(_res1->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "C"), 
			res1_nTotalAtoms + _res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "N"), *combined_graph);
}

/** \brief Looks up an interpolated value from a 2D data table.
 *
 * \param bin1 The lower bin
 * \param bin2 The higher bin
 * \param frac The fraction between bin1 and bin2 where the actual value is located
 * \param index The index of the table to look up
 * \param table The 2D data table to look up
 *
 * \return FLOAT An interpolated value from table
 */
inline FLOAT Rot_Angle_Term(int bin1, FLOAT frac, int index, blitz::Array<FLOAT, 2> const& table)
{
	const FLOAT* bin1_ptr = &table(index, bin1);
	
	if (bin1 >= table.extent(1) - 1)
		return *bin1_ptr;
	const FLOAT* bin2_ptr = &table(index, bin1 + 1);
	
	return *bin1_ptr + frac * ( *bin2_ptr - *bin1_ptr );
}

template <typename TinyVecType>
inline std::vector<TinyVecType> get_IntraRes_InterTypes (std::vector<TinyVecType> atoms)
{
	std::vector<TinyVecType> atom_types;	//all atom types
	for (unsigned int i = 0; i < atoms.size(); i++)
	{
		TinyVecType cur_atom_types;	//current atom types
		for (unsigned int j = 0; j < cur_atom_types.length(); j++)
			cur_atom_types[j] = atypes_->GetType(_res->GetAminoAcid().atomType[atoms[i][j]], AtomType::OPLSaa_INTERACTION_TYPE);
		atom_types.push_back(cur_atom_types);
	}
	
	return atom_types;
}			

template <typename TinyVecType>
inline std::vector<TinyVecType> get_InterRes_InterTypes (std::vector<TinyVecType> atoms)
{
	int res1_nTotalAtoms = _res1->GetAminoAcid().nTotalAtoms;
	
	std::vector<TinyVecType> atom_types;	//all atom types
	for (unsigned int i = 0; i < atoms.size(); i++)
	{
		TinyVecType cur_atom_types;
		for (unsigned int j = 0; j < cur_atom_types.length(); j++)
		{
			if (atoms[i][j] < res1_nTotalAtoms)
				cur_atom_types[j] = atypes_->GetType(_res1->GetAminoAcid().atomType[atoms[i][j]], AtomType::OPLSaa_INTERACTION_TYPE);
			else
				cur_atom_types[j] = atypes_->GetType(_res2->GetAminoAcid().atomType[atoms[i][j] - res1_nTotalAtoms], AtomType::OPLSaa_INTERACTION_TYPE);
		}
		atom_types.push_back(cur_atom_types);
	}
	
	return atom_types;
}
/*
//stores the improper torsion atom indices and the weight associated with each torsion, returns the atom interaction types
inline std::vector<blitz::TinyVector<int, 4> > get_IntraRes_Imptors_InterTypes(std::vector<blitz::TinyVector<int, 4> > atoms, 
						std::vector<blitz::TinyVector<int, 4> >* atoms_index, std::vector<FLOAT>* weights)
{
	std::vector<blitz::TinyVector<int, 4> > all_atom_types = get_IntraRes_InterTypes(atoms);
	std::vector<blitz::TinyVector<int, 4> > atom_types;
	
	int index, previndex = -2;
	FLOAT weight = 1.0;
	for (int i = 0; i < all_atom_types.size(); i++)
	{
		index = getIndex(imptors_tuples, all_atom_types[i], imptors_size, false); 
		std::cout<<index<<"\n";
		if (index != -1)
		{
			atoms_index->push_back(atoms[i]);
			atom_types.push_back(all_atom_types[i]);
			if (index == previndex)
				weight++;
			else
			{
				for (int i = 0; i < weight; i++)
					weights->push_back(weight);
				weight = 1.0;
			}
		}
		previndex = index;
	}

	return atom_types;
}		
*/
template <typename TinyVecType>
inline bool tupleDone (TinyVecType atom_type, std::vector<TinyVecType> atom_types)
{
	//check forward order
	for (unsigned int i = 0; i < atom_types.size(); i++)
	{
		bool isEqual = true;
		for (unsigned int j = 0; j < atom_type.length(); j++)	
			if (atom_types[i][j] != atom_type[j])
				isEqual = false;
		if (isEqual == true)
			return true;
	}
	
	//check reverse order
	TinyVecType reverse;
	for (unsigned int i = 0; i < atom_type.length(); i++)
		reverse[atom_type.length() - i - 1] = atom_type[i];
	for (unsigned int i = 0; i < atom_types.size(); i ++)
	{
		bool isEqual = true;
		for (unsigned int j = 0; j < reverse.length(); j++)
			if (atom_types[i][j] != reverse[j])
				isEqual = false;
		if (isEqual == true)
			return true;
	}	
	
	return false;
}

inline std::vector<blitz::TinyVector<int, 3> > get_InterRes_AngleAtoms(BondGraph combined_graph, IndexMap combined_IndexMap)
{
	int res1_nTotalAtoms = _res1->GetAminoAcid().nTotalAtoms;

	std::vector<blitz::TinyVector<int, 3> > angle_atoms;
	blitz::TinyVector<int, 3> cur_angle_atoms;
	for (VertexRange i = boost::vertices(combined_graph); boost::get(combined_IndexMap, *i.first) != res1_nTotalAtoms; ++i.first)
	{
		cur_angle_atoms[0] = boost::get(combined_IndexMap, *i.first);
		if (_res1->GetAminoAcid().bondLengths[cur_angle_atoms[0]][0] > 1)	//move on if first atom is too far away from C-N bond.
			continue;
			
		for (AdjacencyRange j = boost::adjacent_vertices(*i.first, combined_graph); j.first != j.second; ++j.first)
		{
			cur_angle_atoms[1] = boost::get(combined_IndexMap, *j.first);
			if (cur_angle_atoms[0] == 0 && cur_angle_atoms[1] != res1_nTotalAtoms + 2)	//move on if first atom is C and second is not N
				continue;
			if (cur_angle_atoms[1] < res1_nTotalAtoms && cur_angle_atoms[1] != 0)	//move on if second atom is too far away from C-N bond
				continue;
				
			for (AdjacencyRange k = boost::adjacent_vertices(*j.first, combined_graph); k.first!= k.second; ++k.first)
			{
				cur_angle_atoms[2] = boost::get(combined_IndexMap, *k.first);
				if (cur_angle_atoms[1] == 0 && cur_angle_atoms[2] != res1_nTotalAtoms + 2)	//move on if second atom is C and third is not N
					continue;
				if (cur_angle_atoms[0] != cur_angle_atoms[2])
					if(!tupleDone(cur_angle_atoms, angle_atoms))
						angle_atoms.push_back(cur_angle_atoms);
			}
		}
	}
	
	return angle_atoms;
}					
			
inline std::vector<blitz::TinyVector<int, 4> > get_InterRes_TorsAtoms(BondGraph combined_graph, IndexMap combined_IndexMap)
{		
	int res1_nTotalAtoms = _res1->GetAminoAcid().nTotalAtoms;

	std::vector<blitz::TinyVector<int, 4> > tors_atoms;
	blitz::TinyVector<int, 4> cur_tors_atoms;
	for (VertexRange i = boost::vertices(combined_graph); boost::get(combined_IndexMap, *i.first) != res1_nTotalAtoms; ++i.first)
	{	
		cur_tors_atoms[0] = boost::get(combined_IndexMap, *i.first);
		if (_res1->GetAminoAcid().bondLengths[cur_tors_atoms[0]][0] > 2)	//move on if first atom is too far away from C-N bond.
			continue;

		for (AdjacencyRange j = boost::adjacent_vertices(*i.first, combined_graph); j.first != j.second; ++j.first)
		{
			cur_tors_atoms[1] = boost::get(combined_IndexMap, *j.first);
			if (cur_tors_atoms[0] == 0 && cur_tors_atoms[1] != res1_nTotalAtoms + 2)	//move on if first atom is C and second is not N
				continue;			
			if (cur_tors_atoms[1] < res1_nTotalAtoms && _res1->GetAminoAcid().bondLengths[cur_tors_atoms[1]][0] > 1)	//move on if second atom is too far away from C-N bond
				continue;
			
			for (AdjacencyRange k = boost::adjacent_vertices(*j.first, combined_graph); k.first != k.second; ++k.first)
			{
				cur_tors_atoms[2] = boost::get(combined_IndexMap, *k.first);
				if (cur_tors_atoms[1]== 0 && cur_tors_atoms[2] != res1_nTotalAtoms + 2)	//move on if second atom is C and third is not N
					continue;
				if (cur_tors_atoms[2] < res1_nTotalAtoms && cur_tors_atoms[2] != 0)	//move on if third atom is too far away from C-N bond
					continue;
				
				if (cur_tors_atoms[0] != cur_tors_atoms[2])
				{
					for (AdjacencyRange l = boost::adjacent_vertices(*k.first, combined_graph); l.first != l.second; ++l.first)
					{							
						cur_tors_atoms[3] = boost::get(combined_IndexMap, *l.first);
						if (cur_tors_atoms[2] == 0 && cur_tors_atoms[3] != res1_nTotalAtoms + 2)	//move on if third atom is C and fourth is not N
							continue;
						
						if (cur_tors_atoms[1] != cur_tors_atoms[3])
							if (!tupleDone(cur_tors_atoms, tors_atoms))
								tors_atoms.push_back(cur_tors_atoms);
					}
				}
			}
		}
	}
	
	return tors_atoms;
}

//inter-residue improper torsions only include N-CA-C-O and CA-C-N-H or CA-C-N-CD for proline
inline std::vector<blitz::TinyVector<int, 4> > get_InterRes_ImptorsAtoms(BondGraph combined_graph, IndexMap combined_IndexMap)
{
	int res1_nTotalAtoms = _res1->GetAminoAcid().nTotalAtoms;
	int N = res1_nTotalAtoms + _res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "N"),
		CA1 = _res1->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "CA"),
		CA2 = res1_nTotalAtoms + _res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "CA"),
		C = _res1->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "C"),
		O = _res1->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "O");

	blitz::TinyVector<int, 4> imptors1 = blitz::TinyVector<int, 4> (N, CA1, C, O), imptors2;
	
	if (_res2->GetAminoAcid().name3Char == "PRO" || _res2->GetAminoAcid().name3Char == "NPRO" || _res2->GetAminoAcid().name3Char == "CPRO")
	{
		int CD = res1_nTotalAtoms + _res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "CD");
		imptors2 = blitz::TinyVector<int, 4> (CA2, C, N, CD);
	}
	else
	{
		int H = res1_nTotalAtoms + _res2->GetAminoAcid().indexOf(CHOMP_ATOM_ORDER, "H");
		imptors2 = blitz::TinyVector<int, 4> (CA2, C, N, H);
	}
	
	std::vector<blitz::TinyVector<int,4> > imptors_atoms;
	imptors_atoms.push_back(imptors1);
	imptors_atoms.push_back(imptors2);
	
	return imptors_atoms;
}
	
inline void Rot_AngleInterp(FLOAT angle, int& bin, FLOAT& frac)
{
	if (angle < 0.0)
		angle += 360.0;	

	const FLOAT f_bin = angle / 360.0 * ANGLE_BINS_PER_REV;
	bin = static_cast<int>(f_bin);
	frac = f_bin - static_cast<float>(bin);
}
	
	
	
	
	
	


