// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "rotamer_helpers.h"

//////////////////////////////////////////////////////////////////////////
/////                           PARAMETERS                           /////
//////////////////////////////////////////////////////////////////////////

extern FLOAT parm_torsionWeight;

extern long long TorsionCount;
extern long long ImptorsCount;
extern long long anglebendCount;

//////////////////////////////////////////////////////////////////////////
/////                        PUBLIC INTERFACE                        /////
//////////////////////////////////////////////////////////////////////////

using namespace CHOMPMath;

template < void AddTerm(FLOAT, FLOAT*, FLOAT*, OneBodyEnergyTerm) >
inline void OneBodyTerm_PerRotamer(FLOAT* total, FLOAT* output)
{
	//instantiated version of the rotamer
	Residue temp_res = _res->CloneWithChi(*_chi, _res->GetSystem());

	//torsion energies
	int tbin;
	FLOAT torsE = 0.0, tfrac;
	
	std::vector<blitz::TinyVector<int, 4> > TorsAtomsIndex = temp_res.GetAminoAcid().getTorsionAtoms();
	std::vector<blitz::TinyVector<int, 4> > TorsAtomsInter = get_IntraRes_InterTypes(TorsAtomsIndex);
	
	for (unsigned int i = 0; i < TorsAtomsInter.size(); i++)
	{	
		FLOAT angle = temp_res.GetTorsionAngle(TorsAtomsIndex[i]);
		Rot_AngleInterp(angle, tbin, tfrac);
		
		int index = getIndex(torsion_tuples, TorsAtomsInter[i], torsion_size, true);
		if (index != -1)
		{
			TorsionCount++;
			FLOAT cur_torsE = parm_torsionWeight * Rot_Angle_Term(tbin, tfrac, index, directory_->GetOPLSaa_TorsionTable());
			torsE += cur_torsE;
		}
		else
		{
			std::cout<<"In OneBodyTerm_PerRotamer:\n";
			std::cout<<"Amino Acid: "<<temp_res.GetAminoAcid().name3Char<<"\n";
			std::cout<<"CHOMP error: atom indices:"<<TorsAtomsIndex[i]<<" atom interaction types: "<<TorsAtomsInter[i]<<" not in Torsion data table\n";
			throw std::string("CHOMP error: atom interaction types not in Torsion data table\n");
		}
	}
	
	//improper torsion energies
	int ibin;
	FLOAT imptorsE = 0.0, ifrac;

	std::vector<blitz::TinyVector<int, 4> > AllImptorsAtomsIndex = temp_res.GetAminoAcid().getImptorsAtoms(atypes_);	//list of all possible improper torsion atom sets
	std::vector<blitz::TinyVector<int, 4> > AllImptorsAtomsInter = get_IntraRes_InterTypes(AllImptorsAtomsIndex);	//list of interaction types for all possible improper torsion atom sets
	std::vector<blitz::TinyVector<int, 4> > ImptorsAtomsIndex, ImptorsAtomsInter;	//lists of atom sets and interaction types of the improper torsions
	std::vector<int> ImptorsTableIndex;						//list of improper torsion table lookup indices for the torsions
	std::vector<FLOAT> ImptorsWeights;						//list of improper torsion weights for each improper torsion

	//obtain the actual atom sets, interaction types, and table indices of the improper torsions
	for (unsigned int i = 0; i < AllImptorsAtomsIndex.size(); i++)
	{
		int index = getIndex(imptors_tuples, AllImptorsAtomsInter[i], imptors_size, false);
		if (index != -1)
		{
			ImptorsAtomsIndex.push_back(AllImptorsAtomsIndex[i]);
			ImptorsAtomsInter.push_back(AllImptorsAtomsInter[i]);
			ImptorsTableIndex.push_back(index);
		}
	}

	//obtain the list of improper torsion weights for the different torsions
	//weights for all improper torsions about an atom is equal to the number of improper torsions about that atom
	//this is so that the improper torsion energy about that atom is equal to the average of all improper torsions about it
	for (unsigned int i = 0; i < ImptorsAtomsIndex.size(); i++)
	{
		FLOAT weight = 0.0;
		for (unsigned int j = 0; j < ImptorsAtomsIndex.size(); j++)
			if (ImptorsAtomsIndex[j][2] == ImptorsAtomsIndex[i][2])
				weight++;
		ImptorsWeights.push_back(weight);
	}

	for (unsigned int i = 0; i < ImptorsAtomsInter.size(); i++)
	{
		FLOAT angle = temp_res.GetImptorsAngle(ImptorsAtomsIndex[i]);
		Rot_AngleInterp(angle, ibin, ifrac);
		
		if (ImptorsTableIndex[i] != -1)
		{
			ImptorsCount++;
			FLOAT cur_imptorsE = Rot_Angle_Term(ibin, ifrac, ImptorsTableIndex[i], directory_->GetOPLSaa_ImptorsTable()) / ImptorsWeights[i];
			imptorsE += cur_imptorsE;
		}
		else
		{
			std::cout<<"In OneBodyTerm_PerRotamer:\n";
			std::cout<<"Amino Acid: "<<temp_res.GetAminoAcid().name3Char<<"\n";
			std::cout<<"CHOMP error: atom indices:"<<ImptorsAtomsIndex[i]<<" atom interaction types: "<<ImptorsAtomsInter[i]<<" not in Improper Torsion data table\n";
			throw std::string("CHOMP error: atom interaction types not in Improper Torsion data tables\n");		
		}
	}

	//angle bend energies
	int abin;
	FLOAT angleE = 0.0, afrac;
	
	std::vector<blitz::TinyVector<int, 3> > AngleAtomsIndex = temp_res.GetAminoAcid().getBondAngleAtoms();
	std::vector<blitz::TinyVector<int, 3> > AngleAtomsInter = get_IntraRes_InterTypes(AngleAtomsIndex);
	
	for (unsigned int i = 0; i < AngleAtomsInter.size(); i++)
	{
		int index = getIndex(anglebend_tuples, AngleAtomsInter[i], anglebend_size, true);
		if (index != -1)
		{
			FLOAT angle = temp_res.GetBondAngle(AngleAtomsIndex[i]);
			Rot_AngleInterp(angle, abin, afrac);
		
			anglebendCount++;
			FLOAT cur_angleE = Rot_Angle_Term(abin, afrac, index, directory_->GetOPLSaa_AngleBendTable()); 
			angleE += cur_angleE;
		}
		else
		{
			std::cout<<"In OneBodyTerm_PerRotamer:\n";
			std::cout<<"Amino Acid: "<<temp_res.GetAminoAcid().aminoAcid<<"\n";
			std::cout<<"CHOMP error: "<<"atom indices: "<<AngleAtomsIndex[i]<<" atom interaction types: "<<AngleAtomsInter[i]<<" not in Angle Bend data table\n";
			throw std::string("CHOMP error: atom interaction types not in Angle Bend data tables\n");
		}
	}	
	
	AddTerm(imptorsE, total, output, OBTERM_IMPTORS);
	AddTerm(torsE, total, output, OBTERM_TORS);
	AddTerm(angleE, total, output, OBTERM_AB);
}

//Currently, this function does nothing.  I could possibly put the backbone angle bend energies in here,
//but it will be a little bit uglier to separate backbone angle bends from other angle bends, which are calculated
//in OneBodyTerm_PerRotamer().
template < void AddTerm(FLOAT, FLOAT*, FLOAT*, OneBodyEnergyTerm) >
inline void OneBodyTerm_PerResidue(FLOAT* total, FLOAT* output)
{
	return;
}

