// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "rotamer_helpers.h"
#include "vectors.h"

//////////////////////////////////////////////////////////////////////////
/////                           PARAMETERS                           /////
//////////////////////////////////////////////////////////////////////////

extern FLOAT parm_torsionWeight;

extern long long TorsionCount;
extern long long ImptorsCount;
extern long long anglebendCount;

//////////////////////////////////////////////////////////////////////////
/////                        PUBLIC INTERFACE                        /////
//////////////////////////////////////////////////////////////////////////

using namespace CHOMPMath;

template < void AddTerm(FLOAT, FLOAT*, FLOAT*, PairwiseEnergyTerm) >
inline void Rotamer_PairwiseTerm(FLOAT* total, FLOAT* output)
{	
	//instantiated version of the rotamers
	Residue temp_res1 = _res1->CloneWithChi(*_chi1, _res1->GetSystem());
	Residue temp_res2 = _res2->CloneWithChi(*_chi2, _res2->GetSystem());

	//Calculate rotamer pairwise terms  if residues are consecutive
	if (temp_res2.GetID() - temp_res1.GetID() == 1)
	{
		BondGraph combined_graph;
		IndexMap combined_IndexMap;
		Create_InterRes_BondGraph(&combined_graph, &combined_IndexMap);
		
		//inter-residue torsion energies
		int tbin;
		FLOAT torsE = 0.0, tfrac;
	
		std::vector<blitz::TinyVector<int,4> > TorsAtomsIndex = get_InterRes_TorsAtoms(combined_graph, combined_IndexMap);
		std::vector<blitz::TinyVector<int,4> > TorsAtomsInter = get_InterRes_InterTypes(TorsAtomsIndex);
		
		for (unsigned int i = 0; i < TorsAtomsIndex.size(); i++)
		{

			Vector3 atom_coords[4];
			//obtain atom coordinates for current torsion
			for (int j = 0; j < 4; j++)
			{
				if (TorsAtomsIndex[i][j] < temp_res1.GetAminoAcid().nTotalAtoms)
					atom_coords[j] = temp_res1.GetAtomCoords()[TorsAtomsIndex[i][j]];
				else
					atom_coords[j] = temp_res2.GetAtomCoords()[TorsAtomsIndex[i][j] - temp_res1.GetAminoAcid().nTotalAtoms];
			}
		
			FLOAT angle = CalculateDihedralAngle(atom_coords[0], atom_coords[1], atom_coords[2], atom_coords[3]);
			Rot_AngleInterp(angle, tbin, tfrac);

			int index = getIndex(torsion_tuples, TorsAtomsInter[i], torsion_size, true);
			if (index != -1)
			{
				TorsionCount++;

				FLOAT cur_torsE = parm_torsionWeight * Rot_Angle_Term(tbin, tfrac, index, directory_->GetOPLSaa_TorsionTable()); 
				torsE += cur_torsE;
			}
			else
			{
				std::cout<<"In Rotamer_Pairwise_Term:\n";
				std::cout<<"Amino Acid: "<<temp_res1.GetAminoAcid().aminoAcid<<" and "<<temp_res2.GetAminoAcid().aminoAcid<<"\n";
				std::cout<<"CHOMP error: atom indices:"<<TorsAtomsIndex[i]<<" atom interaction types: "<<TorsAtomsInter[i]<<" not in Torsion data table\n";
				throw std::string("CHOMP error: inter-residue atom interaction types not in Torsion data table\n");
			}		
		}
	
		//Inter-residue improper torsion energies
		int ibin;
		FLOAT imptorsE = 0.0, ifrac;
		
		std::vector<blitz::TinyVector<int, 4> > ImptorsAtomsIndex = get_InterRes_ImptorsAtoms(combined_graph, combined_IndexMap);
		std::vector<blitz::TinyVector<int, 4> > ImptorsAtomsInter = get_InterRes_InterTypes(ImptorsAtomsIndex);
		
		for (unsigned int i = 0; i < ImptorsAtomsIndex.size(); i++)
		{
			Vector3 atom_coords[4];
			//obtain atom coordinates for current torsion
			for (int j = 0; j < 4; j++)
			{
				if (ImptorsAtomsIndex[i][j] < temp_res1.GetAminoAcid().nTotalAtoms)
					atom_coords[j] = temp_res1.GetAtomCoords()[ImptorsAtomsIndex[i][j]];
				else
					atom_coords[j] = temp_res2.GetAtomCoords()[ImptorsAtomsIndex[i][j] - temp_res1.GetAminoAcid().nTotalAtoms];
			}		
		
			FLOAT angle = CalculateOutOfPlaneAngle(atom_coords[0], atom_coords[1], atom_coords[2], atom_coords[3]);
			Rot_AngleInterp(angle, ibin, ifrac);

			int index = getIndex(imptors_tuples, ImptorsAtomsInter[i], imptors_size, false);
			if (index != -1)
			{
				ImptorsCount++;
				FLOAT cur_imptorsE = Rot_Angle_Term(ibin, ifrac, index, directory_->GetOPLSaa_ImptorsTable()); 
				imptorsE += cur_imptorsE;
			}
			else
			{
				std::cout<<"In Rotamer_Pairwise_Term:\n";
				std::cout<<"Amino Acid: "<<temp_res1.GetAminoAcid().aminoAcid<<" and "<<temp_res2.GetAminoAcid().aminoAcid<<"\n";
				std::cout<<"CHOMP error: atom indices:"<<ImptorsAtomsIndex[i]<<" atom interaction types: "<<ImptorsAtomsInter[i]<<" not in Improper Torsion data table\n";
				throw std::string("CHOMP error: inter-residue atom interaction types not in Improper Torsion data table\n");
			}	
		}
			
		//Inter-residue angle bend energies
		int abin;
		FLOAT angleE = 0.0, afrac;
		
		std::vector<blitz::TinyVector<int,3> > AngleAtomsIndex = get_InterRes_AngleAtoms(combined_graph, combined_IndexMap);
		std::vector<blitz::TinyVector<int,3> > AngleAtomsInter = get_InterRes_InterTypes(AngleAtomsIndex);		
		
		for (unsigned int i = 0; i < AngleAtomsIndex.size(); i++)
		{
			Vector3 atom_coords[3];
			//obtain atom coordinates for current bond angle
			for (int j = 0; j < 3; j++)
			{
				if (AngleAtomsIndex[i][j] < temp_res1.GetAminoAcid().nTotalAtoms)
					atom_coords[j] = temp_res1.GetAtomCoords()[AngleAtomsIndex[i][j]];
				else
					atom_coords[j] = temp_res2.GetAtomCoords()[AngleAtomsIndex[i][j] - temp_res1.GetAminoAcid().nTotalAtoms];
			}
			
			FLOAT angle = CalculateVectorAngle(atom_coords[0] - atom_coords[1], atom_coords[2] - atom_coords[1]);
			Rot_AngleInterp(angle, abin, afrac);

			int index = getIndex(anglebend_tuples, AngleAtomsInter[i], anglebend_size, true);
			if (index != -1)
			{
				anglebendCount++;
				FLOAT cur_angleE = Rot_Angle_Term(abin, afrac, index, directory_->GetOPLSaa_AngleBendTable()); 
				angleE += cur_angleE;
			}
			else
			{
				std::cout<<"In Rotamer_Pairwise_Term:\n";
				std::cout<<"Amino Acids: "<<temp_res1.GetAminoAcid().aminoAcid<<" and "<<temp_res2.GetAminoAcid().aminoAcid<<"\n";
				std::cout<<"CHOMP error: atom indices:"<<AngleAtomsIndex[i]<<" atom interaction types: "<<AngleAtomsInter[i]<<" not in Angle Bend data tables\n";
				throw std::string("CHOMP error: inter-residue atom interaction types not in Angle Bend data tables\n");
			}
		}
			
		AddTerm(imptorsE, total, output, PTERM_IMPTORS);
		AddTerm(torsE, total, output, PTERM_TORS);
		AddTerm(angleE, total, output, PTERM_AB);
	}
}

