// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Albert Ng 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#include "stdafx.h"
#include "general.h"
#include "residue.h"
#include "data_table.h"
#include "rotamer_trie.h"

#include <blitz/tinyvec-et.h>

#include "neighbor_map.h"
#include "rotamer_pairwise.h"
#include "applicators.h"
#include "energy_terms.h"
#include "dunbrack.h"

//////////////////////////////////////////////////////////////////////////
/////                         LOCAL VARIABLES                        /////
//////////////////////////////////////////////////////////////////////////

// (these are global to make it easier to use functions in rotamer_pairwise_util.h without
// passing in a ton of parameters, cluttering up the user-defined energy function code)

extern Residue const* _res1;
extern Residue const* _res2;

Vector4 const* _chi1;
Vector3 const* _centroid1;

Vector4 const* _chi2;
Vector3 const* _centroid2;

//////////////////////////////////////////////////////////////////////////
/////                          EXTERNAL DATA                         /////
//////////////////////////////////////////////////////////////////////////

extern DunbrackLibrary::ConstPointer dunLib_;
extern DataDirectory::ConstPointer directory_;
extern AtomTypes::ConstPointer atypes_;
extern NeighborMap::ConstPointer nmap_;

//////////////////////////////////////////////////////////////////////////
/////                            INCLUDES                            /////
//////////////////////////////////////////////////////////////////////////

#include "rotamer_pairwise_util.h"
#include "rotamer_pairwise_ef.h"

//////////////////////////////////////////////////////////////////////////
/////                        PUBLIC INTERFACE                        /////
//////////////////////////////////////////////////////////////////////////

bool Rotamer_Pairwise_TrieTrie(RotamerSet const& trie1, 
		    	       RotamerSet const& trie2,
			       blitz::Array<FLOAT, 2>& output)
{
	_res1 = &trie1.GetBaseResidue();
	_res2 = &trie2.GetBaseResidue();

	// bail out if these residues aren't neighbors
	if (!nmap_->AreNeighbors(*_res1, *_res2))
		return false;

	for (int rot1 = 0; rot1 < trie1.NumRotamers(); ++rot1)
	{
		_chi1 = &trie1.GetChi(rot1);
		_centroid1 = &trie1.GetCentroid(rot1);

		for (int rot2 = 0; rot2 < trie2.NumRotamers(); ++rot2)
		{
			_chi2 = &trie2.GetChi(rot2);
			_centroid2 = &trie2.GetCentroid(rot2);
			
			FLOAT result = 0.0;
			Rotamer_PairwiseTerm<CombinedEnergyTerms>(&result, 0);

			output(rot1, rot2) = result;
		}
	}

	return true;
}

FLOAT Rotamer_Pairwise_ResidueResidue(Residue const& residue1, 
 			       	      Residue const& residue2,
				      blitz::Array<FLOAT, 1>& energiesOut)
{
	FLOAT total = 0.0;
	FLOAT* output = &(energiesOut(0));
	
	_res1 = &residue1;
	_res2 = &residue2;


	_chi1 = &residue1.GetChi();
	_centroid1 = &residue1.GetCentroid();
	_chi2 = &residue2.GetChi();
	_centroid2 = &residue2.GetCentroid();

	// bail out if these residues aren't neighbors
	if (!nmap_->AreNeighbors(*_res1, *_res2))
		return 0.0;

	Rotamer_PairwiseTerm<SeparateEnergyTerms>(&total, output);

	return total;
}


