// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_AA_PROPERTIES
#define IVL_AA_PROPERTIES

#include "general.h"
#include "atom_type.h"

#include <map>
#include <vector>
#include <string>

#include <boost/serialization/serialization.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/shared_ptr.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/string.hpp>

/** \file amino_acid.h 
 *  \brief Contains static data about the physical properties of the amino acids.
 *
 *  This file contains information about amino acids, to be used by energy functions or
 *  any other part of the code that requires chemical data. Any data about amino acids 
 *  that doesn't depend on rotamer/physical location should go here. The data is hardcoded 
 *  into C++ source which is generated automatically from a Python script, so don't edit it
 *  here.
 *
 *
 **/

///////////////////////////////////////////////////////////////////////////
//////                 MAXIMUM NUMBERS OF THINGS                    ///////
///////////////////////////////////////////////////////////////////////////

/// \brief The maximum number of H-bond acceptor atoms allowed in one residue
#define MAX_ACCEPTORS 		5
/// \brief The number of atoms required to extrapolate the position of an atom with unknown coords
#define N_TEMPLATE_ATOMS 	3
/// \brief The number of atoms required to determine a dihedral Chi angle
#define N_CHI_ATOMS		4
/// \brief The maximum number of atoms used to find a crude approximation of a residue's centroid
#define MAX_CENTROID_ATOMS	5

///////////////////////////////////////////////////////////////////////////
//////                         ATOM FLAGS                           ///////
///////////////////////////////////////////////////////////////////////////

// these flags are stored on a per-atom basis. they contain redundant information that can be computed
// from variables in the AminoAcid class, but are used often enough that pre-computing them provides
// a substantial performance benefit.   

/// \brief Set if the atom in question is a hydrogen
#define   ATM_FLAG_H		1		
/// \brief Set if the atom is a polar hydrogen (for hydrogen bonds)
#define   ATM_FLAG_POL_H	2
/// \brief Set if the atom is a hydrogen bond acceptor
#define   ATM_FLAG_HB_ACCEPTOR  4
/// \brief Set if the atom is part of the backbone (C, O, N, HN, CA, HA)
#define   ATM_FLAG_BB		8

// include a Python-generatd enumeration of amino acids
#include "../core/codegen/output/aa_props_enum.hxx"

/** \brief Contains static amino acid data.
 *
 * This struct serves as a repository of data about an amino acid. The actual data is
 * in an .hxx file generated by Python code, which is then loaded into an instance
 * of this class by an AminoAcidDirectory.
 *
 * For struct members that are arrays of atoms, note that the atom ordering is defined
 * by the CHOMP_ATOM_ORDER ordering system (see member indexOf). 
 * 
 */
struct DllExport AminoAcid
{
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/// \brief  The default constructor creates an AminoAcid with garbage data!
	AminoAcid();
	
	///////////////////////////////////////////////////////////////////////////
	//////                      CONVENIENT TYPEDEFS                      //////
	///////////////////////////////////////////////////////////////////////////
	
	typedef		std::vector<int> 		IntVector;
	typedef		std::vector<bool> 		BoolVector; 
	typedef		std::vector<std::string>	StrVector;
	
	typedef 	std::vector<IntVector> 		IntVector2;
	typedef 	std::vector<BoolVector> 	BoolVector2;
	typedef		std::vector<StrVector> 		StrVector2;
	
	///////////////////////////////////////////////////////////////////////////
	//////                             DATA                             ///////
	///////////////////////////////////////////////////////////////////////////
	
	AminoAcidIndex aminoAcid;  	///< Which amino acid is this? (in the enumeration)

	int AA_class;			///< Amino acid class number (used to group N- and C- terminal variants with originals)

	bool polar; 			///< True if this is a polar amino acid

	int nHeavyAtoms;  		///< The number of heavy (non-hydrogen) atoms in this amino acid
	int nTotalAtoms;  		///< The full number of atoms (including hydrogens) in this amino acid
	int nHBAcceptors; 		///< The number of atoms in this AA that can accept hydrogen bonds

	IntVector acceptorList; 	///< The list of atoms that can accept hydrogen bonds (arbitrary order)

	/// \brief For each H-bond acceptor atom, this stores the corresponding 'base' atom(s) for the H-bond
	/// The first index corresponds to the index into acceptorList, and the second index must be 0 for most
	/// amino acids, or else 0 or 1 for those that have acceptors with two bases. 
	IntVector2 acceptorBase; 

	int nChi;			///< The number of chi angles for this amino acid

	/// \brief The list of atoms that define each chi angle.
	/// The first index is the number of the chi angle you want, and the second is the number of the
	/// atom (0 through 4). The result will be an atom index in the CHOM_ATOM_ORDER.
	IntVector2 chiAtoms;		

	/// \brief The atom used to determine whether two residues are neighbors.
	/// To find whether two residues are considered neighbors, we find the distance between their
	/// neighborAtoms and compare it to an amino acid specific cutoff.
	int neighborAtom;		
					
	/// \brief A table to determine which atom positions depend on which chi angles.
	/// The first index is an atom index in the CHOMP_ATOM_ORDER, and the second is the number of a
	/// chi angle. True if the position of the atom depends on this chi angle.	
	BoolVector2 chiDependent;

	/// \brief A table used to regenerate unknown atom positions from 'template' atoms.
	/// The first index is the index of the unknown atom, and the second is 0..3 for the three template
	/// atoms for that atom. Note that not all atoms come with template data (see the Python script that
	/// generates the data for details).
	IntVector2 templateAtom;

	/// \brief The coordinates of the residue in an arbitrary 'template' configuration.
	/// This data is used to fill in missing atoms in a PDB.
	std::vector<Vector3> templateCoords;

	/// \brief The number of atoms used to find a residue's centroid (see centroidAtoms)
	int nCentroidAtoms;

	/// \brief A list of atoms from which to calculate the approximate centroid of a residue.
	/// See the Python sources for details.	
	IntVector centroidAtoms;

	/// \brief A table of atom orderings.
	/// The first index is the atom ordering index, and the second is the atom index in that ordering.
	/// The table entries are standard PDB-style atom name string.
	StrVector2 atomOrders;

	/// \brief The raw type of each atom in this residue.
	/// For information on atom types and how to use them, see atom_types.h
	IntVector atomType;

	/// \brief The ATM_* flags are filled into this table.
	IntVector atomFlags;

	std::string name3Char;		///< The standard 3-character name of this amino acid.

	/// \brief A table of numbers of bond lengths separating atoms in the residue.
	/// The two indices are the atom indices in the CHOMP_ATOM_ORDER
	IntVector2 bondLengths;
	
	/// \brief A topology graph of the amino acid indicating how the atoms bond
	BondGraph topoGraph;
	
	/// \brief A property map of atom indices for the topology graph
	IndexMap tg_IndexMap;

	///////////////////////////////////////////////////////////////////////////
	//////                     CONVENIENCE METHODS                      ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief  Look up the index of an atom in a given ordering, by PDB-style name.
	 * 
   	 * \param atomOrdering One of CHOMP_ATOM_ORDER, PDB_ATOM_ORDER
	 * \param atom The PDB-style name of your atom
	 *
	 * \return The index of your atom. 
	 *
	 * For example, to get the flags on a beta carbon:
	 * \code 
	 * 	atomFlags[ indexOf(CHOMP_ATOM_ORDER, "CB") ]
	 * \endcode
	 */	 
	int indexOf(int atomOrdering, std::string const& atom) const;

	/** \brief Get a list of the atom index triplets that define all of the bond angles in the amino acid
	 */
	std::vector< blitz::TinyVector<int, 3> > getBondAngleAtoms(void) const;
	
	/** \brief Get a list of the atom index quadruplets that define all of the torsions in the amino acid
	 */
	std::vector< blitz::TinyVector<int, 4> > getTorsionAtoms(void) const;

	/** \brief Returns a list of improper torsion atom sets in the amino acid.
	 *		The list contains all possible orders of all atom sets.
	 */
	std::vector<blitz::TinyVector<int, 4> > getImptorsAtoms(AtomTypes::ConstPointer atomTypes) const;

	friend class AminoAcidDirectory;

	// Py++ instantiations
	std::vector<bool>  		_bool_vector_inst;
	std::vector<std::string>  	_str_vector_inst;

private:

	// a reversed lookup table to find atom index by PDB-style name, 
	// automatically generated from atomOrders and used to implement indexOf.
	std::map<std::string, int> indexOf_[N_ATOM_ORDERINGS];

	/** \brief Determines if a atom type tuple has already been accounted for.  Checks both forward and reverse orders
	 *
	 * \param atom_type The atom type tuple to check if currently in vector of atom types already accounted for
	 * \param atom_types Vector of atom type tuples already accounted for
	 *
	 * \return bool True if atom type tuple is accounted for, False otherwise
	 */
	template <typename TinyVecType>
	bool tupleDone (TinyVecType atom_type, std::vector<TinyVecType> atom_types) const
	{
		//check forward order
		for (unsigned int i = 0; i < atom_types.size(); i++)
		{
			bool isEqual = true;
			for (unsigned int j = 0; j < atom_type.length(); j++)	
				if (atom_types[i][j] != atom_type[j])
					isEqual = false;
			if (isEqual == true)
				return true;
		}
		
		//check reverse order
		TinyVecType reverse;
		for (unsigned int i = 0; i < atom_type.length(); i++)
			reverse[atom_type.length() - i - 1] = atom_type[i];
		for (unsigned int i = 0; i < atom_types.size(); i ++)
		{
			bool isEqual = true;
			for (unsigned int j = 0; j < reverse.length(); j++)
				if (atom_types[i][j] != reverse[j])
					isEqual = false;
			if (isEqual == true)
				return true;
		}	
		
		return false;
	}

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(aminoAcid);
		ar & BOOST_SERIALIZATION_NVP(polar);
		ar & BOOST_SERIALIZATION_NVP(nHeavyAtoms);
		ar & BOOST_SERIALIZATION_NVP(nTotalAtoms);
		ar & BOOST_SERIALIZATION_NVP(nHBAcceptors);
		ar & BOOST_SERIALIZATION_NVP(acceptorList);
		ar & BOOST_SERIALIZATION_NVP(acceptorBase);
		ar & BOOST_SERIALIZATION_NVP(nChi);
		ar & BOOST_SERIALIZATION_NVP(chiAtoms);
		ar & BOOST_SERIALIZATION_NVP(neighborAtom);
		ar & BOOST_SERIALIZATION_NVP(chiDependent);
		ar & BOOST_SERIALIZATION_NVP(templateAtom);
		ar & BOOST_SERIALIZATION_NVP(templateCoords);
		ar & BOOST_SERIALIZATION_NVP(nCentroidAtoms);
		ar & BOOST_SERIALIZATION_NVP(centroidAtoms);
		ar & BOOST_SERIALIZATION_NVP(atomOrders);
		ar & BOOST_SERIALIZATION_NVP(atomType);
		ar & BOOST_SERIALIZATION_NVP(atomFlags);
		ar & BOOST_SERIALIZATION_NVP(name3Char);
		ar & BOOST_SERIALIZATION_NVP(indexOf_);
		ar & BOOST_SERIALIZATION_NVP(bondLengths);
//		ar & BOOST_SERIALIZATION_NVP(topoGraph);
//		ar & BOOST_SERIALIZATION_NVP(tg_IndexMap);
	}
};

/** \brief A factory for AminoAcid objects.
 *
 * To retrieve amino acid data, you must first create an AminoAcidDirectory. Then you can
 * fetch amino acids of your choice like this:
 *
 * \code
 *	AminoAcidDirectory::ConstPointer aaDir = AminoAcidDirectory::Create(atomTypes);
 *	AminoAcid const& ala = aaDir->Get("ALA");
 * \endcode
 *
 * Or in Python:
 *
 * \code
 *      aaDir = AminoAcidDirectory(atomTypes)
 *	ala = aaDir['ALA']
 * \endcode
 * 
 */
class DllExport AminoAcidDirectory
{
public:
	typedef boost::shared_ptr<AminoAcidDirectory>		Pointer;
	typedef boost::shared_ptr<const AminoAcidDirectory> 	ConstPointer;

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief This quasi-constructor returns a shared pointer to a fully initialized AminoAcidDirectory.
	 *
 	 * \param atypes The AtomTypes type system to be used by the new AminoAcidDirectory.
	 */
	static Pointer Create(AtomTypes::ConstPointer atypes);

	AminoAcidDirectory(AtomTypes::ConstPointer atomTypes);

	///////////////////////////////////////////////////////////////////////////
	//////                           ACCESSORS                           //////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Look up an AminoAcid object by the AminoAcidIndex enumeration.
	 *
	 * \param The enumerated index of the amino acid you want.
	 * \return The amino acid data.
	 */
	AminoAcid const* Get(AminoAcidIndex index) const;

	/** \brief Look up an AminoAcid object by PDB name.
	 *
	 * \param The 3-character nane of the amino acid you want.
	 * \return The amino acid data.
	 */
	AminoAcid const* Get(std::string const& aaName) const;

private:

	// fills in the atom flags for a given atom in a given amino acid
	int GetAtomFlags(AminoAcid const& aa, int atom, std::string const& atomName, AtomTypes::ConstPointer atomTypes);

	AminoAcid data_[N_AMINO_ACIDS];			///< The actual amino acid data

	std::map<std::string, AminoAcid*> byName_;	///< A pregenerated map for looking up amino acids by PDB name

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(data_);
		ar & BOOST_SERIALIZATION_NVP(byName_);
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void load_construct_data(
	Archive & ar, AminoAcidDirectory* t, const unsigned int file_version
){
	// retrieve data from archive required to construct new instance
	AtomTypes::ConstPointer atypes = AtomTypes::Create();
	// invoke inplace constructor to initialize instance
	::new(t)AminoAcidDirectory(atypes);
}
}} // namespace ...

#endif


