// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef	IVL_ATOM_TYPE_H
#define	IVL_ATOM_TYPE_H

#include "general.h"

#include <map>
#include <vector>
#include <string>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>
#include <boost/serialization/string.hpp>
#include <boost/serialization/binary_object.hpp>

/** \file atom_type.h
*
* \brief Contains static data for classifying atoms into various type systems.
*
* The basic idea here is that different categorizations of atoms are useful
* in different circumstances. For instance, Rosetta uses a 26-type system
* (also implemented in CHOMP), which does not include a specific type for
* beta carbons. Rosetta simultaneously uses a different ad-hoc type system
* to categorize atoms when deciding whether or not to count pairwise interactions
* between neighboring residues (here implemented as the CountPair type system).
* 
* Using this module, the creation and maintenance of multiple conflicting
* atom categorizations is made simple and clean. We start with a fundamental
* 'raw' type system, which at this point has a different type for every atom
* in every amino acid. We then have a many-to-one mapping that translates
* this raw type into every other atom type we may use. This module takes
* care of generating this mapping as well as support code needed for it to work.
* 
* To add a new type system, simply add a new Python script to the atom_types
* directory, following the examples in rosetta.py and countpair.py. Then,
* run gen_types.py and it will automatically detect your addition and include
* it in the generated C++ code.
* 
* This file implements a class for initializing and storing the data generated
* by the atom_types Python scripts. The actual code to construct this data is
* in ../core/codegen.
*
*/

// include generated atom_types declarations
#include "../core/codegen/output/atom_types.hxx"

/** \brief Used to map between raw types used in AminoAcid objects, and the other type
 * systems used in the energy functions and elsewhere.
 *
 * To look up an atom's type within a given type system, you can do something like this:
 *
 * \code
 * 	AtomTypes::ConstPointer atypes = AtomTypes::Create();
 *	int type = atypes->GetType( AtomType::Raw::GLU_OE1, AtomType::ROSETTA_TYPE );
 *	 // 'type' now contains the type of GLU's OE1 in the Roseta type sytsem. You can also do:
 *	int raw_type = atypes->ToType("GLU:OE1", AtomType::RAW_TYPE);
 *	int type = atypes->GetType( raw_type, AtomType::ROSETTA_TYPE);
 *	 // 'type' contains the same thing as before
 * \endcode
 *
 * Or in Python:
 *
 * \code
 *	atypes = AtomTypes()
 *	type = atypes.GetType( Raw.GLU_OE1, TypeSystem.ROSETTA_TYPE )
 *	 # or
 *	raw_type = atypes['GLU:OE1', TypeSystem.RAW_TYPE]
 *	type = atypes.GetType( raw_type, TypeSystem.ROSETTA_TYPE )
 * \endcode
 */
class DllExport AtomTypes
{
public:

	typedef boost::shared_ptr<AtomTypes>		Pointer;
	typedef boost::shared_ptr<const AtomTypes> 	ConstPointer;

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/// \brief This quasi-constructor returns a shared pointer to a fully initialized AtomTypes object.
	static Pointer Create();

	AtomTypes();

	///////////////////////////////////////////////////////////////////////////
	//////                           ACCESSORS                           //////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Translate from a raw atom type to another type system.
	 * 
 	 * \param rawType An entry in the AtomType::Raw enumeration.
	 * \param typeSystem The type system to translate into.
	 *
	 * \return The translated atom type.
 	*/		
	int GetType (int rawType, AtomType::TypeSystem typeSystem) const
	{
		return data_[rawType][typeSystem];
	}

	/** \brief Look up the string atom name by a type index.
	 *
	 * \param type An entry in the enumeration corresponding to the type system passed as the 2nd arg
	 * \param typeSystem The type system in which you're looking up a type name (can be RAW_TYPE).
	 *
	 * \return A string representation of the given atom type.
	 *
	 * Example:
	 * \code
	 *	assert( atypes.ToString(AtomType::Raw::GLU_OE1, AtomType::RAW_TYPE) == "GLU:OE1" );
 	 * \endcode
	*/
	std::string ToString(int type, AtomType::TypeSystem typeSystem) const;

	/** \brief Look up an atom type by its string name.
	 *
	 * \param name A string corresponding to an atom type in the given type system
	 * \param typeSystem The type system in which you're looking up an atom type
	 *
	 * \return The atom type described by the given string.
	 * 
	 * Example:
	 * \code
	 *	assert( atypes.ToType("GLU:OE1", AtomType::RAW_TYPE) == AtomType::Raw::GLU_OE1 );
 	 * \endcode
	*/
	int ToType(std::string const& name, AtomType::TypeSystem typeSystem) const; 

private:
	
	int data_[N_RAW_ATOM_TYPES][N_ATOM_TYPE_SYSTEMS];	///< The actual atom type data

	std::vector<std::string> 	toString_[N_ATOM_TYPE_SYSTEMS]; ///< Look up type name by index
	std::map<std::string, int> 	toType_  [N_ATOM_TYPE_SYSTEMS]; ///< Look up type index by name

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		boost::serialization::binary_object binObj(
			&data_[0], 
			sizeof(int) * N_RAW_ATOM_TYPES * N_ATOM_TYPE_SYSTEMS);
		ar & boost::serialization::make_nvp("data_", binObj); 

		ar & BOOST_SERIALIZATION_NVP( toString_ );
		ar & BOOST_SERIALIZATION_NVP( toType_ );
	}
};

#endif


