// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_DATA_TABLE
#define IVL_DATA_TABLE

#include "general.h"

#include <map>
#include <string>

#include <blitz/array.h>

/** \file data_table.h
 * \brief Contains the interface for loading data tables defined in data_tables.cgs from their files. 
 *
 * \see DataDirectory
 */

/** \file tables/data.dir
 * \brief This file associates each data table defined in data_tables.cgs with a file containing its data.
 *
 * Link new tables to their corresponding data files using the syntax
 * \code <table name> <file name> \endcode
 *
 * with file name relative to the topmost source directory, and the table name as it is used in the C++ code.
 * This file is not automatically generated by the data table python code because it seems useful to be able
 * to change the location of the data files at runtime.
 */

/** \brief Holds a set of tables loaded from external input files.
 *
 * In the most basic usage of this class, you can just:
 * \code 
 * 	DataDirectory::ConstPointer ddir = DataDirectory::Create("tables/data.dir");
 * \endcode
 * 
 * Python:
 * \code
 *  	ddir = DataDirectory('tables/data.dir')
 * \endcode 
 *
 * And from then on, all tables in data.dir and data_tables.cgs will be accessible as 
 * public members in the class, with all their data filled in. When accessed from Python, the 
 * arrays will be automatically converted to Numpy arrays, but will share memory with the C++
 * arrays. So, you can just write new data into them and the DataDirectory will be updated 
 * accordingly.
 *
 * You can also change the file that indexes the locations of the tables with
 * SetDirectoryFile, and you can load individual tables at runtime using
 * ReloadDataTable.
 */
class DllExport DataDirectory
{
public:
	typedef boost::shared_ptr<DataDirectory>		Pointer;
	typedef boost::shared_ptr<const DataDirectory> 		ConstPointer;

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	static Pointer Create(std::string const& directoryFile);

	/** \brief Initializes the data directory with either a TOC file, or a binary
	 * image created with DataDirectory::Compile.
         *
	 * Depending on the type of file you pass in as the directoryFile, one of two
	 * things will happen:
	 *
	 * 1. If the directoryFile is a plaintext TOC file, it will be parsed and all
	 *    the data tables inside will be loaded from external files.
	 * 2. If the directoryFile is a binary image, its contents will be loaded directly
	 *    into memory, and the DataDirectory will be put in "compiled mode". In
	 *    compiled mode, SetDirectoryFile, ReloadDataTable, and ReloadAllDataTables
	 *    are unusable and will throw exceptions.
	 *
	 * \param directoryFile The filename of the directory file or binary image.
	 */
	DataDirectory(std::string const& directoryFile);

	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC INTERFACE                       ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Call this to specify a file containing a list of files to load data from.
	 *
	 * \param fileName The filename of the directory file. Probably data.dir.
	 */
	void SetDirectoryFile(std::string const& fileName);

	/** \brief Loads an individual data table.
	 *
	 * If you have previously called Initialize() or SetDirectoryFile() and the table
	 * you're loading is present in your directory file, you can omit the fileName argument.
	 *
	 * \param tableName This must be a string that exactly matches the C++ symbolic name of the
	 * 		    table you're loading. So if you're loading CountPairTable, give 
	 * 		    "CountPairTable".
	 * \param fileName If the table you're loading isn't present in the directory, or you want
	 * 		   to load the data from an alternate location, specify it here.
	 */
	void ReloadDataTable(std::string const& tableName,
		             std::string const& fileName = "");

	/** \brief Loads all the tables in the current directory file.
	 *
	 * This will not do anything if you haven't previously called Initialize() or SetDirectoryFile().
	 * If you have, it will load all the tables referenced in the file you specified then.
	 */
	void ReloadAllDataTables();

	/** \brief Writes out the current contents of the entire data directory in binary format.
	 *
	 * You can then pass the compiled file into the DataDirectory constructor instead of a 
	 * data table TOC file. This make it much faster to load the data, and a little harder for
	 * someone to edit your data files.
	 *
	 * \param outputFileName The file to store the compiled data into.
	 */
	void Compile(std::string const& outputFileName) const;

	// Individual data table declarations are automatically generated
	#include "../core/codegen/output/data_tables.hxx"

private:
	std::map<std::string, std::string> dataDirectory_;
	bool compiled_;
};

#endif

