// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_DUNBRACK_H
#define IVL_DUNBRACK_H

#include "general.h"

#include <string>
#include <blitz/array.h>

#include "amino_acid.h"

/** \file dunbrack.h
 *  \brief Contains classes for loading and retreiving information from a Dunbrack rotamer library.
 */

/** \brief Loads and provides a convenient interface for a Dunbrack backbone-dependent rotamer library.
 *
 * The Dunbrack library contains a list of the most common chi angle sets for each combination of standard
 * amino acid, phi angle, and psi angle. All angles are discretized into 10-degree bins. In CHOMP, the Dunbrack
 * data is stored in three files, called dunbrack.bbdep, dunbrack.bbdep.maxrot, and dunbrack.bbdep.revlookup.
 * 
 * dunbrack.bbdep contains the same data as the file you can get from the Dunbrack site, with the rotamers sorted
 * according to their 4-integer IDs. It has a header that specifies the total number of entires in the database,
 * the number of angle bins, and the maximum number of entries for any amino acid.
 *
 * dunbrack.bbdep.maxrot is generated using the gen_bbdep_revlookup.py script and contains, for each of the 20
 * standard amino acids, the number of variants in the database for each chi angle (0 if an amino acid doesn't
 * have the chi angle in question).
 *
 * dunbrack.bbdep.revlookup is also generated by gen_bbdep.revlookup.py, and it creates data that can be used to 
 * recover the closest 4-integer rotamer ID from a set of chi angles. This is useful to find the closest rotamer
 * in the dunbrack library for an arbitrary user-specified rotamer.
 *
 * To use this class, you can do something like the following:
 *
 * \code
 *	DunbrackLibrary::ConstPointer dunLib = DunbrackLibrary::Create(aadir, "data/tables/dunbrack.bbdep");
 *	// note that the other two files are expected to be in the same place as dunbrack.bbdep!
 *	// to use this class, you mostly just pass it around to other objects, but for examples of 
 *	// how to use it for actual things, see the RotamerGenerator class, and the energy function code.
 * \endcode
 *
 * Python:
 *
 * \code
 * 	dunLib = DunbrackLibrary(aadir, 'data/tables/dunbrack.bbdep')
 * \endcode
 */
class DllExport DunbrackLibrary
{
public:
	typedef boost::shared_ptr<DunbrackLibrary>		Pointer;
	typedef boost::shared_ptr<const DunbrackLibrary> 	ConstPointer;
	
	//////////////////////////////////////////////////////////////////////////
	/////                           STRUCTURES                           /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief See GetInterpData. */
	struct AngleBinData
	{
		int phi1, phi2;
		int psi1, psi2;
		FLOAT dphi, dpsi;
	};

	/** \brief See GetAngleBins. */
	struct IntAngleBins
	{
		int phi, psi;
	};
	
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	static ConstPointer Create(AminoAcidDirectory::ConstPointer aadir, std::string const& fileName);

	DunbrackLibrary(AminoAcidDirectory::ConstPointer aadir, std::string const& fileName);

	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC INTERFACE                       ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief From a 4-integer Dunbrack rotamer ID, get a number you can feed to the nRot arguments of other fns.
	 *
	 * \param aminoAcid The amino acid to which the rotamer you want belongs.
	 * \param dunIndex  The 4-value ID (columns 5-8 in dunbrack.bbdep)
	 *
	 * \return The rotamer number within a given (aminoacid, phi, psi) bin.
	 */
	int GetNRotByDunbrackID(int aminoAcid, blitz::TinyVector<int, 4> const& dunIndex) const;

	/** \brief Get a rotamer number by percent frequency of appearance.
	 *
	 * \param The amino acid to which the rotamer you want belongs.
	 * \param iPhi The phi angle bin (can be gotten from GetAngleBins)
	 * \param iPsi The psi angle bin (can be gotten from GetAngleBins)
	 * \param percentOrder 0 = highest frequency, 1 = 2nd highest, etc.
	 *
	 * \return The rotamer number within a given (aminoacid, phi, psi) bin.
	 */
	int GetNRotByPercent(int aminoAcid, int iPhi, int iPsi, int percentOrder) const;

	/** \brief Return integer angle bins for a given phi, psi angle pair (in degrees)
	 * 
	 * TODO: This function really duplicates part of the functionality of GetInterpData,
	 *  but I implemented it to match exactly what Rosetta does, for debugging purposes.
	 *  This function can probably just be removed at some point.
	 * 
	 * \param phi Backbone dihedral angle, in degrees
	 * \param psi Backbone dihedral angle, in degrees
	 * \param start True if this is the first residue in a chain
	 * \param end True if this is the last residue in a chain
	 *
	 * \return iPhi, iPsi -- integer angle bins
	 */
	IntAngleBins GetAngleBins(FLOAT phi, FLOAT psi, bool start, bool end) const;
	
	/** \brief Call this to get the interpData argument for the Get* functions.
	 *  
	 *  Example:
	 *  \code
	 *  	AngleBinData adata = dunlib->GetInterpData(phi, psi, false, false);
	 *  	Vector4 chi = dunlib->GetInterpChi(aa, adata, 0);
	 *  	Vector4 SD  = dunlib->GetInterpSD(aa, adata, 0);
	 *  	// ...etc
	 *  \endcode
	 *
	 * \param phi Backbone dihedral angle, in degrees
	 * \param psi Backbone dihedral angle, in degrees
	 * \param start True if this is the first residue in a chain
	 * \param end True if this is the last residue in a chain
	 *
	 * \return Interpolation data
	 */
	AngleBinData GetInterpData(FLOAT phi, FLOAT psi, bool start, bool end) const;

	/** \brief Get interpolated chi values for rotamers from the dunbrack library.
	 *
	 * For each (aa,phi,psi) bin, there is a set of rotamers that most frequently appear in
	 * the PDB. If you have floating-point (phi,psi) values, you can interpolate between 
	 * the chi values for a given rotamer in (iphi, ipsi) and a 'similar' rotamer in (iphi+1, ipsi+1).
	 * This function makes this process transparent; the upshot is that you can get 'likely'
	 * chi values for any floating-point phi,psi combination in order of likelihood by combining
	 * this function with GetInterpData and GetNRotByPercent.
	 *
	 * \param aminoAcid The amino acid to which the rotamer you want belongs
	 * \param interpData Get this from GetInterpData
	 * \param nRot A more or less arbitrary identifier to distinguish between rotamers in the same
	 * 	(phi,psi) bin -- you'll probably usually get this from GetNRotByPercent.
	 *
	 * \return chi angles (degrees)
	 */
	Vector4 GetInterpChi(int aminoAcid, AngleBinData const& interpData, int nRot) const;
	
	/** \brief Get interpolated standard deviations for the chi angles you get from GetInterpChi
	 *
	 * See GetInterpChi. This function simply returns the standard deviations corresponding to those
	 * chi angles.
	 *
	 * \param aminoAcid The amino acid to which the rotamer you want belongs
	 * \param interpData Get this from GetInterpData
	 * \param nRot A more or less arbitrary identifier to distinguish between rotamers in the same
	 * 	(phi,psi) bin -- you'll probably usually get this from GetNRotByPercent.
	 *
	 * \return standard deviations (degrees)
	 */
	Vector4 GetInterpSD(int aminoAcid, AngleBinData const& interpData, int nRot) const;
	
	/** \brief Get interpolated frequency of appearance for the rotamer you get from GetInterpChi
	 *
	 * See GetInterpChi. This function simply returns the interpolated frequency with which you might
	 * expect this particular rotamer to appear in a protein.
	 *
	 * \param aminoAcid The amino acid to which the rotamer you want belongs
	 * \param interpData Get this from GetInterpData
	 * \param nRot A more or less arbitrary identifier to distinguish between rotamers in the same
	 * 	(phi,psi) bin -- you'll probably usually get this from GetNRotByPercent.
	 *
	 * \return percent frequency
	 */
	FLOAT   GetInterpPercent(int aminoAcid, AngleBinData const& interpData, int nRot) const;

	/** \brief Find the (nRot of the) Dunbrack rotamer closest to a given set of chi angles
	 *
	 * Indeed, due to the way the Dunbrack library is constructed, this doesn't depend at all
	 * on your phi and psi angles.
	 *
	 * \param aminoAcid The amino acid to which the rotamer you want belongs
	 * \param chi The chi angles you want to match to a Dunbrack rotamer
	 *
	 * \return The rotamer number within a given (aminoacid, phi, psi) bin.
	 */
	int GetBaseRotamer(int aminoAcid, Vector4 const& chi) const;

	/** \brief Writes out the current contents of the DunbrackLibrary in binary format.
	 *
	 * You can then pass the compiled file into the DunbrackLibrary constructor instead of the
	 * normal plaintext file. This make it much faster to load the data, and a little harder for
	 * someone to edit your data files.
	 *
	 * \param outputFileName The file to store the compiled data into.
	 */
	void Compile(std::string const& outputFileName) const;

private:
	////////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Loads the dunbrack.bbdep.maxrot and dunbrack.bbdep.revlookup files */
	void LoadAuxiliaryFiles(std::string const& fileName);

	/** \brief Loads the DunbrackLibrary from a binary image file */
	void LoadFromBinary(std::ifstream& file);

	/** \brief Saves the DunbrackLibrary to a binary image file */
	void SaveToBinary(std::ofstream& file) const;

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	AminoAcidDirectory::ConstPointer aadir_;

	blitz::Array<Vector4, 1> chiData_;
	blitz::Array<Vector4, 1> sdData_;
	blitz::Array<FLOAT, 1> probData_;
	
	blitz::Array<int, 4> packedIndex_;
	blitz::Array<int, 4> percentOrderIndex_;  

	blitz::Array<int, 5> revLookup_;
	blitz::Array<int, 2> maxRotType_;
};

#endif


