// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Jim Maiolo 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef JRM_ENERGY_COMB_LIST_H
#define JRM_ENERGY_COMB_LIST_H

#include "general.h"

#include <queue>
#include <vector>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/utility.hpp>

/**
 * \file energy_comb_list.h
 *
 * \brief Classes for storing an energy with its associated configuration
 * as well as a priority_queue of these combinations
 *
 * The vector presumably specifies a particular solution, and the energy 
 * stored is the energy associated with that vector.  Two classes are
 * defined: EnergyComb stores a single combination and implements the
 * necessary comparison operators for storing this on a priority_queue, 
 * while EnergyCombList is a wrappper for the priority_queue itself.  
 * Note that all of the function definitions are actually in this file
 * since they are all inline.
 **/

/**
 * \brief A wrapper to store a combination of an energy with a state
 * as specified by std::vector<int>
 **/
class DllExport EnergyComb {
    public:
        ////////////////////////////////////////////////////////////////////////
        //////                          CONSTRUCTORS                      //////
        ////////////////////////////////////////////////////////////////////////
        /**
         * \brief Construct a new combination from the state and the energy
         *
         * \param vec     The integer vector specifying the state
         * \param energy  The energy associated with the state
         **/
        EnergyComb(std::vector<int> vec, FLOAT energy) {
            data_ = std::make_pair(vec, energy);
        }

        /**
         * \brief No argument constructor for declaring variables
         **/
        EnergyComb() { /* does nothing */ }

        // EnergyComb(EnergyComb const& old_EnergyComb);
        //      Use automatically generated copy constructor

        // opertator = (EnergyComb const& old_EnergyComb);
        //      Use automatically generated assignment operator
        
        // ~EnergyComb();
        //      Use automatically generated destructor

        ////////////////////////////////////////////////////////////////////////
        //////                    COMPARISON OPERATORS                    //////
        ////////////////////////////////////////////////////////////////////////
        
        /**
         * \brief Less than operator for storing on a priority queue with the
         * highest energy at the top
         *
         * \param oper1  a in 'a < b'
         * \param oper2  b in 'a < b'
         * \return  true if a.getEnergy < b.getEnergy
         **/
        friend bool operator < (EnergyComb const& oper1,
                                EnergyComb const& oper2) {
            return oper1.getEnergy() < oper2.getEnergy();
        }
        
        /**
         * \brief greater than operator for storing on a priority queue with
         * the lowest energy at the top
         *
         * \param oper1  a in 'a > b'
         * \param oper2  b in 'a > b'
         * \return  true if a.getEnergy > b.getEnergy
         */
        friend bool operator > (EnergyComb const& oper1,
                                EnergyComb const& oper2) {
            return oper1.getEnergy() > oper2.getEnergy();
        }

        ////////////////////////////////////////////////////////////////////////
        //////                    PARAMETER ACCESSORS                     //////
        ////////////////////////////////////////////////////////////////////////
        
        /**
         * \brief Return the state of the combination as an int vector
         *
         * \return int vector representing the state
         **/
        std::vector<int> const& getVector() const {
            return data_.first;
        }

        /**
         * \brief Return the energy of the combination
         *
         * \return FLOAT energy of the combo
         **/
        FLOAT getEnergy() const {
            return data_.second;
        }

    private:
        std::pair< std::vector<int>, FLOAT > data_; ///< The combination

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(data_);
	}
};

/**
 * \brief A class that wraps the priority queue of Energy Combinations.
 * It exposes all of the same functionality as a normal priority_queue, except
 * that the elements may be stored in increasing or decreasing order.
 *
 * All of the queue functions should behave as expected.
 *
 * The function isDescOrder() will return true if the queue is being stored in
 * decreasing order (largest element at the top) and false if it is in
 * increasing order.  The order of the queue can be reversed with reverse(),
 * but that function should be called sparingly as it needs to copy every
 * element from one queue into another.
 **/
class DllExport EnergyCombList { 
    public:    
        // Shared pointers
        typedef boost::shared_ptr<EnergyCombList>       Pointer;
        typedef boost::shared_ptr<const EnergyCombList> ConstPointer;
        
        ////////////////////////////////////////////////////////////////////////
        //////                        CONSTRUCTORS                        //////
        ////////////////////////////////////////////////////////////////////////
        
        /**
         * \brief Make a new empty energy combination list
         *
         * \param descOrder  true if the elements are in decreasing order
         **/
        EnergyCombList(bool descOrder) {
            descOrder_ = descOrder;
        }

        // EnergyCombList(EnergyCombList const& old_EnergyCombList);
        //      Use automatically generated copy constructor

        // EnergyCombList operator = (EnergyCombList const old_EnergyCombList)
        //      Use automatically generated assignment operator

        // ~EnergyCombList()
        //      Use automatically generated destructor (for now...)
        
        ////////////////////////////////////////////////////////////////////////
        //////                      QUEUE FUNCTIONS                       //////
        ////////////////////////////////////////////////////////////////////////
        
        /**
         * \brief push the EnergyComb
         *
         * \param comb The new combination to add
         **/
        void push(EnergyComb const& comb) {
            if (descOrder_)
                queue_dec_.push(comb);
            else
                queue_inc_.push(comb);
        }

        /**
         * \brief return the top element of the queue
         *
         * \return the greatest or smallest element, depending on the type of
         * queue being used
         **/
        EnergyComb const& top() const {
            if (descOrder_)
                return queue_dec_.top();
            else
                return queue_inc_.top();
        }

        /**
         * \brief remove the top element of the queue
         **/
        void pop() {
            if (descOrder_) 
                queue_dec_.pop();
            else
                queue_inc_.pop();
        }

        /**
         * \brief find out the number of elements on the queue
         *
         * \return signed integer size
         **/
        int size() {
            if (descOrder_)
                return static_cast<int>(queue_dec_.size());
            else
                return static_cast<int>(queue_inc_.size());
        }

        /**
         * \brief check for the empty queue
         *
         * \return true if empty
         **/
        bool empty() {
            if (descOrder_)
                return queue_dec_.empty();
            else
                return queue_inc_.empty();
        }

        /**
         * \brief Check which order the queue is stored in decreasing order.
         *
         * \return true if the largest element is at the top
         **/
        bool isDescOrder() {
            return descOrder_;
        }

        /**
         * \brief Reverse the queue storage order.
         *
         * This function should be used sparingly because it can take up to
         * N(log N) time in the worst case.  Every element is popped from one
         * queue and pushed onto the other.
         **/
        void reverse() {
            // wish there was a nicer way to do this...
            if (descOrder_) {
                // make sure the receiving queue is empty (just in case)
                while (!queue_inc_.empty())
                    queue_inc_.pop();

                // move them over
                while (!queue_dec_.empty()) {
                    queue_inc_.push(queue_dec_.top());
                    queue_dec_.pop();
                }

                // flip descOrder_
                descOrder_ = false;
            } else {
                // make sure the receiving queue is empty (just in case)
                while (!queue_dec_.empty())
                    queue_dec_.pop();

                // move them over
                while (!queue_inc_.empty()) {
                    queue_dec_.push(queue_inc_.top());
                    queue_inc_.pop();
                }

                // flip descOrder_
                descOrder_ = true;
            }
        }

    
    private:
        /**
         * \brief The queue for increasing order storage
         **/
        std::priority_queue<EnergyComb, std::vector<EnergyComb>,
                    std::greater<EnergyComb> > queue_inc_;

        /**
         * \brief The queue for decreasing order storage
         **/
        std::priority_queue<EnergyComb, std::vector<EnergyComb>,
                    std::less<EnergyComb> > queue_dec_;

        bool descOrder_;        ///< true for decreasing order

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	/*
	We'll need to implement priority_queue serialization if we want this

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(queue_inc_);
		ar & BOOST_SERIALIZATION_NVP(queue_dec_);
		ar & BOOST_SERIALIZATION_NVP(descOrder_);
	}
	*/
};

#endif          /* JRM_ENERGY_FUNCTION_H */

