// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ENERGY_FUNCTION_H
#define IVL_ENERGY_FUNCTION_H

#include "general.h"
#include "data_table.h"
#include "system.h"
#include "rotamers.h"
#include "energy_graph.h"
#include "rotamer_iteration_map.h"

// platform specifc dynamic linking includes
#ifdef __APPLE__

#include <dlfcn.h>
#define DLL_ACCESS dlsym 

#else
#ifdef __GNUC__

#include <ltdl.h>
#define DLL_ACCESS lt_dlsym

#else
#ifdef _MSC_VER

#include <windows.h>
#define DLL_ACCESS GetProcAddress

#endif
#endif
#endif

/** \file energy_function.h
 *  \brief Contains classes for easing the loading of energy functions from separate shared library files.
 * This code is Linux-specific and uses libtool's dlopen and company.
 */

/** \brief
 * Create an instance of this class for each distinct energy function you use. You can then tweak any 
 * parameters it may have and have it fill in an energy graph for packing.
 *
 * Example:
 *
 * \code
 * 	EnergyFunction ef("libCHOMP_ef_rosetta.so", atypes, directory, dunLib);
 *	ef.SetParameter("some_parameter", 1.2);
 *	EnergyGraph eg;
 *	ef.FillEnergyGraph(rotamers, nmap, imap, &eg);
 * \code
 *
 * Python:
 *
 * \code
 * 	ef = EnergyFunction('libCHOMP_ef_rosetta.so', atypes, directory, dunLib);
 *	ef.SetParameter('some_parameter', 1.2);
 *	eg = EnergyFunction()
 *	ef.FillEnergyGraph(rotamers, nmap, imap, &eg);
 * \code
 */
class DllExport EnergyFunction
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Construct an energy graph from a shared library and a set of data tables.
	 *
	 * \param libName The filename of the compiled energy function .so file.
	 * \param atypes  The atom type set this energy function uses.
	 * \param directory The set of data tables associated with the .so
	 * \param dunLib The dunbrack rotamer library used to generate the rotamers that
	 * \param rstats An optional object for keeping track of the amount of computational resources
	 *	this energy function uses.
	 *  will be evaluated with this energy function.
	 */
	EnergyFunction(std::string const& libName, 
		       AtomTypes::ConstPointer atypes, 
		       DataDirectory::ConstPointer directory, 
		       DunbrackLibrary::ConstPointer dunLib,
		       RuntimeStats::Pointer rstats
			= RuntimeStats::Pointer() );

	/** \brief Releases all resources associated with this energy function, unloads the .so 
	 * from memory.
	 */
	~EnergyFunction();

	///////////////////////////////////////////////////////////////////////////
	//////                      STANDARD INTERFACE                      ///////
	///////////////////////////////////////////////////////////////////////////


	/** \brief Set the path to search for energy function plugins. Defaults to the current directory.
	*/
	static void SetPluginPath(std::string const& path);

	/** \brief Get the current plugin search path (See SetPluginPath).
	*/
	static std::string GetPluginPath();


	/** \brief TODO: Document me
	*/
	void Reinitialize(AtomTypes::ConstPointer atypes, 
		       	  DataDirectory::ConstPointer directory, 
			  DunbrackLibrary::ConstPointer dunLib,
			  RuntimeStats::Pointer rstats 
				= RuntimeStats::Pointer() );

	/** \brief Use this energy function to calculate the energy of a System.
	 *
	 *  \param residues The system to score.
	 *  \param nmap A neighbor map that contains the system to be scored.
	 *  \param energiesOut An array in which each energy component will be stored.
	 */
	FLOAT ScoreSystem(System::ConstPointer residues, NeighborMap::ConstPointer nmap);

	blitz::Array<FLOAT, 3> GetPairwiseEnergies(System::ConstPointer residues, 
			                           NeighborMap::ConstPointer nmap); 

	blitz::Array<FLOAT, 2> GetOneBodyEnergies(System::ConstPointer residues, 
			 	                  NeighborMap::ConstPointer nmap);

	/** \brief Construct an energy graph from a rotamer library.
	 *
	 * \param rotamers The rotamer library for which you want to calculate energies.
	 * \param nmap A neighbor map that contains all the systems you wish to calculate energies for.
	 * \param rmap A data structure that allows you to exclude certain rotamers/systems from the graph.
	 * \param graphOut The newly filled energy graph. NOTE that if the graph you pass in contains data,
	 *  old entries will NOT be deleted! Futhermore, re-scoring the same rotamer/pair into the same graph
	 *  will cause an exception at best, or a silent bug at worst. Delete all graph nodes that are in your
	 *  rmap before calling this function.
	 */
	void FillEnergyGraph(RotamerLibrary::ConstPointer rotamers, 
			     NeighborMap::ConstPointer nmap, 
			     RotamerIterationMap::ConstPointer rmap, 
			     EnergyGraph* graphOut); 	

	//////////////////////////////////////////////////////////////////////////
	/////                       PARAMETER ACCESSORS                      /////
	//////////////////////////////////////////////////////////////////////////
	
	// note: the unusual way these accessors are handled is to make this class
	// work smoothly with boost::python
	
	/** \brief Gets a floating-point energy function parameter.
	 *
	 * This function allows you to retrieve the value of any FLOAT global variable in the energy
	 * function .so, by name. Note that trying to grab a variable that doesn't exist, or one that
	 * is not a FLOAT, will probably cause a crash.
	 *
	 * \param paramName The symbol name of the parameter you want.
	 * \return The FLOATing-point value of the parameter.
	 */
	FLOAT  GetFloatParameter(std::string const& paramName) { return GetParameter<FLOAT>(paramName); }
	
	/** \brief Sets a floating-point energy function parameter.
	 *
	 * This function allows you to set the value of any FLOAT global variable in the energy
	 * function .so, referenced by name. Note that trying to set a variable that doesn't exist, 
	 * or one that is not a FLOAT, will probably cause a crash.
	 *
	 * \param paramName The symbol name of the parameter you want.
	 * \param value The new FLOATing-point value of the parameter.
	 */
	void   SetParameter(std::string const& paramName, FLOAT value) { SetParameter<FLOAT>(paramName, value); }

private:

	///////////////////////////////////////////////////////////////////////////
	//////                      PARAMETER TEMPLATES                      //////
	///////////////////////////////////////////////////////////////////////////

	template<class ParamType>
	ParamType& GetParameter(std::string const& paramName)
	{
		ParamType* ptr = reinterpret_cast<ParamType*>(DLL_ACCESS(module_, paramName.c_str()));
		
		if (!ptr) throw std::string("Energy function parameter " + paramName + " not found or wrong type.");
		
		return *ptr;
	}

	template<class ParamType>
	void SetParameter(std::string const& paramName, ParamType value)
	{
		GetParameter<ParamType>(paramName) = value;
	}

	///////////////////////////////////////////////////////////////////////////
	//////                         HELPER FNS                           ///////
	///////////////////////////////////////////////////////////////////////////
	
	static void UnloadAll();
	static void CreateLoadedModuleDir();
	static void DeleteLoadedModuleDir();
	void LoadModule(std::string const& libName);

	//////////////////////////////////////////////////////////////////////////
	/////                          PRIVATE DATA                          /////
	//////////////////////////////////////////////////////////////////////////

	AtomTypes::ConstPointer atypes_;
	DataDirectory::ConstPointer dataDir_;
	DunbrackLibrary::ConstPointer dunLib_;
#ifdef __APPLE__
        // data structures
        void* module_;
#else
#ifdef __GNUC__
	// libtool data structures
	lt_dlhandle module_;
#else
#ifdef _MSC_VER
	// win32 DLL data structures
	HINSTANCE module_;
#endif
#endif
#endif

	static int efcount_;
	static std::map<std::string, int> loaded_efs_;
	static std::string ef_dirname_;
	static std::string ef_path_;
	
	// function pointers
	typedef void (*InitFn)(	DataDirectory::ConstPointer directory,
			   	DunbrackLibrary::ConstPointer dunLib,
			   	AtomTypes::ConstPointer atypes,
				RuntimeStats::Pointer rstats);

	typedef FLOAT (*ScoreFn)(System::ConstPointer residues, NeighborMap::ConstPointer nmap);
	typedef void (*PWFn)(System::ConstPointer residues, NeighborMap::ConstPointer nmap, 
			     blitz::Array<FLOAT, 3>& energies);
	typedef void (*SBFn)(System::ConstPointer residues, NeighborMap::ConstPointer nmap,
			     blitz::Array<FLOAT, 2>& energies);

	typedef void (*GraphFn)(RotamerLibrary::ConstPointer    rotamers,
				NeighborMap::ConstPointer 	nmap,
				RotamerIterationMap::ConstPointer rmap,
				EnergyGraph* graphOut);
	
	ScoreFn scoreFn_;
	GraphFn graphFn_;
	InitFn  initFn_;
	PWFn pairwiseFn_;
	SBFn onebodyFn_;
};

#endif


