// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ENERGY_GRAPH
#define IVL_ENERGY_GRAPH

#include "general.h"
#include "runtime_stats.h"

#include <vector>

#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/adj_list_serialize.hpp>

#include <boost/archive/text_oarchive.hpp>
#include <boost/serialization/nvp.hpp>
#include <boost/serialization/base_object.hpp>

/** \file energy_graph.h
 *  \brief Contains an energy graph data structure (a see-through wrapper of boost::graph) to hold
 *   precomputed single-body and pairwise energies a rotamer or pose library. 
 *   For most libraries, the pairwise energies would take many gigabytes of RAM if stored in a matrix,
 *   so they are stored in an adjacency-list type graph instead.
 */

// TODO: A lot of the functionality here is specific to the SimulatedAnnealingPacker. Perhaps
// some refactoring is in order here.

typedef std::vector< std::pair<int, int> > IDSet;

/** \brief Represents a node in the energy graph. An edge is a pairwise energy.
 */
struct DllExport EnergyGraphNode
{
	FLOAT energy;		///< The single-body energy for the object
	int   elementID;	///< The ID as used in the corresponding DiscreteSearchSpace
	bool  active;		///< A 'scratch' variable for singling out subsets of the graph

	//////////////////////////////////////////////////////////////////////////
	/////                          CONSTRUCTORS                          /////
	//////////////////////////////////////////////////////////////////////////
	
	EnergyGraphNode (): energy(0.0), elementID(-1), active(false) {}
	EnergyGraphNode (int elementID_, FLOAT energy_): 
		energy(energy_), elementID(elementID_), active(false) {}

	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////
private:

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(elementID);
		ar & BOOST_SERIALIZATION_NVP(energy);
	}
};

typedef boost::adjacency_list<boost::vecS,   		// edge list type
			      boost::vecS,   		// vertex list type
			      boost::undirectedS,	// graph type
			      EnergyGraphNode,		// vertex property type (one-body energy)
			      FLOAT>	     		// edge property type (pairwise energy)
	BoostGraphType;

/** \brief An energy graph data structure (a see-through wrapper of boost::graph) to hold
 *   precomputed single-body and pairwise energies for a rotamer or pose library. 
 *   For most libraries, the pairwise energies would take many gigabytes of RAM if stored in 
 *   a matrix, so they are stored in an adjacency-list type graph instead. In addition to the
 *   basic boost::graph functionality, this class has some helper functions used in the
 *   SimulatedAnnealingPacker.
 */
class DllExport EnergyGraph : public BoostGraphType 
{
public:

	//////////////////////////////////////////////////////////////////////////
	/////                          CONSTRUCTORS                          /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief The default constructor does nothing. */
	EnergyGraph() : energyThreshold(0.01) {}

	virtual ~EnergyGraph() {}

	//////////////////////////////////////////////////////////////////////////
	/////                          GRAPH HELPERS                         /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief Adds a vertex to the graph.
	 * You can use this instead of the boost vertex-adding function from Python.
	 * NOTE: For Simulated Annealing, vertices for rotamers for one residue MUST be added consecutively.
	 * That means you can't AddVertex a rotamer for one position, then AddVertex a rotamer for another, and then
	 * go back to the first residue position. You also can't add intersperse adding vertices with adding edges.
	 * If you need to revise the graph, you must delete ALL the rotamers for the residue and start over.
	 *
	 * \param ID A unique numerical ID by which to refer to this vertex.
	 * \param energy An optional one-body energy for this vertex.
	 */
	void AddVertex(int ID, FLOAT energy = 0.0);

	/** \brief Adds an edge to the graph.
	 * You can use this instead of the boost edge-adding function from Python.
	 *
	 * \param ID1 The ID of one of the vertices to link.
	 * \param ID2 The ID of the other vertex to link.
	 * \param energy The energy of the interaction.
	 */
	void AddEdge(int ID1, int ID2, FLOAT energy);
	
	/** \brief Changes the energy of a pre-exising vertex. This operation is O(|V|)..
	 *
	 * \param ID The ID of the previously-created vertex.
	 * \param newEnergy The new energy for the vertex.
	 */
	void SetVertexEnergy(int ID, FLOAT newEnergy);

	/** \brief Changes the energy of a pre-exising edge. This operation is approximately O(|V|).
	 *
	 * \param ID1 The ID of one of the vertices connected by the edge.
	 * \param ID2 The ID of the other vertex connected by the edge.
	 * \param newEnergy The new energy for the vertex.
	 */
	void SetEdgeEnergy(int ID1, int ID2, FLOAT newEnergy);

        /** \brief Iterates through the graph and sets the magnitude of all vertex and edge energies
         *         that exceed a certain cap to that cap.
         *
         *  \param cap The maximum magnitude allowed for an energy in the graph.
         */
        void CapGraphAt(FLOAT cap);

	/** \brief Returns the list of all vertex IDs of this energy graph.
	 */
	std::vector<int> GetVertexIDs() const;

	/** \brief Returns the list of all active vertex IDs of this energy graph.
	 */
	std::vector<int> GetActiveVertexIDs() const;

	/** \brief Returns the list of all vertex IDs impinged upon by outgoing edges that optionally
	 *	   meet a magnitude cutoff
	 */
	std::vector<int> GetEdgeDestinationIDs(int ID, FLOAT cutoff) const;

	/** \brief Gets the energy of a vertex. This operation is O(|V|).
	 *
	 * \param ID The ID of a vertex in the graph.
	 */
	FLOAT GetVertexEnergy(int ID) const;

	/** \brief Gets the energy of an edge. This operation is approximately O(|V|).
	 *
	 * \param ID1 The ID of one of the vertices connected by the edge.
	 * \param ID2 The ID of the other vertex connected by the edge.
	 */
	FLOAT GetEdgeEnergy(int ID1, int ID2) const;

	/** \brief Returns a list of (ID, energy) pairs for all edges connected to this vertex.
	 *
	 * \param ID The ID of a vertex in the graph.
	 */
	std::vector< std::pair< int, FLOAT > > 
		GetVertexEdges(int vertexID) const;

	//////////////////////////////////////////////////////////////////////////
	/////                        ENERGY FN HELPERS                       /////
	//////////////////////////////////////////////////////////////////////////
	
	/** \brief Determine if a given pairwise energy is worth including in the energy graph.
	 * Energies of 0 should obviously be left out, and so there is a small cutoff you can 
	 * apply to keep the graph as compact as possible.
	 *
	 * \param value The pairwise energy to test for sufficient size
	 * \return true if the energy should be included
	 */
	inline bool IncludeEnergy(FLOAT value) const
	{
		return std::abs(value) > energyThreshold;
	}

	/** \brief Remove a set of nodes from the graph. 
	 * If you have previously filled an energy graph and wish to re-calculate some part of it
	 * (say with a different energy function, or with some additional rotamers), you MUST delete
	 * all nodes that may be overwritten in this process, first.
	 *
	 * \param nodes The set of element IDs to be removed from the graph
	 */
	virtual void DeleteNodes(IDSet const& nodes);

	//////////////////////////////////////////////////////////////////////////
	/////                   SIMULATED ANNEALING HELPERS                  /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief Calculates the energy of a given set of rotamers or poses in this graph.
	 *  The one-body energy for each element in the stateVector is included, as well as the 
	 *  pairwise energies between each element. The 'active' flag is turned on for all 
	 *  vertices in the stateVector, and turned off for all other vertices, unless the
	 *  noActivate flag is set. Note that if you set the noActivate flag, you should be
	 *  careful that the 'active' flags are already as you want them!
	 *
	 * \param stateVector This must be a set of vertex iterators with exactly one vertex for each
	 *  dimension in the search space.
	 * \param noActivate If set to true, the 'active' flags aren't touched
	 * \return The net energy of the given state vector.
	 */
	virtual FLOAT ComputeEnergy(std::vector<EnergyGraph::vertex_iterator> const& stateVector,
				    bool noActivate = false);

	/** \brief Calculates the energy of a given set of rotamers or poses in this graph.
	 *  This version works with just a list of element IDs, so it's less efficient but easier to use
	 *  (for example, from Python). The 'active' flag is turned on for all 
	 *  vertices in the stateVector, and turned off for all other vertices, unless the
	 *  noActivate flag is set. Note that if you set the noActivate flag, you should be
	 *  careful that the 'active' flags are already as you want them!
	 *
	 * \param stateVector This must be a set of element IDs with exactly one ID for each
	 *  dimension in the search space.
	 * \param noActivate If set to true, the 'active' flags aren't touched
	 * \return The net energy of the given state vector.
	 */
	virtual FLOAT ComputeEnergy(std::vector<int> const& stateVector,
				    bool noActivate = false);

	/** \brief Calculates the energy difference caused by substituting one rotamer or pose for another.
	 *  The one-body energies for the incoming and outgoing elements are accounted for, 
	 *  as well as the pairwise energies between both elements and *any element for which the 
	 *  active flag is set*.
	 *
	 * \param IDIn The new element to include.
	 * \param IDOut The element being replaced (must be in the same dimension as IDIn)
	 * \param rstats Information about how many computations this function uses is passed out into
	 *	this optional RuntimeStats object.
	 * \return The difference in energy caused by the substitution.
	 */
	virtual FLOAT DiffEnergy(EnergyGraph::vertex_iterator IDIn, 
				 EnergyGraph::vertex_iterator IDOut,
				 RuntimeStats::Pointer rstats
					= RuntimeStats::Pointer()) const;

protected:
	//////////////////////////////////////////////////////////////////////////
	/////                        HELPER FUNCTIONS                        /////
	//////////////////////////////////////////////////////////////////////////

#ifndef __GC
	void FindVertexIterators(std::vector<int> const& IDs, 
				 std::vector<EnergyGraph::vertex_iterator>& iteratorsOut) const;
	
	EnergyGraph::vertex_iterator FindVertexIterator(int ID) const;

	EnergyGraph::edge_descriptor FindEdgeDescriptor(int vID1, int vID2) const;
#endif
	//////////////////////////////////////////////////////////////////////////
	/////                          SERIALIZATION                         /////
	//////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		using namespace boost::serialization;
		ar & make_nvp("EnergyGraph", base_object<BoostGraphType>(*this));
	}
	
	//////////////////////////////////////////////////////////////////////////
	/////                          PRIVATE DATA                          /////
	//////////////////////////////////////////////////////////////////////////
	
	FLOAT energyThreshold;		///< The minimum energy to be included in this graph

};

/** \brief A variant on the EnergyGraph data structure; it's identical in form, but treats its nodes
 *   in a different way. In a Hypergraph, there are two types of nodes (both stored in the EnergyGraphNode
 *   structure):
 *
 *   1. Data nodes, containing data elements such as rotamers or poses
 *   2. Edge nodes, representing hyper-edges. elementID must be -1.
 *
 *   A hypergraph data structure is thus represented using a normal graph data type by enforcing the constraint
 *   that data nodes connect only to edge nodes, and edge nodes connect only to data nodes.
 *
 *   NOTE that while this structure can trivially emulate the normal EnergyGraph, it is significantly less
 *   efficient both in terms of time and space. This is because a hypergraph is inherently harder to represent
 *   than a simple graph. If N-way interactions are not required, use the normal EnergyGraph variant.
 */
class DllExport EnergyHyperGraph : public EnergyGraph
{
public:
	//////////////////////////////////////////////////////////////////////////
	/////                          CONSTRUCTORS                          /////
	//////////////////////////////////////////////////////////////////////////

	EnergyHyperGraph() {}
	virtual ~EnergyHyperGraph() {}

	//////////////////////////////////////////////////////////////////////////
	/////                       HYPERGRAPH HELPERS                       /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief Adds a vertex to the hypergraph.
	 * Use this instead of the boost vertex-adding function to preserve the hypergraph constraints.
	 * NOTE: For Simulated Annealing, vertices for poses within one block MUST be added consecutively.
	 * That means you can't AddVertex a pose for one block, then AddVertex a pose for another, and then
	 * go back to the first block. You also can't add intersperse adding vertices with adding edges.
	 * If you need to revise the graph, you must delete ALL the poses for a block and start over.
	 *
	 * \param ID A unique numerical ID by which to refer to this vertex.
	 * \param energy An optional one-body energy for this vertex.
	 */
	void AddVertex(int ID, FLOAT energy = 0.0);

	/** \brief Adds an edge to the hypergraph.
	 * Use this instead of the boost edge-adding function to preserve the hypergraph constraints.
	 *
	 * \param IDs The IDs of the vertices to link with a hyperedge.
	 * \param energy The energy of the interaction
	 */
	void AddEdge(std::vector<int> const& IDs, FLOAT energy);

	/** \brief Changes the energy of a pre-exising vertex. This operation is O(|V|).
	 *
	 * \param ID The ID of the previously-created vertex.
	 * \param newEnergy The new energy for the vertex.
	 */
	void SetVertexEnergy(int ID, FLOAT newEnergy);

	/** \brief Changes the energy of a pre-exising edge. This operation is O(|V|+|E|).
	 *
	 * \param IDs The IDs of the vertices connected by the edge.
	 * \param newEnergy The new energy for the vertex.
	 */
	void SetEdgeEnergy(std::vector<int> const& IDs, FLOAT newEnergy);

	/** \brief Returns the list of all vertex IDs of this energy graph.
	 */
	std::vector<int> GetVertexIDs() const;

	/** \brief Gets the energy of a vertex. This operation is O(|V|).
	 *
	 * \param ID The ID of a vertex in the graph.
	 */
	FLOAT GetVertexEnergy(int ID) const;

	/** \brief Gets the energy of an edge. This operation is O(|V|+|E|).
	 *
	 * \param IDs The IDs of the vertices connected by the edge you want.
	 */
	FLOAT GetEdgeEnergy(std::vector<int> const& IDs) const;

	/** \brief Returns a list of (ID_list, energy) pairs for all edges connected 
	 *	to the given vertex. This operation is O(|V|+|E|).
	 *
	 * \param ID The ID of a vertex in the graph.
	 */
	std::vector< std::pair< std::vector<int>, FLOAT > >
		GetVertexEdges(int vertexID) const;

	//////////////////////////////////////////////////////////////////////////
	/////                        ENERGY FN HELPERS                       /////
	//////////////////////////////////////////////////////////////////////////
	
	/** \brief Remove a set of nodes from the graph. 
	 * Analogous to the simple EnergyGraph version; this one has to do some additional bookkeeping.
	 *
	 * \param nodes The set of element IDs to be removed from the graph
	 */
	virtual void DeleteNodes(IDSet const& nodes);

	//////////////////////////////////////////////////////////////////////////
	/////                   SIMULATED ANNEALING HELPERS                  /////
	//////////////////////////////////////////////////////////////////////////

	/** \brief Calculates the energy of a given set of rotamers or poses in this graph.
	 *  The one-body energy for each element in the stateVector is included, as well as the 
	 *  N-wise energies between each element. The 'active' flag is turned on for all 
	 *  vertices in the stateVector, and turned off for all other vertices, unless the
	 *  noActivate flag is set. Note that if you set the noActivate flag, you should be
	 *  careful that the 'active' flags are already as you want them!
	 *
	 * \param stateVector This must be a set of vertex iterators with exactly one vertex for each
	 *  dimension in the search space.
	 * \param noActivate If set to true, the 'active' flags aren't touched
	 * \return The net energy of the given state vector.
	 */
	virtual FLOAT ComputeEnergy(std::vector<EnergyGraph::vertex_iterator> const& stateVector,
								bool noActivate = false);

	/** \brief Calculates the energy of a given set of rotamers or poses in this graph.
	 *  This version works with just a list of element IDs, so it's less efficient but easier to use
	 *  (for example, from Python). The 'active' flag is turned on for all 
	 *  vertices in the stateVector, and turned off for all other vertices, unless the
	 *  noActivate flag is set. Note that if you set the noActivate flag, you should be
	 *  careful that the 'active' flags are already as you want them!
	 *
	 * \param stateVector This must be a set of element IDs with exactly one ID for each
	 *  dimension in the search space.
	 * \param noActivate If set to true, the 'active' flags aren't touched
	 * \return The net energy of the given state vector.
	 */
	virtual FLOAT ComputeEnergy(std::vector<int> const& stateVector,
								bool noActivate = false);

	/** \brief Calculates the energy difference caused by substituting one rotamer or pose for another.
	 *  The one-body energies for the incoming and outgoing elements are accounted for, 
	 *  as well as the N-wise energies between both elements and *any element for which the 'active' flag
	 *  is set*.
	 *
	 * \param IDIn The new element to include.
	 * \param IDOut The element being replaced (must be in the same dimension as IDIn)
	 * \param rstats Information about how many computations this function uses is passed out into
	 *	this optional RuntimeStats object.
	 * \return The difference in energy caused by the substitution.
	 */
	virtual FLOAT DiffEnergy(EnergyGraph::vertex_iterator IDIn, 
				 EnergyGraph::vertex_iterator IDOut,
				 RuntimeStats::Pointer rstats
					= RuntimeStats::Pointer() ) const;
protected:

	//////////////////////////////////////////////////////////////////////////
	/////                        HELPER FUNCTIONS                        /////
	//////////////////////////////////////////////////////////////////////////

#ifndef __GCCXML__
	EnergyGraph::vertex_descriptor FindEdgeDescriptor(std::vector<int> const& vIDs) const;
#endif
};

//////////////////////////////////////////////////////////////////////////
/////                             ALIASES                            /////
//////////////////////////////////////////////////////////////////////////

namespace pyplusplus
{
namespace aliases
{
	typedef std::vector<EnergyGraph::vertex_iterator>		EGVertexVector;

}}

#endif


