// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_INTERPOLATE_H
#define IVL_INTERPOLATE_H

#include "general.h"
#include "vectors.h"

#include <cmath>
#include <blitz/array.h>

namespace CHOMPMath
{

/** \file interpolate.h 
 * \brief Contains free functions for performing various types of interpolation, which is used in all 
 * areas of the code.
 */

/** \brief Gets the interpolation bin for a given floating point angle and a given total number of bins.
 *
 * \param angle The angle to discretize.
 * \param nbins The total number of bins over which the angle is discretized.
 * \param bin1 (output) the lower bin
 * \param bin2 (output) the upper bin
 * \param diff (output) 0.0 - 1.0, the angle's location within the bin
 */
inline void GetAngleInterpBins(FLOAT angle, int nbins, int* bin1, int* bin2, FLOAT* diff)
{
	const FLOAT f_bins = static_cast<FLOAT>(nbins);
	const FLOAT bin_angle = 360.0 / f_bins;

	const FLOAT fbin1_raw = (angle - (0.5 * bin_angle)) / bin_angle;
	const FLOAT fbin1 = fmod( fmod(fbin1_raw, f_bins) + nbins, f_bins );
	
	*bin1 = static_cast<int>( fbin1 );
	*bin2 = *bin1 + 1;

 	*diff = fbin1 - *bin1;
	
	// watch for overflow!
	if (*bin1 == nbins)
	{
		*bin1 = 0;
		*diff = 0.0;
	}
	if (*bin2 == nbins)
	{
		*bin2 = 0;
		// *diff = 0.0;    matching Rosetta's behavior
	}
}

/** \param A helper function to subtract angles in degrees and put them in a standardized form.
 *
 * \param x An angle in degrees.
 * \param y An angle in degrees.
 * \return x - y, in degrees and wrapped to 0-360
 */
inline FLOAT AngleSub(FLOAT x, FLOAT y)
{
	return WrapValue(x - y, 360.0);
}

inline void InterpBilinearAngles(FLOAT const x0y0, FLOAT const x1y0, FLOAT const x0y1, FLOAT const x1y1,
				 FLOAT const xd, FLOAT const yd,
				 int nbins, FLOAT* val, FLOAT* dval_dx, FLOAT* dval_dy)
{
	FLOAT const binrange = 360.0 / static_cast<FLOAT>(nbins);
	
	FLOAT const w1 = ( 1.0f - xd ) * ( 1.0f - yd );
	FLOAT const w2 = xd * ( 1.0f - yd );
	FLOAT const w3 = ( 1.0f - xd ) * yd;
	FLOAT const w4 = xd * yd;

	FLOAT const w12 = w1 + w2;
	FLOAT a12;
	if ( w12 != 0.0 ) {
		a12 = ( w1 * x0y0 + w2 * ( AngleSub(x1y0,x0y0) + x0y0 ) ) / w12;
	} else {
		a12 = 0.0;
	}

	FLOAT const w34 = w3 + w4;
	FLOAT a34;
	if ( w34 != 0.0 ) {
		a34 = ( w3 * x0y1 + w4 * ( AngleSub(x1y1,x0y1) + x0y1 ) ) / w34;
	} else {
		a34 = 0.0;
	}

	*val = w12 * a12 + w34 * ( AngleSub(a34,a12) + a12 );
	*val = WrapValue(*val, 360.0);

	*dval_dx = ( ( 1.0f - yd ) * AngleSub(x1y0,x0y0) +
		 yd * AngleSub(x1y1,x0y1) ) / binrange;
	*dval_dy = ( ( 1.0f - xd ) * AngleSub(x0y1,x0y0) +
		 xd * AngleSub(x1y1,x1y0) ) / binrange;
}

inline void InterpBilinear(FLOAT const x0y0, FLOAT const x1y0, FLOAT const x0y1, FLOAT const x1y1,
			   FLOAT const xd, FLOAT const yd,
			   int nbins, FLOAT* val, FLOAT* dval_dx, FLOAT* dval_dy)
{
	FLOAT const binrange = 360.0 / static_cast<FLOAT>(nbins);
	
	FLOAT const w1 = ( 1.0f - xd ) * ( 1.0f - yd );
	FLOAT const w2 = xd * ( 1.0f - yd );
	FLOAT const w3 = ( 1.0f - xd ) * yd;
	FLOAT const w4 = xd * yd;

	*val = x0y0 * w1 + x1y0 * w2 + x0y1 * w3 + x1y1 * w4;

	*dval_dx = ( ( 1.0f - yd ) * ( x1y0 - x0y0 ) + yd * ( x1y1 - x0y1 ) ) / binrange;
	*dval_dy = ( ( 1.0f - xd ) * ( x0y1 - x0y0 ) + xd * ( x1y1 - x1y0 ) ) / binrange;
}

/** \brief Performs linear interpolation between two discrete values in a data table.
 * \param var The continuous index value
 * \param binsPerVar The number of bins in the data table per one unit of var
 * \param minVar The starting value of var in the data table
 * \param data The data over which to interpolate
 *
 * \return An interpolated value
 */
inline FLOAT LinearInterpolate(const FLOAT var, const FLOAT binsPerVar, 
			       const FLOAT minVar, blitz::Array<FLOAT, 1> const& data)
{
	// note the lack of bounds checking for bin1!
	
	const FLOAT f_bin1 = (var - minVar) * binsPerVar;
	int  bin1   = static_cast<int>(f_bin1);
	int  bin2 = bin1 + 1;
	const FLOAT frac   = f_bin1 - static_cast<FLOAT>(bin1);

	if (bin2 >= data.extent(0)) --bin2;
	// should be
	// if (bin1 >= data.extent(0) - 1)
	// 	return data(bin1);

        return data(bin1) + frac * ( data(bin2) - data(bin1) );
}

inline int LinearInterpolateBinCentered(const FLOAT var, const FLOAT binsPerVar, 
					const FLOAT minVar, FLOAT* fracOut)
{
	FLOAT f_bin1 = (var - minVar - 0.5 / binsPerVar) * binsPerVar;
	
	if (f_bin1 < 0.0) f_bin1 = 0.0;

	const int  bin1   = static_cast<int>(f_bin1);
	*fracOut = f_bin1 - static_cast<FLOAT>(bin1);

	return bin1;
}

}

#endif

