// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_NEIGHBOR_MAP_H
#define IVL_NEIGHBOR_MAP_H

#include "general.h"

#include <blitz/array.h>

#include "system.h"
#include "system_map.h"

/** \file neighbor_map.h
 *
 * \brief Contains a class for keeping track of which Residues are within a
 * given cutoff distance of each other.
 *
 * This file provides two services -- keeping track of which Residues in a System
 * are 'neighbors' according to a per-AminoAcid cutoff table, and recording the 
 * number of 'neighbors' for each Residue to determine its degree of burial.
 */

/** \brief Provides two services -- keeping track of which Residues in a System
 * are 'neighbors' according to a per-AminoAcid cutoff table, and recording the 
 * number of 'neighbors' for each Residue to determine its degree of burial.
 * 
 * Whether a pair of residues are neighbors or not is determined with a rough approximation.
 * Residues are treated as spheres with radii specified in a PairCutoffTable dependent on 
 * the amino acid of each residue. The *number* of neighbors is actually calculated differently,
 * with a single cutoff used for all amino acids.
 *
 * Example:
 * \code
 * 	NeighborMap::ConstPointer nmap = NeighborMap::Create(aSystem, data.PairCutoffTable);
 * \code
 *
 * Python:
 * \code
 *	nmap = NeighborMap(aSystem, data.PairCutoffTable)
 * \code
 */
class DllExport NeighborMap
{
public:
	typedef boost::shared_ptr<NeighborMap>		Pointer;
	typedef boost::shared_ptr<const NeighborMap> 	ConstPointer;

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Constructs a neighbor map for a single system.
	 *
	 * If you are going to evaluate rotamer-rotamer interactions ONLY for a single System,
	 * use this constructor to create a NeighborMap that includes only that one System.
	 * If you will be using multiple Systems together, it will usually NOT work to create
	 * one NeighborMap for each; see the other constructor.
	 *
	 * \param system The System to find neighbors for
	 * \param cutoffs The amino-acid dependent cutoff for neighborness, to be found in a DataDirectory
	 * \return A freshly-created NeighborMap.
	 */
	static Pointer Create(System::ConstPointer system, blitz::Array<FLOAT, 2> const& cutoffs);
	
	/** \brief Constructs a neighbor map for multiple systems.
	 *
	 * If you are going to evaluate rotamer-rotamer interactions for multiple systems at once
	 * (i.e. you want rotamers in one system to interact with rotamers in another), use this 
	 * constructor to create a NeighborMap that includes all of them. 
	 *
	 * \param systems The list of Systems to find neighbors for
	 * \param cutoffs The amino-acid dependent cutoff for neighborness, to be found in a DataDirectory
	 * \return A freshly-created NeighborMap.
	 */
	static Pointer Create(std::vector<System::ConstPointer> const& systems, blitz::Array<FLOAT, 2> const& cutoffs);

	NeighborMap (std::vector<System::ConstPointer> const& systems, blitz::Array<FLOAT, 2> const& cutoffs);

	///////////////////////////////////////////////////////////////////////////
	//////                           ACCESSORS                           //////
	///////////////////////////////////////////////////////////////////////////
	
	/** \brief Determines whether a pair of Residues in this NeighborMap are neighbors.
	 *
	 * \param res1 A Residue, must be in a System in this NeighborMap
	 * \param res1 A Residue, must be in a System in this NeighborMap
	 * \return true if res1 and res2 are neighbors according to this NeighborMap's cutoff table.
	 */
	bool AreNeighbors (Residue const& res1, Residue const& res2) const;

	/** \brief Determines the number of neighbors a residue has (used mostly to identify buried residues).
	 *
	 * \param res A Residue, must be in a System in this NeighborMap
	 * \return true The number of neighbors res has.
	 */
	int  NumNeighbors (Residue const& res) const;

	///////////////////////////////////////////////////////////////////////////
	//////                              I/O                              //////
	///////////////////////////////////////////////////////////////////////////

	friend DllExport std::ostream& operator<< (std::ostream& os, NeighborMap::ConstPointer map);
	
	// note: will only read in neighbor map data w/ the same systems!
	friend DllExport std::istream& operator>> (std::istream& is, NeighborMap::Pointer& map);

private:
	SystemPairMap<bool> data_;
	SystemMap<int> nNeighbors_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(data_);
		ar & BOOST_SERIALIZATION_NVP(nNeighbors_);
	}
};

namespace boost { namespace serialization {

template<class Archive>
inline void load_construct_data(
    Archive & ar, NeighborMap* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    std::vector<System::ConstPointer> systems;
    blitz::Array<FLOAT, 2> cutoffs;
    // invoke inplace constructor to initialize instance
    ::new(t)NeighborMap(systems, cutoffs);
}
}} // namespace ...

///////////////////////////////////////////////////////////////////////////
//////                              I/O                              //////
///////////////////////////////////////////////////////////////////////////

/** \brief WRITEME
 */
DllExport std::ostream& operator<< (std::ostream& os, NeighborMap::ConstPointer map);

/** \brief WRITEME
 */
DllExport std::ostream& operator<< (std::ostream& os, NeighborMap::Pointer map);

/** \brief WRITEME
 */
DllExport std::istream& operator>> (std::istream& is, NeighborMap::Pointer& map);

#endif


