// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_OPT_FASTER_H
#define IVL_OPT_FASTER_H

#include "general.h"

#include <vector>
#include <boost/serialization/vector.hpp>
#include <boost/random.hpp>

#include "discrete_search_space.h"
#include "energy_graph.h"
#include "energy_comb_list.h"
#include "runtime_stats.h"
#include "neighbor_map.h"

#include "opt_base.h"

/** \file opt_faster.h
 *  \brief FASTER algorithm from Desmet, improved by Allen.
 **/

class DllExport FasterPacker : public OptimizationBase
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	FasterPacker();

	///////////////////////////////////////////////////////////////////////////
	//////                      PARAMETER ACCESSORS                      //////
	///////////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////////
	//////                        MAIN INTERFACE                        ///////
	///////////////////////////////////////////////////////////////////////////
 
	/**
         * \brief Using a discrete search space, the function performs a
         * simulated annealing calculation, populating a list of lowest
         * energies and returning the lowest energy.
         *
         * The maximum number of solutions to return in the list is specified
         * by the parameter listSize, which must be greater than 0.  Calling
         * the function with listSize == 1 populates the list with exactly one
         * element - the lowest energy found - as long as the energy list is
         * not a null pointer.  The list pointer should always point to an
         * empty list when passed in.  Any non-empty list will be overwritten.
         *
         * NOTE: If a zero or a negative value of listSize is passed in, the
         * function fails quietly, returning a list of size 1.  It is always a
         * good idea to check on the size of the list after packing.  The
         * lowest energy found is always returned so that the function is
         * backward compatible.
         *
         * \param searchSpace   The space to search - rotamers or poses
         * \param graph         Graph of single and pairwise energies
         * \param initialState  vector specifying the starting point
         * \param energyList    Pointer to an empty list to store solutions
         * \param listSize      maximum number of elements to return
         * \return              the lowest energy solution
         **/
        std::pair< std::vector<int>, FLOAT >
		Pack(DiscreteSearchSpace::ConstPointer searchSpace,
		     EnergyGraph* graph, 
		     std::vector<int> const* initialState = 0,
		     EnergyCombList::Pointer energyList 
			= boost::shared_ptr<EnergyCombList>(),
		     int listSize = 1);

	void SetRandomize(bool Randomize);
	bool GetRandomize();

	void SetEdgeCutoff(FLOAT cutoff);
	FLOAT GetEdgeCutoff();

	void SetSPRrounds(int roundCount);
	int GetSPRrounds();

	void SetSimultaneousFixes(int fixCount);
	int GetSimultaneousFixes();


private:
	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	std::vector<FLOAT>			energies_;

	bool Randomize_;
	FLOAT EdgeCutoff_;
	int SPRrounds_;
	int SimultaneousFixes_;
};

#endif


