// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_OPT_MONTECARLO_FBF_H
#define IVL_OPT_MONTECARLO_FBF_H

#include "general.h"

#include <vector>
#include <boost/serialization/vector.hpp>
#include <boost/random.hpp>

#include "discrete_search_space.h"
#include "energy_graph.h"
#include "energy_comb_list.h"
#include "runtime_stats.h"

class DllExport MonteCarloFBFPacker
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	MonteCarloFBFPacker();

	///////////////////////////////////////////////////////////////////////////
	//////                      PARAMETER ACCESSORS                      //////
	///////////////////////////////////////////////////////////////////////////

	FLOAT GetIterationMultiplier() const;
	void SetIterationMultiplier(FLOAT multiplier);

	int GetMaxFrames();
	void SetMaxFrames(int maxFrames);

	///////////////////////////////////////////////////////////////////////////
	//////                        MAIN INTERFACE                        ///////
	///////////////////////////////////////////////////////////////////////////
 
	/**
         * \brief Using a discrete search space, the function performs a
         * Monte Carlo calculation, populating a list of lowest
         * energies and returning the lowest energy.
         *
         * The maximum number of solutions to return in the list is specified
         * by the parameter listSize, which must be greater than 0.  Calling
         * the function with listSize == 1 populates the list with exactly one
         * element - the lowest energy found - as long as the energy list is
         * not a null pointer.  The list pointer should always point to an
         * empty list when passed in.  Any non-empty list will be overwritten.
         *
         * NOTE: If a zero or a negative value of listSize is passed in, the
         * function fails quietly, returning a list of size 1.  It is always a
         * good idea to check on the size of the list after packing.  The
         * lowest energy found is always returned so that the function is
         * backward compatible.
         *
         * \param searchSpace   The space to search - rotamers or poses
         * \param graph         Graph of single and pairwise energies
         * \param initialState  vector specifying the starting point
         * \param energyList    Pointer to an empty list to store solutions
         * \param listSize      maximum number of elements to return
         * \return              the lowest energy solution
         **/
        std::pair< std::vector<int>, FLOAT >
		Pack(DiscreteSearchSpace::ConstPointer searchSpace,
		     EnergyGraph* graph, 
		     std::vector<int> const* initialState = 0,
		     EnergyCombList::Pointer energyList 
			= boost::shared_ptr<EnergyCombList>(),
		     int listSize = 1);

	void Seed(unsigned int seed);

	void SeedWithTime();

	FLOAT GetTemperature();
	void SetTemperature(FLOAT temperature);

	void RecordRuntimeStats(RuntimeStats::Pointer stats);	

private:
	///////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	void InitState(EnergyGraph& graph, std::vector<int> const& initialState);
	
	std::pair<int, int> PickMove();

	bool KeepChange (FLOAT prevEnergy, FLOAT diff, FLOAT temperature);

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	std::vector<EnergyGraph::vertex_iterator> 	stateVector_;
	std::vector<EnergyGraph::vertex_iterator> 	stateIDStart_;
	std::vector<int>			  	freeDimList_;

	boost::mt19937 				mtrand_;
	std::vector<FLOAT>			energies_;
	DiscreteSearchSpace::ConstPointer 	searchSpace_;
	
	FLOAT				temperature_;
	FLOAT				iterMultiplier_;

	int				maxFrames_;

	RuntimeStats::Pointer		stats_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(iterMultiplier_);
		ar & BOOST_SERIALIZATION_NVP(temperature_);
	}
};

#endif


