// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_PDB_H
#define IVL_PDB_H

#include "general.h"

#include <iostream>

#include <boost/signals.hpp>
#include <boost/function.hpp>

/** \file pdb.h
*
* \brief Contains classes for parsing and generating PDB files.
*
* This file implements a basic I/O system for PDB files. The parsing is
* completely decoupled from constructing Residues and Systems -- all that
* happens in this file is reading/writing PDB files using *Record structs.
*
* TODO: maybe make the Reader/Writer classes into actual streams so as to 
* have a more standard interface? This would be really easy for Writer, at least.
*/

namespace PDB
{

///////////////////////////////////////////////////////////////////////////
//////                          PDB RECORDS                          //////
///////////////////////////////////////////////////////////////////////////

/** \brief Holds all the data entries in a PDB 'ATOM' record. 
 *
 * Although the records are fixed-length, the character ones are all stored
 * in std::strings, just for convenience. Anything that's *always* a numeric
 * type (and is never missing entirely) is an int or a float.
 *
 * For details on what all these members are, see the PDB standard.
 */
struct DllExport AtomRecord
{
	uint 		serial;
	std::string 	name;
	std::string	altLoc;
	std::string	resName;
	std::string	chainID;
	uint		resSeq;
	std::string	iCode;
	Vector3		position;
	FLOAT		occupancy;
	FLOAT		tempFactor;
	std::string	element;
	std::string	charge;

	AtomRecord():
		serial(0),
		resSeq(0),
		position(0.0, 0.0, 0.0),
		occupancy(0.0),
		tempFactor(0.0)
	{

	}
};

////////////////////////////////////////////////////////////////////////////
////				  PARSER 				////
////////////////////////////////////////////////////////////////////////////


/** \brief Reads in the data in a PDB file. Currently supports only ATOM records.
 *
 * To parse a PDB file, you create one of these and then register a callback
 * for each type of record you want to recognize. It uses the boost signals/slots 
 * thing. For example:
 *
 * \code
 * void Generic_ReaderTest(std::string const& inputFile, 
 *			boost::function<bool (PDB::AtomRecord const&)> handler)
 * {
 *	std::ifstream input(inputFile.c_str());
 *	PDB::Reader reader;
 *
 *	connection c = reader.ConnectRecordSignal(handler);
 *	
 *	while (!input.eof())
 *		input >> reader; // handler now receives AtomRecords
 *
 *	c.disconnect();
 * }
 * \endcode
 *
 * Note: Your 'handler' functions should return true if they successfully dealt
 * with the record. If they return false, the same record will be resent the next
 * time operator>> is called.
 */
class DllExport Reader
{
public:
	//////////////////////////////////////////////////////////////////////////
	/////                          CONSTRUCTORS                          /////
	//////////////////////////////////////////////////////////////////////////
	
	/** \brief The default constructor connects dummy record handlers to the reader.
	 */
	Reader();

	////////////////////////////////////////////////////////////////////////////
	////			  PUBLIC MEMBER FNS 				////
	////////////////////////////////////////////////////////////////////////////

	friend DllExport std::istream& operator>> (std::istream& is, Reader& reader);

	/** \brief Peeks at the PDB record type that's coming up next.
	 *
	 * i.e. ATOM, HETATM, etc. The record is *not* consumed by this operation.
	 *
	 * \param is The input stream that's piping the PDB file in.
	 * \return The name of the next record.
	 */
	std::string NextRecordType (std::istream& is) const;

	////////////////////////////////////////////////////////////////////////////
	////			  SIGNAL CONNECTORS 				////
	////////////////////////////////////////////////////////////////////////////

	/** \brief Hook up a handler function for AtomRecords.
	 *
	 * \param fn A function that receives an AtomRecord whenever it comes up in a PDB file, and 
	 * 	returns true if the record has been successfully processed. If fn returns false, the
	 * 	same record will be resent next time operator>> is called.
	 * \return A boost 'connection' object which you can later use to disconnect the handler.
	 */
	boost::signals::connection ConnectRecordSignal ( boost::function<bool (AtomRecord const&)> fn )
	{
		return AtomRecordSignal.connect(fn);
	}

	/** \brief Hook up a handler function for default records, which are anything this class doesn't
	 *  know how to handle yet. These are passed into the hander as raw strings.
	 *
	 * \param fn A function that receives an AtomRecord whenever it comes up in a PDB file, and 
	 * 	returns true if the record has been successfully processed. If fn returns false, the
	 * 	same record will be resent next time operator>> is called.
	 * \return A boost 'connection' object which you can later use to disconnect the handler.
	 */
	boost::signals::connection ConnectRecordSignal ( boost::function<bool (std::string const&)> fn )
	{
		return DefaultRecordSignal.connect(fn);
	}

private:

	/** \brief Fills in an AtomRecord struct with data from the input stream, and calls all the 
	 *  	currently connected handler functions.
	 *
	 * \param is The input stream to read data from
	 * \return True if the handler functions all return true. TODO: check if this is actually the case...
	 */
	bool ParseAtomRecord(std::istream& is);

	///////////////////////////////////////////////////////////////////////////
	//////                         SIGNAL STUFF                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	boost::signal< bool (AtomRecord const&)  > AtomRecordSignal;
	boost::signal< bool (std::string const&) > DefaultRecordSignal;

	///////////////////////////////////////////////////////////////////////////
	//////                        PRIVATE MEMBERS                        //////
	///////////////////////////////////////////////////////////////////////////

	std::string recordName_;	///< the name (PDB ID) of the next record to be parsed

	AtomRecord   atomRecord_;	///< holds the last atom record parsed
	std::string  stringRecord_;	///< holds the last default record parsed

	bool repeatLastRecord_;		///< true if the cached record should be resent
};

////////////////////////////////////////////////////////////////////////////
////				  WRITER				////
////////////////////////////////////////////////////////////////////////////

/** \brief Performs buffered writing in standard PDB format. Currently supports only ATOM records.
 *
 * Example:
 *
 * \code
 *	PDB::Writer writer;
 *	writer.WriteRecord(atomRecord1);
 *	writer.WriteRecord(atomRecord2);
 *
 *	ostream << writer;	// everything in the writer gets sent to the ostream
 * \endcode
 */
class DllExport Writer
{
public:

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief The default constructor does nothing.
	 */
	Writer(): atomRecordIndex_(1), autoSerial_(true)
	{

	}

	////////////////////////////////////////////////////////////////////////////
	////			  PUBLIC MEMBER FNS 				////
	////////////////////////////////////////////////////////////////////////////

	friend DllExport std::ostream& operator<< (std::ostream& os, Writer const& writer);

	/** \brief Turns on automatic serial numbering (on by default).
	 * 
	 * In automatic serial mode, the 'serial' field in AtomRecords is ignored
	 * and atoms are automatically numbered sequentially as they are written out.
	 * If this is off, then you have to specify your own atom serial numbers.
	 *
	 * \param val True to turn on automatic serial numbers.
	 */
	void SetAutoSerial(bool val) 
	{ 
		autoSerial_ = val; 
	}

	/** \brief Checks whether automatic serial numbering is on (default).
	 * 
	 * In automatic serial mode, the 'serial' field in AtomRecords is ignored
	 * and atoms are automatically numbered sequentially as they are written out.
	 * If this is off, then you have to specify your own atom serial numbers.
	 *
	 * \return True if automatic serial numbering is on.
	 */
	bool GetAutoSerial() 
	{ 
		return autoSerial_; 
	}

	/** \brief Writes an ATOM record in PDB format.
	 * This isn't outputted anywhere immediately; call operator<< to flush the output.
	 *
	 * \param record The ATOM record to serialize.
	 */
	void WriteRecord (AtomRecord const& record);	

	/** \brief Writes any record, assuming it's given in proper PDB format.
	 * This isn't outputted anywhere immediately; call operator<< to flush the output.
	 *
	 * \param record The record to serialize... no checking is done to make sure it's valid PDB.
	 */
	void WriteRecord (std::string const& record);

private:

	std::list<std::string> text_;	///< The current contents of the Writer buffer

	uint atomRecordIndex_;		///< For auto serial-numbering
	bool autoSerial_;		///< Auto serial-numbering flag
};

/** \brief WRITEME
 */
DllExport std::istream& operator>> (std::istream& is, Reader& reader);

/** \brief WRITEME
 */
DllExport std::ostream& operator<< (std::ostream& os, Writer const& writer);

}

#endif


