// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_POSES_H
#define IVL_POSES_H

#include "general.h"

#include <vector>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/utility.hpp>

#include "discrete_search_space.h"

/** \file poses.h
 *  \brief A pose library is analogous to a rotamer library
 **/

class DllExport PoseLibrary : public DiscreteSearchSpace
{
public:
	typedef boost::shared_ptr<PoseLibrary>		Pointer;
	typedef boost::shared_ptr<const PoseLibrary> 	ConstPointer;

	//////////////////////////////////////////////////////////////////////////
	/////                          CONSTRUCTORS                          /////
	//////////////////////////////////////////////////////////////////////////

	static Pointer Create(int nBlocks);

	virtual ~PoseLibrary() {}

	PoseLibrary(int nBlocks);

	//////////////////////////////////////////////////////////////////////////
	/////                        PUBLIC INTERFACE                        /////
	//////////////////////////////////////////////////////////////////////////

	std::pair<int, int> GetBlockIDRange(int block) const;

	std::pair<int, int> GeneratePoseIDs(int block, int nPoses);

	//////////////////////////////////////////////////////////////////////////
	/////                 DISCRETE SEARCH SPACE INTERFACE                /////
	//////////////////////////////////////////////////////////////////////////

	virtual int NumDimensions() const;

	virtual int NumElements() const;

	virtual std::pair<int, int> DimensionIDRange(int dimension) const;

private:
	//////////////////////////////////////////////////////////////////////////
	/////                          PRIVATE DATA                          /////
	//////////////////////////////////////////////////////////////////////////

	std::vector< std::pair<int, int> > idRanges_;

	int nBlocks_;
	int nPoses_;
	int idIndex_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(idRanges_);
		ar & BOOST_SERIALIZATION_NVP(nBlocks_);
		ar & BOOST_SERIALIZATION_NVP(nPoses_);
		ar & BOOST_SERIALIZATION_NVP(idIndex_);
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void save_construct_data(
    Archive & ar, const PoseLibrary* t, const unsigned int file_version
){
    // save data required to construct instance
	int nblocks = t->NumDimensions();
	ar << boost::serialization::make_nvp("nblocks", nblocks);
}

template<class Archive>
inline void load_construct_data(
    Archive & ar, PoseLibrary* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    int nblocks;
    ar >> boost::serialization::make_nvp("nblocks", nblocks);
    // invoke inplace constructor to initialize instance
    ::new(t)PoseLibrary(nblocks);
}
}} // namespace ...

#endif


