// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_RESIDUE_H
#define IVL_RESIDUE_H

#include "general.h"

#include <blitz/array.h>
#include <list>
#include <vector>
#include <memory>

#include <boost/weak_ptr.hpp>
#include <boost/serialization/weak_ptr.hpp>
#include <boost/serialization/vector.hpp>

#include "amino_acid.h"
#include "pdb.h"

/** \file residue.h
*
* This file contains the Residue class, which represents an amino acid instantiated
* in a particular location in 3-D space. It handles atom coordinate transforms, 
* chi angles, rebuilding atoms from a template amino acid, etc.
*
* TODO: This file has been undergoing some heavy refactoring, and may still be somewhat
* volatile. There are several parts of the Residue class that still need some cleaning.
*/

class System;

typedef boost::shared_ptr<const System> SystemConstPtr;  // FIXME: is this a SWIG relic? (try compiling without it)

/** \brief This class represents an amino acid instantiated in a particular location 
 * in 3-D space.
 *
 * A Residue is always part of a System, and is associated with a particular AminoAcid.
 * It contains coordinates for all the atoms in the AminoAcid, as well as functions for
 * rotating them about the chi angles, and computing a rough 'centroid'. Along with the
 * atom coordinates, there are 'orientation vectors' used to calculate the strength of
 * hydrogen bonds. Finally, there is some functionality for regenerating atoms that might
 * be missing from a PDB using a 'template' for the amino acid.
 *
 * TODO: should the orientation vectors be moved somewhere else?
 */
class DllExport Residue
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	// FIXME: The default constructor is not suitable for use at this point.
	// Come up with a reasonable construction policy -- i.e. a Residue should
	// always be part of a system, so you shouldn't be able to create one "free"
	
	/** \brief The default constructor does nothing. It probably shouldn't exist at all.
	 * Unless you're absolutely sure you know what you're doing, it's probably not a 
	 * great idea to ever use this constructor, as it makes little sense and will
	 * probably be soon phased out.
	 */
	Residue();
	
	///////////////////////////////////////////////////////////////////////////
	//////                       SIMPLE ACCESSORS                       ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Get the System this Residue is a part of.
	 * \return A pointer to the residue's host System.
	 */
	inline SystemConstPtr GetSystem() const
	{
		return system_.lock();
	}

	/** \brief Get the numeric ID of this Residue.
	 *
	 * An ID is unique within a System, and usually corresponds to the 'residue number'
	 * in a PDB file.
	 *
	 * \return This residue's ID.
	 */
	inline int GetID() const
	{
		return id_;
	}

	/** \brief Sets the numeric ID of this Residue.
	 *
	 * An ID is unique within a System, and usually corresponds to the 'residue number'
	 * in a PDB file.
	 *
	 * FIXME: This function probably shouldn't be used. It was necessary at some point,
	 * but with a more logical construction policy it should be removable.
	 *
	 * \param This residue's new ID.
	 */
	inline void SetID(int id)
	{
		id_ = id;
	}

	/** \brief Returns all of the atom coordinates for this residue.
	 *
	 * NOTE that the size of this vector does not necessarily correspond to the number
	 * of atoms in the Residue. They are all returned essentially as raw data for speed,
	 * as operating on the coordinates will be one of the most common and expensive operations
	 * performed on a Residue.
	 *
	 * ALSO NOTE that *setting* the coordinates should be done with the SetAtomCoords function.
	 * 
	 * Also see AtomIndex()
	 *
	 * \return The Residue's coordinate data
	 */
	inline std::vector<Vector3> const& GetAtomCoords() const
	{
		return atoms_;
	}

	/** \brief Returns this residue's amino acid.
	 *
	 * \return The AminoAcid structure this residue is based on.
	 */
	inline AminoAcid const& GetAminoAcid() const
	{
		return *aminoAcid_;
	}

	///////////////////////////////////////////////////////////////////////////
	//////                       CHECKED ACCESSORS                       //////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Return the 'orientation vector' data, used for calculating hydrogen bonds.
	 *
	 * The non-const version of this method may do some calculations if this Residue's
	 * data was recently changed, but the vectors are thereby guaranteed to be current.
	 *
	 * Also see AtomIndex()
	 *
	 * \return The orientation data
	 */
	std::vector<Vector3> const& GetOrientationVectors();
	
	/** \brief Return the 'orientation vector' data, used for calculating hydrogen bonds.
	 *
	 * The const version of this method throws if this Residue's data was recently changed!
	 * That should never happen, of course, since the Residue is const.
	 *
	 * Also see AtomIndex()
	 *
	 * \return The orientation data
	 */
	std::vector<Vector3> const& GetOrientationVectors() const; // throws if vectors are dirty

	/** \brief Returns a rough centroid approximation for this Residue.
	 *
	 * The centroid is calculated by averaging the positions of a small number of atoms,
	 * depending on the amino acid this Residue is an instance of.
	 *
	 * The non-const version of this method may do some calculations if this Residue's
	 * data was recently changed, but the centroid is thereby guaranteed to be current.
	 *
	 * \return Coordinates that should be fairly close to this Residue's centroid.
	 */
	Vector3 const& GetCentroid();

	/** \brief Returns a rough centroid approximation for this Residue.
	 *
	 * The centroid is calculated by averaging the positions of a small number of atoms,
	 * depending on the amino acid this Residue is an instance of.
	 *
	 * The const version of this method throws if this Residue's data was recently changed!
	 * That should never happen, of course, since the Residue is const.
	 *
	 * \return Coordinates that should be fairly close to this Residue's centroid.
	 */
	Vector3 const& GetCentroid() const;  // throws if centroid is dirty

	/** \brief Returns the backbone dihedral angles (phi, psi).
	 *
	 * The non-const version of this method may do some calculations if this Residue's
	 * data was recently changed, but the angles are thereby guaranteed to be current.
	 *
	 * \return A Vector2 containing (phi, psi).
	 */
	Vector2 GetPhiPsi();

	/** \brief Returns the backbone dihedral angles (phi, psi).
	 *
	 * The const version of this method throws if this Residue's data was recently changed!
	 * That should never happen, of course, since the Residue is const.
	 *
	 * \return A Vector2 containing (phi, psi).
	 */
	Vector2 GetPhiPsi() const; // throws if angles are dirty

	/** \brief Call this if you've moved a neighboring residue and thereby invalidated the phi or
	 *   psi angles for this residue.
	 */
	void UpdatePhiPsi();

	/** \brief Returns the side-chain dihedral angles (chi_n)
	 *
	 * The non-const version of this method may do some calculations if this Residue's
	 * data was recently changed, but the angles are thereby guaranteed to be current.
	 *
	 * \return A Vector4 containing the chi angles. Chi angles that "don't apply" are set to 0.
	 */
	Vector4 const& GetChi();
	
	/** \brief Returns the side-chain dihedral angles (chi_n)
	 *
	 * The const version of this method throws if this Residue's data was recently changed!
	 * That should never happen, of course, since the Residue is const.
	 *
	 * \return A Vector4 containing the chi angles. Chi angles that "don't apply" are set to 0.
	 */
	Vector4 const& GetChi() const; // throws if angles are dirty

	/** \brief Looks up the integer index into the atom coordinate table corresponding to a given atom.
	 *
	 * \param atomName The name of the atom you want (i.e. "CB")
	 * \return The index you should use to get atom coordinates or orientation vectors.
	 */
	int AtomIndex(std::string const& atomName) const;

	/** \brief Returns the backbone dihedral angle omega.
	 * \return The omega dihedral angle.
	 */	
	FLOAT GetOmega() const;
	
	/** \brief Returns the bond angle of that defined by the given atom triplet.
	 * \param angleAtoms  The atom triplet that defines the bond angle.
	 * \return FLOAT  The bond angle of the given atoms.
	 */
	FLOAT GetBondAngle(blitz::TinyVector<int, 3> angleAtoms) const;

	/** \brief Returns the torsion angle of that defined by the given atom quadruplet.
	 * \param torsionAtoms  The atom quadruplet that defines the torsion angle.
	 * \return FLOAT  The torsion angle of the given atoms.
	 */
	FLOAT GetTorsionAngle(blitz::TinyVector<int, 4> torsionAtoms)const ;

	/** \brief Returns the improper torsion angle of that defined by the given atom quadruplet.
	 * \param imptorsAtoms  The atom quadruplet that defines the improper torsion angle.
	 * \return FLOAT  The improper torsion angle of the given atoms.
	 */	
	FLOAT GetImptorsAngle(blitz::TinyVector<int, 4> imptorsAtoms) const;


	///////////////////////////////////////////////////////////////////////////
	//////                     PUBLIC MEMBER FNS                        ///////
	///////////////////////////////////////////////////////////////////////////

	/** FIXME: documentation */
	Residue Clone(SystemConstPtr owner) const;
	
	/** \brief Make a copy of this residue with a new set of chi angles. 
	 *
	 * The cloned Residue will claim to belong to the same System as this one, but the System will
	 * not be aware of it. This is mostly just to make yourself a new Residue for temporary calculations.
	 *
	 * \param chi The chi angles you want for the new residue.
	 * \param owner The system to be this Residue's owner
	 * \return A new residue that's exactly the same as this one except for the chi angles.
	 */
	Residue CloneWithChi(Vector4 const& chi, SystemConstPtr owner) const;
	
	/** \brief Write this residue to the PDBWriter as ATOM records.
	 *
	 * \param writer The PDBWriter to serialize to.
	 */
	void WriteToPDB(PDB::Writer& writer) const;

	/** \brief Replace all side-chain coordinates with those of a 'template' with some arbitrary chi angles.
	 * 
	 * This is useful if you've been fiddling with the backbone coordinates and just want to regenerate the
	 * side-chain atoms without any manual calculations. Note that if you fiddle with the backbone and DON'T
	 * call this, you're responsible for making the atom coordinates make any sort of sense.
	 *
	 * Also see PlaceAtomFromTemplate()
	 */
	void UseTemplateSideChain();

	/** \brief Aligns the backbone of this residue to superimpose it over that of another residue.
	 *
	 * \param target The target residue to superimpose over
	 */
	void AlignBackboneToResidue(Residue const& target);

	/** \brief Put the given chi angle into a unique, standard form that makes sense for this amino acid.
	 *
	 * \param chi The chi angle to standardize
	 * \param chiIndex Which chi angle it is (0 to 4)
	 * \return The same chi angle in standard form
	 */
	FLOAT StandardizeChi(FLOAT chi, int chiIndex) const;

	/** \brief Ensures that all the chi angles for this residue are standardized.
	 *
	 */
	void StandardizeChis();

	/** \brief Sets the chi angles for this residue.
	 *
	 * Also updates the atom coordinates to match the chi angles.
	 *
	 * \param chi The new chi angles for this Residue
	 */
	void SetChi(Vector4 const& chi);

	/** \brief Regenerates a specific atom, if possible, without disturbing the chi angles of this Residue.
	 *
	 * This is useful if side-chain hydrogen atoms are missing from a PDB; all such hydrogens can be 
	 * regenerated this way. Also see UseTemplateSideChain(), AtomIndex()
	 *
	 * \param atomIndex The index of the atom to regenerate
	 * \return True if the atom was successfully placed, false if there was no template data for it
	 */
	bool PlaceAtomFromTemplate(int atomIndex);

	/** \brief Places a OXT atom based on the backbone atoms CA,C, and O.
	 *
	 * \return True if the atom was successfully placed, false otherwise.
	 */
	bool PlaceOXT();

	/** \brief Places a backbone hydrogen based on the surrounding backbone heavy atoms.
	 *
	 * \param prevRes The previous residue in the chain; can be null if, for example,
	 *	this is the first residue or it's on a chain break.
	 * \return True if the atom was successfully placed, false otherwise.
	 */
	bool PlaceHN(Residue const* prevRes);

	/** \brief Changes the amino acid identity of this residue, placing a new sidechain on the
	 *	same backbone as before. Any backbone atoms that weren't previously present are loaded
	 *	from a template.
	 *
	 * \param newAminoAcid The new identity for this residue
	 * \param prevRes The previous residue in the chain; can be null if, for example,
	 *	this is the first residue or it's on a chain break.
	 */
	void Mutate(AminoAcid const* newAminoAcid, Residue const* prevRes);
 
	/** \brief Sets the coordinates of a given atom. 
	 *
	 * Note that orientation vectors, centroids, chi angles, etc. are NOT updated when you do this.
	 * They can be updated manually using UpdateCoords(), or in a non-const Residue they will be 
	 * updated when you access them using a function in the CHECKED ACCESSORS section.
	 *
	 * \param atomIndex The index of the atom to move
	 * \param coords The new coordinates for the atom
	 */
	void SetAtomCoords(int atomIndex, Vector3 const& coords);

	/** \brief Once you've changed atom coordinates using SetAtomCoords, you can recalculate orientation
	 * vectors, centroids, and chi angles all at once using this function.
	 * 
	 * This is useful if you've updated a Residue's coordinates and now want to pass it somewhere as
	 * a const Residue, which will throw if you try to access its 'dirty' data.
	 */
	void UpdateCoords();

	/** \brief True if parent System and ID are the same. No other checks are performed!
	 *
	 * \param rhs The Residue to test for equality with this one
	 * \return True if parent System and ID are the same.
	 */
	bool operator== (Residue const& rhs) const; 

	/** \brief Aligns the backbone atoms of this residue with the given residue, and computes the atom RMSD.
	 *
	 * \param rhs The Residue to compute RMSD with
	 * \return Root Mean Square Deviation (full-atom)
	 */
	FLOAT RMSD(Residue const& rhs) const;

	///////////////////////////////////////////////////////////////////////////
	//////                            FRIENDS                            //////
	///////////////////////////////////////////////////////////////////////////

	friend class PDBResidueFactory;
	friend class System; 		 

private:

	///////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	void ComputeCentroid();
	void ComputeOrientationVectors();
	void ComputePhiPsi();
	void ComputeChi();
	FLOAT ComputeOmega() const;
	FLOAT ComputeBondAngle(blitz::TinyVector<int, 3> angleAtoms) const;
	FLOAT ComputeTorsionAngle(blitz::TinyVector<int, 4> torsionAtoms) const;
	FLOAT ComputeImptorsAngle(blitz::TinyVector<int, 4> imptorsAtoms) const;

	// helper function for AlignBackboneToResidue and UseTemplateSideChain
	void AlignBackbone(
		Vector3 const* src, Vector3* dest,
		AminoAcid const* srcaa, AminoAcid const* destaa);

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	boost::weak_ptr<const System> system_; 
	int id_;				

	AminoAcid const* aminoAcid_;

	std::vector<Vector3> atoms_;
	std::vector<Vector3> hbOrient_;

	Vector3 centroid_;

	Vector4 chi_;
	FLOAT phi_, psi_;

	// below are flags, some of which are set to true when SetAtomCoords is 
	// called and invalidates any cached data
	
	bool phiPsiDirty_;
	bool chiDirty_;
	bool centroidDirty_;
	bool hbOrientDirty_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		AminoAcid* aa = const_cast<AminoAcid*>(aminoAcid_);
		boost::shared_ptr<System> s 
			= boost::const_pointer_cast<System>(system_.lock());

		ar & BOOST_SERIALIZATION_NVP(s);
		ar & BOOST_SERIALIZATION_NVP(id_);
		ar & BOOST_SERIALIZATION_NVP(aa);
		ar & BOOST_SERIALIZATION_NVP(atoms_);
		
		hbOrient_.resize(atoms_.size());
		aminoAcid_ = aa;
		system_ = s;
	}
};

// TODO: move the PDBResidueFactory into its own file, as it clearly wants one

/** \brief This class constructs a single Residue from a PDB file.
 *
 * Example:
 *
 * \code
 * 	PDB::Reader reader;
 * 	System sys;
 * 	std::istream is(file);
 *
 *	while (!is.eof())
 *	{
 *		// ignore all non-ATOM records
 *		std::string nextRecord = reader.NextRecordType(is);
 *		while (nextRecord != "ATOM" && nextRecord != "EOF")
 *		{
 *			is >> reader;
 *			nextRecord = reader.NextRecordType(is);
 *		}
 *
 *		// when we encounter an ATOM record, start parsing into a Residue
 *		if (!is.eof())
 *		{
 *			PDBResidueFactory factory(sys, is, reader);
 *			sys->residues_.push_back( factory.ConstructResidue() );
 *		}
 *	}
 * \endcode
 */
class DllExport PDBResidueFactory
{
public:

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief The constructor sets up the ResidueFactory to produce a single Residue.
	 *
	 * \param system The System the Residue is to belong to
	 * \param is The source input stream
	 * \param reader The PDB reader to extract the Residue from
	 * \param The previous residue in the chain; should be null if this is the
	 *	first residue or it's at a chain break.
	 */
	PDBResidueFactory(
		boost::shared_ptr<const System> system, 
		std::istream& is, 
		PDB::Reader& reader,
		Residue const* prevRes = 0);
	
	/** \brief Releases some resources.
	 */
	~PDBResidueFactory();

	///////////////////////////////////////////////////////////////////////////
	//////                     PUBLIC MEMBER FNS                        ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Creates a residue from the input stream given in the constructor.
	 *
	 * \return A new Residue, constructed from the input PDB, provided that
	 *	Succeeded() returns true.
	 */
	Residue ConstructResidue();		

	/** \brief True if ConstructResidue() has returned a valid Residue.
	 */
	bool Succeeded();

private:

	///////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief The handler method for AtomRecords.
	 *
	 * It implements a state machine that fills in the Residue's atom data.
	 *
	 * \param record Passed in by the boost signal that calls in here
	 * \return True if the ATOM was part of the current Residue, false otherwise.
	 */
	bool HandleAtomRecord( PDB::AtomRecord const& record );
	
	/** \brief A convenience function for producing error messages.
	 *
	 * \return The point in the PDB file where an error occurred.
	 */
	std::string ErrorHeader(); 

	/** \brief Finishes constructing a Residue by filling in data that depends
	 * 	on all the atom coordinates being present.
	 */
	void FinishResidue();

	// TODO: remove the function below once protonated variants are implemented
	/** \brief There are no protonated variants yet, so this allows protonated
	 * variants in the PDB to be converted to the amino acid variants that are
	 * already present in CHOMP.
	 *
	 * \param aa The amino acid to check for variantness
	 * \param atom The atom that might be an extra hydrogen
	 * \return True if this Residue is indeed supposed to be a protonated variant.
	 */
	bool ProtonatedAAHack(AminoAcidIndex aa, std::string const& atom);

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	enum 
	{
		STATE_STARTING,
		STATE_READING_ATOMS,
		STATE_DONE,
		STATE_SKIP
	} state_;
	
	boost::shared_ptr<const System>		system_;
	Residue 				residue_;
	
	int 					atomIndex_;
	int 					atomSerial_;
	AminoAcid const* 			aa_;

	std::vector<bool> 			atomInitialized_;

	boost::signals::connection 		connection_;
	std::istream&  				is_;
	PDB::Reader&   				reader_;

	std::string 				skipResidue_;
	bool					succeeded_;

	Residue const*				prevRes_;
};
#endif


