// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROTAMER_GENERATOR
#define IVL_ROTAMER_GENERATOR

#include "general.h"
#include "residue.h"
#include "system.h"
#include "dunbrack.h"
#include "rotamer_trie.h"
#include "neighbor_map.h"

#include <blitz/array.h>

/** \file rotamer_generator.h
 *
 * \brief Implements a rotamer generation scheme that takes rotamers from the
 * Dunbrack library and adds in extra rotamers depending on amino acid type and
 * burial state.
 */

#define N_EXTRA_CHI_TYPES 	16
#define	N_BURIAL_TYPES		2

///////////////////////////////////////////////////////////////////////////
//////                        DATA STRUCTURES                        //////
///////////////////////////////////////////////////////////////////////////

/** \brief For a given amino acid/burial state combination, these flags let you
 * specify which set of extra rotamers you want to add.
 */
enum ExtraChiType
{
	EXTRA_CHI_1xSD		= 1,	// 1x standard deviation away
	EXTRA_CHI_2xSD		= 2,	// 2x standard deviation away
	EXTRA_CHI_ARO		= 4,	// special extra chis for aromatic residues
	EXTRA_CHI_HALF_ARO	= 8,	// same as above, but fewer
	EXTRA_CHI_CYS		= 16,	// special extra chi for CYS
	EXTRA_CHI_H		= 32,	// hydrogen rotamers
	EXTRA_CHI_CUSTOM	= 64	// you can add more rotamer sets up to 2^32
};

/** \brief A "buried" residue has more than a given cutoff of neighbors, as specified
 * by a NeighborMap.
 */
enum ResidueBurialState
{
	SURFACE			= 0,
	BURIED			= 1
};

/** \brief You can pass this struct to the rotamer generator to change various aspects
 * of how the rotamers are selected.
 */
struct DllExport RotamerGenerationParameters
{
	int   buriedCutoff;		///< # neighbors required to be 'buried'
	int   baseBuriedCutoff;		///< buried cutoff for rotamers straight from the DunLib

	FLOAT maxPercentBuried;		///< maximum % occurrence cutoff for buried residues
	FLOAT maxPercentSurface;	///< maximum % occurrence cutoff for unburied residues
	int   maxRotamersBuried;	///< maximum total # base rotamers for buried residues
	int   maxRotamersSurface;	///< maximum total # base rotamers for unburied residues
	int   minRotamers;		///< minimum # rotamers ever allowed

	FLOAT bumpEnergyCutoff;		///< rotamers with more one-body energy than this are thrown out
	FLOAT extraChiSDCutoff;		///< rotamers with more than this S.D. are thrown out

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(buriedCutoff);
		ar & BOOST_SERIALIZATION_NVP(baseBuriedCutoff);
		ar & BOOST_SERIALIZATION_NVP(maxPercentBuried);
		ar & BOOST_SERIALIZATION_NVP(maxPercentSurface);
		ar & BOOST_SERIALIZATION_NVP(maxRotamersBuried);
		ar & BOOST_SERIALIZATION_NVP(maxRotamersSurface);
		ar & BOOST_SERIALIZATION_NVP(minRotamers);
		ar & BOOST_SERIALIZATION_NVP(bumpEnergyCutoff);
		ar & BOOST_SERIALIZATION_NVP(extraChiSDCutoff);
	}
};

///////////////////////////////////////////////////////////////////////////
//////                       ROTAMER GENERATOR                       //////
///////////////////////////////////////////////////////////////////////////

/** \brief Pass this class to the GenerateRotamers() fns. of a RotamerLibrary to create rotamers
 * using a rotamer selection method based on that of Rosetta.
 *
 * You can customize the generation process by changing the parameters, or the ExtraChi flags.
 * You can set which extra chi sets are used for any amino acid/burial state pair, and you can
 * add your own custom chi sets as well.
 */
class DllExport RotamerGenerator
{
public:

	///////////////////////////////////////////////////////////////////////////
	//////                          CONSTRUCTORS                         //////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Sets up a rotamer generator for a given set of Systems and a Dunbrack library.
	 *
	 * \param nmap The NeighborMap for the Systems you will be generating rotamers for.
	 * \param dunLib The Dunbrack library from which you will be taking rotamers.
	 */
	RotamerGenerator(NeighborMap::ConstPointer nmap, DunbrackLibrary::ConstPointer dunLib);

	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC INTERFACE                       ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Select rotamers according to the current parameter set.
	 *
	 * \param baseResidue The Residue to be rotamerized.
	 * \return The set of rotamers generated for the input Residue.
	 */
	boost::shared_ptr<RotamerSet> GenerateRotamers(Residue const& baseResidue, 
						       RotamerSet const* oldRotamers) const;	

	///////////////////////////////////////////////////////////////////////////
	//////                          PARAMETERS                          ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Apply a given set of parameters to future calls to GenerateRotamers.
	 *
	 * \param params The new parameter set
	 */
	void SetParameters(RotamerGenerationParameters const& params)
	{
		params_ = params;
	}
	
	/** \brief Retrieve the current set of rotamer generation parameters.
	 *
	 * \return The current parameter set
	 */
	inline RotamerGenerationParameters const& GetParameters() const
	{
		return params_;
	}

	/** \brief Customize the selection of extra chi angles in addition to the 'base' rotamers
	 * present in the DunbrackLibrary
	 *
	 * \param aa The amino acid to customize
	 * \param burial The burial state to customize
	 * \param chiAngle The chi angle index (0..4) for which you want to select extra chis
	 * \param flag The extra chi set you want to include
	 */
	inline void SetExtraChiFlag(AminoAcidIndex aa, ResidueBurialState burial, int chiAngle, ExtraChiType flag)
	{
		extraChiFlags_( (int)aa, (int)burial, chiAngle ) |= flag;
	}

	/** \brief Customize the selection of extra chi angles in addition to the 'base' rotamers
	 * present in the DunbrackLibrary
	 *
	 * \param aa The amino acid to customize
	 * \param burial The burial state to customize
	 * \param chiAngle The chi angle index (0..4) for which you want to select extra chis
	 * \param flag The extra chi set you want to exclude
	 */
	inline void ClearExtraChiFlag(AminoAcidIndex aa, ResidueBurialState burial, int chiAngle, ExtraChiType flag)
	{
		extraChiFlags_( (int)aa, (int)burial, chiAngle ) &= ~flag;
	}

	/** \brief Customize the selection of extra chi angles in addition to the 'base' rotamers
	 * present in the DunbrackLibrary
	 *
	 * \param aa The amino acid to customize
	 * \param burial The burial state to customize
	 * \param chiAngle The chi angle index (0..4) for which you want to select extra chis
	 * \return The current extra chi flags for this particular aa/burial/angle index
	 */
	inline int GetExtraChiFlags(AminoAcidIndex aa, ResidueBurialState burial, int chiAngle) const
	{
		return extraChiFlags_( (int)aa, (int)burial, chiAngle );
	}

private:
	///////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	/** \brief Fills in extra chi data with "reasonable" values.
	 */
	void SetupDefaultExtraChiData();
	
	/** \brief Using the extra chi flags, this function combinatorially generates all the rotamers
	 * required to include all the extra chi angles requested.
	 *
	 * \param res The residue to add extra rotamers to
	 * \param baseChi The chi angles of the 'base' rotamer (in the Dunbrack library) to put extra angles around
	 * \param baseSD  The chi S.D.s of the 'base' rotamer
	 * \param rotList A list of rotamers, represented as Residues, to which to add the new rotamers
	 */
	void GetExtraChi(Residue const& res, Vector4 const& baseChi, Vector4 const& baseSD, std::list<Residue>* rotList) const;

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	NeighborMap::ConstPointer nmap_;
	DunbrackLibrary::ConstPointer dunLib_;

	RotamerGenerationParameters params_;

	std::list<FLOAT> 	extraChiData_[N_EXTRA_CHI_TYPES];	
	blitz::Array<uint, 3>  	extraChiFlags_;

	///////////////////////////////////////////////////////////////////////////
	//////                   EXTRA HYDROGEN CHI ANGLES                   //////
	///////////////////////////////////////////////////////////////////////////

	std::list<FLOAT> hydrogenChi1_;
	std::list<FLOAT> extraHydrogenChi1_;
	std::list<FLOAT> hydrogenChi2_;
	std::list<FLOAT> extraHydrogenChi2_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		throw std::string("RotamerGenerator cannot be serialized. Serialize RotamerGenerationParameters instead.");
	}
};

#endif


