// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROTAMER_ITERATION_MAP_H
#define IVL_ROTAMER_ITERATION_MAP_H

#include "general.h"
#include "system_map.h"
#include "rotamers.h"
#include "rotamer_iterators.h"

/** \file rotamer_iteration_map.h
 * 
 * \brief Implements a class that lets you specify which part of a RotamerLibrary you want
 * to include in the generation of a precalculated pairwise energy graph.
 */

/** \brief Pass this data structure to an energy function to include/exclude certain rotamers
 * from the energy graph generation.
 *
 * This is useful when you want to update certain parts of a previously filled energy
 * graph with new rotamers. To do so, you must exclude any rotamers that are already present
 * in the graph, or 'overwriting' old values will result in undefined behavior.
 */
class DllExport RotamerIterationMap
{
public:
	typedef boost::shared_ptr<RotamerIterationMap>		Pointer;
	typedef boost::shared_ptr<const RotamerIterationMap>	ConstPointer;

	////////////////////////////////////////////////////////////////////////////
	////				CONSTRUCTORS 				////
	////////////////////////////////////////////////////////////////////////////

	/** \brief This quasi-constructor returns a shared pointer to a fully initialized RotamerIterationMap
	 *
 	 * \param rlib The RotamerLibrary with which you eventually want to make an energy graph
	 */
	static Pointer Create(RotamerLibrary::ConstPointer rlib);

	/** \brief The constructor does nothing.
	 */
	RotamerIterationMap(RotamerLibrary::ConstPointer rlib)
		: rlib_(rlib), singleData_(rlib->GetSystems(), true), pairData_(rlib->GetSystems(), true)
	{

	}

	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC INTERFACE                       ///////
	///////////////////////////////////////////////////////////////////////////
	
	/** \brief Include all rotamers in the energy graph. Initialize weight to 1.0
	 */
	void EnableAll();

	/** \brief Exclude all rotamers from the energy graph.
	 */
	void DisableAll();

	/** \brief Specify whether *all* rotamer pairs between a given set of systems are included or excluded.
	 *
	 * \param sys1 A system to include/exclude
	 * \param sys2 A system to include/exclude
	 * \param enabled Are these rotamer pairs to be included?
	 */
	void SetSystemPairWeight(System::ConstPointer sys1, System::ConstPointer sys2, FLOAT weight);

	/** \brief Specify the weight of a given rotamer set (i.e. residue) pair.
	 *
	 * \param res1 A rotamer set (residue) to change
	 * \param res2 A rotamer set to change
	 * \param weight What weight to use for that parameter
	 */
	void SetResiduePairWeight(Residue const& res1, Residue const& res2, FLOAT weight);

	/** \brief Specify the weight of single-body energies for rotamers in 'sys'.
	 *
	 * \param sys A system to change
	 * \param weight Weight to use
	 */
	void SetSystemSingleWeight(System::ConstPointer sys, FLOAT weight);
	
	/** \brief Specify the weight of single-body energies for rotamers for this residue.
	 *
	 * \param res A rotamer set (residue) to modify
	 * \param weight 
	 */
	void SetResidueSingleWeight(Residue const& res, FLOAT weight);

	/** \brief Change weight of a given rotamer set (residue) pair
	 *
	 * \param res Residue 
	 * \param weight Weight of residue
	 */
	FLOAT CountPair(Residue const& res1, Residue const& res2) const;

	/** \brief Return weight
	 *
	 * \param res A residue
	 * \return Multiplier for the energy
	 */
	FLOAT CountSingle(Residue const& res) const;

	RotamerLibrary::ConstPointer GetRotamerLibrary() const;

private:

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	RotamerLibrary::ConstPointer rlib_;

	SystemMap<FLOAT> singleData_;
	SystemPairMap<FLOAT> pairData_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(singleData_);
		ar & BOOST_SERIALIZATION_NVP(pairData_);
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void save_construct_data(
    Archive & ar, const RotamerIterationMap* t, const unsigned int file_version
){
    // save data required to construct instance
    	RotamerLibrary::ConstPointer rotlib = t->GetRotamerLibrary();
	ar << boost::serialization::make_nvp("rotlib", rotlib);
}

template<class Archive>
inline void load_construct_data(
    Archive & ar, RotamerIterationMap* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    RotamerLibrary::Pointer rotlib;
    ar >> boost::serialization::make_nvp("rotlib", rotlib);
    // invoke inplace constructor to initialize instance
    ::new(t)RotamerIterationMap(rotlib);
}
}} // namespace ...

#endif


