// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROTAMER_ITERATORS_H
#define IVL_ROTAMER_ITERATORS_H

#include "general.h"
#include "system.h"
#include "rotamer_trie.h"

#include <boost/iterator.hpp>
#include <boost/serialization/vector.hpp>

#include <vector>
#include <map>

/** \file rotamer_iterators.h
 *
 * \brief This file implements some STL-compatible iterators that simplify the process
 * of iterating over the contents of a RotamerLibrary, both over single rotamers and 
 * over rotamer pairs.
 */

///////////////////////////////////////////////////////////////////////////
//////                        DATA STRUCTURES                        //////
///////////////////////////////////////////////////////////////////////////

/** \brief Contains internal RotamerLibrary data. A 'rotamer set' is a set of all
 * the rotamers associated with a single residue. This structure contains all the 
 * rotamer sets for a given system.
 */
struct RotamerSetData
{
	System::ConstPointer 				system;			///< the system these residues belong to
	int 						baseStorageIndex;	///< the storage index of the first residue

	std::vector<boost::shared_ptr<RotamerSet> > 	rotamers;		///< the list of rotamer sets
	std::vector<int>				baseIndices;		///< the list of rotamer IDs associated w/ above

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		System::Pointer s = 
			boost::const_pointer_cast<System>(system);

		ar & BOOST_SERIALIZATION_NVP(s);
		ar & BOOST_SERIALIZATION_NVP(baseStorageIndex);
		ar & BOOST_SERIALIZATION_NVP(rotamers);
		ar & BOOST_SERIALIZATION_NVP(baseIndices);

		system = s;
	}
};

///////////////////////////////////////////////////////////////////////////
//////                     CONVENIENCE TYPEDEFS                     ///////
///////////////////////////////////////////////////////////////////////////

typedef std::vector<boost::shared_ptr<RotamerSet> > 	RSetVector; 
typedef std::map<System const*, RotamerSetData> 	RotamerDataMap;

///////////////////////////////////////////////////////////////////////////
//////                    SINGLE ROTAMER ITERATOR                    //////
///////////////////////////////////////////////////////////////////////////

/** \brief A STL-compatible iterator that lets you iterate over all the rotamers in a RotamerLibrary.
 * 
 * It allows you to use the boost foreach() macro as well.
 *
 * Example:
 * \code
 *	foreach (RotamerSet const& single, rotamerLibrary->SingleResidues())
 *	{
 *		...
 *	}
 * \endcode
 */
class RotamerLibraryIterator: public boost::iterator_facade< RotamerLibraryIterator, RotamerSet const, boost::forward_traversal_tag >
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	RotamerLibraryIterator() {} // don't use this constructor! it's for STL-compatibility only 

	RotamerLibraryIterator(RotamerDataMap::const_iterator sysPos, 
			       RSetVector::const_iterator pos,
			       RotamerDataMap::const_iterator sysEnd)
			: sysPos_(sysPos), sysEnd_(sysEnd), pos_(pos) 
	{ 
	
	} 

private:
	friend class boost::iterator_core_access;

	///////////////////////////////////////////////////////////////////////////
	//////                REQUIRED FOR STL COMPATIBILITY                ///////
	///////////////////////////////////////////////////////////////////////////
	
	bool equal(RotamerLibraryIterator const& rhs) const
	{	
		return sysPos_ == rhs.sysPos_ && pos_ == rhs.pos_;
	}
	
	void increment()
	{
		// traverse the library to the next rotamer
		if (++pos_ == sysPos_->second.rotamers.end())
		{
			if (++sysPos_ != sysEnd_)
				pos_ = sysPos_->second.rotamers.begin();
		}
	}

	RotamerSet const& dereference() const 
	{ 
		return **pos_; 
	}

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	RotamerDataMap::const_iterator sysPos_, sysEnd_;
	RSetVector::const_iterator pos_;
};

///////////////////////////////////////////////////////////////////////////
//////                   PAIRWISE ROTAMER ITERATOR                   //////
///////////////////////////////////////////////////////////////////////////

/** \brief A STL-compatible iterator that lets you iterate over all the 
 * rotamer *pairs* in a RotamerLibrary. Each pair will be counted exactly 
 * once.
 *
 * Note that since this iterator performs an unusual 'flattening' operation,
 * its semantics can be a little strange. When dereferenced, it returns a 
 * reference to an *internal* std::pair object. That means that if you 
 * happen to increment the iterator while holding on to that reference,
 * it will no longer point to where you expect. Basically, you should just
 * copy the std::pair by value just to be completely safe.
 * 
 * This iterator allows you to use the boost foreach() macro as well.
 *
 * Example:
 * \code
 *	std::pair<RotamerSet const*, RotamerSet const*> pair;
 * 	foreach (pair, rotamers->PairResidues())
 *	{
 *		...
 * 	}
 * \endcode
 */
class RotamerLibraryPairIterator: public boost::iterator_facade< RotamerLibraryPairIterator, 
						   std::pair<RotamerSet const*, RotamerSet const*> const, 
						   boost::forward_traversal_tag >
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	RotamerLibraryPairIterator() {} // don't use this constructor! it's for STL-compatibility only 
	
	RotamerLibraryPairIterator(RotamerLibraryIterator pos1,
				   RotamerLibraryIterator pos2, 
				   RotamerLibraryIterator end_sentinel)
		: pos1_(pos1), pos2_(pos2), end_(end_sentinel)
	{
		if (pos1_ != end_ && pos2_ != end_)
		{
			val_ = std::make_pair( &*pos1_, &*pos2_ );
		}
	}

private:
	friend class boost::iterator_core_access;

	///////////////////////////////////////////////////////////////////////////
	//////                REQUIRED FOR STL COMPATIBILITY                ///////
	///////////////////////////////////////////////////////////////////////////

	bool equal(RotamerLibraryPairIterator const& rhs) const
	{	
		return pos1_ == rhs.pos1_ && pos2_ == rhs.pos2_;
	}
	
	void increment()
	{
		if ( ++pos2_ == end_ && ++pos1_ != end_)
		{
			pos2_ = pos1_;

			if (++pos2_ == end_)
				pos1_ = end_;
		}

		if (pos2_ != end_)
		{
			val_ = std::make_pair( &*pos1_, &*pos2_ );
		}
	}

	std::pair<RotamerSet const*, RotamerSet const*> const& dereference() const 
	{
		return val_;
	}

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	RotamerLibraryIterator pos1_, pos2_, end_;
	std::pair<RotamerSet const*, RotamerSet const*> val_;
};

#endif


