// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROTAMER_TRIE
#define IVL_ROTAMER_TRIE

#include "general.h"

#include <blitz/tinyvec.h>
#include <blitz/array.h>

#include "amino_acid.h"
#include "residue.h"
#include "serialization.h"

/** \file rotamer_trie.h
 *
 * \brief Describes a Trie data structure for holding rotamers.
 *
 * This file implements a modified version of the data structure described in
 * Rotamer-Pair Energy Calculations Using a Trie Data Structure (Leaver-Fay, et. al).
 *
 * The idea is that since rotamers are generated from torsional rotations of up to
 * four chi-angles, they share a substantial portion of their atom locations closer
 * to the backbone. So, arranging them in a trie significantly cuts down on energy
 * computation time for atom-atom interaction terms.
*/

/** \brief A node in the Rotamer Trie data structure.
 *
 * This structure contains all the information required in a single atom in a rotamer,
 * along with supplementary data required for navigating the trie structure.
 */
struct RotamerNode
{
	Position  pos;		///< The xyz position of the atom
	Position  hb_orient;    ///< Orientation relative to the atom this atom is bonded to (used for hydrogen bonds)

	BYTE	  depth;	///< The depth in the trie, corresponds to the index of the atom in a simple array
	BYTE	  hv_depth;	///< The depth in a hypothetical trie with all hydrogens excluded
	BYTE	  branch_depth;	///< The number of torsional angles above the atom in question

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(pos);
		ar & BOOST_SERIALIZATION_NVP(hb_orient);
		ar & BOOST_SERIALIZATION_NVP(depth);
		ar & BOOST_SERIALIZATION_NVP(hv_depth);
		ar & BOOST_SERIALIZATION_NVP(branch_depth);
	}
};

/** \brief An array of RotamerNodes representing a trie.
 *
 * TODO: fill this in with updated documentation
 */
class DllExport RotamerSet
{
public:
	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////
	
	RotamerSet(Residue const& baseResidue, std::list<Residue> const& rotList);

	// create a one-rotamer rotamer set
	RotamerSet(Residue const& residue);
	
	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC MEMBER FNS                       //////
	///////////////////////////////////////////////////////////////////////////

	Residue GetRotamer(int index) const;
	void PrintRotamers() const;

	///////////////////////////////////////////////////////////////////////////
	//////                           ACCESSORS                           //////
	///////////////////////////////////////////////////////////////////////////

	inline blitz::Array<RotamerNode, 1> const& GetTrieNodes() const
	{
		return nodes_;
	}

	std::vector<Vector4> GetChi() const
	{
		std::vector<Vector4> result;
		for (int i = 0; i < chi_.extent(0); ++i)
			result.push_back(chi_(i));

		return result;
	}

	inline Vector4 const& GetChi(int rotIndex) const
	{
		return chi_(rotIndex);
	}

	inline Vector3 const& GetCentroid(int rotIndex) const
	{
		return centroid_(rotIndex);
	}

	inline Residue const& GetBaseResidue() const
	{
		return baseResidue_;
	}
	
	inline int NumRotamers() const
	{
		return nRotamers_;
	}

private:
	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	blitz::Array<RotamerNode, 1> nodes_;

	blitz::Array<Vector4, 1> chi_;
	blitz::Array<Vector3, 1> centroid_;

	Residue const& baseResidue_;

	int nRotamers_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(nodes_);
		ar & BOOST_SERIALIZATION_NVP(chi_);
		ar & BOOST_SERIALIZATION_NVP(centroid_);
		ar & BOOST_SERIALIZATION_NVP(nRotamers_);
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void save_construct_data(
    Archive & ar, const RotamerSet* t, const unsigned int file_version
){
    // save data required to construct instance
	Residue const* baseRes = &t->GetBaseResidue();
	ar << boost::serialization::make_nvp("baseResidue", baseRes);
}

template<class Archive>
inline void load_construct_data(
    Archive & ar, RotamerSet* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    Residue *res;
    ar >> boost::serialization::make_nvp("baseResidue", res);
    // invoke inplace constructor to initialize instance
    ::new(t)RotamerSet(*res);
}
}} // namespace ...

#endif


