// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_ROTAMERS_H
#define IVL_ROTAMERS_H

#include "general.h"

#include <list>
#include <map>
#include <vector>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>

#include <blitz/array.h>

#include "dunbrack.h"
#include "residue.h"
#include "rotamer_trie.h"
#include "system.h"
#include "neighbor_map.h"
#include "rotamer_iterators.h"
#include "rotamer_generator.h"
#include "discrete_search_space.h"

typedef std::vector< std::pair<int, int> > RotamerIDSet;

/** \file rotamers.h
 *  \brief Before application of a rotamer generator the rotamer library has only one rotamer per residue
 **/

class DllExport RotamerLibrary : public DiscreteSearchSpace
{
public:
	typedef boost::shared_ptr<RotamerLibrary>		Pointer;
	typedef boost::shared_ptr<const RotamerLibrary> 	ConstPointer;

	///////////////////////////////////////////////////////////////////////////
	//////                         CONSTRUCTORS                         ///////
	///////////////////////////////////////////////////////////////////////////

	static Pointer Create(System::ConstPointer system);
	static Pointer Create(std::vector<System::ConstPointer> const& systems); 

	virtual ~RotamerLibrary() {}

	RotamerLibrary(std::vector<System::ConstPointer> const& systems);

	///////////////////////////////////////////////////////////////////////////
	//////                       PUBLIC MEMBER FNS                       //////
	///////////////////////////////////////////////////////////////////////////

	void GenerateRotamers(RotamerGenerator const& generator, bool discardOld = false);
	RotamerIDSet GenerateRotamers(SystemConstPointer system, RotamerGenerator const& generator, bool discardOld = false);
	RotamerIDSet GenerateRotamers(Residue const& res, RotamerGenerator const& generator, bool discardOld = false);

	RotamerIDSet AddRotamers(Residue const& res, std::vector<Vector4> const& chi);
	RotamerIDSet AddRotamers(RotamerLibrary::ConstPointer rlib, std::vector<int> const& ids);
	
	RotamerIDSet DeleteRotamers(Residue const& res, std::vector<int> const& ids);
	RotamerIDSet DeleteRotamers(std::vector<int> const& ids);
	
	RotamerIDSet ReplaceSystem(System::ConstPointer oldSystem, System::ConstPointer newSystem);

	Residue 		GetRotamerResidue(int rotamerID) const;

	std::pair<int, int> 	GetRotamerIDRange(Residue const& res) const;
	
	int 			GetRotamerID(RotamerSet const& rset, int rotIndex) const;
	
	int 			GetStorageIndex(RotamerSet const& rset) const;

	std::vector<Vector4>  	GetRotamerChis(Residue const& res) const;

	Vector4 		GetClosestChi(Residue const& res) const;

	System::Pointer 	CreateSystem(System::ConstPointer baseSystem, std::vector<int> const& rotamerIDs) const;

	std::vector<int>        TranslateRotamerIDs(RotamerLibrary::ConstPointer sourceLib, std::vector<int> const& rotamerIDs) const;

	RotamerLibrary::Pointer Clone() const;
	
	///////////////////////////////////////////////////////////////////////////
	//////                           ACCESSORS                           //////
	///////////////////////////////////////////////////////////////////////////

	inline int NumRotamers() const
	{
		return nTotalRotamers_;
	}
	
	inline int NumRotamerSets() const
	{
		return nRotamerSets_;
	}

	inline std::vector<System::ConstPointer> const& GetSystems() const
	{
		return systems_;
	}

	///////////////////////////////////////////////////////////////////////////
	//////                           ITERATION                           //////
	///////////////////////////////////////////////////////////////////////////

	std::pair<RotamerLibraryIterator, RotamerLibraryIterator> SingleResidues() const;
	std::pair<RotamerLibraryPairIterator, RotamerLibraryPairIterator> PairResidues() const;

	//////////////////////////////////////////////////////////////////////////
	/////                 DISCRETE SEARCH SPACE INTERFACE                /////
	//////////////////////////////////////////////////////////////////////////

	virtual int NumDimensions() const;

	virtual int NumElements() const;

	virtual std::pair<int, int> DimensionIDRange(int dimension) const;

private:
	///////////////////////////////////////////////////////////////////////////
	//////                      PRIVATE MEMBER FNS                      ///////
	///////////////////////////////////////////////////////////////////////////

	void GenerateRotamersWithoutInvalidation(Residue const& res, 
						 RotamerGenerator const& generator, 
						 bool discardOld);

	RotamerIDSet InvalidateRotamers(System::ConstPointer oldSystem);
	RotamerIDSet InvalidateRotamers(Residue const& res);
	
	RotamerSetData& 	GetRotamerSetData(Residue const& res);
	RotamerSetData const& 	GetRotamerSetData(Residue const& res) const;

	RotamerSetData const& FindRotamer(int rotamerID, int* indexOut) const;

	void GenerateRSetStorageIndices();

	std::map<int, Vector4>  GetChiMap(Residue const& res) const;

	///////////////////////////////////////////////////////////////////////////
	//////                         PRIVATE DATA                         ///////
	///////////////////////////////////////////////////////////////////////////

	std::map< System const*, RotamerSetData > 	rotamerData_;
	std::vector< System::ConstPointer > 		systems_;

	int nRotamerSets_;
	int nTotalRotamers_;
	int rotamerIndex_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		typedef std::map<System const*, RotamerSetData> ConstMap;
		typedef std::map<System*, RotamerSetData*> NonConstMap;

		NonConstMap rdc;
		for (typename ConstMap::iterator i = rotamerData_.begin();
			i != rotamerData_.end(); ++i)
			rdc[ const_cast<System*>(i->first) ] = &i->second;

		ar & BOOST_SERIALIZATION_NVP(rdc);
		ar & BOOST_SERIALIZATION_NVP(nRotamerSets_);
		ar & BOOST_SERIALIZATION_NVP(nTotalRotamers_);
		ar & BOOST_SERIALIZATION_NVP(rotamerIndex_);

		//rotamerData_.clear();
		for (typename NonConstMap::iterator i = rdc.begin();
			i != rdc.end(); ++i)
			rotamerData_[i->first] = *i->second;
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void save_construct_data(
    Archive & ar, const RotamerLibrary* t, const unsigned int file_version
){
    // save data required to construct instance
	ar << boost::serialization::make_nvp("systems", t->GetSystems());
}

template<class Archive>
inline void load_construct_data(
    Archive & ar, RotamerLibrary* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    std::vector<System::Pointer> systems;
    ar >> boost::serialization::make_nvp("systems", systems);

    std::vector<System::ConstPointer> csystems;
    for (size_t i = 0; i < systems.size(); ++i)
	    csystems.push_back( const_pointer_cast<System const>(systems[i]) );

    // invoke inplace constructor to initialize instance
    ::new(t)RotamerLibrary(csystems);
}
}} // namespace ...

#endif


