// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_SERIALIZATION_H
#define IVL_SERIALIZATION_H

#include "general.h"
#include <climits>

//////////////////////////////////////////////////////////////////////////
/////                          SERIALIZATION                         /////
//////////////////////////////////////////////////////////////////////////

// these functions are to make it easy to use boost's Serialization library
// from the Python interface, which doesn't have direct access to it.

#include <fstream>
#include <boost/archive/xml_iarchive.hpp>
#include <boost/archive/xml_oarchive.hpp>
#include <boost/archive/binary_iarchive.hpp>
#include <boost/archive/binary_oarchive.hpp>

#include <boost/serialization/base_object.hpp>
#include <boost/serialization/binary_object.hpp>

//////////////////////////////////////////////////////////////////////////
/////                      GENERIC ARCHIVES                          /////
//////////////////////////////////////////////////////////////////////////

template <class Archive>
class DllExport GenericOutputArchive
{
public:
	template <class T>
	void Save(std::string const& elementName, T const& object)
	{
		using namespace boost::serialization;
		(*ar) << make_nvp(elementName.c_str(), object);
	}

protected:

	Archive* ar;
};

template <class Archive>
class DllExport GenericInputArchive
{
public:
	template <class T>
	void Load(T& object)
	{
		using namespace boost::serialization;
		(*ar) >> make_nvp("", object);
	}

protected:

	Archive* ar;
};

//////////////////////////////////////////////////////////////////////////
/////                        FILE ARCHIVES                           /////
//////////////////////////////////////////////////////////////////////////

template <class Archive>
class DllExport OutputArchive : public GenericOutputArchive<Archive>
{
public:
	explicit OutputArchive(std::string const& filename) : ofs(filename.c_str(), std::ios::binary)
	{
		this->ar = new Archive(ofs);
	}

	~OutputArchive() { delete this->ar; }   // NOTE: can't put this in base class due to 
						// order of calls to destructors...
	
	void Flush()
	{
		ofs.flush();
	}

	unsigned int Tell()
	{
		return (unsigned int)ofs.tellp();
	}

	void Seek(unsigned int pos)
	{
		ofs.seekp(pos);
	}

private:
	std::ofstream ofs;
};

template <class Archive>
class DllExport InputArchive : public GenericInputArchive<Archive>
{
public:
	explicit InputArchive(std::string const& filename) : ifs(filename.c_str(), std::ios::binary)
	{
		this->ar = new Archive(ifs);
	}

	~InputArchive() { delete this->ar; }

	template <class T>
	void Load(T& object)
	{	
		ifs.peek();
		if (!ifs.good())
			throw std::string("EOF");
		else
			GenericInputArchive<Archive>::Load(object);
	}

	unsigned int Tell()
	{
		return (unsigned int)ifs.tellg();
	}

	void Seek(unsigned int pos)
	{
		ifs.seekg(pos);
	}

private:
	std::ifstream ifs;
};

//////////////////////////////////////////////////////////////////////////
/////                        STRING ARCHIVES                         /////
//////////////////////////////////////////////////////////////////////////

template <class Archive>
class DllExport StringOutputArchive : public GenericOutputArchive<Archive>
{
public:
	StringOutputArchive() : oss(std::stringstream::in | std::stringstream::out |
					std::stringstream::binary )
	{
		this->ar = new Archive(oss);
		hdr_pos = (size_t)oss.tellp();
	}

	~StringOutputArchive() { delete this->ar; }

	std::string GetString(bool omit_header = false)
	{
		if (omit_header)
			return oss.str().substr(hdr_pos);
		else
			return oss.str();
	}

	std::string GetSubString(unsigned int start, unsigned int len)
	{
		char* buf = new char[len];
		std::string result;

		oss.seekg(start);
		oss.readsome(buf, len);
		result.assign(buf, len);

		delete[] buf;
		return result;
	}

	unsigned int Tell()
	{
		return (unsigned int)oss.tellp();
	}

	void Seek(unsigned int pos)
	{
		oss.seekp(pos);
	}

private:

	std::stringstream oss;
	size_t hdr_pos;
};

template <class Archive>
class DllExport StringInputArchive : public GenericInputArchive<Archive>
{
public:
	explicit StringInputArchive(std::string str_) : str(str_), iss(str, std::ios::binary)
	{
		this->ar = new Archive(iss);
	}

	~StringInputArchive() { delete this->ar; }

	template <class T>
	void Load(T& object)
	{
		iss.peek();
		if (!iss.good())
			throw std::string("EOF");
		else
			GenericInputArchive<Archive>::Load(object);
	}

	unsigned int Tell()
	{
		return (unsigned int)iss.tellg();
	}

	void Seek(unsigned int pos)
	{
		iss.seekg(pos);
	}

private:
	std::string str;
	std::istringstream iss;
};

//////////////////////////////////////////////////////////////////////////
/////                       PYTHON INTERFACE                         /////
//////////////////////////////////////////////////////////////////////////

typedef GenericOutputArchive<boost::archive::xml_oarchive> GenericXMLOutputArchive;
typedef GenericInputArchive<boost::archive::xml_iarchive> GenericXMLInputArchive;
typedef GenericOutputArchive<boost::archive::binary_oarchive> GenericBinaryOutputArchive;
typedef GenericInputArchive<boost::archive::binary_iarchive> GenericBinaryInputArchive;

typedef OutputArchive<boost::archive::xml_oarchive> XMLOutputArchive;
typedef InputArchive<boost::archive::xml_iarchive> XMLInputArchive;
typedef OutputArchive<boost::archive::binary_oarchive> BinaryOutputArchive;
typedef InputArchive<boost::archive::binary_iarchive> BinaryInputArchive;

typedef StringOutputArchive<boost::archive::xml_oarchive> XMLStringOutputArchive;
typedef StringInputArchive<boost::archive::xml_iarchive> XMLStringInputArchive;
typedef StringOutputArchive<boost::archive::binary_oarchive> BinaryStringOutputArchive;
typedef StringInputArchive<boost::archive::binary_iarchive> BinaryStringInputArchive;

// Py++ doesn't seem to recognize templated member functions, so..
template <class T, class Archive>
void SaveToArchive(Archive& a, std::string const& elementName, T const& object)
{
	a.Save(elementName, object);
}

template <class T, class Archive>
boost::shared_ptr<T> LoadFromArchive(Archive& a)
{
	boost::shared_ptr<T> ptr(new T());
	a.Load(*ptr);
	return ptr;
}

template <class T, class Archive>
T LoadBasicFromArchive(Archive& a)
{
	T object;
	a.Load(object);
	return object;
}

// Manually instantiate a string version, since it's a little different
// we do a bitwise rotation on each character to obfuscate the serialized data a bit

template <typename T>
inline T rotate_left(T val, unsigned char shift=1)
{
    static const int bits = sizeof(T) * CHAR_BIT;
    return (val >> (bits-shift)) | (val << shift);
}

template <typename T>
inline T rotate_right(T val, unsigned char shift=1)
{
    static const int bits = sizeof(T) * CHAR_BIT;
    return (val << (bits-shift)) | (val >> shift);
}

template <class Archive>
void SaveToArchive(Archive& a, std::string const& elementName, std::string const& object)
{
	std::string str = object;
	for (int i = 0; i < (int)str.size(); ++i)
		str[i] = rotate_left((unsigned char)str[i]);

	a.Save(elementName, str);
}

template <class Archive>
std::string LoadStringFromArchive(Archive& a)
{
	std::string str;
	a.Load(str);

	for (int i = 0; i < (int)str.size(); ++i)
		str[i] = rotate_right((unsigned char)str[i]);

	return str;
}

//////////////////////////////////////////////////////////////////////////
/////                        BLITZ SERIALIZATION                     /////
//////////////////////////////////////////////////////////////////////////

#ifndef NO_BLITZ_SERIALIZATION

namespace boost { namespace serialization { 

template <class Archive, class DataType, int N>
void serialize(Archive& ar, blitz::TinyVector<DataType, N>& v, const unsigned int version)
{
	boost::serialization::binary_object binObj(
		v.data(), sizeof(DataType) * N);
	ar & boost::serialization::make_nvp("TinyVector", binObj);
}

template <class Archive, class DataType, int N>
void serialize(Archive& ar, blitz::Array<DataType, N>& a, const unsigned int version)
{
	boost::serialization::binary_object binObj(
		a.data(), sizeof(DataType) * a.size());
	ar & boost::serialization::make_nvp("Array", binObj);
}

} }

#endif

#endif


