// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_SYSTEM_H
#define IVL_SYSTEM_H

#include "general.h"
#include "residue.h"

#include <iostream>
#include <vector>
#include <string>
#include <boost/serialization/vector.hpp>


typedef boost::shared_ptr<System>		SystemPointer;
typedef boost::shared_ptr<const System>		SystemConstPointer;

/** \file system.h
*
* \brief Contains the System class, which is a container for Residues.
*
* This file implements a container for Residues. The container provides
* a boost::range-compatible interface for traversing its contents, as well 
* as the ability to load its data from a PDB and write it back.
*
*/

class DllExport System
{
public:
	// TODO: in the documentation, make sure to put in a warning that using ANY
	// non-const function (including non-const iteration) in this class will 
	// potentially invalidate data in classes that depend on it 
	// (such as NeighborMap and RotamerLibrary).
	//
	// As a rule, it's best to use the non-const functions only in an initialization
	// stage to assemble the System, and then refer to the System ONLY through
	// ConstPointers from then on. 
	
	typedef std::vector<Residue> 		ResidueList; 
	typedef boost::shared_ptr<System>	Pointer;
	typedef boost::shared_ptr<const System> ConstPointer;

	////////////////////////////////////////////////////////////////////////////
	////				ITERATOR TYPES				////
	////////////////////////////////////////////////////////////////////////////

	typedef ResidueList::iterator		iterator;
	typedef ResidueList::const_iterator 	const_iterator;
	typedef ResidueList::size_type 		size_type;

	////////////////////////////////////////////////////////////////////////////
	////				CONSTRUCTORS 				////
	////////////////////////////////////////////////////////////////////////////

	static Pointer Create(AtomTypes::ConstPointer atypes, AminoAcidDirectory::ConstPointer aadir);
	static Pointer Join(std::vector<System::ConstPointer> const& systems);

	System(AtomTypes::ConstPointer atypes, AminoAcidDirectory::ConstPointer aadir);

	////////////////////////////////////////////////////////////////////////////
	////				ITERATOR ACCESSORS			////
	////////////////////////////////////////////////////////////////////////////

	iterator 	begin()	{ return residues_.begin(); }
	const_iterator	begin() const { return residues_.begin(); } 

	iterator 	end() { return residues_.end(); }
	const_iterator	end() const { return residues_.end(); }

	size_type	size() const { return residues_.size(); }

	////////////////////////////////////////////////////////////////////////////
	////				PUBLIC MEMBER FNS			////
	////////////////////////////////////////////////////////////////////////////

	Residue& 	GetResidue(int index);
	Residue const& 	GetResidue(int index) const;
	
	Residue& 	GetResidueByStorageIndex(int index);
	Residue const& 	GetResidueByStorageIndex(int index) const;
	
	int 		GetStorageIndex(Residue const& res) const;

	AminoAcidDirectory::ConstPointer GetAminoAcidData() const;
	AtomTypes::ConstPointer	GetAtomTypes() const;

	bool IsFirstInChain(Residue const& res) const;
	bool IsLastInChain(Residue const& res) const;

	System::Pointer Clone() const;

	std::vector<System::Pointer> Split(std::vector<int> const& splitPoints) const;	

	Residue& NextResidue(Residue const& res);
	Residue& PrevResidue(Residue const& res);

	Residue const& NextResidue(Residue const& res) const;
	Residue const& PrevResidue(Residue const& res) const;

	void Append(Residue const& residue);
	void Clear();

	void MutateResidue(int id, AminoAcid const* newAminoAcid);

	friend DllExport std::ostream& operator<< (std::ostream& os, System::ConstPointer sys);
	friend DllExport std::istream& operator>> (std::istream& is, System::Pointer& sys);

private:
	int firstIndex_;
	ResidueList residues_;

	AtomTypes::ConstPointer atypes_;
	AminoAcidDirectory::ConstPointer aadir_;

	System::Pointer nextSys_;
	System::Pointer prevSys_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		ar & BOOST_SERIALIZATION_NVP(firstIndex_);
		ar & BOOST_SERIALIZATION_NVP(residues_);
		ar & BOOST_SERIALIZATION_NVP(nextSys_);
		ar & BOOST_SERIALIZATION_NVP(prevSys_);

		foreach (Residue& res, residues_)
			res.UpdateCoords();
	}
};

namespace boost { namespace serialization {
template<class Archive>
inline void save_construct_data(
    Archive & ar, const System* t, const unsigned int file_version
){
    // save data required to construct instance
    	AtomTypes::ConstPointer atypes = t->GetAtomTypes();
	AminoAcidDirectory::ConstPointer aadir = t->GetAminoAcidData();

	ar << boost::serialization::make_nvp("atypes", atypes);
	ar << boost::serialization::make_nvp("aadata", aadir);
}

template<class Archive>
inline void load_construct_data(
    Archive & ar, System* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    AtomTypes::Pointer atypes;
    AminoAcidDirectory::Pointer aadir;
    
    ar >> boost::serialization::make_nvp("atypes", atypes);
    ar >> boost::serialization::make_nvp("aadata", aadir);
    // invoke inplace constructor to initialize instance
    ::new(t)System(atypes, aadir);
}
}} // namespace ...

DllExport std::ostream& operator<< (std::ostream& os, System::ConstPointer sys);
DllExport std::ostream& operator<< (std::ostream& os, System::Pointer sys);
DllExport std::istream& operator>> (std::istream& is, System::Pointer& sys);


//////////////////////////////////////////////////////////////////////////
/////                             ALIASES                            /////
//////////////////////////////////////////////////////////////////////////

namespace pyplusplus
{
namespace aliases
{
	typedef std::vector<System::Pointer>				SystemVector;
	typedef std::vector<System::ConstPointer>			ConstSystemVector;
}}

#endif


