// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_SYSTEM_MAP_H
#define IVL_SYSTEM_MAP_H

#include "general.h"

#include <map>
#include <boost/serialization/vector.hpp>
#include <boost/serialization/map.hpp>

#include "system.h"

// GCCXML CVS head as of August 2007 cannot handle boost::FOR_EACH in the headers it parses
// so I've turned the iteration constructs here back to their normal form

template <typename DataType>
class SystemMap
{
public:
	typedef std::map< System const*, blitz::Array<DataType, 1>* > MapType;

	SystemMap(std::vector<System::ConstPointer> const& systems, DataType const& initVal)
		:systems_(systems)
	{
		std::vector<System::ConstPointer>::const_iterator i;
		for (i = systems.begin(); i != systems.end(); ++i)
		{
			blitz::Array<DataType, 1>& data = 
				*data_.insert( std::make_pair( i->get(), new blitz::Array<DataType, 1>() ) ).first->second;
				
			data.resize((int)(*i)->size());
			data = initVal;
		}
	}

	~SystemMap()
	{
		for (typename MapType::const_iterator i = data_.begin(); i != data_.end(); ++i)
			if (i->second) delete i->second;
	}

	blitz::Array<DataType, 1> const& operator()(System::ConstPointer sys) const
	{
		return *(Find(sys)->second);
	}

	DataType const& operator()(System::ConstPointer sys, int index) const
	{
		return (*(Find(sys)->second))(index);
	}

	blitz::Array<DataType, 1>& operator()(System::ConstPointer sys)
	{
		return const_cast< blitz::Array<DataType, 1>& >( 
		     (*const_cast< SystemMap<DataType> const* >(this))(sys) );
	}

	DataType& operator()(System::ConstPointer sys, int index)
	{
		return const_cast< DataType& >( 
		     (*const_cast< SystemMap<DataType> const* >(this))(sys, index) );
	}

	template<class T> friend std::ostream& operator<< (std::ostream& os, SystemMap<T> const& map);
	template<class T> friend std::istream& operator>> (std::istream& is, SystemMap<T>& map);

private:

	typename MapType::const_iterator Find(System::ConstPointer sys) const
	{
		typename MapType::const_iterator i = data_.find(sys.get());
		if (i == data_.end())
			throw std::string("SystemMap: System not found.");

		return i; 
	}

	MapType data_;
	std::vector<System::ConstPointer> systems_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		typedef std::map< System*, blitz::Array<DataType, 1>* > 
			NonConstMapType;

		NonConstMapType datac;
		for (typename MapType::iterator i = data_.begin();
			i != data_.end(); ++i)
			datac[ const_cast<System*>(i->first) ] = i->second;

		ar & BOOST_SERIALIZATION_NVP(datac);
	
		data_.clear();
		for (typename NonConstMapType::iterator i = datac.begin();
			i != datac.end(); ++i)
			data_[i->first] = i->second;
	}
};

namespace boost { namespace serialization {
template<class Archive, class DataType>
inline void save_construct_data(
    Archive & ar, const SystemMap<DataType>* t, const unsigned int file_version
){
    // save data required to construct instance
	ar << boost::serialization::make_nvp("systems", t->systems_);
}

template<class Archive, class DataType>
inline void load_construct_data(
    Archive & ar, SystemMap<DataType>* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    std::vector<System::Pointer> systems;
    ar >> boost::serialization::make_nvp("systems", systems);

    std::vector<System::ConstPointer> csystems;
    for (size_t i = 0; i < systems.size(); ++i)
	    csystems.push_back( const_pointer_cast<System const>(systems[i]) );

    // invoke inplace constructor to initialize instance of my_class
    ::new(t)SystemMap<DataType>(csystems, DataType());
}
}} // namespace ...

template <typename DataType>
class SystemPairMap
{
public:
	typedef std::map< std::pair<System const*, System const*>,
		          blitz::Array<DataType, 2>* > MapType;
	
	SystemPairMap(std::vector<System::ConstPointer> const& systems, DataType const& initVal)
			: systems_(systems)
	{
		std::vector<System::ConstPointer>::const_iterator sys1, sys2;

		for (sys1 = systems.begin(); sys1 != systems.end(); ++sys1)
			for (sys2 = sys1; sys2 != systems.end(); ++sys2)
			{
				blitz::Array<DataType, 2>& data = 
					*data_.insert( std::make_pair( std::make_pair( sys1->get(), sys2->get() ),
								      new blitz::Array<DataType, 2>() )
						     ).first->second;
			
				data.resize((int)(*sys1)->size(), (int)(*sys2)->size());
				data = initVal;
			}
	}

	~SystemPairMap()
	{
		for (typename MapType::const_iterator i = data_.begin(); i != data_.end(); ++i)
			if (i->second) delete i->second;
	}

	blitz::Array<DataType, 2> const& operator()(System::ConstPointer sys1, System::ConstPointer sys2) const
	{
		return *(Find(sys1, sys2)->second);
	}

	DataType const& operator()(System::ConstPointer sys1, System::ConstPointer sys2, int index1, int index2) const
	{
		typename MapType::const_iterator i = Find(sys1, sys2);
	
		if (sys1.get() == i->first.first)
			return (*i->second)(index1, index2);
		else
			return (*i->second)(index2, index1);
	}

	blitz::Array<DataType, 2>& operator()(System::ConstPointer sys1, System::ConstPointer sys2)
	{
		return const_cast< blitz::Array<DataType, 2>& >(
		     (*const_cast< SystemPairMap<DataType> const* >(this))(sys1, sys2) );
	}

	DataType& operator()(System::ConstPointer sys1, System::ConstPointer sys2, int index1, int index2)
	{
		return const_cast< DataType& >( 
		     (*const_cast< SystemPairMap<DataType> const* >(this))(sys1, sys2, index1, index2) );
	}

	template <class T> friend std::ostream& operator<< (std::ostream& os, SystemPairMap<T> const& map);
	template <class T> friend std::istream& operator>> (std::istream& is, SystemPairMap<T>& map);

private:

	typename MapType::const_iterator Find(System::ConstPointer sys1, System::ConstPointer sys2) const
	{
		typename MapType::const_iterator i = data_.find( std::make_pair(sys1.get(), sys2.get()) );

		if (i == data_.end())
			i = data_.find( std::make_pair(sys2.get(), sys1.get()) );
		if (i == data_.end())
			throw std::string("SystemPairMap: System not found.");

		return i;
	}

	MapType data_;
	std::vector<System::ConstPointer> systems_;

	///////////////////////////////////////////////////////////////////////////
	//////                         SERIALIZATION                        ///////
	///////////////////////////////////////////////////////////////////////////

	friend class boost::serialization::access;
	
	template <class Archive>
	void serialize (Archive& ar, const unsigned int version)
	{
		typedef std::map< std::pair<System*, System*>,
		          blitz::Array<DataType, 2>* > NonConstMapType;

		NonConstMapType datac;
		for (typename MapType::iterator i = data_.begin();
			i != data_.end(); ++i)
			datac[ std::make_pair( 
				const_cast<System*>(i->first.first),
				const_cast<System*>(i->first.second))
			] = i->second;

		ar & BOOST_SERIALIZATION_NVP(datac);
	
		data_.clear();
		for (typename NonConstMapType::iterator i = datac.begin();
			i != datac.end(); ++i)
			data_[i->first] = i->second;
	}
};

template <class T>
std::ostream& operator<< (std::ostream& os, SystemPairMap<T> const& map)
{
	typename SystemPairMap<T>::MapType::const_iterator entry;

	//typename std::map< std::pair<System::ConstPointer, System::ConstPointer>,
	//       	   blitz::Array<T, 2>* >::const_iterator entry;

	for (entry = map.data_.begin(); entry != map.data_.end(); ++entry)
		os << *entry->second << std::endl;

	return os;
}

namespace boost { namespace serialization {
template<class Archive, class DataType>
inline void save_construct_data(
    Archive & ar, const SystemPairMap<DataType>* t, const unsigned int file_version
){
    // save data required to construct instance
	ar << boost::serialization::make_nvp("systems", t->systems_);
}

template<class Archive, class DataType>
inline void load_construct_data(
    Archive & ar, SystemPairMap<DataType>* t, const unsigned int file_version
){
    // retrieve data from archive required to construct new instance
    std::vector<System::Pointer> systems;
    ar >> boost::serialization::make_nvp("systems", systems);

    std::vector<System::ConstPointer> csystems;
    for (size_t i = 0; i < systems.size(); ++i)
	    csystems.push_back( const_pointer_cast<System const>(systems[i]) );

    // invoke inplace constructor to initialize instance of my_class
    ::new(t)SystemPairMap<DataType>(csystems, DataType());
}
}} // namespace ...

///////////////////////////////////////////////////////////////////////////
//////                             I/O                              ///////
///////////////////////////////////////////////////////////////////////////

template <class T>
std::istream& operator>> (std::istream& is, SystemPairMap<T>& map)
{
	typename SystemPairMap<T>::MapType::const_iterator entry;
	
	//typename std::map< std::pair<System::ConstPointer, System::ConstPointer>,
	//        	   blitz::Array<T, 2>* >::iterator entry;

	for (entry = map.data_.begin(); entry != map.data_.end(); ++entry)
		is >> *entry->second;

	return is;
}

template <class T>
std::ostream& operator<< (std::ostream& os, SystemMap<T> const& map)
{
	typename SystemMap<T>::MapType::const_iterator entry;
	
	//typename std::map< System::ConstPointer, blitz::Array<T, 1>* >::const_iterator entry;
	
	for (entry = map.data_.begin(); entry != map.data_.end(); ++entry)
		os << *entry->second << std::endl;

	return os;
}

template <class T>
std::istream& operator>> (std::istream& is, SystemMap<T>& map)
{
	typename SystemMap<T>::MapType::const_iterator entry;

	//typename std::map< System::ConstPointer, blitz::Array<T, 1>* >::iterator entry;

	for (entry = map.data_.begin(); entry != map.data_.end(); ++entry)
		is >> *entry->second;
	
	return is;
}

#endif

