// SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
// 
// Contributor(s): Cheng Hong 
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
// (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
// merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
// furnished to do so, subject to the following conditions: 
// 
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
// CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
// THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
// 

#ifndef IVL_VECTORS_H
#define IVL_VECTORS_H

#include "general.h"
#include <math.h>
#include <blitz/tinyvec-et.h>

// This file contains some simple linear algebra operations that are helpful
// for manipulating atom coordinates -- for instance, for aligning one set 
// of vectors to match up with another, for use in regenerating a full residue
// side chain from partial atom coordinate info. These functions are based 
// largely on the Rosetta source, since we use their strategy for aligning 
// partial coordinate data to a known "template".

namespace CHOMPMath
{

inline FLOAT WrapValue(FLOAT val, FLOAT bound)
{
	const float halfBound = 0.5f * bound;

	// wrap the value to the range -bound/2 to +bound/2
	if (val >= halfBound || val <= -halfBound)
		return fmod( 
			fmod(val, bound) + bound + halfBound, 
			bound ) 
		       - halfBound;
	else
		return val;
}

inline void IdentityMatrix(Matrix3* matOut)
{
	(*matOut)(0) = Vector3(1, 0, 0);
	(*matOut)(1) = Vector3(0, 1, 0);
	(*matOut)(2) = Vector3(0, 0, 1);
}

inline Vector3 VectorMatrixMultiply(Vector3 const& vec, Matrix3 const& mat)
{
	return Vector3( blitz::dot(vec, mat(0)),
			blitz::dot(vec, mat(1)),
			blitz::dot(vec, mat(2)) );
}

inline void MakeRotationMatrix(FLOAT phi, FLOAT psi, FLOAT chi, Matrix3* matOut)
{
	if ( std::abs(phi) < SMALL_E ) 
		phi = 0.0;
	if ( std::abs(psi) < SMALL_E ) 
		psi = 0.0;
	if ( std::abs(chi) < SMALL_E )
	{
		IdentityMatrix(matOut);
		return;
	}

	FLOAT const sin_phi = std::sin(phi);
	FLOAT const cos_phi = std::cos(phi);
	FLOAT const sin_psi = std::sin(psi);
	FLOAT const cos_psi = std::cos(psi);
	FLOAT const neg_sin_chi = std::sin(-chi);
	FLOAT const neg_cos_chi = std::cos(-chi);
	FLOAT const neg_cos_chi_1 = neg_cos_chi - 1.0;

	(*matOut)(0) = Vector3(  (cos_phi*cos_phi + sin_phi*sin_phi*cos_psi*cos_psi) * neg_cos_chi
				+ sin_phi*sin_phi*sin_psi*sin_psi,
				 sin_psi*sin_psi*cos_phi*sin_phi*neg_cos_chi_1 
				+ cos_psi*neg_sin_chi,
				 cos_psi*sin_psi*sin_phi*neg_cos_chi_1 
				- cos_phi*sin_psi*neg_sin_chi );

	(*matOut)(1) = Vector3(  sin_psi*sin_psi*sin_phi*cos_phi*neg_cos_chi_1
				- cos_psi*neg_sin_chi,
				 (sin_phi*sin_phi + cos_phi*cos_phi*cos_psi*cos_psi) * neg_cos_chi
				+ cos_phi*cos_phi*sin_psi*sin_psi,
				 -cos_phi*sin_psi*cos_psi*neg_cos_chi_1 
				- sin_phi*sin_psi*neg_sin_chi );

	(*matOut)(2) = Vector3(  sin_phi*cos_psi*sin_psi*neg_cos_chi_1 
				+ cos_phi*neg_sin_chi*sin_psi,
				 -cos_phi*sin_psi*cos_psi*neg_cos_chi_1 
				+ sin_phi*sin_psi*neg_sin_chi, 
				 sin_psi*sin_psi*neg_cos_chi
				+ cos_psi*cos_psi );
}

//calculates the angle between the two given vectors
inline FLOAT CalculateVectorAngle(Vector3 const& v1, Vector3 const& v2)
{
	const FLOAT cosine = blitz::dot(v1, v2)/(std::sqrt(blitz::dot(v1, v1) * blitz::dot(v2, v2)));
        FLOAT angle;
        if (cosine <= -1.0)
                angle = 180.0;
        else if(cosine >= 1.0)
                angle = 0.0;
        else
                angle = std::acos(cosine) / 3.14159 * 180.0;
	return angle;

}

inline void MakeRotationMatrixVector(Vector3 a1, Vector3 a2, FLOAT chi, Matrix3* out_matrix, Vector3* out_vector) {
        FLOAT sinchi = std::sin(chi);
        FLOAT coschi = std::cos(chi);
        Vector3 w = a2 - a1;
        FLOAT mag = std::sqrt(dot(w, w));
       
	if (mag < 0.001) {
		w = 0;
	}
	else {
		w = w/mag;
	}

        FLOAT wx = w(0);
        FLOAT wy = w(1);
        FLOAT wz = w(2);
        (*out_matrix)(0)(0) = coschi+wx*wx*(1-coschi);
        (*out_matrix)(0)(1) = wx*wy*(1-coschi)-wz*sinchi;
        (*out_matrix)(0)(2) = wy*sinchi+wx*wz*(1-coschi);
        (*out_matrix)(1)(0) = wz*sinchi+wx*wy*(1-coschi);
        (*out_matrix)(1)(1) = coschi+wy*wy*(1-coschi);
        (*out_matrix)(1)(2) = -wx*sinchi+wy*wz*(1-coschi);
        (*out_matrix)(2)(0) = -wy*sinchi+wx*wz*(1-coschi);
        (*out_matrix)(2)(1) = wx*sinchi+wy*wz*(1-coschi);
        (*out_matrix)(2)(2) = coschi+wz*wz*(1-coschi);
        Vector3 zero = Vector3(0, 0, 0);
        (*out_vector) = zero;
}

//calculates the out of plane bending angle
inline FLOAT CalculateOutOfPlaneAngle(Vector3 const& a1, Vector3 const& a2, 
				    Vector3 const& a3, Vector3 const& a4)
{
	const Vector3 a2a1 = a2 - a1;
	const Vector3 a3a2 = a3 - a2;
	const Vector3 a4a3 = a4 - a3;
	
	//Vector normals to the planes a1a2a3 and a2a3a4
	const Vector3 t = blitz::cross(a2a1, a3a2);
	const Vector3 u = blitz::cross(a3a2, a4a3);
	
	const Vector3 txu = blitz::cross(t, u);
	
	const FLOAT cosine = blitz::dot(t, u)/(std::sqrt(blitz::dot(t, t) * blitz::dot(u, u)));
	const FLOAT sine = blitz::dot(a3a2, txu)/(std::sqrt(blitz::dot(a3a2, a3a2) * blitz::dot(t, t) * blitz::dot(u, u)));
	
	FLOAT angle;
	if (cosine <= -1.0)
		angle = 180.0;
	else if(cosine >= 1.0)
		angle = 0.0;
	else	
		angle = std::acos(cosine) / 3.14159 * 180.0;
	
	if (sine < 0)
		angle = -1.0*angle;

	return angle;
}

inline void AlignVectors(Vector3 const& a1, Vector3 const& a2,Vector3 const& b1, Vector3 const& b2,Matrix3* out_matrix, Vector3* out_vector) {
        FLOAT phi, psi, chi;
        Vector3 a = a2 - a1;
        Vector3 b = b2 - b1;
        const FLOAT mag_a = std::sqrt(blitz::dot(a,a));
        const FLOAT mag_b = std::sqrt(blitz::dot(b,b));

        Vector3 rot = blitz::cross(a,b);
        const FLOAT mag_rot = std::sqrt(blitz::dot(rot, rot));

        if (mag_rot == 0.0) {
                psi = 0.0;
        } else {
                psi = std::acos(rot(2)/mag_rot);
        }

        if ((rot(0) == 0.0) && (rot(1) == 0.0)) {
                phi = 0.0;
        }
        else {
                phi = std::atan2(-rot(0), rot(1));
        }

	const FLOAT ab = blitz::dot(a,b);
	const FLOAT mag_ab = mag_a * mag_b;
        
	if (std::abs(ab - mag_ab) > 0.001) {
	        chi = std::acos(blitz::dot(a,b)/(mag_a*mag_b));
	}
	else {
		chi = 0.0;
	}

        MakeRotationMatrix(phi, psi, chi, out_matrix);

        Vector3 result = CHOMPMath::VectorMatrixMultiply(a1, *out_matrix);
        *out_vector = b1 - result;
}

inline void RotateAboutVector(Vector3 a1, Vector3 a2, FLOAT chi, Matrix3* out_matrix, Vector3* out_vector) {
        MakeRotationMatrixVector(a1, a2, chi, out_matrix, out_vector);
        Vector3 result = VectorMatrixMultiply(a1, *out_matrix);
        *out_vector = a1 - result;
}


inline FLOAT CalculateDihedralAngle(Vector3 const& a1, Vector3 const& a2, Vector3 const& a3, Vector3 const& a4) {
        Vector3 b1 = a2 - a1;
        Vector3 b2 = a3 - a2;
        Vector3 b3 = a4 - a3;

        FLOAT mag_b2 = std::sqrt(blitz::dot(b2, b2));

        Vector3 b1xb2 = blitz::cross(b1, b2);
        Vector3 b2xb3 = blitz::cross(b2, b3);

        FLOAT angle = std::atan2(mag_b2 * blitz::dot(b1, b2xb3), blitz::dot(b1xb2, b2xb3));

        return angle / 3.14159 * 180;
}

inline void AlignVectorToPlane(Vector3 const& a1, Vector3 const& a2, Vector3 const& a3, Vector3 const& b1, Matrix3* rotOut, Vector3* transOut) {
        // Get perpendicular vectors
        Vector3 c1 = a1 - a2;
        Vector3 c2 = a3 - a2;
        Vector3 c3 = b1 - a2;

        Vector3 v1 = blitz::cross(c2, c1);
        Vector3 v2 = blitz::cross(c3, c1);

        // Find angle between them
        FLOAT mag_v1 = std::sqrt(blitz::dot(v1, v1));
        FLOAT mag_v2 = std::sqrt(blitz::dot(v2, v2));
	FLOAT theta;
	FLOAT v1v2 = blitz::dot(v1, v2);
	FLOAT mag_v1v2 = mag_v1 * mag_v2;
	FLOAT test = 1;

	if (std::abs(std::abs(v1v2) - mag_v1v2) < 0.00001) {
		test = copysignf(test, v1v2);
		if (test < 0.0) {
			theta = 3.14159;
		}
		else { 
			theta = 0.0;
		}
	}
	else {
        	theta = std::acos( v1v2 / mag_v1v2);
	}

        // Check direction
        Vector3 v = blitz::cross(v1, v2);
        if (blitz::dot(v, a2 - a1) < 0) {
                theta = -theta;
        }
	
        RotateAboutVector(a1, a2, theta, rotOut, transOut);
}

}


#endif

