#ifndef IVL_ARRAY_WRAPPER_H
#define IVL_ARRAY_WRAPPER_H

#include <blitz/tinyvec.h>
#include <blitz/array.h>

#include <numpy/arrayobject.h>
#include <boost/python/to_python_converter.hpp>

#include <cstdlib>
#include <vector>

///////////////////////////////////////////////////////////////////////////
//////                    C++ TYPE <-> NUMPY TYPE                    //////
///////////////////////////////////////////////////////////////////////////

template <typename Type> struct PyArray_TypeConverter;

template <> 
struct PyArray_TypeConverter<float>
{
	static int Get() { return PyArray_FLOAT; }
};

template <> 
struct PyArray_TypeConverter<double>
{
	static int Get() { return PyArray_DOUBLE; }
};

template <> 
struct PyArray_TypeConverter<int>
{
	static int Get() { return PyArray_INT; }
};

template <> 
struct PyArray_TypeConverter<long>
{
	static int Get() { return PyArray_INT; }
};

template <typename T, int N>
struct PyArray_TypeConverter< blitz::TinyVector<T, N> > 
{ 
	static int Get() { return PyArray_TypeConverter<T>::Get(); }
};

///////////////////////////////////////////////////////////////////////////
//////                       1D MEMORY BLOCKS                       ///////
///////////////////////////////////////////////////////////////////////////

template <typename Type>
struct Size 
{ 
	static int  Get(Type const& obj) { return obj.size(); }
	static void Set(Type& obj, int size) { obj.resize(size); }
};

template <typename T, int N>
struct Size< blitz::TinyVector<T, N> > 
{ 
	static int  Get( blitz::TinyVector<T, N> const& obj ) { return obj.length(); }
	static void Set( blitz::TinyVector<T, N> const& obj, int size) { /* do nothing */ }
};

template <> struct Size<int>    { static int Get(int    obj) { return 1; } };
template <> struct Size<long>   { static int Get(int    obj) { return 1; } };
template <> struct Size<float>  { static int Get(double obj) { return 1; } };
template <> struct Size<double> { static int Get(double obj) { return 1; } };

// Works for TinyVectors, std::vectors, boost::arrays, etc.
// when the element is a TinyVector or boost::array, will result in a 2D NumPy array.
// The size of the element type must be determinable at compile time... so
// DON'T use this for a std::vector of std::vectors!
// DON'T try to nest the array-ness more than two levels deep!

template < typename Type, typename Element >
struct mem_block_to_numarray
{
	static PyObject* convert( Type const& block )
	{
		int numElements = Size<Type>::Get(block);
		if (numElements == 0) 
			return Py_None;

		int elSize = Size<Element>::Get(block[0]);
		int dims[2] = { numElements, elSize };
		
		Element* data = new Element[numElements];
		memcpy(data, &block[0], sizeof(Element) * numElements);

		return PyArray_FromDimsAndData( elSize == 1 ? 1 : 2, 
			dims, PyArray_TypeConverter<Element>::Get(), (char*)data); 
	}
};

template < typename Type, typename Element >
struct mem_block_from_numarray
{
	mem_block_from_numarray()
	{
		boost::python::converter::registry::push_back(
			&convertible,
			&construct,
			boost::python::type_id<Type>());
	}

	static void* convertible(PyObject* obj_ptr)
	{
		int elSize = Size<Element>::Get(Element());

		PyArrayObject* array = (PyArrayObject*)
			PyArray_ContiguousFromObject(obj_ptr, PyArray_TypeConverter<Element>::Get(), 1, 2);

		// just a bad array
		if (!array) 			  return 0;

		// a malformed 2-D array
		if (elSize > 1)
			if (array->nd == 1 || array->dimensions[1] != elSize) return 0;
			
		// the size of the result is fixed and does not match the input
		int expectedSize = Size<Type>::Get(Type());
		if (expectedSize != 0 && array->dimensions[0] != expectedSize) return 0;
	
		return array;
	}

	static void convert(PyObject* obj_ptr, void* ptr)
	{
		PyArrayObject* array = (PyArrayObject*)
			PyArray_ContiguousFromObject(obj_ptr, PyArray_TypeConverter<Element>::Get(), 1, 2);

		new (ptr) Type();
		Type& result = *(Type*)ptr;

		if (Size<Type>::Get(result) == 0)
			Size<Type>::Set(result, array->dimensions[0]);

		memcpy( &result[0], array->data, sizeof(Element) * array->dimensions[0]);
	}

	static void construct(PyObject* obj_ptr, 
			      boost::python::converter::rvalue_from_python_stage1_data* data)
	{
		void* storage = (
			(boost::python::converter::rvalue_from_python_storage<Type>*)
				data)->storage.bytes;

		convert(obj_ptr, storage);
		data->convertible = storage;
	}
};

///////////////////////////////////////////////////////////////////////////
//////                 BLITZ ARRAY <-> NUMPY ARRAY                   //////
///////////////////////////////////////////////////////////////////////////

// SCALAR ELEMENTS ONLY!

template < typename Element, int N >
struct blitz_array_to_numarray
{
	static PyObject* convert( blitz::Array<Element, N> const& array )
	{
		const int MAX_DIMS = 10;

		int extents[MAX_DIMS];
		int dims = array.dimensions();
		assert( dims < MAX_DIMS );
	
		for (int i = 0; i < dims; ++i)
			extents[i] = array.extent(i);

		PyObject* arr = PyArray_FromDimsAndData(dims, extents, 
					PyArray_TypeConverter<Element>::Get(), (char*)array.data()); 
		PyObject* copy = PyArray_Copy((PyArrayObject*)arr);
		Py_DECREF(arr);

		return copy;
	}
};

template < typename Element, int N >
struct blitz_array_from_numarray
{
	blitz_array_from_numarray()
	{
		boost::python::converter::registry::push_back(
			&convertible,
			&construct,
			boost::python::type_id< blitz::Array<Element, N> >());
	}

	static void* convertible(PyObject* obj_ptr)
	{
		PyArrayObject* array = (PyArrayObject*)
			PyArray_ContiguousFromObject(obj_ptr, PyArray_TypeConverter<Element>::Get(), N, N);

		if (!array || array->nd != N)
			return 0;
		else
			return array;
	}

	static void convert(PyObject* obj_ptr, void* ptr)
	{
		PyArrayObject* array = (PyArrayObject*)
			PyArray_ContiguousFromObject(obj_ptr, PyArray_TypeConverter<Element>::Get(), N, N);

		blitz::TinyVector<int, N> dims;
		for (int i = 0; i < N; ++i)
			dims[i] = array->dimensions[i];
		
		new (ptr) blitz::Array<Element, N>((Element*)array->data, dims, blitz::neverDeleteData);
	}

	static void construct(PyObject* obj_ptr, 
			      boost::python::converter::rvalue_from_python_stage1_data* data)
	{
		void* storage = (
			(boost::python::converter::rvalue_from_python_storage< blitz::Array<Element, N> >*)
				data)->storage.bytes;

		convert(obj_ptr, storage);
		data->convertible = storage;
	}
};

///////////////////////////////////////////////////////////////////////////
//////                      CONVENIENCE MACROS                      ///////
///////////////////////////////////////////////////////////////////////////

#define MEMBLOCK_CONVERT(Type, Element)  						\
{											\
	boost::python::to_python_converter						\
		< Type, mem_block_to_numarray<Type, Element > >();			\
	mem_block_from_numarray< Type, Element >();					\
}

#define BLITZARRAY_CONVERT(Type, N)							\
{											\
	boost::python::to_python_converter						\
		< blitz::Array<Type, N>, 						\
                  blitz_array_to_numarray< Type, N > >();				\
	blitz_array_from_numarray<Type, N>();						\
}

#endif

