# SHARPEN is copyright (c) 2008  Ilya Loksha (primary author), Christopher Snow (project lead).
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files 
# (the "Software") to deal in the Software without restriction, including without limitation the rights to use, copy, modify, 
# merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions: 
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. 
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY 
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
# THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 
# 

import os
import pyplusplus
from pyplusplus import module_builder

def include_misc(mb, float_type):
	# include miscellaneous things that don't get included by default
	mb.decls( name='pair<int, int>' ).include()
	mb.decls( name='pair<int, %s>' % float_type ).include()
	mb.decls( name='pair<std::vector<int, std::allocator<int> >, %s>' % float_type ).include()
	
	for d in xrange(2, 5):
		for type in [float_type, 'int']:
			try:
				mb.decls( name='TinyVector<%s, %d>' % (type, d)).include()
			except: pass # probably no such class

def strip_tinyvec(mb, float_type):
	# exclude parts of TinyVector that make things difficult for boost.python
	for d in xrange(2, 5):
		for type in [float_type, 'int']:
			bv_type = 'TinyVector<%s, %d>' % (type, d)
			try:
				cls = mb.class_(bv_type)
				cls.mem_funs().exclude()
				cls.operators().exclude()

				cls.operators('operator[]')[0].include()
				cls.mem_funs('length')[0].include()
			except: pass # probably no such class

def wrap_code(mb):
	# include all definitions in CHOMP headers
	files = os.listdir('../include') + os.listdir('../core/codegen/output')
	files = [ f for f in files if os.path.splitext(f)[1] in [ '.h', '.hxx' ] ]

	mb.decls( lambda d: d.location and os.path.split(d.location.file_name)[1] in files ).include()

	# include miscellaneous
	try:
		include_misc(mb, 'float')
		strip_tinyvec(mb, 'float')
	except:
		include_misc(mb, 'double')
		strip_tinyvec(mb, 'double')

	# get rid of random noncopyableness
	copyable_types = \
	['vector', 'list', 'pair', 'AminoAcid', 'Residue', 'RotamerGenerationParameters',
	 'Raw', 'Rosetta', 'Hybridization', 'CountPair', 'EnergyComb']

	for type in copyable_types:
		copyable = mb.decls( lambda d: d.name.startswith(type) )
		for decl in copyable.declarations:
			if isinstance(decl, pyplusplus.decl_wrappers.class_wrapper.class_t):
				decl.noncopyable = False

	# exclude the undesirables
	excluded_files = [ 'pdb.h', 'system_map.h', 'rotamer_iterators.h', 'mersenne_twister.h' ]
	mb.decls( lambda d: d.location and os.path.split(d.location.file_name)[1] in excluded_files ).exclude()

	# misc. unnecessary classes
	for cls in [ 'PDBResidueFactory', 'EnergyGraphNode' ]:
		mb.class_(cls).exclude()

	# set up the general call policies
	for rtype in [ '::Residue &', '::Residue const &', '::AminoAcid const *' ]:
		mb.mem_funs( return_type=rtype ).call_policies = module_builder.call_policies.return_internal_reference()
	
	# register all the shared_ptr objects
	shared_ptrs = mb.decls( lambda decl: decl.name.startswith('shared_ptr<') )

	# some classes automatically register their pointers. I haven't yet figured out why...
	exclude_classes = [ 'System', 'DataDirectory', 'RotamerSet', 'RotamerLibrary', 'NeighborMap', 'RotamerIterationMap',
			    'PoseLibrary', 'EnergyCombList', 'RuntimeStats', 'AminoAcidDirectory', 'AtomTypes',
			    'SimulatedAnnealingPacker', 'SimulatedAnnealingPackerMS', 'MonteCarloPacker', 'BruteForcePacker', 
			    'FasterPacker', 'RotamerGenerationParameters', 'Residue', 'EnergyGraph',
			    'MonteCarloFBFPacker', 'EnergyHyperGraph', 'EnergyComb', 'AminoAcid']

	for ptr_type in shared_ptrs:
		pointee = ptr_type.name.split('<')[1].split('>')[0].strip(' ')
		is_const = ( ptr_type.name.find('const') != -1 )

		if not is_const and pointee in exclude_classes: continue

		# a little hack until I put CHOMP into its own namespace
		if pointee == 'void' or pointee.startswith('boost'): continue

		mb.add_registration_code( 'boost::python::register_ptr_to_python< boost::%s >();' % ptr_type.name )

		# add implicit conversion
		if is_const:
			mb.add_registration_code(
				'boost::python::implicitly_convertible< boost::%s, boost::%s >();' % (ptr_type.name.replace('const',''),
												      ptr_type.name) )
	# create instances of the array_wrapper wrappers
	mb.add_registration_code( 'import_array();' )
	for dim in [ 1, 2, 3, 4, 5 ]:
		mb.add_registration_code( 'BLITZARRAY_CONVERT(FLOAT, %d);' % dim )

	# translate operator<< and operator>>
	mb.free_operators( lambda fn: fn.name == 'operator<<' ).exclude()
	mb.free_operators( lambda fn: fn.name == 'operator>>' ).exclude()

	# TODO: convert this to namespace-based operation...
	for io_class in [ 'System' ]:
		cls = mb.class_(io_class)

		# input
		overload_index = 0
		for fn in mb.free_operators( name='operator>>', header_file=cls.location.file_name ):
			fn_name = cls.name + '_parse_' + str(overload_index)
			overload_index += 1

			cls.add_declaration_code(''.join([	\
				'void ', fn_name, ' (', fn.argument_types[1].decl_string, ' obj, std::string const& str)\n',
				'{\n\t std::istringstream ss(str);\n',
				' \t ss >> obj;\n',
				'}\n']))
		
			cls.add_registration_code( 'def( "Parse", ' + fn_name + ' )' )

		# output
		overload_index = 0
		for fn in mb.free_operators( name='operator<<', header_file=cls.location.file_name ):
			fn_name = cls.name + '_print_' + str(overload_index)
			overload_index += 1

			cls.add_declaration_code(''.join([	\
				'std::string ', fn_name, ' (', fn.argument_types[1].decl_string, ' obj)\n',
				'{\n\t std::ostringstream ss;\n',
				' \t ss << obj;\n',
				' \t return ss.str();\n',
				'}\n']))

			cls.add_registration_code( 'def( "__str__", ' + fn_name + ' )' )

	# Adapt iterators to Python

	def adapt_iterators(cls, prop=None, begin='begin', end='end'):
		clsdef = mb.class_(cls)
		if prop:
			clsdef.add_registration_code( ''.join([	\
				'property("',  prop, '", boost::python::range(&', 
				       clsdef.decl_string, '::', begin,
				', &', clsdef.decl_string, '::', end, ' ))']))
		else:
			clsdef.add_registration_code( ''.join([	\
				'def("__iter__", boost::python::iterator< ', clsdef.decl_string, ' >() )']))
			
			# for default iterators, also turn a 'size' fn into a __len__
			clsdef.add_registration_code( ''.join([ \
				'def("__len__", &', clsdef.decl_string, '::size)' ]))
			
			clsdef.mem_funs( name='size' ).exclude()
		
		clsdef.mem_funs( name=begin ).exclude()
		clsdef.mem_funs( name=end   ).exclude()

	# adapt_iterators('System')

	# Turn accessors into properties
	for cls in ['System', 'Residue', 'DataDirectory', 'DunbrackLibrary']:
		mb.class_(cls).add_properties( exclude_accessors=True )

	# Aliases for serialization-load functions
	load_funs = mb.decls( lambda decl: decl.name.startswith('LoadFromArchive') or decl.name.startswith('LoadBasicFromArchive') )
	for load_fun in load_funs:
		if not load_fun.return_type.decl_string.startswith('::boost::shared_ptr'):
			continue

		return_type_name = load_fun.return_type.decl_string.split('<')[1].split('>')[0].strip(' ')
		load_fun.alias = 'Load' + return_type_name + 'FromArchive'
		load_fun.name = load_fun.demangled_name

	# Add exception translation
	mb.add_declaration_code( 
	"""
		void str_exception(std::string const& exc)
		{
			if (exc == "EOF")
				PyErr_SetString( PyExc_EOFError, "" );
			else
				PyErr_SetString( PyExc_RuntimeError, ("CHOMP Error: " + exc).c_str() );
		}
	""")

	mb.add_registration_code('boost::python::register_exception_translator<std::string>(str_exception);')

	# Creating code creator. After this step you should not modify/customize declarations.
	mb.build_code_creator( module_name='PyCHOMP' )

	mb.code_creator.add_include('boost/python/register_ptr_to_python.hpp')
	mb.code_creator.add_include('general.h')
	mb.code_creator.add_include('array_wrapper.hpp')

	# add the stdafx include at the top of the main file
        from pyplusplus.code_creators import include
	creator = include.include_t( "stdafx.h" )
        mb.code_creator.adopt_creator( creator, 0 )

	# Writing code to files.
	mb.split_module( '.' )

	# workaround for mis-generation of code that uses size_t, which causes conflicts on 64-bit vs 32-bit systems
	size_t_files = [ 'EGVertexVector.pypp.cpp', 'EnergyGraph.pypp.cpp', 'EnergyHyperGraph.pypp.cpp' ]

	for f in size_t_files:
		print 'fixing ', f, ' size_t bug'
		cmd_strs = [ "sed 's/counting_iterator<unsigned long/counting_iterator<std::size_t/g' < %s > %s2" % (f,f),
			     "sed 's/counting_iterator<unsigned int/counting_iterator<std::size_t/g' < %s2 > %s3" % (f,f),
			     "sed 's/counting_iterator<long unsigned int/counting_iterator<std::size_t/g' < %s3 > %s" % (f,f),
			     "rm -f %s2 %s3" % (f,f) ]

		for cmd_str in cmd_strs:
			os.system(cmd_str)

	# workaround for gnu_cxx appearing in some files where it's completely unnecessary
	gnu_cxx_files = [
	   ( 'System.pypp.cpp', 
	   	'::__gnu_cxx::__normal_iterator<Residue\\*, std::vector<Residue, std::allocator<Residue> > >',
		'std::vector<Residue>::iterator' ),
	   ( 'System.pypp.cpp', 
	   	'::__gnu_cxx::__normal_iterator<Residue const\\*, std::vector<Residue, std::allocator<Residue> > >',
		'std::vector<Residue>::const_iterator' ),
	]

	for item in gnu_cxx_files:
		f, src_str, dest_str = item

		print 'fixing ', f, ' gnu_cxx non-portable thing'
		cmd_strs = ["sed 's/%s/%s/g' < %s > %s2" % (src_str, dest_str, f, f),
			    "mv -f %s2 %s" % (f,f) ]

		for cmd_str in cmd_strs:
			os.system(cmd_str)
	

